#!/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2023 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

import time
import re
import json
from ..column import ColumnAttr
from .. import utilities
from .. import configuration

class ColumnAlert(ColumnAttr):
    """Mail alert configuration"""
    name = "alert"
    gui_display = "GUI_button"
    action = "configure_alert"
    check_and_set = 'function(value) { if (value.push) return value; return JSON.parse(value);}'
    default_value = ()
    formatter = "function(column, value) { return JSON.stringify(value) ;}"
    def encode(self, value):
        """JSON to Python"""
        return json.loads(value)
    def decode(self, value):
        return json.dumps(value)

WORK_TO_DO = []

def send_mail(page, column, line):
    """Do not do it now because it can take time and locks"""
    WORK_TO_DO.append((page, column, line))
    utilities.start_job(process_job, 1, important="mail_alert")

def process_job():
    """Send the mail"""
    print('='*99)
    def update_filter():
        print('update filter')
        for i, mail in enumerate(column.alert):
            filter_txt = mail.get('filter', '')
            if filter_txt:
                filter_fct = page.table.get_filter(filter_txt, column.type.name)
                really_send[i] = filter_fct(line, line[column.data_col], page.user_name)
                print(mail, really_send[i])

    while WORK_TO_DO:
        page, column, line = WORK_TO_DO.pop(0)
        really_send = []
        for mail in column.alert:
            if mail.get('filter', ''):
                really_send.append(False)
            else:
                really_send.append(True)
        if min(really_send) is False:
            page.table.compute_columns(lines=(line,), hook=update_filter)
        my_mail = parse_mails(page.user_name)[0]
        for mail, send_it in zip(column.alert, really_send):
            if not send_it:
                continue
            recipients = parse_mails(replace(page.table, line, mail.get('to', '')))
            carboncopy = parse_mails(replace(page.table, line, mail.get('cc', '')))
            subject = replace(page.table, line, mail.get('subject', ''))
            body = replace(page.table, line, mail.get('body', ''))
            if recipients or carboncopy:
                utilities.send_mail_in_background(
                    recipients, subject, body, cc=carboncopy, show_to=True, reply_to=my_mail)
        last_check = time.time()
    return last_check

def parse_mails(text):
    """Split mail list and replace logins per mails"""
    mails = []
    for mail in re.split(' ', text):
        if '@' in mail and '.' in mail:
            mails.append(mail)
        else:
            user = configuration.User(mail)
            mail = user.mail
            if mail:
                mails.append(mail)
    return mails

def replace(table, line, text):
    """Replace [Title] #[Title] by the cell content"""
    items = iter(text.split('['))
    result = [next(items)]
    for item in items:
        if ']' not in item:
            result.append('[')
            result.append(item)
            continue
        title, remain = item.split(']')
        column = table.columns.from_title(title)
        if not column:
            result.append('[')
            result.append(item)
            continue
        if result[-1].endswith(('#', '?', '@')):
            what = result[-1][-1]
            result[-1] = result[-1][:-1]
            if what == '#':
                result.append(line[column.data_col].comment)
            elif what == '?':
                result.append(line[column.data_col].date)
            else:
                result.append(line[column.data_col].author)
        else:
            result.append(str(line[column.data_col].value))
        result.append(remain)
    return ''.join(result)
