/*
  TOMUSS: The Online Multi User Simple Spreadsheet
  Copyright (C) 2016-2019 Thierry EXCOFFIER, Universite Claude Bernard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Contact: Thierry.EXCOFFIER@univ-lyon1.fr
*/

/*

Classes:
  * NotationGrade : grade, max and comment
  * NotationQuestion : question, max, steps, type
    + NotationGrade
  * Notation : question list

The question list and the grade list:
  * are updated on each keystroke
  * are stored on student change
  The grades of all the students are computed on popup close.

Type of question:
  * 0 : Q question
  * 1 : D deleted question
  * 2 : B question bonus
  * 3 : C comment

Priorities:
  * 0, 1, 2, 3... for the original question list
  * ...           for the remote created questions
  * 1000, 1001... for the localy created questions (updated after merging)
  * 9999          for the empty question (updated to 1000... if filled)

  XXX BAD: these comments indicate some lines to be modified synchronously
*/

var notation_debug = false;
var notation_types = "QDBC";

// TODO
// XXX Feedback sauvegarde

function trunc(x) {
    return Number(x.toFixed(3));
}

function NotationGrade(txt) {
    this.comment = '';
    this.stored = '?';
    this.max = 1;
    this.set_comment(txt);
    this.local_change = false;
    this.not_graded = true;
}

NotationGrade.prototype.set_comment = function (value) {
    var g;
    try {
        g = RegExp('^ *([-0-9.]*|[?])[-0-9. ]*/([0-9.]*) *([^]*)').exec(value);
        g = [g[1], g[2], g[1] === '' && g[3] === '' ? this.comment : g[3]];
    }
    catch (e) {
        g = RegExp('^ *([-0-9.]*)([^]*)').exec(value);
        g = [g[1], undefined, g[2] || this.comment];
    }
    var error = this.set_grade(g[0], g[1]);
    this.local_change |= this.old_value != value;
    this.not_graded &= !this.local_change;
    this.comment = g[2];
    this.old_value = value;
    return error;
};

NotationGrade.prototype.uncompress_comment = function () {
    var questions = this.question.notation.questions;
    this.comment = this.comment.replace(/{{{([^}]*)}}}/g,
        function (comment_id) {
            the_id = comment_id.substr(3, comment_id.length - 6);
            if (questions[the_id] === undefined)
                return comment_id;
            return questions[the_id].question;
        });
};

NotationGrade.prototype.set_grade = function (value, max) {
    var old_stored = this.stored;
    var error;
    if (value !== '?' && (isNaN(value) || value === ''))
        value = this.stored;
    if (isNaN(max) || max === '')
        max = this.max;
    this.max = Math.max(0.1, Math.min(100, max));
    if (this.max != max)
        error = "MSG_notation_max_error";
    if (value === '?')
        this.stored = '?';
    else
        this.stored = Math.max(this.min === undefined ? -99999 : this.min,
            Math.min(this.max, value));
    if (this.stored != value)
        error = "MSG_notation_value_error";
    if (value === '?')
        this.grade = '?';
    else
        this.grade = this.stored / this.max;
    this.local_change |= old_stored != this.stored;
    this.not_graded &= !this.local_change;
    return error;
};

NotationGrade.prototype.toJSON = function () {
    var comment = this.comment;
    for (var question_id in this.question.notation.questions) {
        var question = this.question.notation.questions[question_id];
        if (question.is_a_comment())
            comment = comment.replace(question.question, '{{{' + question_id + '}}}');
    }
    return (this.stored + "/" + this.max + " " + comment).trim();
};

function NotationQuestion(dict) {
    this.question = "";
    this.max = 1;
    this.steps = 2;
    this.type = 0;
    this.comment_size = 1; // Minimum number of line for comments

    if (dict.substr) {
        g = RegExp('([^ ]*) ([0-9]*)([' + notation_types
            + '])([0-9.]*) (.*)').exec(dict);
        if (g.length != 6)
            alert("Can not parse: " + dict);
        this.id = g[1];
        this.steps = Number(g[2]);
        this.type = myindex(notation_types, g[3]);
        this.max = Number(g[4]);
        this.question = g[5];
    }
    else {
        for (var key in dict)
            this[key] = dict[key];
    }
    this.initial_value = this.hash();
    this.stats = new Stats();
    this.set_grade();
}

NotationQuestion.prototype.toJSON = function () {
    return this.id + ' ' + this.steps + notation_types.substr(this.type, 1)
        + this.max + ' ' + this.question;
};

NotationQuestion.prototype.hash = function () {
    return this.question + this.max + this.steps + this.type;
};

notation_min_digits = { 1: 0, 2: 1, 4: 2, 5: 1, 10: 1, 20: 2, 25: 2, 50: 2, 100: 2 };

NotationQuestion.prototype.nice_grade = function (v, steps) {
    if (v === undefined) {
        if (this.grade.grade === '?')
            return '?';
        v = this.grade.grade * this.max;
    }
    if (steps === undefined)
        steps = this.steps;
    var i = notation_min_digits[steps];
    if (i === undefined)
        i = 3;
    for (; i < 3; i++) {
        var x = v * Math.pow(10, i);
        if (Math.abs(x - Math.round(x)) < 0.0001)
            break;
    }
    return v.toFixed(i);
};

NotationQuestion.prototype.grade_and_comment = function () {
    var v = '       ' + this.nice_grade();
    v = v.substr(v.length - 7, 7); // Right justify
    return v + '/' + this.max + ' ' + this.grade.comment;
};

NotationQuestion.prototype.set_grade = function (txt) {
    var not_graded = !txt || txt.substr(0, 1) === '?';
    if (!txt) {
        if (this.is_a_question())
            txt = "?/" + this.max;
        else
            txt = "0/" + this.max;
    }
    this.new_grade(txt, not_graded);
};

NotationQuestion.prototype.new_grade = function (txt, not_graded) {
    this.grade = new NotationGrade(txt);
    this.grade.question = this;
    this.grade.not_graded = not_graded;
    this.set_min();
};

NotationQuestion.prototype.set_min = function () {
    if (this.is_a_bonus())
        this.grade.min = -this.grade.max;
    else
        this.grade.min = 0;
}

NotationQuestion.prototype.remove_grade = function () {
    if (this.grade.not_graded)
        return;
    this.stats.remove(this.grade.grade);
};

NotationQuestion.prototype.set_grade_to = function (value) {
    this.remove_grade();
    this.grade.set_grade(this.nice_grade(value), this.max);
    if (this.grade.grade != '?') {
        this.grade.not_graded = false;
        this.stats.add(this.grade.grade);
    }
    else
        this.grade.not_graded = true;
};

NotationQuestion.prototype.html = function (modifiable, questions_modifiable) {
    var q, comment;
    var q_class = "question";
    var c_class = "comment_input";
    if (this.is_fully_empty_question()) {
        q = _("MSG_notation_question");
        q_class += ' empty';
        comment = _("MSG_notation_comment");
        c_class += ' empty';
    }
    else {
        q = this.question;
        comment = this.grade_and_comment();
    }
    var textarea = document.getElementById(this.id);
    var style = '';
    if (textarea)
        style = ' style="height:'
            + textarea.getElementsByTagName('TEXTAREA')[0].style.height
            + '"';
    // XXX BAD
    return '<div id="' + this.id + '" class="type' + this.type + '">'
        + '<input value="' + encode_value(q) + '"'
        + ' class="' + q_class + '"'
        + ' onpaste="Notation.on_paste(event)"'
        + ' ondrop="Notation.on_paste(event)"'
        + ' onfocus="Notation.on_focus(event)"'
        + ' spellcheck="true"'
        + (modifiable ? '' : ' disabled')
        + '><div class="bonus_comment">'
        + (questions_modifiable
            ? hidden_txt('<span class="bonus">Ⓑ</span>', _('TIP_bonus_toggle'))
            : '<span class="bonus not_modifiable">Ⓑ</span>'
        )
        + (questions_modifiable
            ? '<span class="move_up">↑</span><span class="move_down">↓</span>'
            : '')
        + '</div><canvas tabindex="0" onfocus="Notation.on_focus(event)"></canvas>'
        + '<div class="incdec">'
        + (questions_modifiable
            ? '<span class="inc">⊕</span><br><span class="dec">⊖</span>'
            : '&nbsp;<br>&nbsp;')
        + '</div>'
        + '<textarea'
        + (modifiable ? '' : ' disabled')
        + ' onpaste="Notation.on_paste(event)"'
        + ' ondrop="Notation.on_paste(event)"'
        + ' onfocus="Notation.on_focus(event)"'
        + ' rows="' + (this.comment_size + 1) + '"'
        + style // override rows if user resized
        + ' spellcheck="true"'
        + ' class="' + c_class + '">' + html(comment) + '</textarea>'
        + (false && notation_debug
            ? '<br>' + this.id
            + ' priority=' + this.priority
            + ' initial_value=' + this.initial_value
            + ' somebody_is_graded=' + this.somebody_is_graded()
            + ' questions_modifiable=' + questions_modifiable
            : '')
        + '</div>';
};

NotationQuestion.prototype.suivi = function () {
    if (this.is_a_bonus() && this.grade.stored == 0 && this.grade.comment === '')
        return '';
    return '<tr><td style="text-align:right">'
        + (this.is_a_question()
            ? this.nice_grade() + '/' + this.grade.max
            : (this.grade.stored > 0 ? '+' : '') + this.nice_grade())
        + '<td>' + html(this.question)
        + (this.grade.comment !== ''
            ? '<br><em>' + html(this.grade.comment).replace(/\n/g, '<br>') + '</em>'
            : '')
        + '</tr>';
};

NotationQuestion.prototype.suivi_modifiable = function () {
    if (this.is_a_bonus())
        return '';
    return '<tr><td>' + _('MSG_T_on') + this.grade.max
        + '<td>' + html(this.question).replace(/⏎/g, '<br>')
        + '<br><textarea info="' + this.id
        + '" rows="' + Math.max(3, this.grade.comment.split('\n').length)
        + '" onblur="notation_save_textarea(this)">'
        + html(this.grade.comment) + '</textarea>'
        + '</tr>';
};

NotationQuestion.prototype.set_comment = function (value, column_modifiable) {
    this.remove_grade();
    var error = this.grade.set_comment(value);
    if (this.max != this.grade.max && !column_modifiable)
        error = "MSG_notation_not_allowed";
    else {
        this.max = this.grade.max;
        this.set_min();
        if (this.grade.grade != '?')
            this.stats.add(this.grade.grade);
    }

    return error;
};

/*
max = 1.2 steps = 2
  -1       -0.5        0        0.5        1        1.5        2
---|---------|---------|---------|---------|---------|---------|
.  |                   |                   |            .      |
.                                                       .
x_min=-1.2             <--span-->                       x_max=1.2 + 1/2
*/

NotationQuestion.prototype.span = function () {
    return 1 / this.steps;
};
NotationQuestion.prototype.x_min = function () {
    return this.is_a_bonus() ? -this.max : 0;
};
NotationQuestion.prototype.x_max = function () {
    return this.max + this.span();
};
NotationQuestion.prototype.floor = function (x) {
    return Math.floor(x * this.steps) / this.steps;
};
NotationQuestion.prototype.ceil = function (x) {
    return Math.ceil(x * this.steps) / this.steps;
};
NotationQuestion.prototype.get_value01 = function (x) {
    return (x - this.x_min()) / (this.x_max() - this.x_min());
};
NotationQuestion.prototype.x2pixel = function (canvas, x) {
    return this.get_value01(x) * canvas.width;
};

NotationQuestion.prototype.draw_canvas = function () {
    var canvas = document.getElementById(this.id);
    if (!canvas)
        return; // deleted question
    canvas = canvas.getElementsByTagName('CANVAS')[0];
    if (!canvas)
        return;
    canvas.width = canvas.offsetWidth;
    canvas.height = canvas.offsetHeight;

    var c = canvas.getContext("2d");
    var THIS = this;

    function line(x1, y1, x2, y2) {
        c.beginPath();
        c.moveTo(THIS.x2pixel(canvas, x1), canvas.height * y1);
        c.lineTo(THIS.x2pixel(canvas, x2), canvas.height * y2);
        c.closePath();
        c.stroke();
    }
    if (this.grade.not_graded) {
        c.fillStyle = "#FFF";
        c.fillRect(0, 0, canvas.width, canvas.height);
    }
    else if (this.grade.grade < -1 || this.grade.grade > 1) {
        c.fillStyle = "#F88";
        c.fillRect(0, 0, canvas.width, canvas.height);
    }
    else {
        c.fillStyle = '#' + ["F44", "DA4", "888", "AD4", "4F4", "4F4"][
            Math.floor(5 * this.get_value01(this.grade.grade * this.max))];
        c.fillRect(this.x2pixel(canvas, this.grade.grade * this.max), 0,
            this.x2pixel(canvas, this.x_min() + this.span()), canvas.height);
    }
    for (var x = this.ceil(this.x_min()); x <= this.floor(this.x_max()); x += this.span()) {
        line(x, 0, x, Math.abs(x % 1.) < 0.001 ? 1 : 0.5);
    }
    c.fillStyle = "#000";
    c.font = "10px sans";
    for (var x = Math.floor(this.x_min()); x <= Math.ceil(this.x_max()); x++)
        c.fillText(x, this.x2pixel(canvas, x) + 2, canvas.height - 1);
    c.strokeStyle = "#000";
    line(this.x_min(), 0, this.x_max(), 0);
    if (this.stats.nr >= 5) {
        var avg = this.stats.average() * this.max;
        var stddev = this.max * this.stats.standard_deviation() / 2;
        if (this.max_graded == this.stats.nr)
            c.fillStyle = c.strokeStyle = "#44F";
        else
            c.fillStyle = c.strokeStyle = "#0FF";
        line(avg - stddev, 0.2, avg + stddev, 0.2);
        c.beginPath();
        c.arc(this.x2pixel(canvas, avg), canvas.height / 5, canvas.height / 6, 0, 2 * Math.PI);
        c.closePath();
        c.fill();
    }
};

NotationQuestion.prototype.is_a_question = function () {
    return this.type === 0;
};

NotationQuestion.prototype.is_a_bonus = function () {
    return this.type === 2;
};

NotationQuestion.prototype.is_a_comment = function () {
    return this.type === 3;
};

NotationQuestion.prototype.is_a_question_or_bonus = function () {
    return this.is_a_question() || this.is_a_bonus();
};

NotationQuestion.prototype.is_an_empty_question = function () {
    return this.is_a_question() && this.question === '';
};

NotationQuestion.prototype.is_not_an_empty_question = function () {
    return this.is_a_question() && this.question !== '';
};

NotationQuestion.prototype.is_not_an_empty_bonus = function () {
    return this.is_a_bonus() && this.question !== '';
};

NotationQuestion.prototype.is_fully_empty_question = function () {
    return this.is_an_empty_question()
        && (this.grade.grade === 0 || this.grade.grade === '?')
        && this.grade.comment === '';
};

NotationQuestion.prototype.somebody_is_graded = function () {
    return this.stats.nr != 0;
};

function Notation() {
}

Notation.prototype.log = function (txt) {
    if (notation_debug)
        console.log(txt);
};

Notation.prototype.start = function () {
    if (this.notation_window_open)
        return;
    this.notation_window_open = true;
    this.jump_old = the_current_cell.jump.bind(the_current_cell);
    this.shortcuts = {};
    the_current_cell.jump = this.jump.bind(this);
    this.init(the_current_cell.column);
    create_popup(
        'notation_content',
        '<style>'
        + 'DIV.notation_content { border: 2px solid black; top: 10em ; right: 1em; bottom: 0px; left: 25%; padding: 0px }'
        + 'DIV.notation_content .the_completions, DIV.notation_content .the_questions {'
        + '    position: absolute; top: 3em; bottom: 0px; overflow-x: hidden; overflow-y: scroll }'
        + 'DIV.notation_content .the_completions { left:60%; right: 0px ; }'
        + 'DIV.notation_content .the_questions { right: 40%; left: 0px ; }'
        + 'DIV.notation_content .empty { color: #888 }'
        + 'DIV.notation_content INPUT { font-size: 100%; background: #EEE; border: none }'
        + 'DIV.notation_content TEXTAREA { font-size: 100%; background: initial; border: none }'
        + 'DIV.notation_content INPUT.question { width: 60% ; font-weight: bold }'
        + 'DIV.notation_content *:focus { background: #FFF }'
        + 'DIV.notation_content .comment_input { width: 100% ; display: block; margin-top:0.5em ; margin-bottom: 0.5em; }'
        + 'DIV.notation_content CANVAS { position: absolute; left: 67% ; width: 30% ; height: 1.4em ; opacity: 0.5 ; cursor: pointer }'
        + 'DIV.notation_content .incdec { left: 97% ; text-align: center; width: 3% }'
        + 'DIV.notation_content .bonus_comment { left: 60%; right: 33% }'
        + 'DIV.notation_content .incdec { line-height: 0.8em }'
        + 'DIV.notation_content .incdec, DIV.notation_content .bonus_comment {'
        + '   display: inline-block; position: absolute ; }'
        + 'DIV.notation_content .incdec, DIV.notation_content .bonus_comment, .edit_comment {cursor: pointer }'
        + '.edit_comment { font-family: emoji }'
        + 'DIV.notation_content .focus CANVAS { opacity: 1 }'
        + 'DIV.notation_content .incdec, .bonus_comment .bonus,'
        + '.bonus_comment .move_up, .bonus_comment .move_down { opacity:0 ; }'
        + '.bonus_comment .bonus, .bonus_comment .move_up, .bonus_comment .move_down'
        + ' { width: 33%; display: inline-block; text-align: center }'
        + '.notation_shortcut { width: 1em; display: inline-block; text-align: right }'
        + 'DIV.notation_content H1 { position: relative ; height: 1.5em}'
        + 'DIV.type2 .bonus { color: #00F }'
        + 'DIV.notation_content .a_completion { border: 1px solid #888; display: block; margin: 4px ; white-space: normal; font-size: 80% }'
        + 'DIV.notation_content .a_completion:hover { background: #FFF }'
        + 'DIV.notation_content .stat { color: #888 ; margin: 0.2em }'
        + '#the_questions P { margin-top: 0.5em ; font-weight: bold }'
        + '#notation_student { }'
        + '#notation_column { position: absolute ; right: 4em;}'
        + '#notation_grade {  }'
        + 'DIV.notation_content .focus DIV.incdec, DIV.notation_content .focus DIV.bonus_comment *, DIV.notation_content .type2 .bonus_comment .bonus { opacity: 1 }'
        + 'DIV.notation_introduction { background: #FFE }'
        + '</style>'
        + '<h1>'
        + '<span id="notation_column">' + html(this.column.title) + '</span>'
        + '<span id="notation_student"></span>: '
        + '<span id="notation_grade"></span><br>'
        + '<span id="notation_other"></span>'
        + '<span id="notation_error"></span>'
        + '</h1>',
        '<div id="notation_content"'
        + ' onmouseup="Notation.on_mouse_up(event)"'
        + ' onmousedown="Notation.on_mouse_down(event)"'
        + ' onkeyup="Notation.on_keyup(event)"'
        + ' onkeydown="Notation.on_keydown(event)"'
        + ' onmousemove="Notation.on_mouse_move(event)"'
        + '></div>',
        '', false);
    this.local_priority = 1000;
    this.popup_close = popup_close;
    this.notation_student = document.getElementById("notation_student");
    this.notation_grade = document.getElementById("notation_grade");
    this.notation_content = document.getElementById("notation_content");
    this.notation_error = document.getElementById("notation_error");
    this.notation_other = document.getElementById("notation_other");
    popup_close = this.close.bind(this);
    this.select_current_line();
    this.update_popup();
    this.focus(this.question_list()[0], 2 /* XXX BAD */);
};

Notation.prototype.init = function (column) {
    this.column = column;
    this.column_modifiable = column_change_allowed(column);
    this.parse_questions(column.comment);
    this.scrolltop = 0;
};

Notation.prototype.parse_questions = function (txt) {
    this.log("parse questions");
    var questions;
    try {
        questions = JSON.parse(txt);
    } catch (e) {
        questions = [];
    }
    this.questions = {};
    for (var i in questions) {
        var question = new NotationQuestion(questions[i]);
        question.priority = Number(i);
        question.notation = this;
        this.questions[question.id] = question;
    }
    this.nr_empty_questions_added = 0;
};

Notation.prototype.save_current_line = function () {
    this.log("save_current_line " + (this.line ? this.line[0].value : "?"));
    this.save_questions();
    this.save_grades();
};

Notation.prototype.clear_current_line = function () {
    this.log("clear_current_line");
    this.save_current_line();
    the_current_cell.tr.className =
        the_current_cell.tr.className.replace(/ *currentformline/, '');
    this.scrolltop = document.getElementById("the_questions").scrollTop;
};

Notation.prototype.parse_grades = function (txt) {
    var grades;
    try { grades = JSON.parse(txt); }
    catch (e) { grades = {}; }
    var questions = this.questions;
    for (var i in questions) {
        var question = questions[i];
        if (question.is_a_comment())
            continue;
        question.set_grade(grades[i]);
        question.grade.uncompress_comment();
    }
};

Notation.prototype.get_the_commented_line = function (line) {
    if (this.column.groupcolumn === '')
        return line;

    var col = data_col_from_col_title(this.column.groupcolumn);
    if (!col)
        return line;

    var group = line[col].value.toString();
    if (group === '')
        return line;
    var group_lines = [];
    for (var line_key in lines) {
        if (lines[line_key][col].value.toString() == group
            && lines[line_key][this.column.data_col].comment !== '')
            group_lines.push(lines[line_key]);
    }
    if (group_lines.length > 1) {
        var message = _("ALERT_notation_conflict") + '\n';
        for (var i in group_lines)
            message += group_lines[i][0].value + ' ' + group_lines[i][1].value
                + ' ' + group_lines[i][2].value + '\n';
        alert(message);
        return group_lines[0];
    }
    return group_lines.length ? group_lines[0] : line;
};

Notation.prototype.init_cell = function (line) {
    this.line = this.get_the_commented_line(line);
    this.cell = this.line[this.column.data_col];
    this.parse_grades(this.cell.comment);
    this.modifiable = this.cell
        && this.cell.changeable(this.line, this.column) === true;
};

Notation.prototype.select_current_line = function () {
    this.log("select_current_line " + the_current_cell.line[0].value);
    the_current_cell.tr.className += ' currentformline';
    this.init_cell(the_current_cell.line);
    this.notation_error.innerHTML = "";
    this.notation_student.innerHTML = html(this.line[0].value)
        + ' ' + html(this.line[1].value) + ' ' + html(this.line[2].value);
    if (this.line != the_current_cell.line)
        this.notation_other.innerHTML = '→ '
            + html(the_current_cell.line[1].value) + ' '
            + html(the_current_cell.line[2].value) + ' ';
    else
        this.notation_other.innerHTML = '';
};

Notation.prototype.question_list = function (with_deleted) {
    var questions = [];
    for (var i in this.questions)
        if (this.questions[i].is_a_question_or_bonus()
            || with_deleted
        )
            questions.push(this.questions[i]);
    this.sort_questions(questions);
    return questions;
};

Notation.prototype.jump = function (lin, col, do_not_focus, line_id, data_col) {
    this.log("jump");
    this.clear_current_line();
    this.jump_old(lin, col, do_not_focus, line_id, data_col);
    this.select_current_line();
    this.update_popup();
    if (!this.jump_from_notation)
        this.stop_event = false;
};

Notation.prototype.update_all_grades = function () {
    this.log("update_all_grades");
    var cell, v, answer;
    var save = this.column.min;
    this.column.min = -1e9;
    for (var line_id in lines) {
        cell = lines[line_id][this.column.data_col];
        if (cell.comment === '')
            continue;
        this.parse_grades(cell.comment);
        if (!this.fully_graded())
            continue;
        v = this.get_grade() + this.get_bonus();
        if (answer === undefined && v < save)
            answer = confirm(_('ALERT_record_negative'));
        if (v !== cell.value) {
            if(v < save && ! answer )
                v = save;
            cell_set_value_real(line_id, this.column.data_col, v.toString());
            update_line(line_id, this.column.data_col);
        }
    }
    this.column.min = save;
};

Notation.prototype.close = function () {
    this.log("close");
    if (this.cancel_save)
        this.cancel_save = false;
    else
        this.clear_current_line();

    the_current_cell.jump = this.jump_old;
    popup_close = this.popup_close;
    popup_close();
    this.update_all_grades();
    table_fill(true, true, true, true);
    this.notation_window_open = false;
};

Notation.prototype.merge_question_changes = function () {
    var current = this.questions;
    var current_sorted = this.question_list();
    this.parse_questions(this.column.comment);
    var remotes = this.question_list(true);
    this.questions = current;
    this.log("Merge before: " + JSON.stringify(this.questions));
    this.log("Merge remote: " + JSON.stringify(remotes));
    current = this.question_list(true);
    var done = {};
    for (var i in remotes) {
        var remote = remotes[i];
        var local = this.questions[remote.id];
        done[remote.id] = true;
        if (!local) {
            this.log('New question from somebody: ' + remote.id);
            this.questions[remote.id] = remote;
        }
        else if (local.initial_value == local.hash()) {
            this.log('Not changed by local user: ' + remote.id);
            remote.grade = this.questions[remote.id].grade; // Keep grades
            this.questions[remote.id] = remote;
            this.questions[remote.id].priority = local.priority;
        }
        else {
            this.log('Changed by local user: ' + remote.id);
            this.log('initial_hash: ' + local.initial_value);
            this.log('current hash: ' + local.hash());
            this.questions[remote.id].priority = local.priority;
        }
    }
    // Only local questions
    for (var i in current_sorted)
        if (!done[current_sorted[i].id]
            && !this.is_the_last(current_sorted[i]))
            current_sorted[i].priority = this.local_priority++;

    this.log("Merge after: " + JSON.stringify(this.questions));
};

Notation.prototype.save_questions = function () {
    if (!this.column_modifiable)
        return;
    this.log("save questions");
    this.merge_question_changes();
    var questions = [];
    for (var question in this.questions) {
        question = this.questions[question];
        if (!this.is_the_last(question)) {
            questions.push(question);
            question.initial_value = question.hash();
        }
    }
    this.sort_questions(questions);
    var v = JSON.stringify(questions);
    this.log("save questions old: " + this.column.comment);
    this.log("save questions new: " + v);
    if (v != this.column.comment) {
        column_attr_set(this.column, 'comment', v);
        column_attr_set(this.column, 'minmax', '[0;' + this.maximum() + ']');
        the_current_cell.do_update_column_headers = true;
        the_current_cell.update_column_headers();
    }
};

Notation.prototype.is_the_last = function (question) {
    if (question.question !== '')
        return false;
    var questions = this.question_list();
    return question == questions[questions.length - 1];
};


Notation.prototype.merge_grade_changes = function () {
    var grades;
    try { grades = JSON.parse(this.cell.comment); }
    catch (e) { return; }
    this.log("Remote: " + this.cell.comment);
    for (var grade in grades) {
        var question = this.questions[grade];
        if (!question)
            continue;
        if (question.grade.not_graded || !question.grade.local_change) {
            this.log("Merge remote grade: " + grade);
            question.new_grade(grades[grade]);
        }
        else
            this.log("Keep local grade: " + grade);
    }
    for (var i in this.questions) {
        var question = this.questions[i];
        if (!question.grade.local_change && grades[question.id] === undefined) {
            this.log("Remote grade removed: " + question.id);
            if (question.is_a_question())
                question.grade.set_grade('?');
            else
                question.grade.set_grade(0);
        }
    }
};

Notation.prototype.get_json_grades = function () {
    var grades = {};
    for (var i in this.questions) {
        var question = this.questions[i];
        if (question.grade.not_graded && question.grade.comment === '')
            continue;
        if (question.is_a_comment() && question.grade.grade === 0)
            continue;
        grades[i] = question.grade;
    }
    return JSON.stringify(grades);
};

Notation.prototype.save_grades = function () {
    if (!this.modifiable)
        return;
    this.log("save student");
    this.merge_grade_changes();
    var v = this.get_json_grades();
    if (v == '{}')
        v = '';
    this.log("Old: " + this.cell.comment);
    this.log("New: " + v);
    if (v != this.cell.comment) {
        if (this.column.modifiable == 2) {
            alert(_("MSG_annotate_not_allowed") + '\n' + _("ALERT_column_not_saved"));
            return;
        }
        this.log("Store into " + this.line.line_id + ' ' + this.column.data_col);
        comment_change(this.line.line_id, this.column.data_col, v);
        for (var i in this.questions)
            this.questions[i].grade.local_change = false;
    }
};

Notation.prototype.contain_empty_question = function () {
    this.log("contain_empty_question");
    for (var i in this.questions)
        if (this.questions[i].is_an_empty_question())
            return this.questions[i];
    return false;
};

Notation.prototype.add_empty_question_if_needed = function (txt, max) {
    this.log("add_empty_question_if_needed");
    if (!this.column_modifiable)
        return;
    var empty = this.contain_empty_question();
    if (empty)
        return empty;
    var id = this.unused_id();
    this.questions[id] = new NotationQuestion(
        { id: id, priority: 9999 + this.nr_empty_questions_added / 1000 });
    this.questions[id].notation = this;
    this.nr_empty_questions_added++;
    if (txt)
        this.questions[id].question = txt;
    switch (max) {
        case _("MSG_Bonus").toLowerCase():
            this.questions[id].type = 2;
            this.questions[id].max = 5;
            break;
        case undefined:
            break;
        default:
            this.questions[id].max = Number(max.replace('±', ''));
            if (max.substr(0, 1) == '±')
                this.questions[id].type = 2;
            break;
    }

    if (this.notation_error) {
        this.notation_error.innerHTML = '';
        this.update_error();
    }

    return this.questions[id];
};

Notation.prototype.maximum = function () {
    var sum = 0;
    for (var i in this.questions)
        if (this.questions[i].is_not_an_empty_question())
            sum += this.questions[i].max;
    return trunc(sum);
};

Notation.prototype.get_grade = function () {
    var sum = 0;
    for (var i in this.questions)
        if (this.questions[i].is_not_an_empty_question()
            && this.questions[i].grade.grade !== '?'
        )
            sum += this.questions[i].grade.grade * this.questions[i].max;
    return trunc(sum);
};

Notation.prototype.fully_graded = function () {
    for (var i in this.questions)
        if (this.questions[i].is_not_an_empty_question()
            && this.questions[i].grade.not_graded)
            return false;
    return true;
};

Notation.prototype.get_bonus = function () {
    var sum = 0;
    for (var i in this.questions)
        if (this.questions[i].is_not_an_empty_bonus(i)
            && this.questions[i].grade.stored != '?')
            sum += this.questions[i].grade.stored;
    return trunc(sum);
};

Notation.prototype.focus = function (question, what) {
    if (!question)
        return;
    this.log("focus " + question.id + ' what=' + what);
    var e = document.getElementById(question.id);
    if (!e || !e.childNodes)
        return;
    if (!this.modifiable)
        what = 2; // XXX BAD
    if (what != -1)
        e = e.childNodes[what];
    e.focus();
};

Notation.prototype.is_an_empty_value = function (value) {
    return value === '' || value == _("MSG_notation_question")
        || value == _("MSG_notation_comment");
};

Notation.prototype.is_active = function (value) {
    return !!document.getElementById("the_questions");
};

Notation.prototype.queue_update_completion = function (event) {
    var t = this;
    setTimeout(function () { t.update_completions(event); }, 100);
};

Notation.prototype.on_focus = function (event) {
    if (!this.is_active())
        return; // closed
    this.log("on_focus");
    event = this.get_event(event);
    if (this.old_focused)
        this.old_focused.className = this.old_focused.className.replace(/ *focus/,
            '');
    this.old_focused = event.line;
    event.line.className += ' focus';
    if (this.is_an_empty_value(event.target.value)) {
        event.target.value = '';
        event.target.className = event.target.className.replace(/ *empty/, "");
        return;
    }
    if (event.target.tagName == 'TEXTAREA' && millisec() - this.keyboard_focus < 100) {
        var i = myindex(event.target.value, '/')
        if (i != -1)
            set_selection(event.target, 0, i);
        // Not now because the cursor position is not yet correct.
        this.queue_update_completion(event);
    }
};

Notation.prototype.update_title = function () {
    this.log("update title");
    var s = this.get_grade() + '/' + this.maximum();
    var b = this.get_bonus();
    if (b)
        s += ' → ' + trunc(this.get_grade() + this.get_bonus()) + 'Ⓑ/' + this.maximum();
    this.notation_grade.innerHTML = s;
};

Notation.prototype.sort_questions = function (questions) {
    // JS dictionary order is random
    questions.sort(function (a, b) { return a.priority - b.priority; });
};

function get_scroll_size(e) {
    try { return e.lastChild.offsetTop + e.lastChild.offsetHeight; }
    catch (e) { return 1; }
}

Notation.prototype.update_popup = function () {
    this.log("update");
    var i;
    this.add_empty_question_if_needed();
    this.compute_stats();
    this.update_title();
    var questions = [];
    for (i in this.questions)
        if (this.questions[i].is_a_question_or_bonus())
            questions.push(this.questions[i]);
    this.sort_questions(questions);
    var s = [];
    s.push('<div id="the_questions" class="the_questions">');
    for (i in questions)
        s.push(questions[i].html(this.modifiable, this.column_modifiable));
    s.push('<div class="notation_introduction">');
    s.push(_("MSG_notation_introduction"));
    s.push('</div>');
    s.push('</div>');
    s.push('<div class="the_completions" id="the_completions"></div>');

    questions = [];
    for (i in this.questions)
        if (this.questions[i].is_a_comment())
            questions.push(this.questions[i]);
    questions.sort(function (a, b) {
        return a.question < b.question
            ? -1
            : (a.question > b.question ? 1 : 0);
    });
    this.notation_content.innerHTML = s.join('');
    document.getElementById("the_questions").scrollTo(0, this.scrolltop);
    this.the_completions = document.getElementById("the_completions");
    for (i in this.questions)
        if (this.questions[i].is_a_question_or_bonus())
            if (this.questions[i].is_a_question_or_bonus())
                this.questions[i].draw_canvas();
};

Notation.prototype.unused_id = function () {
    this.log("unused_id");
    var used_ids = {};
    for (var i in this.questions)
        used_ids[this.questions[i].id] = true;
    for (var i = 0; ; i++)
        if (!used_ids[page_id + '_' + i])
            return page_id + '_' + i;
};

Notation.prototype.get_event = function (event) {
    last_user_interaction = millisec();
    event = the_event(event);
    event.line = event.target;
    while (!event.line.id)
        event.line = event.line.parentNode;
    event.question = this.questions[event.line.id];
    event.what = '';
    if (event.question && event.line.childNodes.length >= 5
        && event.line.childNodes[2].tagName == 'CANVAS'
    ) // XXX BAD
    {
        // On a question line
        event.question_input = event.line.childNodes[0]; // XXX BAD
        event.canvas = event.line.childNodes[2]; // XXX BAD
        event.comment_input = event.line.childNodes[4]; // XXX BAD
        event.child_nr = myindex(event.line.childNodes, event.target);
        switch (event.target) {
            case event.question_input: event.what = 'question'; break;
            case event.canvas: event.what = 'canvas'; break;
            case event.comment_input: event.what = 'comment'; break;
        }
    }
    else {
        event.child_nr = -1;
        if (event.target.className == 'stat')
            event.target = event.target.parentNode;
    }
    if (event.what === '')
        event.what = event.target.className.split(' ')[0];
    return event;
};

Notation.prototype.on_mouse_move = function (event, force) {
    event = this.get_event(event);
    if (!force && event.what != 'canvas')
        return;
    if (!event.question)
        return;
    if (this.button_pressed_on != event.question.id)
        return;
    if (!this.modifiable)
        return;

    if (event.what == 'canvas' && event.question.question !== "") {
        // -4 because of borders
        var x = (event.x - findPosX(event.target) - 4) / event.target.offsetWidth;
        x = x * (event.question.x_max() - event.question.x_min()) + event.question.x_min();
        x = event.question.floor(x);
        event.question.set_grade_to(x);
        if (event.comment_input)
            event.comment_input.value = event.question.grade_and_comment();
    }
    event.question.draw_canvas();
    this.update_title();
};

Notation.prototype.move = function (question, direction) {
    var questions = this.question_list();
    var i = myindex(questions, question);
    if (direction == 'move_up') {
        if (i == 0)
            return;
        if (i == 1)
            question.priority = questions[i - 1].priority - 1;
        else
            question.priority = (questions[i - 2].priority + questions[i - 1].priority) / 2;
    }
    else {
        if (i == questions.length - 1)
            return;
        if (i == questions.length - 2)
            question.priority = questions[i + 1].priority + 1;
        else
            question.priority = (questions[i + 1].priority + questions[i + 2].priority) / 2;
    }
    this.update_popup();
    var t = this;
    setTimeout(function () { t.focus(question, 0); }, 200);
};

Notation.prototype.insert_completion = function (completion, event) {
    event.target = this.old_focused.childNodes[4]; // XXX BAD
    event.question = this.questions[this.old_focused.id];
    event.comment_input = event.target;
    event.line = this.old_focused;
    completion = completion.replace(/^<[^ ]* /, "");
    event.target.value = this.update_completions(event, completion);
    event.question.grade.comment = event.target.value;
    event.question.grade.local_change = true;
    event.target.value = event.question.grade_and_comment();
    var cursor = event.target.value.indexOf(completion) + completion.length + 1;
    function do_focus() {
        event.target.focus();
        set_selection(event.target, cursor, cursor);
    }
    setTimeout(do_focus, 100);
};

Notation.prototype.on_mouse_down = function (event) {
    this.log("on_mouse_down");
    event = this.get_event(event);
    this.stop_event = true;

    GUI.add("notation_mouse_down", event, event.what);

    if (!event.question) {
        if (event.target.className == 'a_completion') {
            if (this.modifiable)
                this.insert_completion(event.target.textContent.replace(/[^ ]* /, ''), event);
            else
                alert(_("ERROR_value_defined_by_another_user"));
        }
        return;
    }
    this.button_pressed_on = event.question.id;
    if (event.what == 'canvas') {
        event.target.focus();
        this.queue_update_completion(event);
    }
    else {
        if (!this.column_modifiable)
            return;
        switch (event.what) {
            case 'inc':
                event.question.steps = Math.min(event.question.steps + 1, 10);
                event.question.draw_canvas();
                break;
            case 'dec':
                event.question.steps = Math.max(event.question.steps - 1, 1);
                event.question.draw_canvas();
                break;
            case 'bonus':
                if (event.question.is_a_question()) {
                    event.question.type = 2;
                    if (event.question.grade.stored == '?')
                        event.question.grade.stored = 0;
                }
                else if (event.question.is_a_bonus())
                    event.question.type = 0;
                event.line.className = event.line.className.replace(
                    /type[0-9]/, 'type' + event.question.type);
                event.question.set_min();
                break;
            case 'comment':
                this.queue_update_completion(event);
                break;
            case 'move_up':
            case 'move_down':
                this.move(event.question, event.what);
                break;
            case 'edit_comment':
                var v;
                if (event.question.somebody_is_graded())
                    v = prompt(_("MSG_notation_fix_comment"), event.question.question);
                else
                    v = prompt(_("MSG_notation_edit_comment"), event.question.question);
                if (v !== null) {
                    this.save_grades(); // Compress with the old value
                    event.question.question = v;
                    if (v === '' && !event.question.somebody_is_graded())
                        event.question.type = 1;
                    this.init_cell(the_current_cell.line);
                    this.update_popup();
                }
                break;
        }
    }
    if (this.modifiable)
        this.on_mouse_move(event, true);
};

Notation.prototype.on_mouse_up = function (_event) {
    this.log("on_mouse_up");
    this.button_pressed_on = false;
};

Notation.prototype.on_paste = function (event) {
    this.log("on_paste");
    event = this.get_event(event);
    GUI.add("notation_paste", event, event.what);
    var me = this;
    if (event.what == 'comment')
        setTimeout(function () { me.on_comment_change(event); }, 100);
    else if (event.what == 'question')
        setTimeout(function () { me.on_question_change(event); }, 100);
};

Notation.prototype.compute_stats = function () {
    this.global_comments = {};
    this.global_comments_id = {};
    for (var question_id in this.questions) {
        question = this.questions[question_id];
        question.stats = new Stats();
        if (question.is_a_comment()) {
            this.global_comments[question.question] = 0;
            this.global_comments_id[question.question] = question_id;
        }
        question.comments = {};
    }
    this.max_graded = 0;
    for (var line_id in lines) {
        try {
            g = JSON.parse(lines[line_id][this.column.data_col].comment);
        }
        catch (e) { continue; }
        var a_grade = false;
        for (var id in g) {
            var grade = new NotationGrade(g[id]);
            var question = this.questions[id];
            if (question === undefined) {
                console.log(lines[line_id][0].value + ' ' + id + ':' + g[id]);
                continue;
            }
            grade.question = question;
            grade.uncompress_comment();
            if (question.is_a_question() || question.is_a_bonus()) {
                if (grade.stored != '?') {
                    question.stats.add(grade.stored / grade.max);
                    a_grade = true;
                }
                var comments = grade.comment.split(/\n+/);
                question.comment_size = Math.max(question.comment_size,
                    Math.min(10, comments.length));
                for (var i = 0; i < comments.length; i++) {
                    var comment = comments[i];
                    if (comment === '')
                        continue;
                    if (comment.indexOf('{{{') != -1)
                        continue;
                    if (this.global_comments[comment] === undefined)
                        this.global_comments[comment] = 0;
                    this.global_comments[comment]++;

                    if (question.comments[comment] === undefined)
                        question.comments[comment] = [];

                    if (grade.stored != '?') {
                        var gg = grade.stored / grade.max;
                        question.comments[comment].push(gg);
                        if (this.global_comments_id[comment])
                            this.questions[this.global_comments_id[comment]].stats.add(gg);
                    }
                }
            }
            else if (question.is_a_comment()) {
                question.stats.add(1);
                this.global_comments[question.question]++;
            }
        }
        if (a_grade)
            this.max_graded++;
    }
    this.global_comments_sorted = [];
    for (var i in this.global_comments)
        this.global_comments_sorted.push(i);
    this.global_comments_sorted.sort();
    for (var question in this.questions) {
        var sorted_comments = [];
        for (var comment in this.questions[question].comments)
            sorted_comments.push(comment);
        sorted_comments.sort();
        this.questions[question].sorted_comments = sorted_comments;
    }
    for (var question in this.questions)
        this.questions[question].max_graded = this.max_graded;
    // Create new global comments if allowed
    if (!this.column_modifiable)
        return;
    var empty_question = this.add_empty_question_if_needed();
    var created_comments = {};
    for (var i in this.global_comments)
        if (this.global_comments[i] > 1 && !this.global_comments_id[i]
            && i.length > 10 && i.indexOf('{{{') == -1) {
            empty_question.type = 3;
            empty_question.grade.stored = 0;
            empty_question.question = i;
            empty_question = this.add_empty_question_if_needed();
            created_comments[i] = empty_question.id;
        }
    // Change existing comments.
    // Do not use all the framework in order to not modify the current state
    // XXX So it is really not clean...
    if (Object.keys(created_comments).length) {
        var cell, grade, text, new_grades;
        for (var line_id in lines) {
            if (lines[line_id] == this.line)
                continue;
            cell = lines[line_id][this.column];
            if (!this.cell.changeable(cell, this.column))
                continue;
            try { grades = JSON.parse(cell.comment); }
            catch (e) { continue; }
            for (var question_id in grades) {
                grade = grades[question_id].replace(/ */, '');
                text = grades[question_id].substr(value.length);
                if (text !== '') {
                    for (var comment in created_comments)
                        text = text.replace(comment, '{{{' + created_comments[comment] + '}}}');
                    grades[question_id] = (grade + ' ' + text).trim();
                }
            }
            new_grades = JSON.stringify(grades);
            if (new_grades != cell.comment)
                comment_change(line_id, this.column, new_grades);
        }
    }
};

Notation.prototype.update_error = function (error, element) {
    if (error) {
        this.notation_error.style.color = '#F00';
        if (element)
            element.style.color = this.notation_error.style.color;
        this.notation_error.style.fontSize = "100%";
        this.notation_error.innerHTML = _(error);
    }
    else {
        if (element)
            element.style.color = '#000';
        this.notation_error.innerHTML = '';
    }
};

Notation.prototype.on_comment_change = function (event) {
    this.log("on_comment_change " + event.target.value);
    var error = event.question.set_comment(event.target.value,
        this.column_modifiable);
    this.update_error(error, event.target);
    event.question.draw_canvas();
    this.update_title();
    var completion = this.update_completions(event);
    if (event.keyCode != 8 && completion) {
        do_autocompletion(event.target, event.target.value + completion);
        event.question.set_comment(event.target.value, this.column_modifiable);
    }
};

Notation.prototype.toggle_local_completion = function () {
    this.hide_local_completion = !this.hide_local_completion;
    this.update_completions(this.last_event, this.last_replacement);
};

Notation.prototype.update_completions = function (event, replacement) {
    // Cherche la ligne du commentaire à modifier
    var selection = get_selection(event.comment_input);
    var lines = event.comment_input.value.split(/[\r\n]/);
    var current;
    this.last_event = event;
    this.last_replacement = replacement;
    for (var line = 0, length = 0; ; length += lines[line].length + 1, line++)
        if (selection.start < length)
            break;
    line -= 1;
    if (line == 0)
        current = event.question.grade.comment.split('\n')[0];
    else
        current = lines[line];
    if (replacement) {
        if (line !== 0)
            lines[0] = event.question.grade.comment.split('\n')[0];
        lines[line] = replacement + '\n';
        return '\n'.join(lines);
    }
    current = current.toLowerCase();
    var completions = ['<p>', _('MSG_notation_here'),
        ' <span onclick="Notation.toggle_local_completion()" style="font-style:emoji">',
        this.hide_local_completion ? '⊕' : '⊖',
        '</span></p>'];
    var done = {};
    var nb = 0;
    var shortcuts = {};
    this.shortcuts = shortcuts;
    function shortcut() {
        var k = "0123456789BDEFGHIJKLMOPRSU".substr(nb++, 1);
        if (k !== '') {
            shortcuts[k] = comment;
            k = hidden_txt('<b class="notation_shortcut">' + k + '</b>', 'Ctrl+' + k);
        }
        return k;
    }
    if (!this.hide_local_completion)
        for (var comment in event.question.sorted_comments) {
            comment = event.question.sorted_comments[comment];
            if (comment.substr(0, current.length).toLowerCase() == current) {
                var grades = event.question.comments[comment];
                grades.sort(function (a, b) { return a - b; });
                var m = event.question.grade.max;
                var min = event.question.nice_grade(m * grades[0]);
                var max = event.question.nice_grade(m * grades[grades.length - 1]);
                var g;
                if (min != max)
                    g = min + ';' + max;
                else
                    g = min;
                g = g.toString();
                if (g == 'NaN')
                    g = '';
                completions.push('<div class="a_completion">' + shortcut()
                    + '<span class="stat">['
                    + g + ']</span>'
                    + (this.column_modifiable && this.global_comments_id[comment]
                        ? '<span class="edit_comment" id="'
                        + this.global_comments_id[comment] + '">✍</span>'
                        : '')
                    + ' ' + html(comment) + '</div>');
                done[comment] = 1;
            }
        }
    completions.push('<p>' + _('MSG_notation_other') + '</p>');

    for (var comment in this.global_comments_sorted) {
        comment = this.global_comments_sorted[comment];
        if (comment.substr(0, current.length).toLowerCase() == current)
            completions.push('<div class="a_completion">' + shortcut()
                + '<span class="stat">'
                + this.global_comments[comment]
                + '</span>'
                + (this.column_modifiable && this.global_comments_id[comment]
                    ? '<span class="edit_comment" id="'
                    + this.global_comments_id[comment] + '">✍</span>'
                    : '')
                + ' ' + html(comment) + '</div>');
    }
    this.the_completions.innerHTML = completions.join("");
    if (completions.length == 1)
        return completions[0].split("</span> ")[1]
            .split("<")[0]
            .substr(current.length);
};

Notation.prototype.on_question_change = function (event) {
    this.log("on_question_change " + event.target.value);
    event.question.question = event.target.value;
    if (this.is_the_last(event.question) && event.question.question !== '')
        event.question.priority = this.local_priority++;
};

Notation.prototype.on_keydown = function (event) {
    event = this.get_event(event);
    if (event.keyCode == 27 // Escape
        && !confirm(_("ALERT_notation_save")))
        this.cancel_save = true;

    if (event.ctrlKey) {
        if (event.keyCode == 38) {
            this.move(event.question, "move_up");
            stop_event(event);
            return;
        }
        else if (event.keyCode == 40) {
            this.move(event.question, "move_down");
            stop_event(event);
            return;
        }
        else {
            var completion = this.shortcuts[String.fromCharCode(event.keyCode)];
            if (completion) {
                this.insert_completion(completion, event);
                stop_event(event);
                return;
            }
        }
    }

    if (event.keyCode == 13
        || event.keyCode == 38
        || event.keyCode == 40
        || event.keyCode == 33
        || event.keyCode == 34
        || ((event.keyCode == 37 || event.keyCode == 39)
            && event.what == 'canvas')
    ) {
        if (event.target.tagName == 'TEXTAREA') {
            this.keydown_selection = get_selection(event.target);
            if (event.keyCode == 13 && this.column_modifiable && !this.contain_empty_question()) {
                this.update_popup();
                var t = this;
                function focus_on_last() {
                    var questions = t.question_list();
                    t.focus(questions[questions.length - 1], 0); // XXX BAD
                };
                setTimeout(focus_on_last, 100);
            }
            return;
        }
        stop_event(event);
    }
};

Notation.prototype.on_keyup = function (event) {
    event = this.get_event(event);
    if (event.question === undefined)
        return;
    if (event.ctrlKey)
        return;
    this.log("keyup what=" + event.what + ' question=' + event.question.id
        + " keycode=" + event.keyCode);
    var questions = this.question_list();
    var question_index = myindex(questions, event.question);
    if (event.keyCode <= 40 && event.keyCode != 37 && event.keyCode != 39)
        GUI.add_key(event, "notation_key");

    this.keyboard_focus = millisec();
    switch (event.keyCode) {
        case 38: // Cursor up
        case 40: // Cursor down
        case 13: // Return
            if (event.what == 'comment') {
                if (event.keyCode == 38 && this.keydown_selection.end == 0
                    || event.keyCode == 40 && this.keydown_selection.start == event.target.value.length
                    || event.keyCode != 13 && this.keydown_selection.start == 0 && this.keydown_selection.end
                ) {
                }
                else {
                    this.update_completions(event);
                    break;
                }
            }
            if (event.what == 'question') {
                if (!this.contain_empty_question())
                    this.update_popup();
            }
            questions = this.question_list();
            this.focus(questions[question_index + (event.keyCode == 38 ? -1 : 1)],
                event.child_nr);
            stop_event(event);
            break;
        case 37: // Cursor left
        case 39: // Cursor right
            if (event.what == 'canvas' && event.question.question !== "") {
                if (!this.modifiable)
                    break;
                var x = event.question.grade.grade;
                if (x === '?')
                    x = (event.keyCode == 37 ? event.question.max : 0);
                else
                    x += (event.keyCode == 37 ? -1 : 1)
                        * 1 / (event.question.max * event.question.steps);
                event.question.set_grade_to(x * event.question.max);
                event.comment_input.value = event.question.grade_and_comment();
                event.question.draw_canvas();
                this.update_title();
            }
            break;
        case 34: // Next page
        case 33: // Previous page
            this.jump_from_notation = true;
            if (event.keyCode == 34)
                the_current_cell.cursor_down();
            else
                the_current_cell.cursor_up();
            this.jump_from_notation = false;
            setTimeout(function () {
                Notation.focus(event.question, event.child_nr);
            }, 100);
            break;
        default:
            this.log("keyup default " + event.what);
            if (event.target.value == event.target.old_value)
                break;
            event.target.old_value = event.target.value;
            if (event.what == 'comment')
                this.on_comment_change(event);
            else if (event.what == 'question')
                this.on_question_change(event);
    }

    if (event.what == 'question' && event.target.value === ''
        && !this.is_the_last(event.question)
    ) {
        if (confirm(_("ALERT_delete_line"))) {
            event.question.type = 1;
            this.update_popup();
        }
    }
};

var notation_suivi_first_time = true;

Notation.prototype.init_style = function () {
    if (notation_suivi_first_time) {
        notation_suivi_first_time = false;
        var style = document.createElement('STYLE');
        style.textContent =
            ".DisplayTypeNotation .DisplayCellValue .tipped .text { display: block }"
            + "TABLE.notation_suivi { border-spacing: 0px }"
            + "TABLE.notation_suivi TD"
            + "      { border: 1px solid #888 ; font-size:120%; vertical-align:top ; }"
            + "DIV#tip.notation_suivi_tip { margin-top: -3em }"
            + 'DIV.notation_suivi_modifiable { right: 0px; top:0px;  bottom: 0px; left: 0px; position: fixed;'
            + '   border: 2px solid black ; overflow:hidden }'
            + '.notation_top > DIV {'
            + '  height: calc(100% - 4em); overflow-x: hidden; margin-bottom: 0.5em;}'
            + '.notation_top { height: 100% ; }'
            + ".notation_top TABLE { border-spacing: 0px 1em; width: calc(100% - 1em) }"
            + "DIV.import_export .notation_top TEXTAREA { margin-bottom: 0.5em; width: calc(100% - 1em) }"
            + '.notation_top TD { vertical-align: top ;}'
            + '.notation_top TD:first-child { width: 1%; padding-right: 1em; white-space: nowrap; opacity: 0.5 }'
            + '.notation_footer { text-align: right }'
            + '.Textual .notation_footer { text-align: left }'
            + '.notation_record { background: #8F8 ; }'
            + '.notation_empty { background: #CCC }'
            + '.notation_feedback, .notation_stats { display: inline-block ; }'
            + '.notation_stats { display: inline-block ; }'
            + '.notation_feedback DIV.feedback { display: inline-block ; }'
            ;
        document.getElementsByTagName('HEAD')[0].appendChild(style);
    }
};

Notation.prototype.suivi = function () {
    var questions = this.question_list();
    this.init_style();
    s = ['<div style="overflow:auto;max-height:20em">', // MUST BE INLINE STYLE !
        '<table class="notation_suivi">'];
    for (var question in questions)
        s.push(questions[question].suivi());
    s.push("</table>");
    var c = [];
    for (var question in this.questions) {
        question = this.questions[question];
        if (question.is_a_comment() && question.grade.stored)
            c.push(html(question.question));
    }
    return s.join("") + c.join('<br>') + '</div>';
};

Notation.prototype.suivi_modifiable = function (key) {
    var questions = this.question_list();
    this.init_style();
    this.key = key;
    s = ['<div class="notation_top" info="', key, '"><div><table>'];
    for (var question in questions)
        s.push(questions[question].suivi_modifiable());
    s.push("</table></div>");
    s.push('<div class="notation_footer">');
    s.push('<div class="notation_feedback"></div>');
    s.push('<button id="notation_record" class="notation_record" onclick="notation_record(this)">' + _("LABEL_MCQ_save") + '</button>');
    s.push(' <div class="notation_stats"></div>');
    s.push('</div></div>');
    return s.join('');
};

Notation.prototype.get_question_from_question = function (question) {
    question = question.toLowerCase().replace(/ /g, ' ').replace(/ +/g, ' ');
    for (var j in this.questions)
        if (this.questions[j].question.toLowerCase().replace(/ /g, ' ').replace(/ +/g, ' ') == question)
            return this.questions[j];
};

Notation.prototype.import_do = function (csv) {
    if (csv.length < 2)
        return "???";
    while (csv[0] === '')
        csv.splice(0, 1);
    var question_texts = csv[0].split(/\t/);
    var questions = [];
    var column = popup_column();
    var messages = "";
    this.nr_empty_questions_added = 0;

    this.init(column)
    for (var i = 3; i < question_texts.length; i += 2) {
        var question = question_texts[i + 1];
        var max = question_texts[i].toLowerCase();
        if (question === undefined)
            return _("ALERT_notation_missing_column");
        var q = this.get_question_from_question(question);
        if (q) {
            if (max == _("MSG_Bonus").toLowerCase())
                max = q.max;
            else
                max = Number(max.replace('±', ''));
            if (isNaN(max))
                return q.question + "\n" + max + ' isNaN';
            if (q.max != max)
                return q.question + "\n" + q.max + ' != ' + max;
        }
        else {
            q = this.add_empty_question_if_needed(question, max);
            messages += "\n/" + max + ' ' + question;
        }
        questions.push(q);
    }
    if (messages !== "")
        if (!confirm(_("MSG_notation_confirm_import") + '\n' + messages))
            return _("MSG_tablelinear_cancel");
    csv.splice(0, 2);
    var g;
    for (var i in csv) {
        var grades = csv[i].split(/ *\t */);
        if (grades[0] === '' || !grades[0])
            continue;
        var lin_id = login_to_line_id(grades[0]);
        if (!lin_id)
            return grades[0] + ' ???';
        this.init_cell(lines[lin_id]);
        for (var j in questions) {
            g = grades[3 + 2 * j];
            if (g === '' || g === '???')
                continue;
            if (g != '?') {
                if (isNaN(g))
                    return questions[j].question + ": " + g + "=NaN";
                if (g > questions[j].max)
                    return questions[j].question + ": " + g + ">" + questions[j].max;
            }
            questions[j].set_comment(g + '/' + questions[j].max
                + ' ' + decode_lf_tab(grades[4 + 2 * j]), true);
        }
        this.save_grades();
    }
    this.save_questions(); // Must be after
    this.update_all_grades();
    table_fill(true, true, true, true);
}

Notation = new Notation(); // Only one instance

function notation_open(value, _column) {
    Notation.start();
    return value;
}

function notation_get_key(element) {
    while (element) {
        if (element.tagName == 'DIV' && element.getAttribute('info')) {
            Notation.element_top = element;
            Notation.element_footer = element.lastChild;
            Notation.element_feedback = Notation.element_footer.firstChild;
            Notation.element_record = Notation.element_footer.getElementsByTagName('BUTTON')[0];
            Notation.element_stats = Notation.element_footer.lastChild;
            return element.getAttribute('info');
        }
        element = element.parentNode;
    }
}

function notation_init_state(key, use_textarea) {
    if (!key)
        key = year + '/' + semester + '/' + DisplayGrades.ue.ue
            + '/' + DisplayGrades.column.the_id
            + '/' + DisplayGrades.ue.line_id;

    var info = key.split('/');
    DisplayGrades.column = init_suivi_state(info[2], info[4], info[3]);
    Notation.column = DisplayGrades.column;
    Notation.the_year = info[0];
    Notation.the_semester = info[1];
    Notation.the_ue = info[2];
    Notation.the_line_id = info[4];
    Notation.parse_questions(DisplayGrades.column.comment);
    if (use_textarea)
        notation_update_modifiable();
    else
        Notation.parse_grades(DisplayGrades.cell.comment);
    return key;
}

function notation_save_cell() {
    var cell = DisplayGrades.ue.line[DisplayGrades.column.data_col];
    while (cell.length < 4)
        cell.push('');
    cell[3] = Notation.get_json_grades();
}


function notation_update_modifiable() {
    var nr_full = 0, nr_empty = 0, what;
    var elements = Notation.element_top.getElementsByTagName('TEXTAREA');
    for (var i = 0; i < elements.length; i++) {
        var element = elements[i];
        var q = Notation.questions[element.getAttribute('info')];
        q.set_comment('?/1 ' + element.value.trim(), true);
        if (q.grade.comment === '') {
            nr_empty++;
            what = 'notation_empty';
        }
        else {
            nr_full++;
            what = '';
        }
        element.parentNode.parentNode.className = what;
    }
    Notation.element_stats.innerHTML = nr_full + '/' + (nr_empty + nr_full);
}

function notation_update_all_modifiable() {
    t = document.getElementsByTagName('DIV');
    for (var i = 0; i < t.length; i++)
        if (t[i].getAttribute('info')) {
            notation_init_state(notation_get_key(t[i]));
            notation_update_modifiable();
        }
}

function notation_save_textarea(textarea) {
    var key = notation_init_state(notation_get_key(textarea), true);
    var question_id = textarea.getAttribute('info');
    Notation.questions[question_id].set_comment('?/1 ' + textarea.value, true);
    localStorage[key] = Notation.get_json_grades();
    Notation.element_feedback.innerHTML = '';
    notation_update_modifiable();
}

function notation_record_ok(key) {
    notation_init_state(key, true);
    delete localStorage[key];
    notation_save_cell();
}

function notation_record(element) {
    var key = notation_get_key(element);
    notation_init_state(key, true);
    Notation.element_record.previousSibling.innerHTML = _("LABEL_MCQ_saved") + ':';
    _cell(Notation.element_record.previousSibling.lastChild,
        Notation.the_year + '/' + Notation.the_semester
        + '/' + Notation.the_ue + '/comment/' + Notation.column.the_id
        + '/' + Notation.the_line_id,
        Notation.get_json_grades(),
        function () { notation_record_ok(key); }
    );
}

function notation_suivi_popup(key) {
    notation_init_state(key);
    var column = DisplayGrades.column;
    if (localStorage[key]
        && localStorage[key] != line[column.data_col].comment
        && confirm(_("CONFIRM_MCQ_retrieve"))
    )
        Notation.parse_grades(localStorage[key]);
    else
        Notation.parse_grades(line[column.data_col].comment);

    create_popup('notation_suivi_modifiable',
        html(DisplayGrades.ue.ue) + '/' + html(column.title),
        Notation.suivi_modifiable(key), '', false);
    notation_get_key(popup_get_element().getElementsByTagName('TABLE')[0]);
    notation_update_modifiable();
    if (column.comment.match(/\\\(.*\\\)|\\\[.*\\\]/)) {
        var script = document.createElement('SCRIPT');
        script.id = "mcqmath";
        script.async = '';
        script.src = mathjaxurl;
        document.getElementsByTagName('HEAD')[0].appendChild(script);
        setTimeout('MathJax.Hub.Queue(["Typeset",MathJax.Hub])', 100);
    }
    Notation.popup_close = popup_close;
    var cell = DisplayGrades.ue.line[DisplayGrades.column.data_col];
    function close() {
        if (cell[3] != Notation.get_json_grades()) {
            if (confirm(_("LABEL_MCQ_save"))) {
                notation_record(document.getElementById('notation_record'));
                return;
            }
        }
        popup_close = Notation.popup_close;
        popup_close();
    }
    popup_close = close;
}

function notation_format_suivi() {
    var key = notation_init_state();
    if (cell_modifiable_on_suivi()) {
        if (display_update.top == 'Textual') {
            setTimeout(notation_update_all_modifiable, 500);
            return Notation.suivi_modifiable(key);
        }
        return '<button onclick="notation_suivi_popup(' + js2(key) + ')">'
            + _('Answer') + '</button>';
    }
    return hidden_txt(
        html(DisplayGrades.cell.value === ''
            ? DisplayGrades.column.empty_is
            : DisplayGrades.cell.value + '/' + DisplayGrades.column.max),
        Notation.suivi(), undefined, undefined,
        "get_tip_element().tip_target=undefined;get_tip_element().className += ' notation_suivi_tip'")
        .replace('<div class="text">', '<div class="text" instanttip="1">')
}

function notation_export() {
    Notation.parse_questions(the_current_cell.column.comment);
    Notation.column = the_current_cell.column;
    var questions = Notation.question_list();
    var data_col = the_current_cell.data_col;
    var csv = [];
    var g;

    t = [_("COL_TITLE_ID"), _("COL_TITLE_surname"), _("COL_TITLE_firstname")];
    for (var question in questions) {
        question = questions[question];
        t.push((question.is_a_bonus() ? '±' : '') + question.max);
        t.push(html(question.question));
    }
    csv.push('<tr><th>' + t.join('<th>') + '</tr>\n');

    t = ['', '', ''];
    for (var question in questions) {
        t.push(_("B_Note"));
        t.push(_("TH_comment"));
    }
    csv.push(['<tr><th>' + t.join('<th>') + '</tr>\n']);

    for (var lin in filtered_lines) {
        line = filtered_lines[lin];
        t = [line[0].value, line[2].value, line[1].value];
        try {
            g = JSON.parse(Notation.get_the_commented_line(line)[data_col].comment);
        }
        catch (e) { g = {}; }
        if (g) {
            for (var question in questions) {
                question = questions[question];
                if (g[question.id]) {
                    var grade = new NotationGrade(g[question.id]);
                    if (grade.grade == '?')
                        t.push('?');
                    else
                        t.push(trunc(grade.grade * question.max));
                    grade.question = question;
                    grade.uncompress_comment();
                    t.push(html(encode_lf_tab(grade.comment)));
                }
                else {
                    t.push("???");
                    t.push("");
                }
            }
            for (var question in g) {
                if (!Notation.questions[question]
                    || !Notation.questions[question].is_a_comment())
                    continue;
                if (g[question] && g[question].substr(0, 1) != '0')
                    t.push(Notation.questions[question].question);
            }
        }
        csv.push('<tr><td>' + t.join('<td>') + '</tr>\n');
    }
    create_popup('notation_export', _("MSG_print_popup_title"),
        '',
        '<style>DIV.import_export.notation_export { bottom: 3em; left: 3em; right: 3em; top: 3em ; border: 4px solid green }</style>'
        + '<table class="colored" style="height: 20em;font-size: 50%;">'
        + csv.join('') + '</table>',
        false);
}

function notation_import_do() {
    var error = Notation.import_do(popup_value(true));
    if (error)
        alert(error);
    else
        popup_close();
}

function notation_import() {
    if (!column_change_allowed(the_current_cell.column)) {
        alert(_("ERROR_value_not_modifiable"));
        return;
    }
    if (the_current_cell.column.modifiable == 2) {
        alert(_("MSG_annotate_not_allowed") + '\n' + _("ALERT_column_not_saved"));
        return;
    }
    var c = _("TH_comment").substr(0, 7);
    var g = _("B_Note").substr(0, 7);
    var example = [
        [_("COL_TITLE_ID"), _("COL_TITLE_surname"), _("COL_TITLE_firstname"),
            2, "Quest X", 2, "Quest Y", _("MSG_Bonus"), "Z"],
        ['', '', '', g, c, g, c, g, c],
        [],
        ["1160000", "", "", 2, "Good!", 1, "Bien", 1, "Super!"],
        ["1160001", "", "", 0.5, "False", 1, "Faux", 0, ""],
        ["1160002", "", "", 0.25, "Bad :-(", 0, "Mauvais", -0.5, "Spelling"]
    ];
    for (var i = 0; i < 3 && i < filtered_lines.length; i++) {
        example[i + 3][0] = filtered_lines[i][0].value.substr(0, 7);
        example[i + 3][1] = filtered_lines[i][2].value.substr(0, 4) + '…';
        example[i + 3][2] = filtered_lines[i][1].value.substr(0, 4) + '…';
    }
    for (var i = 0; i < example.length; i++)
        example[i] = example[i].join("\t");

    var title = _("MSG_notation_import_title") + ' «'
        + the_current_cell.column.title + '»';
    create_popup('import_div', title,
        caution_message() + _("MSG_notation_import"),
        '<button onclick="notation_import_do()">'
        + title + '</button><p>'
        + _("MSG_notation_import_warning"),
        example.join('\n'));
}
