#!/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2008-2015 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

import collections
import re
from .. import utilities
from .. import configuration
from .. import document
from .. import sender
from .. import data

###############################################################################
# This part is for column import.
# column import can by done on any column type.
###############################################################################

cookies = None

def read_url_not_cached(url):
    """Read an URL content to import in a column"""
    try:
        if configuration.real_regtest:
            timeout = 0.1
        else:
            timeout = 10 # pragma: no cover
        c = utilities.read_url(url, timeout)
        if '<html>' in c.lower(): # pragma: no cover
            try:
                from ..LOCAL import fakeuser
                global cookies
                c, cookies = fakeuser.connection(url, cookies)[:2]
            except ImportError:
                utilities.send_backtrace('Missing fakeuser.connection')
        
        return c
    except:
        utilities.send_backtrace('IMPORT ' + url)
        return
read_url = utilities.add_a_cache(read_url_not_cached, timeout=5)    

def error(column, message, more=None): # pragma: no cover
    if column.table.loading:
        return
    if more is None:
        more = column.url_import
    for page in column.table.active_pages:
        if page.user_name == column.author:
            sender.append(
                page.browser_file,
                "<script>Alert('%s', %s);</script>"
                % (message,
                   utilities.js("\n" + column.title + ' : ' + more)))

getters = {
    ':': lambda x: x.history_string(),
    '?': lambda x: x.date,
    '@': lambda x: x.author,
    '#': lambda x: x.comment,
    }

def get_column_from_a_table(column, table_list):
    """It is possible to extract multiple table/columns.
    In this case, values are concatened
    """
    values = collections.defaultdict(list)
    try:
        getter = getters[table_list[0]]
        get_value = False
    except KeyError:
        getter = lambda x: x.value
        get_value = True
    table_list = table_list.lstrip(''.join(getters)).strip()
    for url in re.split("  *", table_list):
        table = None
        col = None
        for prepend in (
                '',
                '%d/' % column.table.year,
                '%d/%s/' % (column.table.year, column.table.semester),
                '%d/%s/%s/' % (column.table.year, column.table.semester,
                               column.table.ue)
                ):
            splited = (prepend + url).split('/')
            if len(splited) != 4:
                continue
            year, semester, table_name, column_name = splited
            try:
                year = int(year)
            except ValueError:
                continue
            semester = utilities.safe(semester)
            table_name = utilities.safe(table_name)
            table = document.table(year, semester, table_name, create=False)
            if not table:
                continue
            col = table.columns.from_title(column_name.split('[')[0])
            if col:
                break # Got the good path
        if not table:
            error(column, 'ALERT_url_import_table', url)
            return
        if not col:
            error(column, 'ALERT_url_import_column', url)
            return
        colkey = None
        if '[' in column_name:
            v = column_name.split('[')[1]
            if ']' in v:
                v = v.split(']')[0]
                colkey = column.table.columns.from_title(v)
                if colkey is None:
                    error(column, 'MSG_suivi_extract_no_col', v)
                    return

        if col.private:
            error(column, 'ALERT_url_import_private', url)
            return
        if table.private: # pragma: no cover
            if (not column.table.private
                    or set(column.table.masters) != set(table.masters)
                    or column.author not in column.table.masters and column.author != data.no_user
               ):
                error(column, 'ALERT_url_import_private', url)
                return
            # The masters list are identical and a master created the column
        table.compute_columns()
        for line in table.lines.values():
            if line[0].value != '':
                v = getter(line[col.data_col])
                if v == '':
                    if get_value:
                        v = col.empty_is
                    if v == '':
                        continue
                v = (v, url)
                values[line[0].value].append(v)
                login = utilities.the_login(line[0].value)
                if line[0].value != login:
                    values[utilities.the_login(line[0].value)].append(v) # pragma: no cover
    nobody_page = None
    table = column.table
    if column.cell_writable:
        cell_writable = table.get_filter(column.cell_writable, column.type.name)
    else:
        cell_writable = lambda _line, _value, _user_name, _is_a_teacher: True # pragma: no cover
    for line_id, line in table.lines.items():
        if colkey:
            key = line[colkey.data_col].value
        else:
            key = line[0].value
        new_val = values.get(key, None)
        cell = line[column.data_col]
        if not cell_writable(line, cell,
                             data.no_user if cell.author == data.no_user else data.ro_user,
                             True):
            continue
        if new_val is None:
            new_val = values.get(utilities.the_login(key), None)
            if new_val is None:
                # Erase values no more needed: give them to 'nobody'
                if cell.author == data.ro_user and cell.value != '':
                    table.lock()
                    try:
                        if nobody_page is None:
                            nobody_page = table.get_nobody_page()
                        table.cell_change(nobody_page,
                                          column.the_id,
                                          line_id, '',
                                          force_update=True)
                    finally:
                        table.unlock()
                continue
        if len(new_val) == 1:
            new_val = new_val[0][0]
        else:
            new_val = ' '.join("%s(%s)" % (v, u) for v, u in new_val)
        if not column.type.equals(cell.value, new_val):
            table.lock()
            try:
                table.cell_change(table.pages[0], column.the_id, line_id, new_val)
            finally:
                table.unlock()
    restore_user_rights(column, all_lines=False)

@utilities.add_a_lock
def update_column_content(column, url):
    utilities.warn('IMPORT %s' % url)

    url_base = url.split('#')[0]

    if configuration.regtest:
        c = read_url_not_cached(url_base)
    else:
        c = read_url(url_base) # pragma: no cover
    if c is None:
        error(column, 'ALERT_url_import_bad')
        return

    c = c.replace('\r','\n').split('\n')

    utilities.warn('READ %d LINES IN %s' % (len(c), url))

    try:
        col = int(url.split('#')[1]) - 1
    except (IndexError, ValueError):
        error(column, 'ALERT_url_import_column')
        return

    if col <= 0:
        return # pragma: no cover
    if url.endswith('#'):
        import_in_comment = True
    else:
        import_in_comment = False

    import csv
    if '\t' in c[0]:
        delimiter = '\t'
    elif ';' in c[0]:
        delimiter = ';' # pragma: no cover
    else:
        delimiter = ','
    utilities.warn('DELIMITER IS %s' % delimiter)
    table = column.table
    if column.cell_writable:
        cell_writable = table.get_filter(column.cell_writable, column.type.name)
    else:
        cell_writable = lambda _line, _value, _user_name, _is_a_teacher: True # pragma: no cover

    for line in csv.reader(c, delimiter=delimiter):
        if len(line) <= col:
            continue
        for line_id, cells in  table.get_items(line[0].strip()):
            cell = cells[column.data_col]
            if not import_in_comment and not cell_writable(cells, cells[column.data_col], data.ro_user, True):
                continue
            new_value = line[col].strip()
            try:
                table.lock()
                if import_in_comment:
                    if not column.type.equals(cell.comment, new_value):
                        table.comment_change(table.pages[0], column.the_id, line_id, new_value)
                else:
                    if not column.type.equals(cell.value, new_value):
                        table.cell_change(table.pages[0], column.the_id, line_id, new_value)
            finally:
                table.unlock()
    restore_user_rights(column, all_lines=False)

def restore_user_rights(column, all_lines=True):
    column.table.lock()
    no_user = column.table.get_nobody_page()
    try:
        for line_id, line in column.table.lines.items():
            if (line[column.data_col].author == data.ro_user
                and (all_lines or line[0].value == '')
                ):
                column.table.cell_change(no_user,
                                         column.the_id,
                                         line_id, line[column.data_col].value,
                                         force_update=True
                                     )
    finally:
        column.table.unlock()

###############################################################################
# Text column definition
###############################################################################

class Text(object):

    human_priority = 0 # To define menu order
    
    # Default tips
    tip_column_title = "TIP_column_title"
    tip_filter = "TIP_filter_Text"
    tip_cell = "TIP_cell_Text MSG_no_private_data"

    # The value is a float most of the time
    should_be_a_float = 0

    # check value to be stored in the cell
    cell_test = 'test_nothing'
    # possible completions
    cell_completions = 'test_nothing'
    # What to do on mouse click
    onmousedown = 'cell_select'
    # How to display the cell value
    formatte = 'text_format'
    # How to display the cell value in the suivi
    formatte_suivi = 'text_format_suivi'
    # What to do on cell double click
    ondoubleclick = 'undefined'
    # How to compute the cell value
    cell_compute = 'undefined'
    # Cell is modifiable ? (a computed cell may be modifiable)
    cell_is_modifiable = 1
    # In which column display the type
    type_type = 'data'
    # Called after the user sets the column type
    type_change = "undefined"
    # Hide cell comment
    hide_cell_comment = 0
    # Hide column comment
    hide_column_comment = 0

    # The columns attributes that should be displayed for this type
    # DO NOT INDICATES ATTRIBUTES VISIBLE BY ALL THE COLUMN TYPES
    attributes_visible = ('completion', 'repetition', 'url_import',
                          'groupcolumn', 'multiline')

    # This function is called when the column is no more of this type
    # For example to stop file sharing for the 'document upload' type
    # because the file sharing is done by an external server.
    def leave_this_type(self, table, page, column, value):
        return

    def value_range(self, v_min, v_max):
        """Display the range of the possible values"""
        return '' # pragma: no cover

    def stat(self, column, value, cell, lines, grp_seq):
        """Returns a dict of various values, currently:
        'rank', 'rank_grp', 'average'
        """
        return {}

    def __str__(self):
        return '"' + self.name + '"'

    def attribute_js_value(self, k):
        if k.startswith('set_') or k in (
                'cell_test', 'cell_completions','onmousedown', 'formatte',
                'formatte_suivi', 'ondoubleclick', 'cell_compute',
                'type_change'):
            return getattr(self, k)
        else:
            return utilities.js(getattr(self, k))

    def update_one(self, the_table, line_id, column):
        """Do some server side compute on the cell.
        This function is used when the user input interactivly one cell.
        """
        pass

    def update_all(self, the_table, column, attr=None, line_ids=None):
        """Do some server side compute on the cell.
        This function update the full column content for every line.
        It is called :
           * on page load to check changs.
           * when the 'attr' column attribute is changed.
           * when a column used by this one changed
        """
        if (attr is not None
            and attr.name != 'comment'
            and attr.name != 'url_import'):
            return

        if 'IMPORT(' in column.comment:
            url = re.sub(r'.*IMPORT\(', '', column.comment)
            url = re.sub(r'\).*', '', url)
        elif column.url_import:
            url = column.url_import
        else:
            if getattr(column, 'url_import_cleared', None):
                # The URL was erased, allow the user to change the value
                column.url_import_cleared = False
                restore_user_rights(column)
            return

        # Reload even if the value is the same
        # if column.import_url == url:
        #    return
        # column.import_url = url

        if ':' not in url[1:]:
            get_column_from_a_table(column, url)
            return

        if not (url.startswith('http:')
                or url.startswith('https:')
                or url.startswith('ftp:')
                or (url.startswith('file:') and configuration.regtest)
                ):
            return

        update_column_content(column, url)

    def update_for_suivi(self, dummy_column):
        """Do some update before column computation.
        Needed by notation.py
        """
        return

    def equals(self, a, b):
        """Returns True if the 2 values are equals for this column type"""
        if a == b:
            return True
        if self.should_be_a_float:
            try:
                if isinstance(a, str):
                    a = float(a.replace(',', '.'))
                if isinstance(b, str):
                    b = float(b.replace(',', '.'))
                return a == b
            except ValueError:
                return False

Text.keys = sorted([i for i in Text.__dict__ if not i.startswith('_')])
