#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2008-2014 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

import time
from .. import plugin
from .. import utilities
from .. import abj
from ..files import files
from .. import document
from .. import cell
from .. import column
from .. import configuration

css = """
TR#dates TD {
  text-align: center ;
  vertical-align: middle ;
}
TR#ampm TD { text-align: center ; }

TABLE.colored TD.weekend { background-color: #FBB; }

TABLE.display_abjs TH { text-align: center ; }
TABLE.display_abjs TD, TABLE.display_da TD { background-color: #EEE ; }
TABLE.display_abjs TH, TABLE.display_abjs TD, TABLE.display_da TD, TABLE.display_da TH { padding: 1px }

IFRAME#datasend {
  height: 15em ;
  width: 100% ;
}

DIV#student_display IMG {
  width: 10% ;
  float: right ;
}

INPUT#sendabj { width: 15em ; }
INPUT#sendmessage { width: auto ; text-align: left }

INPUT#sendabj, INPUT#sendmessage { 
  background: #EEE; 
}

SPAN.arrow { 
  color: blue ;
  font-size: 200% ;
}

BODY#BODY P, BODY#BODY H1 { margin: 0px ; }

BODY.abj_body { background: #EEE }

P.wait { background-color: #8F8 ; }

DIV.abj_comments { border: 3px solid black; line-height: 1.5em; top:6em }
.abj_comments A { border: 1px solid black ; }

#abjcomment, #dateda { background: white; }

"""

def abj_home(server):
    """Display the home page for ABJ management"""
    utilities.warn('Start')
    d = (document.table_head(int(server.the_year), server.the_semester,
                             the_ticket=server.ticket.ticket,
                             user_name = server.ticket.user_name,
                             hide_more=True,
                             )
         + '<script src="%s/all_ues.js"></script>' % configuration.url_files
         + '<title>%s</title>' % server._("TITLE_abj")
         + files['abj.html'].get_content()
         )
    server.the_file.write(d)

def abj_html(server):
    """Send an HTML file containing all the ABJ and DA information
    for the students registered to the specified UE regular expression.

    Add to 'config_home' table some link :
       * with container 'abj_master';
       * the 'verysafe' class;
       * the group name needing to see the link 'abj_masters'

    For example:

       Master: Full JLV+DA report           javascript:go('abj/html/.*M$24')
       Licence: Full JLV+DA report          javascript:go('abj/html/.*L$24')
       EPU: Full JLV+DA report              javascript:go('abj/html/UE-EI')

    $24 because the dollar must be escaped.
    """
    if server.the_path[0] == "REGEXP":
        server.the_file.write(server._("HELP_abjhtml"))
    else:
        abj.alpha_html(server, server.the_year, server.the_semester,
                       match=server.the_path[0])

def abj_alpha_author(server):
    """Send a CSV file containing all the ABJ and DA information
    for all the students"""
    abj.alpha_html(server, server.the_year, server.the_semester,
                   author=server.ticket.user_name)
    
def abj_send_mail(server):
    """When triggered this function send all the mail
    generated by the plugin <a href="#plugin_abjlistmail">listmail</a>."""
    abj.send_mail(server.the_file)
    
def abj_list_mail(server):
    """Generate and display all the message to the UE teachers
    given them the names of the students with ABJ, DA and TT.
    <p>At the end of the page, there is a link to really send the messages</p>
    """
    abj.list_mail(server.the_file, server.the_year, server.the_semester)
    
def abj_action(server):
    """This plugin perform an action on the ABJ and DA list.
    These actions can be adding or removing an ABJ or a DA.
    The client receive a feedback image indicating the status
    of the change.
    """

    path = server.the_path
    if path[0] == 'addabj':
        abj.add_abjs(server.the_year, server.the_semester, server.ticket,
                     server.the_student,
                     '/'.join(path[1:4]) + path[4],
                     '/'.join(path[5:8]) + path[8], path[9].strip())
    elif path[0] == 'add_da':
        abj.add_abjs_da(server.the_year, server.the_semester,
                        server.ticket, server.the_student,
                        path[1], path[2] + '/' + path[3] + '/' + path[4],
                        path[5].strip())
    elif path[0] == 'rem_da':
        abj.rem_abjs_da(server.the_year, server.the_semester,
                        server.ticket, server.the_student, path[1])
    elif path[0] == 'delabj':
        abj.rem_abjs(server.the_year, server.the_semester, server.ticket,
                     server.the_student,
                         '/'.join(path[1:4]) + path[4],
                         '/'.join(path[5:8]) + path[8])
    elif path[0] == 'display':
        pass
    else:
        server.the_file.write('bug') # pragma: no cover
    abj.a_student(server.the_file, server.the_year, server.the_semester,
                  server.ticket, server.the_student)
    to_update.append(server.the_student)

    if configuration.regtest:
        do_update()
    else:
        utilities.start_job(do_update, 60) # pragma: no cover

to_update = []

def do_update():
    """Update the ABJS"""
    while to_update:
        students = set()
        while to_update:
            students.add(to_update.pop(0))
        for table in document.tables.values():
            if (table.year, table.semester) == configuration.year_semester:
                for student in students:
                    if student in table.the_key_dict:
                        table.update_the_abjs()
                        break # Only update the table once
        last_check = time.time()
    return last_check

def abj_no(server):
    """Display an error message becauses this action is unauthorized"""
    server.the_file.write('BUG: abj_no')


plugin.Plugin('abj', '/{Y}/{S}/abj', function=abj_home, group='abj_masters',
              link=plugin.Link(
                  url="javascript:go('abj')",
                  where='abj_master', html_class="safe",
                  ),
              css=css,
              priority = -9, unsafe=False,
              )

plugin.Plugin('abjhtml', '/{Y}/{S}/abj/html/{*}',
              function=abj_html, group='abj_masters',
              launch_thread = True,
              priority = -3,
              link=plugin.Link(
                  url="javascript:go('abj/html/REGEXP')",
                  where='abj_master', html_class="verysafe",
                  ),
              )

plugin.Plugin('abjalphaauthor', '/{Y}/{S}/abj/alpha_author.xls',
              function=abj_alpha_author, group='abj_masters',
              launch_thread = True,
              link=plugin.Link(
                  url="javascript:go('abj/alpha_author.xls')",
                  where='abj_master', html_class="verysafe",
                  ),
              priority = -3,
              )

plugin.Plugin('abjlistmail', '/{Y}/{S}/abj/list_mail',
              function=abj_list_mail, group='abj_masters',
              launch_thread = True,
              link=plugin.Link(
                  url="javascript:go('abj/list_mail')",
                  where='abj_master', html_class="safe",
                  ),
              priority = -3,
              )

plugin.Plugin('abjhacker', '/{Y}/{S}/abjs', function=abj_no,
              priority = -20,
              )

plugin.Plugin('abjaction', '/{Y}/{S}/abj/{P}/{I}/{*}',
              function=abj_action, group='abj_masters',
              priority = -20,
              )
plugin.Plugin('abjsendmail', '/{Y}/{S}/abj/send_mail',
              function=abj_send_mail, group='abj_masters',
              launch_thread = True,
              priority = -3,
              )

def abj_table(server): # pylint: disable=too-many-locals
    """Returns ABJ/DA information for all the students"""
    year = server.the_year
    semester = server.the_semester
    columns = [
        column.Column('0', '', title=server._('COL_TITLE_ID'), width=2),
        column.Column('1', '', title=server._('COL_TITLE_0_2')),
        column.Column('2', '', title=server._('COL_TITLE_0_1')),
        column.Column('3', '', title=server._('TH_what'), width=1),
        column.Column('4', '', title=server._('B_Date'), width=2),
        column.Column('5', '', title=server._('TH_end_or_ue'), width=2),
        column.Column('6', '', title=server._('TH_comment')),
        column.Column('7', '', title=server._('UE')),
        column.Column('8', '', title=server._('TH_abj_author'), width=2),
        column.Column('9', '', title=server._('Etape'), width=2),
    ]
    progress_bar = utilities.ProgressBar(
        server, message='<h1>' + server._("MSG_suivi_student_wait")+'</h1>')
    lines = []
    students = tuple(abj.Abjs(year, semester).students(progress_bar))
    nbr = 0
    for login in students:
        nbr += 1
        progress_bar.update(nbr, len(students))
        user = configuration.User(login)
        def write():
            lines.append(cell.Line([
                cell.CellValue(x, date=mdate)
                for x in (login, user.surname, user.firstname,
                          what, date1, date2, comment, ues, author, ' '.join(user.etapes_list))]))
        ues = ' '.join(user.ues_short)
        student = abj.Abj(year, semester, login, get_date=True)
        what = 'ABJ'
        for date1, date2, author, comment, mdate in student.abjs:
            write()
        what = 'DAS'
        for date1, date2, author, comment, mdate in student.da:
            date1, date2 = date2, date1
            write()
    progress_bar.hide()
    document.virtual_table(server, columns, lines,
                           table_attrs={'default_nr_columns': 10})

plugin.Plugin('abj_table', '/{Y}/{S}/abj/table',
              function=abj_table, group='abj_masters',
              launch_thread=True,
              priority=-3,
              link=plugin.Link(
                  url="javascript:go('abj/table')",
                  where='abj_master', html_class="safe",
                  ),
              )

def do_abi2abj(server):
    """Replace an ABI by ABJ"""
    table = document.table(server.the_year, server.the_semester, server.the_ue, create=False)
    if not table:
        return 'BUG table'
    lines = tuple(table.get_items(server.the_student))
    if len(lines) != 1:
        return 'BUG student'
    lin_id, line = lines[0]
    if table.semester == 'P':
        col = table.columns.from_title(server.something)
        if not col:
            return 'BUG column'
        value = line[col.data_col].value
        if value not in configuration.all_abi and not value.startswith(configuration.abi):
            return server._("ALERT_abj_duplicate") + ':' + value + ' ≠ ' + configuration.abi
    else:
        for a_cell, col in zip(line, table.columns):
            if a_cell.date == server.something:
                value = line[col.data_col].value
                if value in configuration.all_abi or value.startswith(configuration.abi):
                    break
        else:
            return 'BUG cell not found'
    with table.the_lock:
        page = table.new_page(
            server.ticket.ticket, server.ticket.user_name,
            server.ticket.user_ip, server.ticket.user_browser, recycle=True)
        new_value = configuration.abj
        if '@' in value:
            new_value += '@'
        return table.cell_change(page, col.the_id, lin_id, new_value, force_update=True)

def abi2abj(server):
    """Replace an ABI by ABJ"""
    result = do_abi2abj(server)
    if result == 'ok.png':
        background = '#0F0'
        result = 'OK'
    else:
        background = '#F00'
    server.the_file.write('<style>BODY { margin: 0px; background:%s }</style>%s'
        % (background, result))

plugin.Plugin('abi2abj', '/abi2abj/{I}/{Y}/{S}/{U}/{?}',
              function=abi2abj, group='abj_masters',
              launch_thread=True,
              )
