#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2021 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

"""
Every seconds displays the tomuss state.

The functions must returns a positive value or (value, text)

The function doc string indicates the color and the help message.

In the browser interface:
   * click on the graphic to see the values for this time (you can let the button down)
   * user cursor keys to move step by step
   * use 'space' key to toogle the history display
"""


# pylint: disable=missing-function-docstring

import sys
import time
import json
import re
import os
import threading
import traceback
import random

from .. import plugin
from .. import utilities
from .. import configuration
from .. import document
from .. import ticket
from .. import sender
from .. import inscrits
from .. import authentication

def value_and_texts(table):
    """Helper function to return the text only if there is one"""
    table = tuple(table)
    if table:
        return len(table), '\n'.join(table) # Random # pragma: no cover
    return 0



def timestamp():
    "#000 The time of the sample"
    return time.time()


def threads():
    "#F00 Number of running threads"
    return len(tuple(threading.enumerate()))

def mainthread():
    "#F44 Main thread working"
    texts = []
    main = threading.main_thread().ident
    frames = traceback.format_stack(sys._current_frames()[main]) # pylint: disable=protected-access
    if 'socketserver' not in frames[-2]:
        for frame in frames: # pragma: no cover
            if 'do_GET_real(' in frame:
                texts = []
            else:
                texts.append(frame)
    if texts: # pragma: no cover
        return 1, ''.join(texts)
    return 0 # pragma: no cover

def senders():
    "#F80 Number of writing thread active"
    return sender.File.nr_active_thread
def plugin_call(init=False): # pylint: disable=unused-argument # MUST BE PRESENT
    "#088 Number of plugin call"
    nr_plugin_call = plugin.nr_plugin_call
    plugin.nr_plugin_call = 0
    return nr_plugin_call
def importants():
    "#F08 Number of important jobs running"
    return value_and_texts(utilities.current_jobs)
def analyser():
    "#F08 Analyser"
    return len(configuration.analyser_in_use)
def locks():
    "#F40 Number of locks"
    return value_and_texts(
        [
            "{} {}".format(fct.fct.__name__, fct.fct.__module__)
            for fct in utilities.lock_list
            if fct.the_lock.locked()
        ]
        + [
            str(table)
            for table in document.tables_values()
            if table.the_lock.locked()
        ])

def users():
    "#0A0 Number of 'User' in memory"
    return len(configuration.Users.instances)
def tables():
    "#4A0 Number of 'Table' in memory"
    return len(document.tables)
def tickets():
    "#0A4 Number of 'Ticket' in memory"
    return len(ticket.tickets)


def requests():
    "#00F Number of table actions queued"
    return value_and_texts(
        '{} {} {} {}'.format(page, timestamp, actions)
        for page, timestamp, actions, _output_file in document.request_list)
def authentications():
    "#08F Number of authentication queued"
    return value_and_texts(
        server.ticket and str(server.ticket) or ticket.client_ip(server)
        for server in authentication.authentication_requests)
def batchs():
    "#88F Number of batch requests queued"
    return value_and_texts(
        '{} {} {}'.format(i[1][0].ticket, i[1][0].path, i[1][1].name)
        for i in plugin.background_todo)
def fastupdate():
    "#88F Number of recompute cell queued"
    return value_and_texts(
        '{} {} {}'.format(the_table, lin, a_column.the_id)
        for the_table, lin, a_column in document.cell_changed_list_fast)
def slowupdate():
    "#88F Number of recompute column queued"
    return value_and_texts(
        '{} {} {}'.format(the_table, a_column.the_id, attr)
        for the_table, a_column, attr in document.column_changed_list)
def cleanups():
    "#80F Number of remove files queued"
    return len(utilities._cleanup_list) # pylint: disable=protected-access
def updatestudentlists():
    "#84F Number of student lists update queued"
    return value_and_texts(
        str(table)
        for table in document.update_students)
def indexesupdates():
    "#48F Number of student index update queued"
    return len(document.indexes_to_update)

def loadavg():
    "#444 The 5 minutes load average of the host"
    return os.getloadavg()[0]
def diskusage(init=False):
    "#844 The disk usage of the host"
    try:
        with open('/proc/diskstats', 'r', encoding="utf-8") as datafile:
            lines = [re.split('  *', ' ' + i) for i in datafile.readlines()]
    except IOError: # pragma: no cover
        return 0
    current = sum([int(i[14])
                   for i in lines
                   if not i[3][-1].isdigit()])
    if init:
        diskusage.previous = current
        return 0 # unused
    difference = current - diskusage.previous
    diskusage.previous = current
    return difference
def netusage(init=False):
    "#448 The network usage of the host"
    try:
        with open('/proc/net/dev', 'r', encoding="utf-8") as datafile:
            datafile.readline()
            datafile.readline()
            values = []
            for line in datafile:
                line = re.split(' +', line.strip())
                values.append(int(line[1]) + int(line[9]))
    except IOError: # pragma: no cover
        return 0
    current = max(values) # Assume maximum traffic on ethernet
    if init:
        netusage.previous = current
        return 0 # unused
    difference = current - netusage.previous
    netusage.previous = current
    return difference
def virtual():
    "#488 The virtual memory used by the process"
    try:
        with open('/proc/self/statm', 'r') as datafile:
            line = datafile.readline()
    except IOError: # pragma: no cover
        virtual.infos = (0, 0)
        return 0
    virtual.infos = [int(4*int(v)/1024) for v in line.split(' ')]
    return virtual.infos[0]
def resident():
    "#884 The resident memory used by the process"
    return virtual.infos[1]
def openfiles():
    "#F08 Number of open files"
    return len(tuple(os.listdir('/proc/self/fd')))
def pagein(init=False):
    "#448 Page in"
    try:
        with open('/proc/vmstat', 'r') as datafile:
            data = datafile.read()
    except IOError: # pragma: no cover
        data = 'pswpin 0\npswpout 0\n'

    page_in = int(data.split('pswpin ', 1)[1].split('\n', 1)[0])
    page_out = int(data.split('pswpout ', 1)[1].split('\n', 1)[0])

    if init:
        pagein.page_in = page_in
        pagein.page_out = page_out
        return 0 # unused

    pagein.old_page_in = pagein.page_in
    pagein.old_page_out = pagein.page_out
    pagein.page_in = page_in
    pagein.page_out = page_out
    return pagein.page_in - pagein.old_page_in
def pageout():
    "#44A Page out"
    return pagein.page_out - pagein.old_page_out

def ldap():
    "#F0F LDAP"
    return value_and_texts(
        connection.search
        for connection in (inscrits.L_fast, inscrits.L_slow, inscrits.L_batch)
        if connection.search)

DISPLAY = (
    timestamp, threads, mainthread, users, tables, tickets,
    updatestudentlists, indexesupdates, fastupdate, slowupdate,
    requests, authentications, senders, batchs, plugin_call, importants, analyser, locks, ldap, cleanups,
    loadavg, diskusage, netusage, virtual, resident, openfiles, pagein, pageout
)

#REDEFINE
# Add your own functions to display values and texts on the dashboard
DISPLAY_MORE = ()

CLIENTS = []
KEYS = {} # key → last usage time

def dashboard_run():
    display = DISPLAY + DISPLAY_MORE
    for fct in display:
        if 'init' in fct.__code__.co_varnames:
            fct(True)
    while CLIENTS:
        try:
            data = [fct() for fct in display]
        except: # pylint: disable=bare-except # pragma: no cover
            data = [0] * len(display)
            data[0] = [0, ''.join(traceback.format_exc())]
        data = json.dumps(data) + '\n'
        for client in tuple(CLIENTS):
            try:
                client.the_file.write(data)
            except OSError: # pragma: no cover
                CLIENTS.remove(client)
                client.close_connection_now()
            if configuration.real_regtest:
                client.close_connection_now()
                CLIENTS.clear()
                return
        time.sleep(1) # pragma: no cover

def dashboard_data(server):
    if server.the_path[-1] in KEYS:
        KEYS[server.the_path[0]] = time.time()
        CLIENTS.append(server)
        utilities.start_job(dashboard_run, 0)
    else:
        server.the_file.write("Bad key: " + server.the_path[-1])
        server.close_connection_now()

def remove_old_keys():
    too_old = time.time() - 86400
    for key, last_access in tuple(KEYS.items()):
        if last_access < too_old: # pragma: no cover
            KEYS.pop(key)

def dashboard(server):
    """Display the dashboard"""
    remove_old_keys()
    key = hex(random.randrange(1<<128))
    KEYS[key] = time.time()
    server.the_file.write(utilities.read_file('PLUGINS/dashboard.html'))
    server.the_file.write('<script>initialize({},{});</script>\n'.format(
        json.dumps([[fct.__name__, fct.__doc__] for fct in DISPLAY + DISPLAY_MORE]),
        json.dumps(key)))

plugin.Plugin('dashboard', '/dashboard', group='roots', function=dashboard,
              link=plugin.Link(url="/dashboard", where='debug', html_class='verysafe'))

plugin.Plugin('dashboard_data', '/dashboard_data/{*}', function=dashboard_data,
              authenticated=False, keep_open=True)
