#!/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2021 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr


# MISSING FEATURES:
#   * Rounding
#   * One tab per TD group
#   * Column choice
#   * Line choice
#   * Invert name/surname
#   * Invert Seq/Grp
#   * URL Image

import io
import time
import xlsxwriter
from TOMUSS import document
from TOMUSS import plugin
from TOMUSS import cell
from TOMUSS.PYTHON_JS import tomuss_python

def get_columns(table, user_name):
    """Get the viewable columns and compile their coloring filters"""
    table.compute_columns()
    columns = [
        column
        for column in table.columns.left_to_right()
        if column.readable_by(user_name)
    ]

    with document.Table.compute_columns.the_lock: # pylint: disable=no-member
        tomuss_python.columns = table.columns.columns
        tomuss_python.C = cell.Cell
        for column in columns:
            if column.red:
                column.red_filter = table.get_filter(column.red, column.type.name)
            else:
                column.red_filter = lambda _a, _b, _c, _d: False
            if column.green:
                column.green_filter = table.get_filter(column.green, column.type.name)
            else:
                column.green_filter = lambda _a, _b, _c, _d: False
            if column.redtext:
                column.redtext_filter = table.get_filter(column.redtext, column.type.name)
            else:
                column.redtext_filter = lambda _a, _b, _c, _d: False
            if column.greentext:
                column.greentext_filter = table.get_filter(column.greentext, column.type.name)
            else:
                column.greentext_filter = lambda _a, _b, _c, _d: False
    return columns

def get_lines(table, sort_columns):
    """Get all the sorted table lines"""
    lines = [line for line in table.lines.values() if line[0].value]
    for data_col in sort_columns[::-1]:
        data_col = int(data_col)
        if data_col >= 0:
            lines.sort(key=lambda line: str(line[data_col].value)) # pylint: disable=cell-var-from-loop
        else:
            data_col = -data_col + 1
            lines.sort(key=lambda line: str(line[data_col].value), reverse=True) # pylint: disable=cell-var-from-loop
    return lines

def get_formats(workbook, wrap):
    """All the format combination"""
    formats = []
    for url in range(2): # pylint: disable=too-many-nested-blocks
        for bold in range(2):
            for grey in range(2):
                for fg_color in ('', '#A00000', '#00A000', '#DD6000'):
                    for bg_color in ('', '#FF8080', '#80FF80', '#FFC000'):
                        attributes = {}
                        if bg_color:
                            attributes['bg_color'] = bg_color
                        if fg_color:
                            attributes['color'] = fg_color
                        if grey:
                            attributes['color'] = '#808080'
                        if bold:
                            attributes['bold'] = True
                        if wrap:
                            attributes['text_wrap'] = True
                        if url:
                            attributes['color'] = '#0000FF'
                            attributes['underline'] = 1
                        formats.append(workbook.add_format(attributes))
    return formats

def create_headers(workbook, worksheet, table, columns):
    """Create XLS top line. Return the first empty line number"""
    line = 0
    header_format = workbook.add_format({'bold': True})
    worksheet.write(line, 0, table.year, header_format)
    worksheet.write(line, 1, table.semester, header_format)
    worksheet.write(line, 2, table.ue, header_format)
    worksheet.write(line, 3, ' '.join(table.masters), header_format)
    line += 1
    worksheet.write(line, 0, time.strftime('%Y-%m-%d %H:%M:%S'), header_format)
    line += 1
    worksheet.write(line, 0, table.table_title, header_format)
    line += 1
    if table.comment:
        worksheet.write(line, 0, table.comment, header_format)
        line += 1
    for i, column in enumerate(columns):
        worksheet.write(line, i, column.title, header_format)
        if column.comment:
            worksheet.write_comment(line, i, column.comment)
        worksheet.set_column(i, i, 2.5 * column.width)
    return line + 1

# pylint: disable=too-many-arguments
def fill_data(worksheet, line_nr, formats, columns, lines, user_name):
    """Fill the sheet with the table data"""
    today = time.strftime('%Y%m%d')
    for i, line in enumerate(lines):
        i += line_nr
        for j, column in enumerate(columns):
            item = line[column.data_col]
            if item.value == '':
                value = column.empty_is
            else:
                value = item.value
            if str(value).isdigit():
                value = int(value)
            else:
                try:
                    value = float(value)
                except ValueError:
                    pass
            fmt = formats[
                column.red_filter(line, item, user_name, True)
                + 2 * column.green_filter(line, item, user_name, True)
                + 4 * column.redtext_filter(line, item, user_name, True)
                + 8 * column.greentext_filter(line, item, user_name, True)
                + 16 * (item.value != value)
                + 32 * item.date.startswith(today)
                + 64 * (column.type.name == 'URL')
                ]
            if column.type.name == 'URL':
                worksheet.write_url(i, j, column.url_base + str(value), fmt, str(value))
            else:
                worksheet.write(i, j, ' ' if value == '' else value, fmt)

            if item.comment:
                worksheet.write_comment(i, j, item.comment)

def export_xls(server):
    """Generate an XLS file
    """
    table = document.table(server.the_year, server.the_semester, server.the_ue, create=False)
    if not table:
        return
    if not table.readable_by(server.ticket):
        server.the_file.write(b'PRIVATE') # server._("MSG_suivi_extract_unreadable") + table.ue)
        return
    server.the_path.pop() # The last arg is the filename

    file_content = io.BytesIO()
    workbook = xlsxwriter.Workbook(file_content, {'nan_inf_to_errors': True})
    worksheet = workbook.add_worksheet(server.the_ue[:31])
    columns = get_columns(table, server.ticket.user_name)
    line = create_headers(workbook, worksheet, table, columns)
    fill_data(
        worksheet,
        line,
        get_formats(workbook, 'wrap' in server.the_path),
        columns,
        get_lines(table, server.the_path[0].split(',')),
        server.ticket.user_name)

    workbook.close()
    server.the_file.write(file_content.getvalue())

plugin.Plugin('export_xls', '/{Y}/{S}/{U}/export_xls/{*}', function=export_xls,
              mimetype='application/vnd.ms-excel',
              group='staff', launch_thread=True)
