#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2008-2021 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

import os
import time

from .. import plugin
from .. import document
from ..utilities import warn
from .. import configuration
from .. import utilities

initial_content = '''
<script>
var    Xcell_change  = window.parent.Xcell_change    ;
var Xcomment_change  = window.parent.Xcomment_change ;
var  Xcolumn_delete  = window.parent.Xcolumn_delete  ;
var  Xcolumn_attr    = window.parent.Xcolumn_attr    ;
var  Xtable_attr     = window.parent.Xtable_attr     ;
var  change_abjs     = window.parent.change_abjs     ;
var  update_mail     = window.parent.update_mail     ;
var  update_portail  = window.parent.update_portail  ;
var  login_list      = window.parent.login_list      ;
var  set_updating    = window.parent.set_updating    ;
</script>
    '''

def new_page(server):
    """Create a new page and send the table editor to the client."""

    if not server.table_access_allowed():
        server.the_file.write(
            '<script>window.location = '
            + utilities.js(server.ticket.url(configuration.suivi.url(
                server.the_year, server.the_semester) + '/' + server.the_ue))
            + '</script>')
        server.close_connection_now()
        return

    if configuration.table_redirection(server): # pragma: no cover
        server.close_connection_now()
        return
    server.the_file.write(
        document.the_head.get_content() + '''
<div id="intro" style="position:absolute;left:0;top:0;right:0;bottom:0;
    background:#FFF;z-index:1000;opacity:0.8;text-align:center"
><div style="font-size:200%;padding-top:10%">''' + configuration.intro_message + '''</div>
<button style="font-size:200%;margin-top:2em" onclick="hide_intro()">👍 OK !</button>
</div>
<script>
function hide_intro() {
    var e = document.getElementById('intro');
    e.parentNode.removeChild(e);
    intro_message = undefined;
 }
try {
var intro_message = document.getElementById('intro').firstChild.innerHTML;
if ( localStorage['intro'] == intro_message || intro_message === '' )
    hide_intro();
else
    localStorage['intro'] = intro_message;
}
catch(err) {}
</script>
''')

    document.create_link(server.the_year, server.the_semester, server.the_ue)
    start_load = time.time()
    try:
        table, page = document.table(server.the_year, server.the_semester,
                                     server.the_ue, None, server.ticket)
    except IOError: # pragma: no cover
        server.the_file.write(server._("TIP_violet_square"))
        utilities.send_backtrace(repr(server.the_path), 'Newpage IOError')
        server.close_connection_now()
        return

    if table == None:
        utilities.send_backtrace(repr(server.the_path),
                                 'Newpage Unauthorized %s/%s/%s by %s' %
                                  (server.the_year, server.the_semester,
                                   server.the_ue, server.ticket.user_name),
                                 exception=False)
        server.the_file.write(server._("MSG_new_page_unauthorized"))
        server.close_connection_now()
        warn('No Table', what="error")
        return

    if not table.on_disc:
        server.the_file.write("%s/%s/%s" % (
                server.the_year, server.the_semester, server.the_ue))
        server.the_file.write(server._("MSG_new_page_in_past"))
        server.close_connection_now()
        return
    
    warn('New page, do_not_unload=%s' % table.do_not_unload, what="table")

    if configuration.regtest_sync:
        # We want immediate update of the table content (abjs for example)
        document.update_students.append(table)
        document.check_new_students_real()

    # Update the number of access for the user
    # No lock because it is not important (there is a lock in manage_key)
    if server.the_semester in configuration.semesters:
        d = server.ticket.user.manage_key('pages')
        if d is False:
            d = {}
        else:
            d = eval(d)
        if server.the_ue not in d:
            d[server.the_ue] = 1
        else:
            d[server.the_ue] += 1
        server.ticket.user.manage_key('pages', content=utilities.stable_repr(d))
    # With this lock cell modification can't be lost
    # between page content creation and the page activation
    server.the_file.write(table.content(page, server, hide_head=True))
    warn('Actives=%s do_not_unload=%s' % (
        table.active_pages, table.do_not_unload), what="table")

    page.start_load = start_load # For end_of_load computation
    server.close_connection_now()

plugin.Plugin('emptyname', '/{Y}/{S}/', response=307,
              headers = lambda x: (('Location', x.ticket.url(configuration.server_url)),
                                  ),
              documentation = "Bad url, redirect user to the home page")

def answer_page(server):
    """Connect the browser IFRAME to the page"""
    if not server.table_access_allowed():
        return
    try:
        table, page = document.table(server.the_year, server.the_semester,
                                     server.the_ue, server.the_page,
                                     server.ticket)
        if server.request_number:
            # Do not revert request sent
            # It is possible to go in the future because page.request
            # may forgot incrementation.

            request_page_creation = server.request_number.split('.')
            try:
                page.index = int(request_page_creation[1])
            except:
                pass
            if page.index > len(table.sent_to_browsers):
                 # Server rebooted
                page.index = 0 # pragma: no cover
            if (len(request_page_creation) >= 5
                and request_page_creation[3] != "None"
                # Reloaded since last sync, so the history is lost
                and int(request_page_creation[3]) != table.load_time
                # Modified since the last sync
                and int(request_page_creation[4]) < table.mtime
                ): # pragma: no cover
                # In some case table.mtime > request_page_creation[4]
                # because the page itself wrote things after the reconnection.
                # So we look all the page time stamps except ours
                t = max(p.mtime
                        for p in table.pages
                        if p is not page
                        )
                if  int(request_page_creation[4]) < t:
                    with open(os.path.join('LOGS', 'desync.log'), 'a') as f:
                        f.write('{}\n'.format(page))
                    server.the_file.write('<script>window.parent.desync()</script>')
                    server.the_file.flush()
                    if configuration.regtest:
                        server.close_connection_now()
                    return

            if (len(request_page_creation) >= 3
                and request_page_creation[2] != table.pages[0].date
                and request_page_creation[2] != "None"
                ): # pragma: no cover
                server.close_connection_now()
                raise Exception("The table has been copied or deleted...") # pragma: no cover
    except ValueError:
        server.the_file.write(
            '''
<script>
if (window.parent.click_to_revalidate_ticket)
      window.parent.click_to_revalidate_ticket() ;
else
      Alert("ALERT_new_page_too_old") ;
</script>''')
        server.close_connection_now()
        return
    except:
        # Reconnection of an old non modifiable page on: a stopped server
        # or with an old ticket after a connection lost.
        # See REGTEST_SERVER/tests.py 'lostpage'
        server.the_file.write('<script>window.parent.location = "%s/%s/%s/%s";</script>' % (configuration.server_url, server.the_year, server.the_semester, server.the_ue))
        server.close_connection_now()
        utilities.send_backtrace('', 'Page not found')
        return

    server.the_file.write("<!-- page.index=%d sent_to_browser=%d browser=%s closed=%s sync=%s\n -->\n" % (
        page.index,
        len(table.sent_to_browsers),
        type(page.browser_file),
        page.browser_file and page.browser_file.closed,
        configuration.regtest_sync ))

    server.the_file.write(initial_content)
    server.the_file.write(''.join(table.sent_to_browsers[page.index:]))
    server.the_file.write("<script>window.parent.page_index = %d</script>" % len(table.sent_to_browsers))
    server.the_file.flush()
    if configuration.regtest_sync:
        if configuration.regtest:
            server.close_connection_now()
    else:
        table.active_page(page, server.the_file)

plugin.Plugin('answer_page', '/{Y}/{S}/{U}/{P}/{R}',
              function=answer_page, group='',
              keep_open = True,
              launch_thread=True,
              priority = -7, # Before student redirect
              )

plugin.Plugin('pagenew', '/{Y}/{S}/{U}/{=}', function=new_page, group='',
              keep_open = True,
              priority = -7, # Before student redirect
              launch_thread = True, unsafe=False)
