#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2008 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

import time
import os
import json
import urllib.parse
from .. import configuration
from .. import plugin
from .. import document
from .. import abj
from .. import utilities
from .. import files

def page_rewrite(server):
    """Generate the minimal python module defining the current
    table state. So there is no more history."""
    # do_not_unload : not import because it is readonly
    table = document.table(server.the_year, server.the_semester,
                           server.the_ue, None, None)
    server.the_file.write(table.rewrite())

def page_action(server):
    """Receive a Resquester set of actions"""
    timestamp, server.the_year, server.the_semester, server.the_ue, page_id, actions = json.loads(server.content)
    if not actions:
        raise ValueError("Empty actions list") # pragma: no cover
    if not server.table_access_allowed():
        return
    actions = [[urllib.parse.unquote(i)
                for i in action.replace('$', '%').split('/')
               ]
               for action in actions
              ]
    try:
        _table, page = document.table(server.the_year, server.the_semester, server.the_ue, int(page_id),
                                     server.ticket, do_not_unload='page_action')
    except:
        server.the_file.write(b'"bug.png"')
        utilities.send_backtrace(repr(actions), "page_action")
        return
    page.add_request(int(timestamp), actions, server.the_file)


def page_resume(server):
    """Display the list of ABJ, DA and TT for the students in the table.
    It is the same list being sended by mail."""
    lines, dummy_mails = abj.ue_resume(server.the_ue, server.the_year,
                                       server.the_semester, server.the_file)
    server.the_file.write(''.join(lines))


plugin.Plugin('pagerewrite', '/{Y}/{S}/{U}/rewrite',
              function=page_rewrite, group='staff',
              mimetype = "text/plain; charset=UTF-8",
              launch_thread = True)

plugin.Plugin('pageresume', '/{Y}/{S}/{U}/resume',
              function=page_resume, group='staff',
              mimetype = "text/plain; charset=UTF-8",
              launch_thread = True)

plugin.Plugin('pageaction', '/pageaction', function=page_action,
              group='', mimetype = "image/png",
              keep_open=True,
              cached = True, # We don't want browser reloading actions
              priority = -8, # Before student redirect
              )

def page_unload(server):
    """Unload page from memory"""
    table = document.table(server.the_year, server.the_semester,
                           server.the_ue, create=False)
    if table:
        server.the_file.write(server._("MSG_page_unload_before") + '<br>\n')
        table.unload(force=True)
        server.the_file.write(server._("MSG_page_unload_after") + '<br>\n')
    else:
        server.the_file.write(server._("MSG_page_unload_no_table") + "<br>\n")

plugin.Plugin('page_unload', '/{Y}/{S}/{U}/page_unload',
              function=page_unload, group='roots')

def extension(server):
    """Extend the current table to the next semester.
    A symbolic link is created and the current table is accessible
    in the 2 semesters.
    """

    next_year, next_semester = utilities.next_year_semester(
        server.the_year, server.the_semester)

    new_filename = document.table_filename(next_year, next_semester,
                                          server.the_ue)
    if os.path.islink(new_filename):
        server.the_file.write(server._("MSG_extension_yet_done"))
        return

    table = document.table(next_year, next_semester, server.the_ue, create=False)
    if table:
        empty, message = table.empty(empty_even_if_used_page=True,
                                    empty_even_if_created_today=True,
                                    empty_even_if_column_created=True)
        if not empty: # pragma: no cover
            server.the_file.write(server._("MSG_extension_not_empty")
                                % table.location() + message)
            return

    old_filename = document.table_filename(server.the_year,
                                           server.the_semester,
                                           server.the_ue)
    if not os.path.exists(old_filename):
        server.the_file.write(server._("MSG_extension_no_table")
                              % (server.the_year, server.the_semester))
        return

    t = document.table(server.the_year, server.the_semester, server.the_ue,
                       None, None)
    if server.ticket.user_name not in t.masters:
        server.the_file.write(server._("MSG_extension_not_master"))
        return

    old_filename = os.path.join(*old_filename.split(os.path.sep)[1:])
    if table:
        table.unload(force=True)
    utilities.unlink_safe(new_filename)
    os.symlink(os.path.join('..', '..', old_filename), new_filename)

    server.the_file.write(server._("MSG_extension_ok")
                          % (server.the_semester, next_semester))

    utilities.start_new_thread(document.update_indexes,
                               (next_year, next_semester, server.the_ue))

plugin.Plugin('extension', '/{Y}/{S}/{U}/extension', group='staff',
              function=extension,
              )

def undo_extension(server):
    """Remove the symbolic link merging the tables.
    """
    year, semester = utilities.university_year_semester(
        server.the_year, server.the_semester)
    first_semester = semester
    while True:
        link = f'{configuration.server_url}/{year}/{semester}/{server.the_ue}'
        server.the_file.write(f'<a target="_blank" href="{link}">{link}</a> ')
        filename = document.table_filename(year, semester, server.the_ue)
        if os.path.islink(filename):
            table = document.table(year, semester, server.the_ue)
            if server.ticket.user_name in table.masters:
                table.unload(force=True)
                os.unlink(filename)
                table = document.table(year, semester, server.the_ue)
                assert table is not None
                server.the_file.write(server._("TITLE_create_table") + ' !<br>')
            else:
                server.the_file.write(server._("MSG_extension_not_master") + '<br>')
        elif not os.path.exists(filename):
            table = document.table(year, semester, server.the_ue)
            server.the_file.write(f'{server._("TITLE_create_table")}<br>')
        else:
            server.the_file.write(server._("MSG_unextension_yet_done") + '<br>')
        year, semester = utilities.next_year_semester(year, semester)
        if semester == first_semester:
            break # Tested all the semesters of the year

plugin.Plugin('undo_extension', '/{Y}/{S}/{U}/undo_extension', group='staff',
              function=undo_extension,
              )

def bookmark(server):
    """Toggle the bookmarked state of a table"""
    bookmarked = server.ticket.user.manage_key('bookmarked')
    if bookmarked is False:
        bookmarked = []
    else:
        bookmarked = eval(bookmarked)

    bookmark = (server.the_year, server.the_semester, server.the_ue)
    if bookmark in bookmarked:
        bookmarked.remove(bookmark)
    else:
        bookmarked.append(bookmark)
    server.ticket.user.manage_key('bookmarked', content=utilities.stable_repr(bookmarked))
    server.the_file.write(files.files['ok.png'].bytes())

plugin.Plugin('bookmark', '/{Y}/{S}/{U}/bookmark', group='staff',
              function=bookmark, mimetype = 'image/png'
              )


def key_history(server):
    """Store the received data in the key history.
    It is used in 'linear' interface to analyse how it is used."""
    if not server.table_access_allowed():
        return
    utilities.append_file(os.path.join('LOGS', 'key_history'),
                          '%s %d/%s/%s %s\n' % (
        server.ticket, server.the_year, server.the_semester, server.the_ue,
        server.the_path[0]))

plugin.Plugin('key_history', '/{Y}/{S}/{U}/{P}/key_history/{*}',
              function=key_history,
              group='', mimetype = None,
              priority=-7
              )

def end_of_load(server):
    """Keep track of the end of page loading to do stats.
    If the page takes too long to load, then 'newpage.py' assumes
    that something blocked TOMUSS because the page loading does not finish.
    """
    if not server.table_access_allowed():
        return
    dummy_table, page = document.table(server.the_year, server.the_semester,
                                       server.the_ue, server.the_page,
                                       server.ticket)
    page.end_of_load = time.time()
    try:
        server.log_time('page_load_time', start_time=page.start_load)
    except AttributeError: # pragma: no cover
        pass

plugin.Plugin('end_of_load', '/{Y}/{S}/{U}/{P}/end_of_load',
              function=end_of_load,
              group='',
              priority=-7 # Before student_redirect
              )

def getline(server):
    """Get a student line from a table"""
    if not server.table_access_allowed():
        return # pragma: no cover
    table = document.table(server.the_year, server.the_semester,
                           server.the_ue, create=False)
    if not table:
        server.the_file.write('[]')
        return
    lines = tuple(table.get_lines(server.something))
    if lines:
        cells = []
        for column, cell in zip(table.columns, lines[0]):
            if cell.value and (
                column.data_col <= 2 or len(cell.author) != 1
                ) and column.readable_by(server.ticket.user_name):
                cells.append((column.title, cell.value))
        server.the_file.write(json.dumps(cells))
    else:
        server.the_file.write('[]')

plugin.Plugin('getline', '/{Y}/{S}/{U}/getline/{?}',
              function=getline, group='staff',
              mimetype = "application/json; charset=UTF-8",
              launch_thread = True)
