#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2011-2014 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

"""
Use this plugin in order to make TOMUSS send student pictures.
"""

import os
import socket
import io
import time
import PIL.Image
from .. import plugin
from .. import inscrits
from .. import configuration
from .. import display
from .. import files
from .. import utilities

#REDEFINE
# Read/compute the student picture from a database.
# Store it in the filename indicated if it needs to be cached.
# Returns the filename containing the picture to display
# This function is deprecated and should not be used.
def get_the_picture(student_id, filename):
    return filename

#REDEFINE
# Read/compute the student picture from a database.
# It is possible to replace this by a code not storing data in a file.
# The image will be cached in User and not in PICTURES
# So PICTURES will only contains uploaded pictures
def get_the_picture_data(user):
    """Get the JPEG picture"""
    name = os.path.join('PICTURES', user.ID_ + '.JPG')
    if (not os.path.exists(name)
            or time.time() - os.path.getmtime(name) > configuration.picture_ttl):
        # Try to update the cache: store the picture in 'name'
        try:
            name = configuration.get_the_picture(user.ID_, name)
            if not os.path.exists(name):
                name = os.path.join('FILES', 'bug.png')
        except: # pylint: disable=bare-except # pragma: no cover
            name = os.path.join('FILES', 'bug.png')
    return utilities.read_file(name, "bytes")

configuration.get_the_picture = get_the_picture
configuration.get_the_picture_data = get_the_picture_data
configuration.icon_picture_width = 64
configuration.picture_ttl = 24*3600
configuration.picture_extension = '.JPG'

def find_size(i, max_width):
    height = (max_width * i.size[1]) // i.size[0]
    if height > 2 * max_width: # pragma: no cover
        height = 2 * max_width
    if max_width > i.size[0] and height > i.size[1]:
        return i.size[0], i.size[1]
    return max_width, height

def picture(server, user=None, icon=''):
    """Display the picture of a student"""
    # Remove .JPG and get the student number
    if '/' in server.something:
        return # pragma: no cover
    if user is None:
        user = configuration.User(server.something[:-4]) # remove .JPG
    data = user.picture_icon if icon else user.picture
    try:
        server.the_file.write(data)
    except (socket.error, TypeError): # pragma: no cover
        pass

plugin.Plugin('picture', '/picture/{?}',
              function=picture,
              mimetype='image/jpeg',
              group='staff',
              cached = True,
              unsafe=False,
              priority = -10, # Before student_redirection
              launch_thread = True,
              )

plugin.Plugin('picture_icon', '/picture-icon/{?}',
              function=lambda server: picture(server, icon='-'),
              documentation = "The icon sized picture of an user",
              mimetype='image/jpeg',
              group='staff',
              cached = True,
              unsafe=False,
              launch_thread = True,
              )

def my_picture(server, icon=''):
    """Display the connected user picture or a teacher picture"""
    target = server.something[:-4] # Remove .JPG
    if configuration.is_a_student(target):
        picture(server, server.ticket.user, icon)
    else:
        picture(server, configuration.User(target), icon)

plugin.Plugin('my_picture', '/picture/{?}',
              function=my_picture,
              mimetype='image/jpeg',
              group='!staff',
              cached = True,
              unsafe=False,
              launch_thread = "batch",
              priority = -10 # Before student_redirection
              )

plugin.Plugin('my_picture_icon', '/picture-icon/{?}',
              function=lambda server: my_picture(server, icon='-'),
              documentation = "The icon sized picture of the connected user",
              mimetype='image/jpeg',
              group='!staff',
              cached = True,
              unsafe=False,
              launch_thread = "batch",
              priority = -10 # Before student_redirection
             )

# Picture uploading

def uploadable(server, allow_if_teacher=True):
    return (not configuration.regtest or configuration.real_regtest) and (
        configuration.allow_picture_upload
        or (configuration.allow_teacher_picture_upload
            and server.ticket.is_a_teacher
            and allow_if_teacher))

display.Display("PictureUpload", "LinksTable", 20,
                lambda server: (server.is_a_student
                                or (server.the_path[0].startswith(' ')
                                    and configuration.allow_picture_upload)
                ) and uploadable(server,
                                 getattr(server, 'teacher_as_a_student', 0)))

files.files["display.js"].append("picture.py", """
function DisplayPictureUpload(node)
{
  if ( ! node.data )
    return ;
  return hidden_txt(
       '<a href="javascript:picture_upload()">' + _("MSG_picture_upload")
        + '</a>',_("TIP_picture_upload")) ;
}
function picture_upload()
{
   create_popup("picture_upload", '<h1>' + _("MSG_picture_upload") + '</h1>',
                _("MSG_picture_upload_choose") + ' ' + username
                + '<p>' + _("TIP_picture_upload")
                + '<form action="' + add_ticket('my_picture_upload') + '"'
                + ' target="_blank"'
                + ' enctype="multipart/form-data" method="post">'
                + '<input type="file" name="datafile" size="40"><br>'
                + '<input type="submit" value="'
                + encode_value(_("MSG_picture_upload_send")) + '">'
                + '</form>',
                '', false) ;
}
""")

def my_picture_upload(server):
    """Allow an user to upload its picture.
    TOMUSS will store it in the PICTURES directory"""
    if not uploadable(server):
        server.the_file.write(
            server._("config_table_allow_picture_upload")
            + ' : ' + str(configuration.allow_picture_upload)
            + '<p>' + server._("config_table_allow_teacher_picture_upload")
            + ' : ' + str(configuration.allow_teacher_picture_upload)
        )
        return
    posted_data = server.get_posted_data(size=10000000)
    image = posted_data['datafile'][0]
    f = io.BytesIO(image)
    i = PIL.Image.open(f).convert('RGB')

    student_id = inscrits.login_to_student_id(server.ticket.user_name)

    j = i.resize(find_size(i, 4 * configuration.icon_picture_width))
    j.save(os.path.join('PICTURES', student_id + '.JPG'))

    url = configuration.server_url + '/picture'
    server.the_file.write(
        '{}<p><img src="{}"><p>{}<p><img src="{}">'.format(
            server._("MSG_picture_upload_icon"),
            server.ticket.url(url + '-icon/' + server.ticket.user_name + '.JPG'),
            server._("MSG_picture_upload_large"),
            server.ticket.url(url + '/' + server.ticket.user_name + '.JPG'),
        ))

plugin.Plugin('my_picture_upload', '/my_picture_upload',
              function=my_picture_upload,
              launch_thread = True,
              priority = -10, # Before student_redirection
              link=plugin.Link(html_class="verysafe",
                               where="root_rw",
                               url="javascript:picture_upload()"
                               ),
             )
