#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2010,2020 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr
"""
Referent : add or remove refered students
"""

import os
import collections
from .. import plugin
from .. import utilities
from .. import referent
from .. import configuration
from .. import files
from .. import document

files.files['types.js'].append(
    'referent_get',
    utilities.read_file(os.path.join('PLUGINS', 'referent_get.js')))


def replace_table_by_students(codes):
    """Replace a table code by student list :
            UE-XXX     will be replaced by all the registered students of the table
            UE-XXX:A   restricted to the group A
            UE-XXX:A:1 restricted to the group A and sequence 1
    """
    year, semester = configuration.year_semester
    students = []
    for code in codes:
        code, grp, seq = (code + '::').split(':')[:3]
        if not os.path.exists(document.table_filename(year, semester, code)):
            students.append(code)
            continue
        table = document.table(year, semester, code)
        for line in table.lines_of_grp_seq(grp, seq):
            students.append(line[0].value)
    return students

def referent_get_students(server, login, new_students, allow_referent_change=True):
    """Referent 'login' take all the 'new_students'.
    If 'allow_referent_change' is False then the student having yet
    a referent are not modified.
    The old referent receives a mail.
    """
    new_students = replace_table_by_students(new_students)
    current_students = configuration.Users(referent.students_of_a_teacher(login))
    mails_to_old = collections.defaultdict(list)
    for student in new_students:
        if student == '':
            continue
        user = configuration.User(student)
        firstname, surname = (user.firstname, user.surname)
        if firstname is None and surname is None: # pragma: no cover
            server.the_file.write("%s%s\n" % (
                student, server._("MSG_referent_get_not_a_student")))
        elif user.student_id in current_students or student in current_students:
            server.the_file.write("%s%s\n" % (
                student, server._("MSG_referent_get_yet_yours")))
        else:
            old_referent = user.referent
            if old_referent:
                if allow_referent_change:
                    referent.remove_student_from_referent(old_referent, student)
                    mails_to_old[configuration.User(old_referent).mail].append(
                        (student, firstname, surname))
                else:
                    server.the_file.write('%s%s\n' % (
                        student, server._("MSG_referent_get_yet_referent")))
                    continue
            try:
                referent.add_student_to_referent(login, student)
            except: # pragma: no cover # pylint: disable=bare-except
                utilities.send_backtrace('referent get %s' % student)
                server.the_file.write('%s : %s %s %s\n' % (
                    student, firstname, surname,
                    server._("MSG_referent_get_problem")))
                continue
            server.the_file.write('%s : %s %s\n' % (
                student, firstname, surname))
    for mail, students in mails_to_old.items():
        try:
            mail_body = ['\n\n']
            for student, firstname, surname in students:
                mail_body.append(student + ' ' + firstname + ' ' + surname + '\n')
            utilities.send_mail_in_background(
                mail,
                utilities._("MSG_referent_get_subject") % ('...', '', ''),
                utilities._("MSG_referent_get_message")
                % (''.join(mail_body), '', '\n', login),
                frome=configuration.maintainer, show_to=True)
        except:  # pragma: no cover # pylint: disable=bare-except
            utilities.send_backtrace("Sendmail failed")


def referent_get(server):
    """Add a student to its refered students"""
    login = utilities.the_login(server.ticket.user_name)
    server.the_file.write("%s\n" % server._("MSG_referent_get_done"))
    referent_get_students(server, login, server.the_path)

plugin.Plugin('referent_get', '/referent_get/{*}',
              mimetype='text/plain; charset=UTF-8',
              function=referent_get, group='referents')


def referent_set(server, force=False):
    """Add a student to a referent :
            ...../referent_set/user.name/student1/student2/...
    """
    login = utilities.the_login(server.the_path[0])
    server.the_file.write(server._("MSG_referent_get_set_done") % login + "\n")
    referent_get_students(server, login, server.the_path[1:],
                          allow_referent_change=force)


plugin.Plugin('referent_set', '/referent_set/{*}',
              mimetype='text/plain; charset=UTF-8',
              link=plugin.Link(html_class="verysafe", where="referents",
                               url="javascript:go_referent_set()",
                              ),
              function=referent_set, group='referent_masters'
             )

plugin.Plugin('referent_set_force', '/referent_set_force/{*}',
              mimetype='text/plain; charset=UTF-8',
              function=lambda s: referent_set(s, force=True),
              group='referent_masters'
             )

def referent_set_pairs(server):
    """Add lines as:
          teacher1 student1 student2 ...
          teacher2 student90 student91 ...
    """
    data = server.uploaded
    if data is None: # pragma: no cover
        server.the_file.write("BUG")
        return
    server.the_file.write('<pre>')
    for line in data.getfirst('lines').split('\n'):
        line = line.strip().split(' ')
        if len(line) < 2:
            continue
        referent_get_students(server, line[0], line[1:], allow_referent_change=True)

plugin.Plugin('referent_set_pairs', '/referent_set_pairs',
              function=referent_set_pairs,
              link=plugin.Link(html_class="safe", where="referents",
                               url="javascript:go_referent_set_pairs()",
                              ),
              upload_max_size=1000000,
              launch_thread=True,
              group='referent_masters'
             )

def orphan_students(server):
    """Remove the teacher of students"""

    year, semester = configuration.year_semester

    for student in server.the_path:
        if student == '':
            continue
        teacher = referent.referent(year, semester, student)
        if teacher is None:
            server.the_file.write(server._("MSG_referent_get_orphan_none") % student
                                  + "\n")
            continue
        referent.remove_student_from_referent(teacher, student)
        server.the_file.write(student + server._("MSG_referent_get_orphan_quit")
                              + teacher + '\n')

plugin.Plugin('orphan_students', '/orphan_students/{*}',
              mimetype='text/plain; charset=UTF-8',
              link=plugin.Link(html_class="safe", where="referents",
                               url="javascript:go_orphan_students()",
                              ),
              function=orphan_students, group='referent_masters'
             )
