#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet)
#    Copyright (C) 2008-2018 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

import time
import os
from .. import plugin
from .. import utilities
from .. import files
from .. import tablestat
from .. import document
from .. import configuration

try:
    import PIL.Image
    ok = True
except ImportError: # pragma: no cover
    utilities.warn("No Python module PIL")
    ok = False

def student_icone_no_index(server):
    """Generate an icon summarising the student information
    from all the tables."""
    
    login = server.the_student
    
    prst = []
    note = []
    ues = tablestat.the_ues(server.year, server.semester, login)
    if not ues:
        return prst, note
    for t in ues:
        for line in t.get_lines(login):
            prst += t.lines.line_indicator(line, what='Prst')
            note += t.lines.line_indicator(line, what='Note')

    prst.sort()
    prst = [(255 - v[1]*255, v[1]*255, 128) for v in prst]
    
    note.sort()
    note = [(int(255 - v[1]*255), int(v[1]*255), 128) for v in note]

    return prst, note

colors = {
        configuration.pre: (0  , 255,   0),
        configuration.abj: (0  ,   0, 255),
        configuration.ppn: (0  ,   0, 255),
        configuration.abi: (255,   0,   0),
        configuration.tnr: (255,   0, 255),
}

def student_icone_index(server):
    """Generate an icon summarising the student information
    from all the tables."""
    prst = []
    note = []
    content = server.student.manage_key('grades')
    if content:
        content = content.strip('\n').split('\n')
    else:
        return  student_icone_no_index(server)
    y = str(server.year)
    for grade in content:
        _date, year, semester, _ue, grade = grade.split(' ')
        if semester != server.semester or year != y:
            continue # pragma: no cover
        if grade in configuration.allowed_grades:
            grade = configuration.allowed_grades[grade][0] # pragma: no cover
        if grade in colors:
            prst.append(colors[grade])
        else:
            try:
                grade = float(grade)
                note.append((int((1 - grade) * 255),
                            int(grade * 255),
                            128))
            except ValueError: # pragma: no cover
                continue # Not a valid value
    return prst, note

def student_icone(server):
    if not ok: # pragma: no cover
        return

    prst, note = student_icone_index(server)

    if not prst and not note:
        server.the_file.write(files.files['tip.png'].bytes())
        return

    n = max(int(len(prst)**0.5), int(len(note)**0.5), 3) + 1
    s = [(238, 238, 238)] * ((2*n+1) * n)
    for i in range(n):
        s[n+i*(2*n+1)] = (255, 255, 255)

    for i, vv in enumerate(prst):
        s[int(i/n)*(2*n+1) + i%n] = vv

    for i, vv in enumerate(note):
        s[int(i/n)*(2*n+1) + i%n + n+1] = vv

    im = PIL.Image.new("RGB", (2*n+1, n))
    im.putdata(s)
    im.save(server.the_file, 'PNG')

plugin.Plugin('icone', '/{_I}',
              function=student_icone,
              priority=-20,
              mimetype = 'image/png',
              unsafe=False,
              cached = True,
              group='staff',
              launch_thread="batch",
              )

files.add('PLUGINS', 'timeline.js')

def timeline(server):
    """Display a multi year timeline of the student"""
    prefs_table = document.get_preferences(server.ticket.user_name,
                                           create_pref=False,
                                           the_ticket=server.ticket)
    server.the_file.write(
        document.the_head.get_content()
        + document.translations_init(prefs_table['language'])
        + '''<script src="{0}/timeline.js"></script>
        <SCRIPT src="{0}/bilan.js"></SCRIPT>
    <script>'''.format(configuration.url_files))
    content = server.student.manage_key('grades')
    resume = server.student.manage_key('resume')
    if not content:
        content = ""
    student = server.student
    server.the_file.write('timeline({},{},{},{}, {});</script>'.format(
        utilities.js(content),
        configuration.external_bilan(server.the_student),
        resume or "{}",
        utilities.js(server.the_student),
        utilities.js([student.firstname, student.surname, student.mail])
        ))

plugin.Plugin('timeline', '/timeline/{I}', function=timeline, group='staff')

