#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet)
#    Copyright (C) 2008-2012,2021 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

import os
from .. import plugin
from .. import column
from .. import document
from ..cell import CellValue, Line
from .. import configuration
from .. import utilities

def table_statistics(server):
    """Create a table of statistics about all the tables."""
    stats = os.path.join(
        'TMP', 'SCANNER', 'Y' + str(server.year), 'S' + server.semester, 'Table')
    if not os.path.exists(stats): # pragma: no cover
        server.the_file.write(server._("F_empty"))
        return
    with open(stats, 'r') as f: # pragma: no cover
        tables = [utilities.literal_eval(line.strip()) for line in f]
    max_cels = max(t['nr'] for _, t in tables)
    max_cols = max(t['nr_cols'] for _, t in tables)
    max_lines = max(t['nr_lines'] for _, t in tables)
    max_pages = max(t['nr_pages'] for _, t in tables)
    max_comment = max(t['nr_comments'] for _, t in tables)
    max_teachers = max(len(t['teachers']) for _, t in tables)

    columns = [
        column.Column('c0', '', freezed='F', type='Text', width=4,
                      title=server._('COL_TITLE_table'),
                      comment=server._('COL_COMMENT_table'),
                      ),
        column.Column('c1', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_cells_entered'),
                      comment=server._('COL_COMMENT_nb_cells_entered'),
                      minmax="[0;%s]" % max_cels,
                      ),
        column.Column('c2', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_teachers'),
                      comment=server._('COL_COMMENT_nb_teachers'),
                      minmax="[0;%s]" % max_teachers,
                      ),
        column.Column('c3', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_lines'),
                      comment=server._('COL_COMMENT_nb_lines'),
                      minmax="[0;%s]" % max_lines,
                      ),
        column.Column('c4', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_columns'),
                      comment=server._('COL_COMMENT_nb_columns'),
                      minmax="[0;%s]" % max_cols,
                      ),
        column.Column('c5', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_pages'),
                      comment=server._('COL_COMMENT_nb_pages'),
                      minmax="[0;%s]" % max_pages,
                      ),
        column.Column('c6', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_comments'),
                      comment=server._('COL_COMMENT_nb_comments'),
                      minmax="[0;%s]" % max_comment,
                      ),
        column.Column('c7', '', type='Date', width=4,
                      title=server._('COL_TITLE_first_change'),
                      ),
        column.Column('c8', '', type='Date', width=4,
                      title=server._('COL_TITLE_last_change'),
                      ),
        column.Column('c9', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_registered'),
                      comment=server._('COL_COMMENT_nb_registered'),
                      ),
        column.Column('ca', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_unregistered'),
                      comment=server._('COL_COMMENT_nb_unregistered'),
                      ),
        column.Column('cc', '', type='Text', width=4,
                      title=server._('COL_TITLE_empty'),
                      comment=server._('COL_COMMENT_empty'),
                      ),
        column.Column('cd', '', type='Bool', width=2,
                      title=server._('COL_TITLE_nb_grp'),
                      comment=server._('COL_COMMENT_nb_grp'),
                      ),
        column.Column('ce', '', type='Text', width=1,
                      title=server._('COL_TITLE_formula'),
                      comment=server._('COL_COMMENT_formula'),
                      ),
        column.Column('cf', '', type='Bool', width=1,
                      title=server._('COL_TITLE_modifiable'),
                      comment=server._('COL_COMMENT_modifiable'),
                      ),
        column.Column('cg', '', type='Text', width=4,
                      title=server._('COL_TITLE_grade_without_date'),
                      comment=server._('COL_COMMENT_grade_without_date'),
                      ),
        ]
    table_attrs = {
        'comment': server._('LINK_tables'),
        'default_nr_columns': 16
        }

    lines = []
    for code_ue, stats in tables:
        lines.append(Line((
            CellValue(code_ue),
            CellValue(stats['nr']),
            CellValue(len(stats['teachers'])),
            CellValue(stats['nr_lines']),
            CellValue(stats['nr_cols']),
            CellValue(stats['nr_pages']),
            CellValue(stats['nr_comments']),
            CellValue(stats['date_min']),
            CellValue(stats['date_max']),
            CellValue(stats['nr_inscrits']),
            CellValue(stats['nr_not_inscrits']),
            CellValue(stats['empty'][1]),
            CellValue(len([g for g in stats['group_and_seq'] if g != ''])),
            CellValue(stats['problem_in_column_name']),
            CellValue(stats['modifiable'] and configuration.yes or configuration.no),
            CellValue(' '.join(stats['grade_without_date'])),
            )))
    document.virtual_table(server, columns, lines, table_attrs=table_attrs)

plugin.Plugin('tables', '/*2/{=}', function=table_statistics,
              group='staff', launch_thread=True, unsafe=False,
              link=plugin.Link(url="javascript:go_suivi('*2')",
                               where="informations", html_class="verysafe",
                               ),
              )
