#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet)
#    Copyright (C) 2008-2012,2021 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

import os
from .. import plugin
from .. import document
from .. import utilities
from .. import referent
from ..cell import CellValue, Line
from .. import column

def teachers_statistics(server):
    """Create a table of statistics about all the teachers"""
    stats = os.path.join(
        'TMP', 'SCANNER', 'Y' + str(server.year), 'S' + server.semester, 'Teacher')
    if not os.path.exists(stats): # pragma: no cover
        server.the_file.write(server._("F_empty"))
        return
    with open(stats, 'r') as file:
        teachers = [utilities.literal_eval(line.strip()) for line in file]
    teachers = dict(teachers)

    year = utilities.university_year(server.year, server.semester)
    stats = os.path.join(
        'TMP', 'SCANNER', 'Y' + str(year), 'SReferents', 'Teacher')
    if os.path.exists(stats):
        with open(stats, 'r') as file: # pragma: no cover
            for line in file:
                teacher, stats = utilities.literal_eval(line.strip())
                if teacher not in teachers:
                    teachers[teacher] = {}
                teachers[teacher].update(
                    {'nr_blocnote': stats['nr'],
                     'nr_pages': stats['nr_pages'] + teachers[teacher].get('nr_pages', 0),
                    })

    for teacher in referent.referents_login(server.year, server.semester):
        if teacher not in teachers:
            teachers[teacher] = {}
        teachers[teacher]['nr_students'] = len(referent.students_of_a_teacher(teacher))

    max_cels = max([stats.get('nr', 0) for stats in teachers.values()]) or 1
    max_cols = max([stats.get('nr_cols', 0) for stats in teachers.values()]) or 1
    max_pages = max([stats.get('nr_pages', 0) for stats in teachers.values()]) or 1
    max_students = max([stats.get('nr_students', 0) for stats in teachers.values()]) or 1
    max_blocnote = max([stats.get('nr_blocnote', 0) for stats in teachers.values()]) or 1
    max_comment = max([stats.get('nr_comments', 0) for stats in teachers.values()]) or 1

    columns = [
        column.Column('c0', '', freezed='F', type='Text', width=6,
                      title=server._('COL_TITLE_teacher'),
                      comment=server._('COL_COMMENT_teacher'),
                      ),
        column.Column('cfn', '', freezed='F', type='Text', width=4,
                      title=server._('COL_TITLE_firstname'),
                      ),
        column.Column('csn', '', freezed='F', type='Text', width=4,
                      title=server._('COL_TITLE_surname'),
                      ),
        column.Column('c1', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_cells_entered'),
                      comment=server._('COL_COMMENT_nb_cells_entered'),
                      minmax="[0;%s]" % max_cels,
                      ),
        column.Column('c2', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_columns'),
                      comment=server._('COL_COMMENT_nb_columns'),
                      minmax="[0;%s]" % max_cols,
                      ),
        column.Column('c3', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_pages'),
                      comment=server._('COL_COMMENT_nb_pages'),
                      minmax="[0;%s]" % max_pages,
                      ),
        column.Column('c4', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_comments'),
                      comment=server._('COL_COMMENT_nb_comments'),
                      minmax="[0;%s]" % max_comment,
                      ),
        column.Column('c5', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_suivi_students'),
                      comment=server._('COL_COMMENT_nb_suivi_students'),
                      minmax="[0;%s]" % max_students,
                      ),
        column.Column('c6', '', type='Note', width=2,
                      title=server._('COL_TITLE_nb_notepad_cells'),
                      comment=server._('COL_COMMENT_nb_notepad_cells'),
                      minmax="[0;%s]" % max_blocnote,
                      ),
        column.Column('c7', '', type='Date', width=4,
                      title=server._('COL_TITLE_first_change'),
                      ),
        column.Column('c8', '', type='Date', width=4,
                      title=server._('COL_TITLE_last_change'),
                      ),
        ]
    table_attrs = {
        'comment': server._('LINK_teachers'),
        'default_nr_columns': 11
        }

    lines = []

    for name, stats in teachers.items():
        fn_sn = name.split('.')
        if len(fn_sn) == 1:
            fn_sn = ('', fn_sn[0]) # pragma: no cover
        lines.append(Line((
            CellValue(name),
            CellValue(fn_sn[0]),
            CellValue(fn_sn[1]),
            CellValue(stats.get('nr', 0)),
            CellValue(stats.get('nr_cols', 0)),
            CellValue(stats.get('nr_pages', 0)),
            CellValue(stats.get('nr_comments', 0)),
            CellValue(stats.get('nr_students', 0)),
            CellValue(stats.get('nr_blocnote', 0)),
            CellValue(stats.get('date_min', 0)),
            CellValue(stats.get('date_max', 0)),
            )))
    document.virtual_table(server, columns, lines, table_attrs=table_attrs)

plugin.Plugin('teachers', '/*', function=teachers_statistics,
              group='staff', launch_thread=True,
              link=plugin.Link(url="javascript:go_suivi('*')",
                               where="informations", html_class="verysafe",
                               ),
              )
