#!/usr/bin/python3
# TOMUSS: The Online Multi User Simple Spreadsheet
# Copyright (C) 2023 Guillaume FOURCHE, Universite Claude Bernard
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Contact: Thierry.EXCOFFIER@univ-lyon1.fr

"""
Competences Catalog functions and classes
"""

ERR_DUPLICATE = -1
ERR_OBSERVATION = -2
ERR_INVALID = -3

def is_vocabulary(char):
    """Return True if the element is a key letter (a-z, +)"""
    return char == '+' or is_letter(char)
def is_letter(char):
    """Return True if the element is a letter (a-z / A-Z)"""
    # str() is here to fix a RapydScript bug
    is_lowcase = char >= str('a') and char <= str('z')
    is_uppercase = char >= str('A') and char <= str('Z')
    return is_lowcase or is_uppercase or char == '+'
def text_simplify(text):
    """Returns the text without the non letters characters"""
    simple_text = ""
    for char in flat(text):
        if is_letter(char) or char.isdigit() or char == ' ':
            simple_text += char.lower()
    return simple_text
def key_split(key):
    """ Desc:
            Transform a str key to a words list: "m0b10a2" -> [['m', 0], ['b', 10], ['a', 2]]

        Args:
            str: key
        Return:
            list[list[chr, int]]: splitted key
    """
    result = []

    for char in key:
        if char.isdigit() and len(result) > 0:
            if result[-1][1] is None:
                result[-1][1] = int(char)
            else:
                result[-1][1] = 10 * result[-1][1] + int(char)
        elif is_vocabulary(char):
            result.append([char, None])

    return result
def filter_split(range_key):
    """Transform a regular expression to the list of all keys targeted
       "m1pb0,1,2eO" -> [('m', [1]), ('p', []), ('b', [0, 1, 2]), ('e', [0])]

        Args:
            str: regular expression
        Return:
            list[tuple(chr, list[int])]: splitted keys grouped in one list
    """
    result = []
    new_index = None
    for char in range_key:
        if char.isdigit() and len(result) > 0:
            if new_index is None:
                new_index = int(char)
            else:
                new_index = 10 * new_index + int(char)
        elif char == ',':
            indexes.append(new_index)
            new_index = None
        elif is_letter(char):
            if new_index is not None:
                indexes.append(new_index)
                new_index = None
            indexes = []
            result.append((char.lower(), indexes))
    if new_index is not None:
        indexes.append(new_index)

    return result

def key_match_word(key, word):
    """Returns True if the  key is included in the filter"""
    filtre_voc, filtre_indexes = word
    if len(filtre_indexes) == 0:
        for key_word in key:
            if key_word[0] == filtre_voc:
                return True
    else:
        for filtre_voc_f, filtre_index_f in key:
            if filtre_voc == filtre_voc_f:
                for filtre_index in filtre_indexes:
                    if filtre_index == filtre_index_f:
                        return True
    return False

def key_match_filter(key, filter_words):
    """Return True if the key contains all filter words

        Args:
            list[(chr, int)]: key splitted
            list[(chr, list[int])]: a splitted filter
        Returns:
            bool: True if key in filter
    """
    if len(filter_words) == 0:
        return True
    if filter_words[-1][0] == 'o':
        cutted_filter = filter_words[:-1]
    else:
        cutted_filter = filter_words

    for word in cutted_filter:
        if not key_match_word(key, word):
            return False
    return True
def key_last_word(key):
    """Get the last word of a str key: "m0b10a2" -> ['a', 2]

        Args:
            str: key
        Returns:
            tuple(str, index): (voc key, index)
    """
    list_voc = key_split(key)
    if len(list_voc) > 0:
        return list_voc[-1]
    return []
def key_subtract_word(key, words):
    """ Desc:
            Returns the key without the words

        Args:
            str: key
            str: word (b0m10)
        Return:
            str: key without the targeted words
    """
    cutted_key = ""
    targets = key_split(words)
    for voc_key, index in key_split(key):
        not_in = True
        for target_key, target_index in targets:
            if voc_key == target_key and index == target_index:
                not_in = False

        if not_in:
            cutted_key += voc_key + str(index)
    return cutted_key
def choices_filter(choices, key, simplify=False):
    """ Desc:
            find the list of keys having all the words of the expected key
            [p1e1, p1e2, m2p2e1], e1 -> [p1e1, m2p2e1]

        Args:
            list[str]: choices
            str: a key
            bool: if true, the filtre key is deleted to keys results
        Return:
            list[str]: filtred choices
    """
    filtered_choices = []
    words = filter_split(key)
    for choice in choices:
        if key_match_filter(key_split(choice), words):
            if simplify:
                # simplification of choice
                if choice != key:
                    filtered_choices.append(key_subtract_word(choice, key))
            else:
                filtered_choices.append(choice)

    return filtered_choices
def choices_sum(choices):
    """ Desc:
            Return for all choices, the number of keys which contains it
            [p1, e1, p1e1, m2p2e1] -> [p1: 2, e1: 3, p1e1: 1, m2p2e1: 0]

        Args:
            list[str]: choices
        Return:
            dictionnary{str: int}: key -> number of occurrences
    """
    words_usage = {}
    for choice in choices:
        words_usage[choice] = len(choices_filter(choices, choice))

    return words_usage
def string_distance(inputed, expected):
    """ Desc:
            find the textual distance of 2 strings

        Args:
            str: any simplified text
            str: any simplified text
        Return:
            int: textual distance
    """
    current_line = [i for i in range(len(inputed)+1)]
    previous_line = current_line[:]

    for expec_letter in expected:

        current_line[0] += 1
        for i, input_letter in enumerate(inputed):

            # cost action calculation
            if expec_letter == input_letter:
                cost = 0
            else:
                cost = 1

            add_action = current_line[i] + 1
            delete_action = current_line[i+1] + 1
            replace_action = previous_line[i] + cost

            current_line[i+1] = min(add_action, delete_action, replace_action)
        previous_line = current_line[:]

    return current_line[-1]
def text_in_list(text, strings, distance=4): # pragma: no cover
    """Returns True if 'text' is close (<distance) to one of the strings"""
    for string in strings:
        if string_distance(text, string) < distance:
            return True
    return False
def all_texts_in_list(texts, strings, distance=4): # pragma: no cover
    """Returns True if all 'texts' elements are close (<distance) to one of the strings"""
    for text in texts:
        if not text_in_list(text, strings, distance):
            return False
    return True
def extract_rating(description):
    """ Desc:
            Get the rating of the description (0 to 1, -1 if unrated, -2 if no rating)

        Args:
            str: any description
        Return:
            float: rating
    """
    rating = description.split('(')[-1].split(')')[0].split('/')

    if len(rating) == 2 and rating[0].isdigit() and rating[1].isdigit():
        return int(rating[0]) / int(rating[1])
    if len(rating) == 1 and rating[0] == 'NE':
        return -1
    return -2
def get_tree_of(observations):
    """Desc:
            Return the tree of the observations. The words among keys match with branches of competence,
            all dictionaries have a "o" key equal to the observation of the section (default 0).

        Args:
            list : a list of observations [m0b0c0o2, m0b1f5o3, ...]
        Return
            dict : a tree of observations {"o": 0, "m0": {"o": 2, ...}, "m1": {...}, ...}
    """
    root = {}
    observations.sort()
    for obs in observations:
        competence = root
        for voc, ind in key_split(obs):
            if voc == "o":
                competence["o"] = ind
            else:
                full_word = voc + str(ind)
                if full_word not in competence:
                    competence[full_word] = {"o": 0}
                competence = competence[full_word]

    if "o" not in root:
        root["o"] = 0
    return root
def make_array_tree(dictionnary, word = None):
    """ Desc:
            Transform all dictionnaries to arrays. and complete observations with aggregations.

        Args :
            dict : a dict tree made by "get_tree_of"
        Return :
            list : the same tree with ordered structure, aggregations and complete vocabularies
                    ["m0", 0, ["b0", 2, [...], ...], ...]
    """
    if word is not None:
        array = [word, []]
        grade_ind = 1
    else:
        array = [[]]
        grade_ind = 0
    for key in dictionnary:
        if key == "o":
            grade = dictionnary["o"]
            array[grade_ind] = grade
        else:
            array.append(make_array_tree(dictionnary[key], key))

    # Determine all observations of the tree with branches grades
    # observations are used by the sunburst printer
    if len(array) > grade_ind+1:
        sum_grades = 0
        no_grade_count = 0
        for branch in array[(grade_ind+1):]:
            if branch[1] != 0:
                sum_grades += branch[1]
            else:
                no_grade_count +=1 # pragma: no cover
        divide = len(array[(grade_ind+1):]) - no_grade_count
        if divide > 0:
            array[grade_ind] = rint(sum_grades/( len(array[(grade_ind+1):]) - no_grade_count))
        else :
            array[grade_ind] = 0 # pragma: no cover

    return array
def get_arr_tree_leaves(arr_tree, has_name = True):
    """Desc:
            Return the list of all leaves of the array tree, in a tuple value-grade

        Args:
            list : the tree in array form ["value", 0, [...], [...]]
            bool : is false if the tree don't have name
        Return
            list : a list of tuple [("val", 2), ("toto", 4), ...]
    """
    head = 1
    if has_name:
        head = 2
    if len(arr_tree) == head:
        return [arr_tree]
    leaves = []
    for branch in arr_tree[head:]:
        leaves.extend(get_arr_tree_leaves(branch))
    return leaves

# --------------------------------------------------------------------------------------------------
class Catalog:
    """
        vocabulary      : dictionnary{str: list[str]}
        choices         : list[key]
        vocabulary_name : dictionnary{chr: str}
        trees           : dictionnary{str: list[str]}
        trees_ue        : dictionnary{str: list[str]}
        merges          : dictionnary{  'method': 'median'|'average',
                                        'fail': list[key],
                                        'possible_results': list[key]}
        errors          : list[tuple(str, str)]
        observations    : dictionnary{str: (str, str, str, str, str, str)}
                          The key is 'm0' or 'm0b1c1'
    """
    def __init__(self):
        self.vocabulary      = {}
        self.choices         = []
        self.vocabulary_name = {}
        self.trees           = {}
        self.trees_ue        = {}
        self.merges          = {}
        self.errors          = []
        self.observations    = {}

# ----- Initialisation functions -------------------------------------------------------------------
    def parse_lines(self, lines): # pylint: disable=too-many-branches
        """Fill the catalog with a list of lines, one line have one key and one description

            Args:
                list[tuple(str, str)]: a list of lines
            Return:
                bool: True
        """
        last_word_keys = []
        # separate keys, trees and merges
        for line in lines:
            key = str(line[0])
            description = str(line[1])
            if len(key) == 0 and len(description) == 0: # empty line verification:
                pass # pragma: no cover
            elif len(key) <= 1 or len(description) == 0:
                self.errors.append(_("ALERT_comp_unvalid_line") + key + " \t " + description)
            elif key.startswith("^"): # DEPRECATED
                if key[1:] not in self.trees:
                    self.trees[key[1:]] = description.split(' ')
            elif key.startswith(":"): # DEPRECATED
                if key[1:] not in self.merges:
                    self.merges[key[1:]] = parseJSON(description)
            elif key.startswith("#"):
                self.update_vocabulary(key[1], description)
            else: # vocabulary / choices
                is_valid = True
                for word in key_split(key):
                    if word[1] is None:
                        is_valid = False # pragma: no cover
                if is_valid:
                    if key not in self.choices:
                        last_word_keys.append([key_split(key)[-1], description])
                        self.choices.append(key)
                else: # pragma: no cover
                    self.errors.append(_("ALERT_comp_unvalid_line") + key + " \t " + description)
                if len(line) > 2 and line[2]:
                    # Disciplines using this competence subtree
                    disciplines = '(' + '|'.join(REsplit(' +', line[2].strip())) + ').*'
                    if disciplines not in self.trees:
                        self.trees[disciplines] = [key]
                    else:
                        self.trees[disciplines].append(key)
                if len(line) > 3 and line[3]:
                    # Disciplines using this competence subtree
                    disciplines = '(' + '|'.join(REsplit(' +', line[3].strip())) + ').*'
                    if disciplines not in self.trees_ue:
                        self.trees_ue[disciplines] = [key]
                    else:
                        # XXX tests.py must cover this case. Why not ?
                        self.trees_ue[disciplines].append(key) # pragma: no cover
                if len(line) > 4 and line[4]:
                    self.merges[key] = parseJSON(line[4])
                if len(line) > 5 and line[5]:
                    self.observations[key] = parseJSON(line[5])
                    if len(self.observations[key]) != 6:
                        self.errors.append(_("ALERT_comp_unvalid_line") + key + " \t " + line[5])

        self.choices.sort()
        self.errors.sort()
        last_word_keys.sort()

        # vocabulary table filling
        for key, description in last_word_keys:
            voc_key = key[0]
            if voc_key not in self.vocabulary:
                self.vocabulary[voc_key] = []

            while key[1] >= len(self.vocabulary[voc_key]):
                self.vocabulary[voc_key].append('')

            self.vocabulary[voc_key][key[1]] = description

        return True
    def parse_dict(self, dict):
        """ Desc:
                fill the catalog with a catalog_dictionnary

            Args:
                str: a dictionary with catalog structure
            Return:
                bool: True
        """
        self.vocabulary = dict["vocabulary"]
        self.choices = dict["choices"]
        self.vocabulary_name = dict["vocabulary_name"]
        self.trees = dict["trees"]
        self.trees_ue = dict["trees_ue"]
        self.merges = dict["merges"]
        self.errors = dict["errors"]
        self.observations = dict["observations"]

    def parse_json(self, json):
        """ Desc:
                fill the catalog with a json file string

            Args:
                str: a catalog json file
            Return:
                bool: True
        """
        self.parse_dict(parseJSON(json))
        return True

# ----- Set functions ------------------------------------------------------------------------------
    def choices_add(self, key):
        """ Desc:
                Add a key to the choices list and sort the list

            Args:
                str: any key
            Return:
                bool: True if key added
        """
        if key in self.choices:
            return False

        self.choices.append(key)
        self.choices.sort()
        return True
    def choices_del(self, key):
        """Delete a key of the choices list

            Args:
                str: key in the choices list
            Return:
                bool: True if the key is ine the list and correctly deleted
        """
        for i, choice in enumerate(self.choices):
            if choice == key:
                del self.choices[i]
                return True
        return False
    def get_index_from_description(self, voc_key, description):
        """Returns:
                * None if no matching description
                * or the index in the vocabulary
                * or negative error code, see "MSG_comp_error-..." in .po file

            Args:
                str: any existing vocabulary key
                str: any description
        """
        if len(description) == 0:
            return ERR_INVALID

        simplified_description = text_simplify(description)
        if voc_key in self.vocabulary_name:
            if text_simplify(self.vocabulary_name[voc_key]) == simplified_description:
                return ERR_DUPLICATE # pragma: no cover

        if voc_key in self.vocabulary:
            for i, voc_desc in enumerate(self.vocabulary[voc_key]):
                if text_simplify(voc_desc) == simplified_description:
                    return i

        # descriptions in "o" vocabulary needs to contain a rating
        if voc_key == 'o' and extract_rating(description) == -2:
            return ERR_OBSERVATION
        return None
    def add_description(self, voc_key, description):
        """Add a description to the vocabulary dictionnary and give the index

            Args:
                str: key in vocabulary
                str: any description text
            Return:
                int: the index off the new description
                     if negative error code, see "MSG_comp_error-..." in .po file
        """
        index = self.get_index_from_description(voc_key, description)
        if index is not None:
            if index >= 0:
                index = ERR_DUPLICATE
            return index
        voc = self.vocabulary[voc_key]
        voc.append(description)
        return len(voc) - 1
    def replace_description(self, voc_key, desc_index, description):
        """ Desc:
                Replace a description by another

            Args:
                chr: key in vocabulary
                int: index of description
                str: any description text
            Return:
                int: 0 if replacing was done or -1, -2 if duplicate/invalid description
        """
        index = self.get_index_from_description(voc_key, description)
        if index is None or index == desc_index:
            self.vocabulary[voc_key][desc_index] = description
            return 0
        if index < 0:
            return index # pragma: no cover
        return ERR_DUPLICATE
    def update_vocabulary(self, voc_key, description):
        """ Desc:
                Add a letter in the vocabulary, if the vocabulary exist, replace his description

            Args:
                chr: any vocabulary key
                str: any description text
            Return:
                int: 1 if added, 0 if replaced or negative if error
        """
        if len(voc_key) != 1 or not is_letter(voc_key) or len(description) == 0:
            return ERR_INVALID

        if voc_key in self.vocabulary_name:
            self.vocabulary_name[voc_key] = description
            return 0
        self.vocabulary[voc_key] = []
        self.vocabulary_name[voc_key] = description
        return 1

# ----- Get functions ------------------------------------------------------------------------------
    def get_formation_vocabs(self, form_key):
        """ Desc:
                Return all choices of the catalog in a formation expected with key

            Arg:
                tuple(str, int): a splited word that define the local formation to describe
            Return:
                list[str]: list of keys
        """
        keys = []
        for choice in self.choices:
            key = key_split(choice)
            if key[0] == form_key :
                keys.append(choice)
        return keys
    def get_key(self, key):
        """ Desc:
                Find all descriptions referenced in a str key,
                with the vocabulary key and the index

            Args:
                str: a key
            Return:
                list[list[chr, int, str]]: list of (voc_key, index, description)
        """
        return [
            [voc_key, desc_index, self.vocabulary[voc_key][desc_index]]
            for voc_key, desc_index in key_split(key)
            ]
    def filter_is_valid(self, key_range):
        """ Desc:
                return false if the key_range is not valid for this catalog

            Args:
                list[(chr, list[int])]: a key_range (-> use filter_split)
            Return:
                bool: True if the key_range is valid
        """
        is_valid = True
        for voc_key, indexes in key_range:
            if voc_key != 'o':
                if voc_key not in self.vocabulary:
                    self.errors.append(_("ALERT_comp_voc_not_in_catalog_1") + voc_key
                        + _("ALERT_comp_voc_not_in_catalog_2") + str(self.vocabulary))
                    is_valid = False
                elif len(indexes) > 0 and len(self.vocabulary[voc_key]) <= max(indexes): # pragma: no cover
                    self.errors.append(_("ALERT_comp_key_not_in_catalog_1")
                        + voc_key+str(max(indexes))
                        + _("ALERT_comp_key_not_in_catalog_2") + str(self.vocabulary[voc_key]))
                    is_valid = False
        return is_valid

    def get_allowed_keys_real_add(self, formation_filter, result):
        if formation_filter == '':
            return # pragma: no cover

        filters = []
        # filters structure : [filter, filter, ...]
        #   or [ filter[(voc_key, [index 1, index 2, ...]), ..., ('o': [...])], ...]
        for reg_key in formation_filter:
            key_range = filter_split(reg_key)
            if self.filter_is_valid(key_range):
                filters.append(key_range)

        for choice in self.choices:
            split_choice = key_split(choice)
            for key_range in filters:
                if key_match_filter(split_choice, key_range) and split_choice[-1][0] != 'o':
                    if key_range[-1][0] != 'o' or len(key_range[-1][1]) == 0:
                        result[choice] = ['o'+str(index) for index in range(len(self.vocabulary['o']))]
                    else:
                        result[choice] = ['o'+str(index) for index in key_range[-1][1]]

    def get_allowed_keys_real(self, comp_ue, trees):
        """ Desc:
                Return a dictionnary of choices filtred by the regular expression. The value for all
                choices is a list of possible observations (only 1 if the rating is defined)

            Args:
                str: a regular expression
            Return:
                dict[str: list[str]]: list of keys with her observations
        """
        result = {}
        for reg_exp in trees:
            if regexp_match(reg_exp, comp_ue):
                self.get_allowed_keys_real_add(trees[reg_exp], result)

        return result
    def get_allowed_keys(self, comp_ue):
        """Keys allowed for formation"""
        return self.get_allowed_keys_real(comp_ue, self.trees)
    def get_allowed_keys_ue(self, comp_ue): # pragma: no cover
        """Keys allowed for UE"""
        return self.get_allowed_keys_real(comp_ue, self.trees_ue)
    def get_allowed_keys_observations(self, key): # pragma: no cover
        """Observation defined for the UE"""
        result = None
        key_length = 0
        for starts_by in self.observations:
            if key.startswith(starts_by) and len(starts_by) > key_length:
                result = self.observations[starts_by]
                key_length = len(starts_by)
        return result or ('NE (NE)', 'No (5/20)', 'Bad (5/20)', 'Ok (10/20)', 'Good (15/20)', 'Perfect (20/20)')
    def get_lines(self, key_filter):
        """ Desc:
                Return lines having key with all words in the expected key

            Args:
                list[tuple(str, str)]: list of lines with (key, description)
                str: a key
            Return:
                list[tuple(str, str)]: list of filtered lines
        """
        lines = []
        for choice in choices_filter(self.choices, key_filter):
            last_word = key_last_word(choice)
            lines.append([choice, self.vocabulary[last_word[0]][last_word[1]]])
        return lines
    def index_rate(self, choices, rating):
        """Find the index of the closest rating among choices

            Args:
                list[str]: a list of keys with the last word 'o'
                float: rating between 0 and 1
            Return:
                int: index among those of the 'o' vocabulary
        """
        best = 1
        index = 0
        for choice in choices:
            voc_index = key_last_word(choice)[1]
            ch_rating = extract_rating(self.vocabulary['o'][voc_index])
            distance = abs(ch_rating - rating)
            if distance <= best:
                index = voc_index
                best = distance
        return index
    def get_completion(self, voc_key, description):
        """ Desc:
                Find for a vocabulary key and a inputed description
                the most similar description in the vocabulary

            Args:
                str: a key in the vocabulary
                str: any text
            Return:
                list[list[int, int, str]]: list of (textual distance, index, description)
        """
        if voc_key not in self.vocabulary:
            return []
        response_list = [
            # distance add 10 000 for sorting the list
            [10000 + string_distance(description, vocabulary), i, vocabulary]
            for i, vocabulary in enumerate(self.vocabulary[voc_key])]
        response_list.sort()
        return response_list
    def aggregate(self, dictionnary, param): # pylint: disable=too-many-branches
        """ Desc:
                Aggregate all results key in one,
                and calculate the observation with a choosed parameter (average or median)

                BEWARE the dictionnary can be a list if there are no weights.

            Args:
                dictionnary{str: int}: dict{key: weight} or list[key (default weight = 1)]
                dictionnary{str: str}: {'method': 'average'|'median', 'fail': list[key],
                                        'possible_results': list['o' word]}
            Return:
                str: aggregate key with the observation word
        """
        # Filling key words and ratings analysis
        aggregation = ''
        ratings = []        #rating -> [(rating, weight), (...),  ...]
        if hasattr(dictionnary, 'append'): # So it is a list
            real_dict = {}
            for key in dictionnary:
                real_dict[key] = 1
            dictionnary = real_dict

        if len(dictionnary) == 0:
            return "o0" # XXX A revoir

        sum_weight = 0
        for key in dictionnary:
            o_voc, o_index = key_last_word(key)
            if o_voc != 'o':
                raise ValueError('bug') # pragma: no cover
            result = extract_rating(self.vocabulary['o'][o_index])
            if result != -1:
                weight = dictionnary[key]
                sum_weight += weight
                ratings.append((result, weight))

        aggregation = key_subtract_word(key, 'o0') # XXX A revoir
        for key in dictionnary:
            # Remove what is not present     A - (A - W) == A & W
            aggregation = key_subtract_word(aggregation, key_subtract_word(aggregation, key))

        if sum_weight == 0:
            o = -1
        else:
            # fail ?
            for key in dictionnary:
                rating = key_last_word(key)
                if rating[0] + str(rating[1]) in param['fail']:
                    return aggregation + "o1" # XXX A revoir
            o = 0
            if param['method'] == 'median':
                # median
                ratings.sort()
                tot_coef = 0
                for result, weight in ratings:
                    if tot_coef <= sum_weight / 2:
                        o = result
                    tot_coef += weight
            else:
                # average
                for result, weight in ratings:
                    o += result*weight
                o = o / sum_weight

        return aggregation + "o" + str(self.index_rate(param['possible_results'], o))
    def get_json(self):
        """Return the json string of the catalog"""
        return dumpJSON({'vocabulary': self.vocabulary,
                         'choices': self.choices,
                         'vocabulary_name': self.vocabulary_name,
                         'trees': self.trees,
                         'trees_ue': self.trees_ue,
                         'merges': self.merges,
                         'errors': self.errors,
                         'observations': self.observations,
                         })
    def __str__(self):
        return self.get_json() # pragma: no cover
