#!/usr/bin/python3
# -*- coding: utf-8 -*-
# TOMUSS: The Online Multi User Simple Spreadsheet
# Copyright (C) 2013-2016 Thierry EXCOFFIER, Universite Claude Bernard
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Contact: Thierry.EXCOFFIER@univ-lyon1.fr

# In order to translate to pure JavaScript:
#    Define classes before using them
#    No else: after for:
#    No format or %

# Set to True if 'B' filter must not found 'b' value.
# This value must never be changed once tables have been created by users
contextual_case_sensitive = False

nan = float('NaN')

# Dates are defined as string : YYYYMMDDHHMMSS
# We assume that date are all valid.
# YYYYMM = YYYYMM00000000
# User dates:
#    * Day (current month)
#    * Day/Month (current year)
#    * Day/Month/Year
#    * Date[-_ ] Hour
#    * Date[-_ ] Hour:Minute
#    * Date[-_ ] Hour:Minute:Seconds

def two_digits(value):
    if value >= 10:
        return str(value)
    return '0' + str(value)

letter_to_duration = {
    'y': 365*24,
    'm':  30*24,
    'w':   7*24,
    'd':     24,
    'h':      1,
    "'":   1/60.
    }
# French translations
letter_to_duration['a'] = letter_to_duration['y']
letter_to_duration['s'] = letter_to_duration['w']
letter_to_duration['j'] = letter_to_duration['d']

def is_a_relative_date(txt):
    return txt != '' and txt[-1] in letter_to_duration

def get_relative_span(txt):
    try:
        nb = float(txt[:-1])
    except: # pylint: disable=bare-except
        return -1
    return nb * letter_to_duration[txt[-1]] * 60*60

def get_relative_date(txt):
    span = get_relative_span(txt)
    if span < 0:
        return LAST_DATE
    return seconds_to_date(millisec()/1000 - span)

def user_date_to_date(txt):
    if txt == '':
        # Today
        year_month_day = localtime()
        return (str(year_month_day[0])
                + two_digits(year_month_day[1])
                + two_digits(year_month_day[2])
                )
    txt = str(txt)
    if is_a_relative_date(txt):
        return get_relative_date(txt)

    txt = REsplit("[- _]", txt)
    t = ''
    if len(txt) == 2:
        hour = txt[1].split(':')
        for hms in hour:
            if hms == '':
                break # pragma: no cover
            try:
                t += two_digits(int(hms))
            except: # pylint: disable=bare-except # pragma: no cover
                break
    invalid = REsplit('[0-9/]+', txt[0])
    if invalid != ['', '']:
        return '9999' # Invalid character in the date

    year_month_day = localtime()
    txt = txt[0].split('/')
    try:
        the_day = int(txt[0])
    except: # pylint: disable=bare-except
        return '9999'
    if isNaN(the_day):
        return '9999' # pragma: no cover
    the_month = year_month_day[1]
    if len(txt) >= 2:
        try:
            the_month = int(txt[1])
        except: # pylint: disable=bare-except  # pragma: no cover
            pass
    the_year = year_month_day[0]
    if len(txt) >= 3:
        try:
            the_year = int(txt[2])
        except: # pylint: disable=bare-except  # pragma: no cover
            pass
    return str(the_year) + two_digits(the_month) + two_digits(the_day) + t

class FilterContext:
    def __init__(self, line, cell, username, is_a_teacher):
        self.line = line
        self.cell = cell
        self.username = username
        self.is_a_teacher = is_a_teacher

class CellAttr:
    def __init__(self):
        pass
    def is_empty(self):
        return False
class CellAttrCell(CellAttr):
    def __init__(self, what):
        self.what = what
    def fct(self, fc):
        return getattr(fc.cell, self.what)
    def js(self):
        return 'cell.' + self.what
    def editor(self):
        return _('F_cell_' + self.what)
class CellAttrOther(CellAttr):
    def __init__(self, data_col, what):
        self.what = what
        self.data_col = data_col
    def fct(self, fc):
        return getattr(fc.line[self.data_col], self.what)
    def js(self):
        return 'line[' + str(self.data_col) + '].' + self.what
    def editor(self):
        return _('F_col_' + self.what) + ' «' + html(columns[self.data_col].title) + '»'

class CellAttrConst(CellAttr):
    def __init__(self, value):
        self.value = value
    def fct(self, _fc):
        return self.value
    def js(self):
        return js_str(self.value)
    def editor(self):
        if str(self.value) == '':
            return _("F_empty")
        return '«' + html(self.value) + '»'
    def is_empty(self):
        return self.value == ''
class CellAttrConstDate(CellAttrConst):
    def fct(self, _fc):
        return get_relative_date(self.value)
    def js(self):
        return "get_relative_date(" + js_str(self.value) + ")"
    def editor(self):
        return html(self.value[:-1]) + ' ' + _('F_' + self.value[-1])

class CellAttrConstFixedDate(CellAttrConst):
    def editor(self):
        f = "%a %d/%m/%Y"
        if len(self.value) >= 10:
            f += ' %H'
        if len(self.value) >= 12:
            f += ':%M'
        if len(self.value) >= 14:
            f += ':%S'
        return get_date_tomuss(self.value).formate(f)

class CellAttrConstSeconds(CellAttr):
    def fct(self, _fc):
        return seconds_to_date()
    def js(self):
        return "seconds_to_date()"
    def editor(self):
        return _('F_date')

class CellAttrConstUserName(CellAttr):
    def fct(self, fc):
        return fc.username
    def js(self):
        return "(window.display_data && is_a_teacher === 0 ? display_data['Login'] : username)"
    def editor(self):
        return _('F_login')

class CellAttrAsDate(CellAttr):
    def __init__(self, cellattr):
        self.cellattr = cellattr
    def fct(self, fc):
        return user_date_to_date(self.cellattr.fct(fc))
    def js(self):
        return "user_date_to_date(" + self.cellattr.js() + ')'
    def editor(self):
        return self.cellattr.editor()
class CellAttrAsFloat(CellAttrAsDate):
    def fct(self, fc):
        return to_float_or_nan(self.cellattr.fct(fc))
    def js(self):
        return "to_float_or_nan(" + self.cellattr.js() + ')'
class CellAttrAsString(CellAttrAsDate):
    def fct(self, fc):
        return str(self.cellattr.fct(fc))
    def js(self):
        return self.cellattr.js() + '.toString()'
class CellAttrAsLower(CellAttrAsDate):
    def fct(self, fc):
        return str(self.cellattr.fct(fc)).lower()
    def js(self):
        return self.cellattr.js() + '.toString().toLowerCase()'
class CellAttrAsFlat(CellAttrAsDate):
    def fct(self, fc):
        return flat(self.cellattr.fct(fc))
    def js(self):
        return "flat(" + self.cellattr.js() + ')'
class CellAttrAsFixed(CellAttrAsDate):
    def fct(self, fc):
        return as_fixed(self.cellattr.fct(fc))
    def js(self):
        return "as_fixed(" + self.cellattr.js() + ")"
class CellAttrTruncate(CellAttr):
    def __init__(self, cellattr, length):
        self.cellattr = cellattr
        self.length = length
    def fct(self, fc):
        return self.cellattr.fct(fc)[:self.length]
    def js(self):
        return self.cellattr.js() + ".substr(0," + str(self.length) + ')'
    def editor(self):
        return self.cellattr.editor()
class CellAttrAppend(CellAttr):
    def __init__(self, cellattr, value):
        self.cellattr = cellattr
        self.value = value
    def fct(self, fc):
        return self.cellattr.fct(fc) + self.value
    def js(self):
        return '(' + self.cellattr.js() + '+' + js_str(self.value) + ')'

class FilterCommon:
    def __init__(self): # Needed for RapydScript
        pass
    data_col_left = None
    data_col_right = None

FIRST_DATE = '0000'
LAST_DATE = '9999'

def nice_span_date(date, end):
    start = get_date_tomuss((date + '000000')[:14])
    if end:
        if len(date) == 8:
            date_add(start, 24 * 60 * 60)
        elif len(date) == 10:
            date_add(start, 60 * 60)
        elif len(date) == 12:
            date_add(start, 60)
    return start.formate('%a %d/%m/%Y %H:%M:%S')

class Dates:
    def __init__(self, dates):
        self.dates = dates
        self.cleanup()
    def cleanup(self):
        # ]A, A[ → ø
        for date in self.dates[:]:
            if not date[1] and date[0] == date[2] and not date[3]:
                self.dates.remove(date) # pragma: no cover
        if len(self.dates) <= 1:
            return
        dates = [self.dates[0]]
        for date in self.dates[1:]:
            start, include_start, end, include_end = date
            last_end = dates[-1][2]
            if start > last_end:
                # [2, 4] [6, 8] → [2, 4] [6, 8]
                dates.append(date)
            elif start == last_end:
                if include_start or dates[-1][3]:
                    # [2, 4] [4, 6] → [2, 6]
                    dates[-1][2] = end
                    dates[-1][3] = include_end
                else:
                    # [2, 4[ ]4, 6] → [2, 4[ ]4, 6]
                    dates.append(date)
            else:
                if start == dates[-1][0] and not dates[-1][1]:
                    # ]2, 5[ [2, 4] → [2, 5[
                    dates[-1][1] = include_start
                if end == last_end and not dates[-1][3]:
                    # ]2, 5[ [3, 5] → ]2, 5]
                    dates[-1][3] = include_end
                if end > last_end:
                    # ]2, 5[ [3, 6] → ]2, 6]
                    dates[-1][2] = end
                    dates[-1][3] = include_end
                # else:
                #   [2, 5[ ]3, 4] → [2, 5[
        self.dates = dates

    def negate(self):
        # [A, B] → ]0000, A[ + ]B, 9999[
        # [A, B] [C, D] → ]0000, A[ + ]B, C[ + ]D, 9999[
        if not self.dates:
            return ALL_DATES.clone()
        dates = []
        last_end = FIRST_DATE
        last_include_end = True
        for start, include_start, end, include_end in self.dates:
            if start != FIRST_DATE:
                dates.append([last_end, not last_include_end, start, not include_start])
            last_end = end
            last_include_end = include_end
        if last_end != LAST_DATE:
            dates.append([last_end, not last_include_end, LAST_DATE, False])
        return Dates(dates)

    def add(self, dates):
        for date in dates.dates:
            self.dates.append(date)
        self.dates.sort()
        self.cleanup()

    def sub_date(self, start, include_start, end, include_end):
        dates = []
        for date in self.dates:
            if date[2] < start:
                # [2, 5] - [6, 9] : [2, 5]
                dates.append(date)
            elif date[0] > end:
                # [2, 5] - [0, 1] : [2, 5]
                dates.append(date)
            elif date[0] == end:
                # [2, 5] - [0, 2] : ]2, 5]
                dates.append([date[0], False, date[2], date[3]])
            elif date[2] == start:
                if date[3] and include_start:
                    # [2, 5] - [5, 9] : [2, 5[
                    date[3] = False
                dates.append(date)
            elif date[0] >= start:
                if date[0] == start and date[1] and not include_start:
                    # [2, x] - ]2, y] : [2, 2]
                    dates.append([start, True, start, True])
                if end < date[2]:
                    # [2, 5] - [1, 4] : ]4, 5]
                    dates.append([end, not include_end, date[2], date[3]])
                elif end == date[2]:
                    if not include_end and date[3]:
                        # [2, 5] - [1, 5[ : [5, 5]
                        dates.append([end, True, end, True]) # pragma: no cover
                else:
                    # [2, 5] - [1, 6] : nothing
                    pass # pragma: no cover
            else:
                dates.append([date[0], date[1], start, not include_start])
                if end < date[2]:
                    # [2, 5] - [3, 4] : [2, 3[ + ]4, 5]
                    dates.append([end, not include_end, date[2], date[3]]) # pragma: no cover
                elif end == date[2]:
                    if not include_end and date[3]:
                        # [2, 5] - [3, 5[ : [2, 3[ + [5, 5]
                        dates.append([end, True, end, True]) # pragma: no cover
                else:
                    # [2, 5] - [3, 6] : nothing more
                    pass
        self.dates = dates

    def sub(self, dates):
        for start, include_start, end, include_end in dates.dates:
            self.sub_date(start, include_start, end, include_end)

    def intersection(self, dates):
        self.sub(dates.negate())

    def clone(self):
        # Deep copy because many changes in place
        return Dates([date[:] for date in self.dates])

    def forever(self):
        if len(self.dates) != 1:
            return False
        return self.dates[0][0] == FIRST_DATE and self.dates[0][2] == LAST_DATE

    def __str__(self):
        texts = []
        for start, include_start, end, include_end in self.dates:
            if start == FIRST_DATE:
                s = ''
            else:
                if include_start:
                    s = nice_span_date(start, 0)
                else:
                    s = nice_span_date(start, 1)
            s += ' → '
            if end != LAST_DATE:
                if include_end:
                    s += nice_span_date(end, 1)
                else:
                    s += nice_span_date(end, 0)
            texts.append(s)
        return '\n'.join(texts)





ALL_DATES = Dates([[FIRST_DATE, False, LAST_DATE, False]])
NO_DATES = Dates([])

class FilterNegate(FilterCommon):
    # The negation node in the filter tree"""
    def __init__(self, node):
        self.node = node
        self.data_col_left = node.data_col_left
        self.data_col_right = node.data_col_right
    def evaluate(self, fc):
        return not self.node.evaluate(fc)
    def js(self):
        return '!' + self.node.js()
    def editor(self):
        negation = self.node.editor_negate()
        if negation:
            return negation
        return _('F_!') + ' ' + self.node.editor()
    def dates(self, fc):
        return self.node.dates(fc).negate()

class FilterTrue(FilterCommon):
    # True
    def evaluate(self, fc):
        return True
    def js(self):
        return 'true'
    def editor(self):
        return _('F_true')
    def dates(self, _fc):
        return ALL_DATES.clone()

class FilterFalse(FilterCommon):
    # False
    def evaluate(self, fc):
        return False
    def js(self):
        return 'false'
    def editor(self):
        return _('F_false')
    def dates(self, _fc):
        return NO_DATES.clone()

def search_operator(string):
    for operator in filterOperators:
        if string.startswith(operator[0]):
            return string[len(operator[0]):], operator
    raise ValueError('bug') # pragma: no cover

class FilterAny(FilterCommon):
    operator = None
    def evaluate(self, fc):
        return self.python(self.left.fct(fc), self.right.fct(fc))

    def dates(self, fc):
        if self.evaluate(fc):
            return ALL_DATES.clone()
        return NO_DATES.clone()

    def js(self):
        return self.js_string(self.left.js(), self.right.js())

    def editor(self):
        operator = self.operator[0] or 'start'
        operator_help = _('F_' + operator)
        if getattr(self, 'relative_date', False):
            if operator in ('<', '<=', '>', '>='):
                operator_help = _('F_?' + operator)
        text = self.left.editor() + ' ' + operator_help + ' ' + self.right.editor()
        if self.right.is_empty():
            if operator == 'start' or operator == '~':
                return text + ': <b class="attribute_error">' + _('F_true') + '</b>'
            elif operator == '=':
                return self.left.editor() + ' ' + _('F_is_empty')
        return text

    def editor_negate(self):
        if not self.operator:
            return False
        if self.operator[0] == '=':
            if self.right.is_empty():
                return self.left.editor() + ' ' + _('!F_is_empty')
            return self.left.editor() + ' ' + _('!F_=') + ' ' + self.right.editor()
        if self.operator[0] == '':
            text = self.left.editor() + ' ' + _('!F_start') + ' ' + self.right.editor()
            if self.right.is_empty():
                return text +  ': <b class="attribute_error">' + _('F_false') + '</b>'
            return text
        if self.operator[0] == '~':
            text = self.left.editor() + ' ' + _('!F_~') + ' ' + self.right.editor()
            if self.right.is_empty():
                return text +  ': <b class="attribute_error">' + _('F_false') + '</b>'
            return text
        return False

class FilterDate(FilterAny):
    def __init__(self, operator, what, value, column_type, left):
        if operator[0] == '':
            dummy, operator = search_operator('=')
        self.left = left
        if what != "date":
            self.left = CellAttrAsDate(self.left)
        if is_a_relative_date(value):
            dummy, operator = search_operator(operator[3])
            self.right = CellAttrConstDate(value)
            self.relative_date = True
        else:
            value = user_date_to_date(value)
            self.left = CellAttrTruncate(self.left, len(value))
            self.right = CellAttrConstFixedDate(value)
        self.operator = operator

    def dates(self, fc):
        # print('====', self, self.left.fct(fc), self.operator[0], self.right.fct(fc), self.left, self.right)
        operator = None
        if isinstance(self.left, CellAttrTruncate):
            #print('=======', self.left.cellattr.__class__.__name__)
            if isinstance(self.left.cellattr, CellAttrConstSeconds):
                date = self.right.fct(fc)
                operator = self.operator[0]
        elif isinstance(self.left, CellAttrConstSeconds):
            date = self.right.fct(fc)
            operator = self.operator[0]
        elif isinstance(self.right, CellAttrConstDate):
            left = get_date_tomuss(self.left.fct(fc)).getTime()/1000
            span = get_relative_span(self.right.value)
            date = seconds_to_date(left + span)
            operator = self.operator[3]
        elif isinstance(self.right, CellAttrConstSeconds):
            date = self.left.fct(fc)
            operator = self.operator[3]
        if operator is None:
            if self.evaluate(fc):
                return ALL_DATES.clone()
            return NO_DATES.clone()
        if operator == '<':
            return Dates([[FIRST_DATE, False, date, False]])
        if operator == '<=':
            return Dates([[FIRST_DATE, False, date, True]])
        if operator == '>':
            return Dates([[date, False, LAST_DATE, False]])
        if operator == '>=':
            return Dates([[date, True, LAST_DATE, False]])
        if operator == '=':
            return Dates([[date, True, date, True]])
        bug # pragma: no cover
class FilterFloat(FilterAny):
    def __init__(self, operator, value, column_type, left):
        self.left = CellAttrAsFloat(left)
        self.right = CellAttrConst(value)

class FilterStr(FilterAny):
    def __init__(self, operator, value, column_type, left):
        self.left = CellAttrAsString(left)
        value_lower = value.lower()
        hide_upper = value_lower == value  or  not contextual_case_sensitive
        if hide_upper:
            value = value_lower
            self.left = CellAttrAsLower(self.left)

        value_flat = flat(value)
        hide_diacritics = (value_flat == value
                           or operator[0] == '<'
                           or operator[0] == '>'
                           )
        if hide_diacritics:
            value = value_flat
            self.left = CellAttrAsFlat(self.left)

        self.right = CellAttrConst(value)

class FilterAnyStr(FilterAny):
    def __init__(self, operator, what_right, data_col, value, column_type,left):
        if data_col >= 0:
            self.right = CellAttrAsString(CellAttrOther(data_col, what_right))
        else:
            if what_right == 'author':
                self.right = CellAttrConstUserName()
            else:
                # Undefined behavior
                self.right = CellAttrConst('') # pragma: no cover
        if value:
            self.right = CellAttrAppend(self.right, value)
        self.right = CellAttrAsFlat(CellAttrAsLower(self.right))
        self.left = CellAttrAsFlat(CellAttrAsLower(left))

class FilterAnyDate(FilterDate):
    def __init__(self, operator, what, what_right, data_col, column_type, left):
        if what_right == 'date' and data_col >= 0:
            self.right = CellAttrOther(data_col, what_right)
        elif what_right == 'date':
            self.right = CellAttrConstSeconds()
        else:
            self.right = CellAttrAsDate(CellAttrOther(data_col, what_right))

        self.left = left
        if what != 'date':
            self.left = CellAttrAsDate(self.left)

class FilterIsTeacher(FilterAny):
    def __init__(self):
        pass
    def evaluate(self, fc):
        return fc.is_a_teacher
    def js(self):
        return "is_a_teacher"
    def editor(self):
        return _('F_teacher')

class FilterAnyType(FilterAny):
    def __init__(self, operator, what_right, data_col, column_type, left):
        self.right = CellAttrOther(data_col, what_right)
        self.left = left
        if operator[0] == '=':
            self.js_oper = "=="
        else:
            self.js_oper = operator[0]

    def evaluate(self, fc):
        left = to_float_or_nan(self.left.fct(fc))
        if not isNaN(left):
            right = to_float_or_nan(self.right.fct(fc))
            if isNaN(right):
                right = -1e50
            return self.python(left, right)
        right = to_float_or_nan(self.right.fct(fc))
        if isNaN(right):
            return self.python(flat(str(self.left.fct(fc)).lower()),
                               flat(str(self.right.fct(fc)).lower()))
        left = -1e50
        return self.python(left, right)

    def js(self):
        left = self.left.js()
        right = self.right.js()
        return (
            "((isNaN(to_float_or_nan(" + left
            + ')) && isNaN(to_float_or_nan(' + right
            + '))) ? '
            + "flat(" + left  + ".toString().toLowerCase())"
            + self.js_oper
            + "flat(" + right  + ".toString().toLowerCase())"
            + ":to_float_or_small(" + left + ')'
            + self.js_oper
            + "to_float_or_small(" + right + '))'
            )


def FilterOperator(operator, what, value, column_type, left, errors, columns):
    if value and value[0] in filterAttributes:
        what_right = filterAttributes[value[0]]
        elsewhere = from_another_column(value[1:], errors, columns)
    else:
        what_right = "value"
        elsewhere = from_another_column(value, errors, columns)

    value = replace_all(value, "\\\\", "￾")
    value = replace_all(value, "\\", "")
    value = replace_all(value, "￾", "\\")
    if elsewhere is None:
        if (operator[0] == "~"
            or (operator[0] == "" and what != "date")
            or what == "history"
            or value.lower() == 'nan'
            or (value == '' and (what == 'value' or what == 'comment'))
        ):
            f = FilterStr(operator, str(value), column_type, left)
        elif what == 'date' or (column_type == 'Date' and what == "value"):
            f = FilterDate(operator, what, str(value), column_type, left)
        elif what == 'author':
            f = FilterStr(operator, str(value), column_type, left)
        else:
            try:
                f = FilterFloat(operator, to_float(value), column_type, left)
            except:
                f = FilterStr(operator, str(value), column_type, left)
    else:
        data_col, string = elsewhere
        if data_col == -1:
            if what_right == 'author':
                f = FilterAnyStr(operator, what_right, -1, string,
                                 column_type, left)
            elif what_right == 'date':
                f = FilterAnyDate(operator, what, what_right, -1, column_type,
                                  left)
            else:
                return FilterFalse()
        elif operator[0] == "~" or operator[0] == "" or string != "":
            f = FilterAnyStr(operator, what_right, data_col, string,
                             column_type, left)
        elif what == 'date' or  (column_type == 'Date' and what == "value"):
            f = FilterAnyDate(operator, what,what_right, data_col, column_type,
                              left)
        else:
            f = FilterAnyType(operator, what_right, data_col, column_type,
                              left)
        if data_col >= 0:
            f.data_col_right = data_col

    if f.operator is None:
        f.operator = operator
    f.python = f.operator[1]
    f.js_string = f.operator[2]
    return f

# Should be subclasses, but not possible with PythonJS
# The arguments are:
#    * the filter operation,
#    * the python eval
#    * a function creating the JavaScript code to do the same evaluation.
#    * The reverse numeric operator

def LE(a, b): return a <= b
def LE_str(a, b): return '(' + a + '<=' + b + ')'
def LT(a, b): return a < b
def LT_str(a, b): return '(' + a + '<' + b + ')'
def GE(a, b): return a >= b
def GE_str(a, b): return '(' + a + '>=' + b + ')'
def GT(a, b): return a > b
def GT_str(a, b): return '(' + a + '>' + b + ')'
def EQ(a, b): return a == b
def EQ_str(a, b): return '(' + a + '==' + b + ')'
def TILDE(a, b): return b in a
def TILDE_str(a, b): return '(' + a + '.indexOf(' + b + ') != -1)'
def START(a, b): return a.startswith(b)
def START_str(a, b): return '(' + a + '.substr(0,' + b + '.length)==' + b + ')'
def AUTHOR(a, b): return a == b or a in major_of(b)
def AUTHOR_str(a, b): return ('(' + a + '==' + b + '||myindex(minors,'
                              + a + ')!=-1)')

filterOperators = [
    ['<=', LE, LE_str, '>='],
    ['<' , LT, LT_str, '>' ],
    ['>=', GE, GE_str, '<='],
    ['>' , GT, GT_str, '<' ],
    ['=' , EQ, EQ_str, '=' ],
    ['~' , TILDE, TILDE_str, None],
    [''  , START, START_str, None],
    ]
filterAttributes = {'@': 'author',
                    '#': 'comment',
                    ':': 'history',
                    '?': 'date'}

def from_another_column(string, errors, columns):
    if not string.startswith('['):
        return None
    s = string.split(']')
    if len(s) == 1:
        return None
    if s[0][1:] == '':
        return -1, string[len(s[0])+1:]
    data_col = data_col_from_col_title(s[0][1:], columns)
    if data_col or data_col == 0:
        return data_col, string[len(s[0])+1:]
    errors["«" + s[0][1:] + '» ' + _("ALERT_url_import_column")] = True
    return None

class Filter:
    # Parse the filter to create a list of nodes
    def __init__(self, string, column_type, the_columns=None):
        if the_columns is None:
            try:
                the_columns = columns # Global value for JavaScript
            except: # pragma: no cover
                the_columns = None # Home page
        self.errors = {}
        self.string = string
        while string.startswith(" "):
            string = string[1:]
        if string == '':
            self.filters = [[FilterTrue()]]
            return
        self.filters = [] # 'ORed' list of 'ANDed' items
        # OR is replaced by its translation
        string = replace_all(string," " + or_keyword() + " ", " | ")
        mode = "|"
        while string:
            node, string = self.parse(string, column_type, the_columns)
            if mode == '|':
                self.filters.append([node])
            else:
                self.filters[-1].append(node)
            if len(string) and (string[0] == '|' or string[0] == '&'):
                mode = string[0]
                string = string[1:].lstrip()
            else:
                mode = '&'

    def parse(self, string, column_type, columns):
        negate = False
        if string[0] == '!':
            if string == '!':
                return FilterFalse(), ''
            negate = True
            string = string[1:]
        if string[0] in filterAttributes:
            attr = filterAttributes[string[0]]
            if attr != 'value':
                string = string[1:]
        else:
            attr = "value"
        elsewhere = from_another_column(string, self.errors, columns)
        if elsewhere is None:
            string, operator = search_operator(string)
            left = CellAttrCell(attr)
            if attr == 'value' and operator[0] == '':
                left = CellAttrAsFixed(left)
        else:
            data_col, string = elsewhere
            if data_col == -1:
                string, operator = search_operator(string)
                if attr == 'author':
                    left = CellAttrConstUserName()
                elif attr == 'date':
                    left = CellAttrConstSeconds()
                elif attr == 'comment':
                    left = FilterIsTeacher()
                    if negate:
                        left = FilterNegate(left)
                    if operator[0] != '':
                        self.errors['«' + operator[0] + '#[] » → « #[] »'] = True
                    return left, string.lstrip()
                else:
                    return FilterFalse(), ''
            else:
                string, operator = search_operator(string)
                left = CellAttrOther(data_col, attr)
                column_type = columns[data_col].type.name
                if attr == 'value' and operator[0] == '':
                    left = CellAttrAsFixed(left)
        value = ''
        bs_protected = False
        i = 0
        for char in string:
            i += 1
            if not bs_protected:
                if char == ' ':
                    # Search non space char
                    next_char = ""
                    for next_char in string[i:]:
                        if next_char != ' ':
                            break
                    if next_char == ' ':
                        break

                if (char == ' '
                    and operator[0] == ''
                    and attr == 'value'
                    and next_char not in filterAttributes # the value
                    and search_operator(next_char)[1][0] == ''    # starting by
                    and next_char not in '&|@#:?<>=~!['
                    and negate == False
                    and elsewhere is None
                    ):
                    # Filter 'a b' is to translated as 'a\ b' because
                    # starting by both 'a' and 'b' is impossible
                    # End users WANTS this
                    pass
                elif " |&".find(char) >= 0:
                    i -= 1
                    break
                if char == '\\':
                    bs_protected = True
            else:
                bs_protected = False
            value += char

        if len(value) == 0 and operator[0] == '':
            if attr == 'author':
                value = "@[]"
                if elsewhere is None:
                    operator = [_('is'), AUTHOR, AUTHOR_str, None]
                else:
                    left = CellAttrCell("value")
            elif attr == 'comment' or attr == 'history':
                # Special filters : '#' and ':'
                dummy, operator = search_operator('=')
                negate = not negate

        node = FilterOperator(operator, attr, value, column_type,
                              left, self.errors, columns)

        if elsewhere is not None and elsewhere[0] >= 0:
            node.data_col_left = elsewhere[0]

        if negate:
            node = FilterNegate(node)

        return node, string[i:].lstrip()

    def evaluate(self, line, cell, username, is_a_teacher=False):
        fc = FilterContext(line, cell, username, is_a_teacher)
        result = False
        for f in self.filters:
            bad = False
            for ff in f:
                if not ff.evaluate(fc):
                    bad = True
                    break
            if not bad:
                result = True
                break

        if debug:
            error = False
            if not python_mode: # pragma: no cover
                error = "Can not compile js: " + self.js()
                js = self.compiled_js()
                error = "Can not evaluate compiled filter: " + js
                value = js(line, cell, username)
                if value == result:
                    error = False
                else:
                    error = ("Unexpected value: '" + value
                             + "' in place of '" + result
                             + "' for:\n\tcell: " + cell.js()
                             + '\n\tjs: ' + js + "'"
                             + '\n\teditor: ' + self.editor()
                             )
            if error: # pragma: no cover
                print(error)
                raise ValueError(error)

        return result

    def dates(self, line, cell, username, is_a_teacher=False):
        fc = FilterContext(line, cell, username, is_a_teacher)
        all_dates = NO_DATES.clone()
        for f in self.filters:
            dates = f[0].dates(fc)
            for ff in f[1:]:
                dates.intersection(ff.dates(fc))
            all_dates.add(dates)
        return all_dates

    def js(self):
        return '(' + '||'.join(['&&'.join([ff.js()
                                           for ff in f])
                                for f in self.filters
                             ]) + ')'

    def compiled_js(self): # pragma: no cover
        return eval('(function x(line, cell, username) { return '+self.js()+';})')

    def get_errors(self):
        s = ""
        for i in self.errors:
            s += i + "<br>\n"
        if s != "":
            return s

    def other_data_col(self):
        cols = []
        for ored in self.filters:
            for f in ored:
                if f.data_col_left not in cols:
                    cols.append(f.data_col_left)
                if f.data_col_right not in cols:
                    cols.append(f.data_col_right)
        cols = [i
                for i in cols
                if i or i == 0
                ]
        cols.sort()
        return cols

    def editor(self):
        html = []
        for ored in self.filters:
            html.append('<div>'
                        + _("TIP_tablecopy_and").upper().join(['<div>' + f.editor() + '</div>' for f in ored])
                        + '</div>'
                       )
        return '<div class="filter_editor">' + _('or').upper().join(html)  + '</div>'
