#!/usr/bin/python3
# -*- coding: utf-8 -*-
# TOMUSS: The Online Multi User Simple Spreadsheet
# Copyright (C) 2022 Thierry EXCOFFIER, Universite Claude Bernard
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Contact: Thierry.EXCOFFIER@univ-lyon1.fr

def find_ue_grade():
    """Find all the columns needed to compute the UE grade and report errors"""
    def info(col):
        """Information about a bad column"""
        return (
            '<TR><TD>' + html(col.title)
            + '<TD>' + _("SELECT_column_grade_type_" + str(col.grade_type))
            + '<TD>' + _("SELECT_column_grade_session_" + str(col.grade_session))
            + '<TD>' + _("B_print_attr_weight") + str(col.real_weight)
            + '</TR>')
    # XXX 1000 not nice. To fix JS string sort
    cols = [[1000 + column.position, column.data_col] for column in columns]
    cols.sort()
    errors = []
    by_grade_type = [[], [], [], []]
    previous = None
    for _position, data_col in cols:
        column = columns[data_col]
        if not column.grade_type:
            previous = column
            continue
        if not column.real_weight_add:
            errors.append(column.title + ' ' + _('ERROR_addition_not_allowed'))
        if column.type == 'Ue_Grade':
            continue # No recursion allowed # pragma: no cover
        liste = by_grade_type[column.grade_type]
        if column.grade_session == 0:
            liste.append([data_col])
        else:
            if column.grade_type != previous.grade_type:
                error = _("SELECT_column_grade_type_0").split(' ')[-1]
            elif column.real_weight != previous.real_weight:
                error = _("BEFORE_column_attr_weight")
            elif column.grade_session != previous.grade_session + 1:
                error = _("SELECT_column_grade_session_x")
            else:
                error = ''
            if error != '':
                errors.append(
                    '<P>'
                    + _("ALERT_invalid_value")
                    + error
                    + '<TABLE class="colored" style="margin:0.5em; border:1px solid #F00">'
                    + info(previous)
                    + info(column)
                    + '</TABLE>')
            else:
                liste[-1].append(column.data_col)
        previous = column
    return by_grade_type, '<br>'.join(errors)

LAST_COMPUTED_UE_GRADE = ['', 0, '']

def ue_grade(by_grade_type, line, abi_is_def):
    """Compute the UE grade"""
    def get_grade_and_weight(data_cols):
        for data_col in data_cols[::-1]:
            cell = line[data_col]
            cell_str = str(cell.value)
            if cell_str:
                column = columns[data_col]
                if column.visibility == 5 and python_mode:
                    continue # Compute without the grade for the suivi
                value = to_float_or_nan(cell.value)
                if isNaN(value):
                    if cell_str.lower() == 'nan':
                        continue # NaN does no count
                    if abi_is_def and column.grade_type != 1: # CI and CT → DEF
                        return None
                    value = column.min
                value = do_round(value, column.round_by, column.table.rounding, None)
                return [(value - column.min) / (column.max - column.min),
                        column.real_weight]
        return ''

    sum_grades = 0
    weight_grades = 0
    miss_grade = False
    for data_cols in by_grade_type[3]: # CT
        grade_weight = get_grade_and_weight(data_cols)
        if grade_weight is None:
            return None, True
        if grade_weight == '':
            miss_grade = True
            continue
        sum_grades += grade_weight[1] * grade_weight[0]
        weight_grades += grade_weight[1]
    if weight_grades:
        ct_grade = sum_grades / weight_grades
    else:
        ct_grade = 0
    for data_cols in by_grade_type[2]: # CI
        grade_weight = get_grade_and_weight(data_cols)
        if grade_weight is None:
            return None, True
        if grade_weight == '':
            miss_grade = True
            continue
        sum_grades += grade_weight[1] * grade_weight[0]
        weight_grades += grade_weight[1]
    for data_cols in by_grade_type[1]: # CC
        grade_weight = get_grade_and_weight(data_cols)
        # Can not be None
        # if grade_weight is None:
        #     return None, True
        if grade_weight == '':
            miss_grade = True
            continue
        sum_grades += grade_weight[1] * max(grade_weight[0], ct_grade)
        weight_grades += grade_weight[1]
    if weight_grades:
        return sum_grades / weight_grades, miss_grade
    return '', True

def compute_ue_grade(data_col, line, _username):
    """Compute the grade.
    Use a cache to get the list of columns to average
    """
    column = columns[data_col]
    if (column is not LAST_COMPUTED_UE_GRADE[0]
            or millisec() - LAST_COMPUTED_UE_GRADE[1] > 100):
        LAST_COMPUTED_UE_GRADE[0] = column
        LAST_COMPUTED_UE_GRADE[1] = millisec()
        LAST_COMPUTED_UE_GRADE[2] = find_ue_grade()
        cols_data_col = []
        cols_title = []
        for used in LAST_COMPUTED_UE_GRADE[2][0][1:]:
            for column_type in used:
                for session in column_type:
                    cols_title.append(columns[session].title)
                    cols_data_col.append(session)
        column.columns = ' '.join(cols_title)
        column.average_columns = cols_data_col
        try:
            if len(LAST_COMPUTED_UE_GRADE[2][1]): # pylint: disable=len-as-condition
                set_message('ue_grade', 3, LAST_COMPUTED_UE_GRADE[2][1])
            else:
                set_message('ue_grade', 3)
        except: # pylint: disable=bare-except
            pass

    by_grade_type, errors = LAST_COMPUTED_UE_GRADE[2]
    if len(errors): # pylint: disable=len-as-condition
        return ''
    grade, miss_grade = ue_grade(by_grade_type, line, column.abi_is == 1)
    if grade is None:
        return "DEF"
    if grade == '':
        return ''
    grade = grade * (column.max - column.min) + column.min
    if miss_grade:
        return '0' + as_fixed(grade)
    return grade
