#!/usr/bin/python3

"""
The current directory must contains:

  YourScript.py : The grading script
  students.csv  : One line per student, 10 columns per value :
                         New Value Author Comment Date UploadName and old one
  other_name    : uploaded files

The stdout is a Python dictionnary with this structure,
competence and students have the same format than Notation (JSON).

  'competences': List["Key 1QMax text" | "Index 1BMax text"]
  'students': Dict[str, Dict[str, "Grade/Max comment"]]

If the student grade dict contains : 'message_green' key or 'message_red',
then the value is a text message to display to the student on upload


Helper script to create batches logins from /home/batch template:

cd /home
for I in 0 1 2 3 4 5 6 7 8 9
do
B=batch$I
echo "$B:x:200$I:1001:,,,:/home/batch$I:/bin/bash" >>/etc/passwd
echo "$B::18936:0:99999:7:::" >>/etc/shadow
mkdir $B
cp -r batch/.ssh $B # Must be root owner
chown $B $B
chmod 700 $B
done
# Account list to put in config_table of the Analyser :
echo batch{0,1,2,3,4,5,6,7,8,9}@example.com


Useful package :

python3-xlrd
python3-docx

"""

import csv
import json
import traceback
from typing import List, IO, Optional, Tuple
import zipfile
import tarfile

class Comment:
    """Define a grade and a comment"""
    def __init__(self, competence: "Competence", grade: float = 0, comment: str = '') -> None:
        self.competence = competence
        self.grade = grade
        self.comment = comment

    def add(self, grade: float = 0, comment: str = '') -> None:
        """Add the grade and append the comment"""
        self.grade += grade
        self.comment += comment

    def __str__(self):
        text = [f'\t{self.grade:5.2f}/']
        if self.competence.bonus:
            text.append('±')
        text.append(f'{self.competence.grade_max}')
        if self.comment:
            text.append(' ')
            text.append(self.comment)
        text.append(f' -- {self.competence.question}')
        return''.join(text)

class Value: # pylint: disable=too-many-instance-attributes
    """The content of a TOMUSS cell"""
    upload_zip: Optional[zipfile.ZipFile] = None
    upload_tar: Optional[tarfile.TarFile] = None

    # pylint: disable=too-many-arguments
    def __init__(self, cell: "Cell", value: str, author: str, comment: str,
                 date: str, filename: str) -> None:
        self.cell = cell
        self.value = value
        self.author = author
        self.comment = comment
        self.date = date
        self.filename = filename

    def get_upload_handle(self) -> IO[bytes]:
        """The file handle to manipulate the uploaded file"""
        return open(self.filename, 'rb')

    def get_upload_content(self) -> bytes:
        """Get the uploaded bytes or None"""
        with self.get_upload_handle() as file:
            return file.read()

    def get_upload_size(self) -> int:
        """Get the uploaded size or None"""
        return len(self.get_upload_content())

    def upload_is_zip(self) -> bool:
        """Is the uploaded file a ZIP?"""
        if self.upload_zip:
            return True
        try:
            self.upload_zip = zipfile.ZipFile(self.get_upload_handle(), mode="r")
        except: # pylint: disable=bare-except
            return False
        return True

    def upload_is_tar(self) -> bool:
        """Is the uploaded file a TAR?"""
        if self.upload_tar:
            return True
        try:
            self.upload_tar = tarfile.TarFile(fileobj=self.get_upload_handle(), mode="r")
        except: # pylint: disable=bare-except
            return False
        return True

    def upload_is_an_archive(self) -> bool:
        """Returns True if the archive is usable"""
        return self.upload_is_zip() or self.upload_is_tar()

    def upload_filelist(self) -> List[str]:
        """Return the full file list content"""
        if self.upload_zip:
            return [file.filename for file in self.upload_zip.filelist]
        if self.upload_tar:
            return self.upload_tar.getnames()
        raise ValueError("You have not tested if the uploaded file was an archive")

    def archive_handle(self, filename: str) -> Optional[IO[bytes]]:
        """Return a file handle for an archive member"""
        if self.upload_zip:
            return self.upload_zip.open(filename)
        if self.upload_tar:
            return self.upload_tar.extractfile(filename)
        raise ValueError("You have not tested if the uploaded file was an archive")

    def archive_size(self, filename: str) -> int:
        """Return the file size of an archive member"""
        if self.upload_zip:
            with self.upload_zip.open(filename) as file:
                return len(file.read())
        if self.upload_tar:
            return self.upload_tar.getmember(filename).size
        raise ValueError("You have not tested if the uploaded file was an archive")

    def archive_content(self, filename: str) -> bytes:
        """Return the content of an archive member"""
        if self.upload_zip:
            with self.upload_zip.open(filename) as file:
                return file.read()
        if self.upload_tar:
            with self.upload_tar.extractfile(filename) as file: # type: ignore
                return file.read()
        raise ValueError("You have not tested if the uploaded file was an archive")

    def with_upload(self) -> bool:
        """True if there is an upload"""
        return self.filename != ''

    def __str__(self):
        if self.with_upload():
            more = f' UPLOAD'
        else:
            more = ''
        return f'{self.value:8.8} {self.author:8.8} {self.comment:8.8} {self.date:14}{more:8}'

class Cell: # pylint: disable=too-few-public-methods
    """The content of a TOMUSS cell"""
    def __init__(self, student: "Student",
                 data: Tuple[str, str, str, str, str, str, str, str, str, str]) -> None:
        self.student = student
        self.new = Value(self, *data[:5])
        self.old = Value(self, *data[5:])

    def __str__(self):
        return f'{self.old} → {self.new}'


class Competence: # pylint: disable=too-few-public-methods
    """Define a grading item"""
    def __init__(self, question: str, grade_max: float, bonus: bool, index: int) -> None:
        self.question = question
        self.grade_max = grade_max
        self.bonus = bonus
        self.index = index

    def __str__(self):
        return f'/ {"±" if self.bonus else ""}{self.grade_max} {self.question}'

class Student:
    """All the cells in the columns to analyse for the student and the grading done.
    """
    comments: List[Comment] = []
    def __init__(self, analyser: "Analyser", student_id: str) -> None:
        self.analyser = analyser
        self.cells: List[Cell] = []
        self.student_id = student_id
        self.green = ''
        self.red = ''

    def __str__(self):
        values = '\n'.join(str(cell) for cell in self.cells)
        comments = '\n'.join(f'{comment}' for comment in self.comments)
        return f'\n{values}\n{comments}'

    def grade(self, grade: Competence, value: float, comment: str = '') -> None:
        """Add a grade and a comment"""
        if not self.comments:
            self.comments = [Comment(grade) for grade in self.analyser.competences]
        self.comments[grade.index].add(value, comment)

    def grade01(self, grade: Competence, value: float, comment: str = '') -> None:
        """Add a grade 0=min and 1=max and a comment"""
        self.grade(grade, value * grade.grade_max, comment)

    def add_cell(self, data):
        """Add a column to the student"""
        self.cells.append(Cell(self, data))

    def message_green(self, message):
        """Add a green message"""
        self.green += message

    def message_red(self, message):
        """Add a red message"""
        self.red += message

    def __iter__(self):
        return iter(self.cells)

    def __getitem__(self, key):
        return self.cells[key]

class Analyser:
    """Parse the data assuming files are in the good order"""
    students: List[Student] = []
    def __init__(self):
        self.competences = []

    def read_dir(self):
        """Read a directory"""
        with open('students.csv', 'r') as file:
            self.read_csv(file)

    def read_csv(self, file):
        """Read the CSV file"""
        reader = csv.reader(file)
        for line in reader:
            student = Student(self, line[0])
            self.students.append(student)
            for i in range(1, len(line), 10):
                student.add_cell(line[i: i+10])

    def grading(self): # pylint: disable=no-self-use
        """Execute the provided grading script"""
        with open('YourScript.py', 'r') as file:
            grading = file.read()
        exec(compile(grading, 'YourScript.py', mode='exec'), dict(globals())) # pylint: disable=exec-used

    def add_competence(self, question: str, grade_max: float, bonus: bool = False) -> Competence:
        """Add a new competence"""
        grade = Competence(question, grade_max, bonus, len(self.competences))
        self.competences.append(grade)
        return grade

    def __iter__(self):
        return iter(self.students)

    def send_result(self):
        """Returns the result to TOMUSS"""
        result = {}
        result['competences'] = json.dumps([
            f"c{i:03d} 1{'B' if competence.bonus else 'Q'}{competence.grade_max} {competence.question}"
            for i, competence in enumerate(self.competences)
        ], sort_keys=True)

        def grades_dict(student):
            grades = {
                f'c{i:03d}': f'{comment.grade}/{competence.grade_max} {comment.comment}'
                for (i, (comment, competence)) in enumerate(zip(student.comments, self.competences))
            }
            if student.red:
                grades['message_red'] = student.red
            if student.green:
                grades['message_green'] = student.green
            return grades

        result['students'] = {
            student.student_id: json.dumps(grades_dict(student), sort_keys=True)
            for student in self
            }
        print(json.dumps(result, sort_keys=True, indent=1))

try:
    STUDENTS = Analyser()
    STUDENTS.read_dir()
    STUDENTS.grading()
    STUDENTS.send_result()
except: # pylint: disable=bare-except
    print(json.dumps({'error': traceback.format_exc()}))
