#!/usr/bin/python3

"""
Scanner harvesting data each night.

The goal is to do an incremental harvesting and merging.
The stored files are sorted so the merging of all the semesters
does not imply to read the whole file and so it uses little memory.

The scanner is based on plugin defined as Python files

All 'scanner_*.py' and 'LOCAL/scanner_*.py' files are imported
and all 'Plugin' subclasses are instanciated.

Plugins attributes default values are defined in scanner_plugin.py
so read it in order too choose the tables to analyse.
"""

# pylint: disable=wrong-import-position

import glob
import os
import sys
import time
import itertools
import importlib

ERRORS_FILE = os.path.join("LOGS", "scanner.log")
OLD_STDERR = sys.stderr
sys.stderr = open(ERRORS_FILE, 'w')
print('Errors messages are stored in LOGS/scanner.log')

import tomuss_init # pylint: disable=unused-import
from .. import configuration
from .. import document
from .. import utilities
from . import scanner_plugin

def get_part(text, position):
    """Split the text by '_' and get the position part"""
    text = text.split('_')
    if position < len(text):
        return text[position]
    return ''

def help_message():
    """Display usage message"""
    sys.stdout.write(f"""
Usage: {sys.argv[0]} 'SEMESTERS' 'PLUGINS'

SEMESTERS is:
  * a shell protected pattern has 'Y2021/S*'
  * or an empty string to not read any semester and to only do the merging
  * or 'update' : current university year and never scanned semesters

PLUGINS is:
  * a list of plugins name has 'Index,Grade'
  * or an empty string for all

Semester scanning results are in TMP/SCANNER/Y..../S..../PluginName
as a line sorted file of pairs: (key, data)

Defined plugins:
""")
    header = ('name       ',
              'only_real_file', 'only_official_ue', 'only_suivi_semester',
              'only_current_year', 'only_current_semester', 'only_weekday',
              'one_merge_per_day', 'merge_manually', 'not_on_tables')
    formatter = ''.join(
        "|{{:{0}.{0}}}".format(max(len(part) for part in title.split('_')))
        for title in header) + '|'
    line = formatter.format(*(['-----------------'] * len(header)))
    line = line.replace('|', '+')
    print(line)
    for part in range(4):
        print(formatter.format(*[get_part(title, part) for title in header]))
    print(line)
    def text(value):
        if value is False or value is None:
            return ''
        if isinstance(value, tuple):
            return ''.join(str(i) for i in value)
        return str(value)
    for plugin in Scanner.plugins:
        print(formatter.format(*[text(getattr(plugin, attr.strip()))
                                 for attr in header]))
    print(line)


class Scanner:
    """Run the plugins on all the tables"""
    plugins = []
    def __init__(self, year, semester, tables=None, plugins=None):
        """Parameters: Y and S before year and semester and '.py' for tables"""
        print(year, semester, tables or '*', end=' ')
        university_year = utilities.university_year()
        last = 0
        scanner_dir = os.path.join('TMP', 'SCANNER', year, semester)
        day = time.localtime(self.start_time).tm_wday
        utilities.mkpath(scanner_dir)
        if plugins is None:
            plugins = self.plugins
        for plugin in plugins:
            plugin.year = year[1:]
            plugin.semester = semester[1:]
            plugin.state_filename = os.path.join(scanner_dir, plugin.name)
            plugin.start()
        if tables is None:
            tables = utilities.python_files(
                os.path.join(configuration.db, year, semester))
        for file in tables:
            if file in ('abj.py', "undefined.py"):
                continue # pragma: no cover
            is_link = os.path.islink(os.path.join(configuration.db, year, semester, file))
            table = document.table(year[1:], semester[1:], file[:-3], ro=True)
            if not table:
                continue # pragma: no cover
            print('{}{}{}{}'.format(
                '\010'*last, ' '*last, '\010'*last, table.ue), end='', flush=True)
            last = len(table.ue)
            table_university_year = utilities.university_year(table.year, table.semester)
            for plugin in plugins:
                if plugin.only_real_file and is_link:
                    continue
                if plugin.only_official_ue and not table.official_ue:
                    continue
                if plugin.only_suivi_semester and semester[1:] not in configuration.semesters:
                    continue # pragma: no cover
                if plugin.only_current_year and university_year != table_university_year:
                    continue
                if plugin.only_current_semester:
                    if configuration.year_semester != (table.year, table.semester):
                        continue
                if day not in plugin.only_weekday:
                    continue # pragma: no cover
                if plugin.not_on_tables:
                    continue
                plugin.table(table)
            table.unload()
        print()
        for plugin in plugins:
            plugin.do_record()
    @classmethod
    def load_plugins(cls, plugins):
        """The plugin register itself with this function"""
        for plugin in glob.glob(os.path.join('SCRIPTS', 'scanner_*.py')
                               ) + glob.glob(os.path.join('LOCAL', 'scanner_*.py')):
            module = importlib.import_module(plugin.split(os.path.sep)[-1][:-3])
            for value in module.__dict__.values():
                if value is scanner_plugin.Plugin:
                    continue # pragma: no cover # Should be covered?
                if not isinstance(value, type):
                    continue
                if not issubclass(value, scanner_plugin.Plugin):
                    continue
                if plugins:
                    if value.__name__ not in plugins:
                        continue
                else:
                    if value.__name__.startswith('Debug'):
                        # Run Debug scanner only if asked
                        continue
                cls.plugins.append(value(on_command_line=plugins != ''))


    @classmethod
    def merge(cls):
        """Merging the data incrementally"""
        for plugin in cls.plugins:
            plugin.do_merge(cls.start_time)
            plugin.done()

def main():
    """Scan semesters and merge all the data"""
    Scanner.start_time = time.time()
    Scanner.load_plugins(sys.argv[2])
    if sys.argv[1]:
        if sys.argv[1] == 'update':
            # Never scanned semester
            semesters = set(
                semester
                for semester in glob.glob(os.path.join(configuration.db, 'Y*', 'S*'))
                if not os.path.exists(semester.replace(
                    configuration.db, os.path.join('TMP', 'SCANNER')))
                )
            # Current semesters
            uyear = utilities.university_year()
            if time.localtime()[0] != uyear:
                year_before = 0 # pragma: no cover
            else:
                year_before = 1 # pragma: no cover
            for year in range(uyear - year_before, uyear + 2):
                semesters.update(glob.glob(os.path.join(
                    configuration.db, f'Y{year}', 'S*')))
        else:
            semesters = glob.glob(os.path.join(configuration.db, sys.argv[1]))
            print('       DB =',  configuration.db)
            print('  Pattern =', sys.argv[1])
            print("Semesters =", ' '.join(sorted(semesters)))

        # Remove DB name
        semesters = sorted(semester.split(os.path.sep)[-2:]
                           for semester in semesters)

        # Run Scanner on full semester
        print("Scan the semesters")
        for year, semester in semesters:
            Scanner(year, semester)

        # Add special table names, only if not in the analysed semesters
        print("Find special tables not yet analysed")
        special_tables = set()
        special_plugins = set()
        for plugin in Scanner.plugins:
            for y_s_ue in plugin.tables_to_analyse():
                if list(y_s_ue[:2]) not in semesters:
                    special_tables.add(y_s_ue)
                    special_plugins.add(plugin)
        # Run on these tables
        print("Analyse special tables")
        special_tables = sorted(special_tables)
        for (year, semester), group in itertools.groupby(special_tables, key=lambda x: x[:2]):
            Scanner(year, semester,
                    tables=[code_ue for _year, _semester, code_ue in group],
                    plugins=special_plugins)

    Scanner.merge()

if __name__ == '__main__':
    if len(sys.argv) >= 3:
        try:
            main()
        except KeyboardInterrupt: # pragma: no cover
            pass
        except: # pylint: disable=bare-except # pragma: no cover
            sys.stderr = OLD_STDERR
            sys.stderr.write(utilities.read_file(ERRORS_FILE))
            raise
    else:
        Scanner.load_plugins('') # Load all plugins
        help_message()
