"""
ABI scanner for the last week for the current semester

The computed dict is : Student → List[abi]

The finalization send the mail to the referent teachers each monday.
Not the merging because the key is the teacher, not the student.

The environment variable SCANNER_ABI_TEACHERS may contains:
  * '' : no mail sent
  * '*' : mail sent to all the teachers
  * 'login1 login2 login3...' : only these teachers

The environment variable SCANNER_SPECIAL_TABLES may contains a dictionnary:
  Table code → data_col if student ID if not 0
"""

import os
import time
import collections
from TOMUSS.SCRIPTS import scanner_plugin
from TOMUSS import configuration
from TOMUSS import utilities

SCANNER_SPECIAL_TABLES = utilities.literal_eval(
    os.environ.get('SCANNER_SPECIAL_TABLES', '{}'))

class ABI(scanner_plugin.Plugin):
    """Register all last week ABI"""
    students = None # type: Dict[List[str]]]
    only_current_semester = True
    only_real_file = False
    one_merge_per_day = True
    if not configuration.regtest:
        only_weekday = (0,) # pragma: no cover
    last_time = None
    merged = None
    def start(self):
        self.students = collections.defaultdict(list)
        self.last_time = time.strftime("%Y%m%d%H%M%S", time.localtime(self.start_time))
        self.abis = configuration.all_abi
    def stop(self):
        self.students = None

    def table(self, table):
        id_column = SCANNER_SPECIAL_TABLES.get(table.ue, 0)
        for line in table.lines.values():
            student_id = line[id_column].value
            if not student_id:
                continue
            student_id = utilities.the_login(student_id)
            infos = []
            for col, cell in zip(table.columns, line):
                if cell.value in self.abis and cell.date > self.last_time:
                    infos.append([cell.date, col.title, cell.author, cell.comment])
            if infos:
                infos.sort()
                self.students[student_id].append([table.ue, table.table_title, infos])

    def record(self):
        return self.students.items()
    def merge(self, data):
        """Create 'merged' Dict[student, ues]"""
        if not self.merged:
            self.merged = collections.defaultdict(list)
        for (login, ues), _year, _semester, _ in data:
            self.merged[login].extend(ues)

    def send_message(self, teacher, students, all_students): # pylint: disable=too-many-locals
        """Send the message with all the ABI to the teacher"""
        message = [
            time.strftime("%c", time.localtime(self.start_time)),
            ' → ',
            time.strftime("%c", time.localtime()),
            '\n\n'
            ]
        for student, ues_with_abi in students.items():
            user = all_students[student]
            message.append(f'{student} {user.firstname.title()} {user.surname} {user.mail}\n')
            for ue_with_abi in ues_with_abi:
                for ue_code, ue_title, infos in ue_with_abi:
                    message.append(f'    {ue_code} «{ue_title}»\n')
                    for date, title, author, comment in infos:
                        date = time.strftime("%c", time.strptime(date, '%Y%m%d%H%M%S'))
                        message.append(f'        {date} «{title}» «{author}» {comment}\n')
            message.append('\n')

        utilities.send_mail(
            to=teacher.mail,
            subject="TOMUSS " + configuration.abi + ' (' + teacher.ID_ + ')',
            message=''.join(message),
            frome=configuration.maintainer,
            show_to=True)

    def finalize(self):
        if not self.merged:
            return "No merge, so can't finalize" # pragma: no cover
        selector = os.environ.get('SCANNER_ABI_TEACHERS', '')
        if selector == '':
            return 'SCANNER_ABI_TEACHERS is empty, does not send mail' # pragma: no cover
        if selector != '*':
            selector = set(selector.split(' ')) # pragma: no cover
        all_students = configuration.Users(self.merged)
        teachers = collections.defaultdict(lambda: collections.defaultdict(list))
        for student, ues in self.merged.items():
            teachers[all_students[student].referent][student].append(ues)
        all_teachers = configuration.Users(teachers)
        message = []
        for teacher, students in teachers.items():
            if selector != '*' and teacher not in selector:
                continue # pragma: no cover
            teacher = all_teachers[teacher]
            if teacher.mail:
                self.send_message(teacher, students, all_students)
                message.append(teacher.ID_)
            else:
                message.append(f'\n{teacher} no mail address') # pragma: no cover
        return '\n'.join(message)
