"""
Grade scanner

The computed dict is : Student → List[Tuple[date, grade]]

Store results in:
  * LOGINS/.../grades
  * LOGINS/.../resume
"""

import collections
from TOMUSS import configuration
from TOMUSS.SCRIPTS import scanner_plugin

class Grade(scanner_plugin.Plugin):
    """Statistics on the grades of the students"""
    students = None

    def start(self):
        self.students = collections.defaultdict(list)

    def stop(self):
        self.students = None

    def analyse_note(self, table, column, texts, what):
        """Analyse a column containing grades
        what = 0 : A normal grade
        what = 1 : A top average
        what = 2 : Nothing (to store the table name once)
        """
        mini, maxi = column.min_max()
        for login in table.the_keys():
            for line in table.get_lines(login):
                cell = line[column.data_col]
                if cell.date:
                    try:
                        value = float(cell.value)
                        if mini <= value <= maxi:
                            value = (value - mini) / (maxi - mini)
                        else:
                            continue
                    except ValueError:
                        if cell.value in texts:
                            value = texts[cell.value]
                        else:
                            continue
                    self.students[login].append((cell.date, value, table.ue, what))

    def analyse_prst(self, table, column, texts):
        """Analyse a column containing attendance"""
        for login in table.the_keys():
            for line in table.get_lines(login):
                cell = line[column.data_col]
                if cell.value in texts:
                    self.students[login].append((cell.date, texts[cell.value], table.ue, 0))

    def table(self, table):
        texts = configuration.all_all
        for column in table.columns:
            if column.type.name in self.grading_types:
                if column.weight[0] in '+-':
                    continue # pragma: no cover
                self.analyse_note(table, column, texts, 0)
            elif column.type.name == 'Prst':
                self.analyse_prst(table, column, texts)
        result = table.columns.result_column()
        if result:
            table.compute_columns()
            self.analyse_note(table, result, texts, 1)
        for login in table.the_keys():
            if login not in self.students or self.students[login][-1][-2] != table.ue:
                # In case of no grade for this table
                self.students[login] += [(None, None, table.ue, 2)]

    def record(self):
        return self.students.items()

    def merge(self, data): # pylint: disable=too-many-locals,too-many-branches
        table = []
        resume = collections.defaultdict(list)
        login = '' # To please pylint
        for (login, date_value_ue_what), year, semester, _ in data:
            year = int(year)
            for date, value, code_ue, what in date_value_ue_what:
                # For 'resume'
                stats = resume[code_ue]
                if not stats or (stats[-1][0] != year or stats[-1][1] != semester):
                    # Starts a new stat
                    stats.append([year, semester, 0, 0, 0, 0, 0])
                if what == 0 or isinstance(value, str):
                    if value == configuration.pre:
                        stats[-1][2] += 1
                    elif value == configuration.abi:
                        stats[-1][3] += 1
                    elif value == configuration.abj:
                        stats[-1][4] += 1
                elif what == 1:
                    # Replace the sum by the top level average
                    # The top level average is the last stat.
                    stats[-1][5] = value * stats[-1][6]
                    continue
                else:
                    # Table without grade
                    continue # pragma: no cover

                # For 'grades'
                if not isinstance(value, str):
                    stats[-1][5] += value
                    stats[-1][6] += 1
                    value = '{:5.3f}'.format(value)
                table.append('{} {} {} {} {}\n'.format(date, year, semester, code_ue, value))

        if len(login) < 3:
            return
        user = configuration.User(login)
        if user.is_a_student:
            print('\r' + login, end='         ')
            table.sort()
            user.manage_key('grades', content=''.join(table))

            for code_ue, stats in resume.items():
                formatted = []
                for year, semester, pre, abi, abj, avg, nr_grades in stats:
                    if nr_grades:
                        avg /= nr_grades
                    else:
                        avg = -1
                    formatted.append(
                        f"[{year},'{semester}',{pre},{abi},{abj},{avg:.3f},{nr_grades}]")
                formatted.sort()
                resume[code_ue] = f"'{code_ue}':[{','.join(formatted)}],\n"
            user.manage_key('resume', content='{\n%s}\n' % ''.join(sorted(resume.values())))

        user.forget() # To not eat memory (not reused)
