"""
Root class of scanner Plugins
"""

# pylint: disable=no-self-use

import os
import time
import glob
from TOMUSS import configuration

def read(states, files):
    """
    Read one line of the merged files.
    Remove the file from 'states' if nothing more to read.

    'states' and 'files' types are:
        List[Tuple[Tuple[key,data], year, semester, file]]

    """
    for state in tuple(files):
        line = state[3].readline().strip()
        if line:
            state[0] = eval(line) # pylint: disable=eval-used
        else:
            state[3].close()
            states.remove(state)

class Plugin:
    """The base class for all scanners"""
    only_real_file = True             # Not symbolic links
    only_official_ue = True           # Not invisible tables
    only_suivi_semester = True        # Not table not on the suivi page
    only_current_year = False         # The semesters of current university year
    only_current_semester = False     # The current  year semester
    one_merge_per_day = False         # If True only one MERGE per day
    merge_manually = False            # Only if the plugin is on the command line
    not_on_tables = False             # .table() method is not called
    only_weekday = (0, 1, 2, 3, 4, 5, 6) # Monday is 0

    start_time = 0                    # The scanner start time of the last successful merge

    name = 'not initialized'
    state_filename = 'not initialized'
    year = -1
    semester = 'not initialized'
    def __init__(self, on_command_line):
        self.grading_types = configuration.grading_types
        self.name = self.__class__.__name__
        self.on_command_line = on_command_line
        self.merge_report = os.path.join('TMP', 'SCANNER', 'merged.' + self.name)
        if os.path.exists(self.merge_report):
            with open(self.merge_report, 'r') as file:
                merge_report = eval(file.read()) # pylint: disable=eval-used
            self.start_time = merge_report['start_time']
        else:
            self.start_time = 0

    def merge_allowed(self):
        """Return True if the plugin merging is allowed"""
        if self.merge_manually and not self.on_command_line:
            return False
        if self.one_merge_per_day:
            start_time = time.localtime(self.start_time)
            if time.strftime('%Y%m%d') == time.strftime('%Y%m%d', start_time):
                return False
        return True

    def do_merge(self, start_time):
        """Merge all the semester data"""
        print('='*66, 'MERGE', self.name)
        merge_start = time.time()

        if not self.merge_allowed():
            print("SKIP IT: A merge has yet be done today")
            return

        messages = []
        states = [
            ['',
             name.split(os.path.sep)[2][1:],
             name.split(os.path.sep)[3][1:],
             open(name, 'r'),
            ]
            for name in glob.glob(os.path.join('TMP', 'SCANNER', '*', '*', self.name))
            ]
        read(states, states)
        while states:
            min_key = min(state[0][0] for state in states)
            to_merge = [
                state
                for state in states
                if state[0][0] == min_key
            ]
            message = self.merge(to_merge)
            if message:
                messages.append(message) # pragma: no cover
            read(states, to_merge)
        print('\nDONE')
        message = self.finalize()
        if message:
            messages.append(message)
        print('FINALIZED')
        with open(self.merge_report, 'w') as file:
            file.write(repr(
                {'elapsed_time': time.time() - merge_start,
                 'start_time': start_time,
                 'end_time': time.time(),
                 'messages': messages,
                 }) + '\n')

    def do_record(self):
        """Record the computed state"""
        if self.state_filename is None:
            return
        if os.path.exists(self.state_filename):
            # Create backup.
            os.rename(self.state_filename, self.state_filename + '~')
        with open(self.state_filename, 'w') as file:
            for line in sorted(self.record()):
                file.write(repr(line) + '\n')
        self.stop()

    def get_state(self):
        """Return the last recorded state as a dictionnary"""
        state = {}
        if os.path.exists(self.state_filename):
            with open(self.state_filename, 'r') as file:
                for line in file:
                    line = eval(line) # pylint: disable=eval-used
                    state[line[0]] = line[1]
        return state

    def start(self):
        """Called before any scan to initialize state"""
        raise ValueError("You must define an initialization") # pragma: no cover
    def stop(self):
        """Free the memory"""
        raise ValueError("You must free the memory") # pragma: no cover
    def tables_to_analyse(self):
        """Return a list of tables to be analysed with prefixes and suffixes
        List[Tuple[Yyear, Ssemester, ue.py]]
        """
        return ()
    def table(self, table):
        """Called for each table"""
        raise ValueError("You must define how the table content is scanned") # pragma: no cover
    def record(self):
        """
        Caller when all the semester tables have been scanned.
        It returns Tuple[key, data]"""
        raise ValueError("You must define how to record data") # pragma: no cover
    def merge(self, _data):
        """
        Merge data from all the semesters.
        List[Tuple[Tuple[key, data], year, semester, file]]
        """
        return "You must define how to merge the semester data" # pragma: no cover
    def finalize(self):
        """
        Called when all merges are done.
        If no merge are done, it is not called.
        """
        return ''
    def done(self):
        """
        Called when all the work is done.
        It is called even if no merge are done.
        """