#!/usr/bin/python3

"""
This script run TOMUSS server and launch a browser test with Chrome and Firefox

Arguments:
    * '1' : run until all the tests pass once (if not: run forever)
    * 'hidden' : run in a virtual X11 server

Required packages:

    apt install python3-selenium firefox-geckodriver chromium-chromedriver xnest xvfb i3

"""

# pylint: disable=missing-function-docstring,line-too-long,too-many-statements
import os
import sys
import subprocess
import time
import traceback
import contextlib
import selenium.webdriver
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.common.by import By

WAIT_KEY = 0.2

IN_DOCKER = not os.getenv('DISPLAY')

PORT = 8888

os.environ.pop('http_proxy', None)
os.environ.pop('https_proxy', None)

def retry(test):
    for _ in range(200):
        test_result = test()
        if test_result:
            return test_result
        time.sleep(0.01)
    raise ValueError('Failed test')

def test_fine(name, end=' '):
    print(name, end=end, flush=True)

class Tests: # pylint: disable=too-many-public-methods
    """Test for one browser"""
    body = input = line_filter = column_title = messages = None
    def __init__(self, driver):
        url = f"http://127.0.0.1:{PORT}/=super.user"
        base = url.split('/=', 1)[0]
        start = time.time()
        test_fine(driver.name.ljust(7, ' ') + ':')
        self.driver = driver
        # For debugging without problem: reload plugins
        self.driver.get(url)
        base = self.driver.current_url.split('/?')[0]
        if '&ticket' in self.driver.current_url:
            ticket = self.driver.current_url.split('&ticket')[1]
        else:
            ticket = '=' + url.split('/=')[1]
        time.sleep(2*WAIT_KEY)
        year_semester = self.driver.find_element(By.CSS_SELECTOR, '#s OPTION[selected]').get_attribute('innerText')
        self.driver.get(f"{base}/{ticket}/save_preferences/language=fr")
        time.sleep(WAIT_KEY)
        self.driver.get(f"{base}/{ticket}/reload_plugins")
        retry(lambda: len(self.driver.find_elements(By.TAG_NAME, 'TABLE')) == 3)


        with open('DBregtest/Y0/SDossiers/config_table.py', 'a', encoding='utf-8') as f:
            f.write("cell_change(1,'0_2','language','fr','20220321185042')\n")
        if not os.path.exists('DBregtest/Y0/SCompetences'):
            os.mkdir('DBregtest/Y0/SCompetences')
        with open('DBregtest/Y0/SCompetences/test_gui.py', 'w', encoding='utf-8') as f:
            with open('SCRIPTS/test_gui_competences.py', 'r', encoding='utf-8') as g:
                f.write(g.read())

        try:
            self.test_competence_table(base, ticket, year_semester)
            self.test_observations(base, ticket, year_semester)
            self.test_normal_table(base, ticket, year_semester)
            self.test_presence(base, year_semester.split('/')[0], ticket)
        except: # pylint: disable=bare-except
            self.driver.save_screenshot("xxx-test-gui-failure.png")
            print('\n\nScreen shot of the failure: xxx-test-gui-failure.png\n')
            raise # Keep driver open for debugging
        finally:
            pass
        self.driver.close()
        print(f'OK ({time.time()-start:.1f} secs)')

    def cursor_move_until_this_id(self, key, value):
        for _j in range(10):
            for _i in range(10):
                if self.cell_value() == value:
                    return
                time.sleep(0.1)
            self.input.send_keys(key)

    def click(self, selector, index=0):
        retry(lambda: len(self.driver.find_elements(By.CSS_SELECTOR, selector)) > index)
        try:
            self.driver.find_elements(By.CSS_SELECTOR, selector)[index].click()
        except selenium.common.exceptions.StaleElementReferenceException:
            time.sleep(0.1)
            self.driver.find_elements(By.CSS_SELECTOR, selector)[index].click()

    def send_keys(self, selector, text, index=0):
        retry(lambda: self.driver.find_elements(By.CSS_SELECTOR, selector))[index].send_keys(text)

    def hide_tip_popup(self):
        self.driver.execute_script('hide_the_tip_real(true)')

    def double_click(self):
        action = selenium.webdriver.ActionChains(self.driver)
        action.move_to_element_with_offset(self.input, 10, 10) # offset because FF bug
        action.double_click()
        action.perform()

    @contextlib.contextmanager
    def wait_table_refreshed(self, cell_value_change=True, goal=None):
        def ids():
            """All first column ID displayed"""
            return self.driver.execute_script('''
            var content = '';
            for(var i = nr_headers; i < nr_headers + table_attr.nr_lines; i++)
                content += '/' + table.rows[i].cells[0].textContent;
            return content''')
        content = ids()
        cell = self.cell_value()
        yield
        for _ in range(1000):
            if (goal or ids()) != content and (not cell_value_change or self.cell_value() != cell):
                return
            time.sleep(0.01)
        raise ValueError(f'''
content={content}/{ids()}
cell={cell}/{self.cell_value()}''')

    # COMPETENCES STARTS

    def test_competence_table(self, base, ticket, year_semester):
        year = year_semester.split('/')[0]
        try:
            os.unlink(f"DBregtest/Y{year}/SCompetences/test8.py")
        except OSError:
            pass
        self.driver.get(f"{base}/{ticket}/{year}/Competences/test8")
        self.elements()
        self.deactivate_autosave()
        self.input.send_keys(Keys.UP)
        self.button_plus_bot = self.driver.find_element(By.CSS_SELECTOR, '#button_plus_bot')
        with self.wait_table_refreshed(goal='/#m/m0'):
            self.test_add_first_voc("m","mention","mention informatique") # creates 2 lines
        self.input.send_keys(Keys.DOWN)
        self.test_add_new_voc("a","desc du voc a","voca a index 0", "m0")
        self.input.send_keys(Keys.DOWN)
        self.test_add_new_voc("e","desc du voc e","voca e index 0", "m0")
        self.input.send_keys(Keys.DOWN)
        self.input.send_keys(Keys.DOWN)
        self.input.send_keys(Keys.DOWN)
        self.test_add_new_voc("c","desc du voc c","voca c index 0", "m0e0")
        self.input.send_keys(Keys.UP)
        self.test_insert_new_line("a","ddesc de a index 1", "m0")
        self.input.send_keys(Keys.DOWN)
        self.test_insert_new_line("c","desc de c index 1", "m0a0")
        self.input.send_keys(Keys.DOWN)
        self.input.send_keys(Keys.DOWN)
        self.test_insert_new_line("e","desc de e index 1", "m0a1")
        self.test_insert_new_line("c","desc de c index 1", "m0a1")

        test_fine('Cadd')
        self.input.send_keys(Keys.DOWN)
        self.test_modif_key_desc("modif de c1")
        self.test_del_line()
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.LEFT)
        self.test_add_line_z0()
        self.cursor_move_until_this_id(Keys.DOWN, '')
        self.test_add_new_voc("o","observation(xx/20)","o_0(0/20)")
        self.cursor_move_until_this_id(Keys.DOWN, '')
        self.test_insert_new_line("o","o_un(10/20)")
        self.input.send_keys(Keys.DOWN)
        self.cursor_move_until_this_id(Keys.DOWN, '')
        self.test_insert_new_line("o","o_deux(20/20)")
        self.cursor_move_until_this_id(Keys.UP, 'm0e0')
        self.test_add_aggregation()
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.input.send_keys(Keys.UP)
        self.cursor_move_until_this_id(Keys.UP, ':m0a0,1')
        test_fine('Cagg')
        self.cursor_move_until_this_id(Keys.DOWN, 'm0e0')
        self.test_modif_aggregation()
        test_fine('COMP')


    def test_add_first_voc(self,vocValue,descVoc,descIndex):
        # ça créer 2 lignes : un voc #VocValue avec sa desc descVpc et une première clé du voc tel que VocValue0 descIndex
        self.button_plus_bot.click()
        self.click('#input_new_voc option[value="' + vocValue + '"]')
        time.sleep(2*WAIT_KEY)
        self.send_keys('#input_new_desc_voc', descVoc)
        self.driver.find_element(By.CSS_SELECTOR, '#button_save_voc').click()
        time.sleep(2*WAIT_KEY)
        self.send_keys('#recherche', descIndex)
        with self.wait_table_refreshed():
            self.driver.find_element(By.CSS_SELECTOR, '#button_inser').click()

    def test_add_new_voc(self,vocValue, descVoc, descIndex, expected=''):
        # insère une nouvelle ligne avec un nouveau voc (et donc une autre ligne avec la desc du nouveau voc)
        retry(lambda: self.cell_value() == expected)
        time.sleep(WAIT_KEY)
        self.button_plus_bot.click()
        self.click('#vocabulary_select #opt_add_new_voc')
        self.click('#input_new_voc option[value="' + vocValue + '"]')
        self.send_keys('#input_new_desc_voc', descVoc)
        self.driver.find_element(By.CSS_SELECTOR, '#button_save_voc').click()
        retry(lambda: self.driver.find_element(By.CSS_SELECTOR, '#vocabulary_select option:checked'))
        self.send_keys('#recherche', descIndex)
        with self.wait_table_refreshed():
            self.driver.find_element(By.CSS_SELECTOR, '#button_inser').click()

    def test_insert_new_line(self, vocValue, desc, expected=''):
        # insère une nouvelle ligne sous celle que l'on veut, vocValue est le voc voulu,
        # la desc de recherche est a changer a chaque fois sauf si voc different
        retry(lambda: self.cell_value() == expected)
        self.button_plus_bot.click()
        self.click('#vocabulary_select option[value="' + vocValue + '"]')
        time.sleep(WAIT_KEY)
        self.send_keys('#recherche', desc)
        with self.wait_table_refreshed(cell_value_change=False):
            self.driver.find_element(By.CSS_SELECTOR, '#button_inser').click()

    def test_modif_key_desc(self,newDesc):
        retry(lambda: self.cell_value() == 'm0a1c1')
        self.input.send_keys(Keys.RIGHT)
        self.input.send_keys(newDesc)
        self.input.send_keys(Keys.ENTER)

    def test_del_line(self):#permet la suppression la desc de la ligne courante
        self.input.send_keys(Keys.BACKSPACE)
        self.input.send_keys(Keys.ENTER)

    def test_add_line_z0(self) :
        retry(lambda: self.cell_value() == 'm0a1c1')
        self.button_plus_bot.click()
        self.click('#vocabulary_select option[value="e"]')
        self.send_keys('#recherche', "desc de e index 1")
        self.driver.find_element(By.CSS_SELECTOR, '#button_inser').click()

    def test_add_aggregation(self):#Ajoute une ligne discipline
        self.goto(13, 5)
        self.click('input[type="checkbox"][value="o1"]')
        self.click('input[type="checkbox"][value="o2"]', 1)
        self.click('#button_insert_line_aggregation')
        self.goto(13, 1)

    def test_modif_aggregation(self): #cette fonction ouvre la popup de modification de l'aggregation, puis l'on change des données et on enregirstre
        self.goto(13, 5)
        self.click('#id_method_choice_select option[value="median"]')
        self.click('input[type="checkbox"][value="o1"]')
        self.click('input[type="checkbox"][value="o2"]', 1)
        self.click('input[type="checkbox"][value="o0"]')
        self.click('input[type="checkbox"][value="o1"]', 1)
        self.click('#button_insert_line_aggregation')
        self.goto(13, 1)

    def test_show_error_button(self):
        retry(lambda: self.driver.find_element(By.CSS_SELECTOR, '#button_add_discipline'))
        self.driver.find_element(By.CSS_SELECTOR, '#button_add_discipline').click()
        self.driver.save_screenshot("xxx.png")

    # COMPETENCES END

    # Observation start

    def test_observations(self, base, ticket, year_semester):##

        # Competences tests initialisation

        year, semester = year_semester.split('/')
        try:
            os.unlink(f"DBregtest/Y{year}/S{semester}/test.py")
        except OSError:
            pass
        # self.deactivate_autosave()
        self.driver.get(f"{base}/{ticket}/{year_semester}/test")
        self.elements()
        self.deactivate_autosave()
        self.input.send_keys(Keys.ARROW_UP)
        time.sleep(WAIT_KEY)
        self.input.send_keys('k01')
        self.input.send_keys(Keys.ARROW_DOWN)
        time.sleep(WAIT_KEY)
        self.input.send_keys('k02')
        self.input.send_keys(Keys.ARROW_RIGHT)
        time.sleep(WAIT_KEY)
        self.input.send_keys('Zac')
        self.input.send_keys(Keys.ARROW_UP)
        time.sleep(WAIT_KEY)
        self.input.send_keys('Doe')
        self.input.send_keys(Keys.ARROW_RIGHT)
        time.sleep(WAIT_KEY)
        self.input.send_keys('John')
        self.input.send_keys(Keys.ARROW_DOWN)
        time.sleep(WAIT_KEY)
        self.input.send_keys('Zac')
        self.input.send_keys(Keys.ARROW_UP)

        self.input.send_keys(Keys.ARROW_RIGHT)
        self.input.send_keys(Keys.ARROW_RIGHT)
        self.input.send_keys(Keys.ARROW_RIGHT)
        self.set_column_type('Competences')
        self.set_column_title('Eval')
        self.input.send_keys(Keys.ARROW_UP)

        # self.driver.find_element(By.ID, 'current_input_div').click()

        for _ in range(10): # In some cases the first double click does not works
            self.hide_tip_popup()
            self.double_click()
            if self.popup_visible():
                break
            print('<R>', end=' ', flush=True)
        else:
            raise ValueError('No popup on double click')

        test_fine("INIT")

        # Competences table gui tests :

        self.test_comp_table_display()
        self.test_comp_line()
        self.test_detailed_line()
        self.test_subcomp_line()
        self.test_comp_jump()
        self.test_top_FUC()

        os.unlink('DBregtest/Y0/SCompetences/test_gui.py')

    def popup_visible(self):
        for _ in range(100):
            popup = self.driver.find_elements(By.CSS_SELECTOR, '#popup_id')
            if len(popup) == 1:
                return popup[0]
            time.sleep(0.01)
        return False

    def test_comp_table_display(self) :

        # Start : Popup open and category active is "formation"
        # End :   Only line m0 displayed

        # Table infos verify
        retry(lambda: self.table_inner_verify("Eval\nJohn Doe\n\nk01", "Informatique\n\nDescription creer aplication\n\n"
                                            + "savoir coder\n\nCoder en java"))

        # Formation verify
        formation_button = self.driver.find_elements(By.CSS_SELECTOR, 'button.top_FUC')[0]
        self.filter_verify(formation_button)
        retry(lambda: self.count_lines(4))

        # Searchbar test
        self.driver.find_element(By.CSS_SELECTOR, '#comp_searchbar').send_keys('Infoormatique')
        retry(lambda: self.count_lines(1))
        line = self.driver.find_elements(By.CSS_SELECTOR, 'DIV.comp_line')[0].get_attribute('innerText')
        if line != "Informatique" :
            raise ValueError('Search bar problem. expected line 1 : "Informatique"; find : "'+line+'" !')

        test_fine("TABLE")
    def test_comp_line(self) :
        self.hide_tip_popup()

        # Start : Only line m0 displayed and no params checked
        # End :   All params and greatest observation checked
        time.sleep(WAIT_KEY)
        for checkbox in self.driver.find_elements(By.CSS_SELECTOR, 'DIV.toggle_observations input'):
            checkbox.click()
        searchbar = self.driver.find_element(By.CSS_SELECTOR, '#comp_searchbar')
        self.driver.find_elements(By.CSS_SELECTOR, '.top_FUC')[0].click()
        for i in range(20):
            searchbar.send_keys(Keys.BACKSPACE)
        searchbar.send_keys(Keys.ENTER)
        time.sleep(WAIT_KEY)

        for i in range(1, 4):
            self.driver.find_element(By.CSS_SELECTOR, f'DIV.toggle_FUC input:nth-child({i})').click()
            time.sleep(WAIT_KEY) # Let recreate checboxes on change

        self.driver.find_element(By.CSS_SELECTOR, f'DIV.toggle_observations div:nth-child(2)').click()
        time.sleep(WAIT_KEY)

        #popup close
        self.driver.find_element(By.CSS_SELECTOR, 'DIV.import_export BUTTON.close').click()
        time.sleep(WAIT_KEY)

        retry(lambda: self.cell_value() == 'm0b0c0f0o1')
        # cell verification test
        self.input.send_keys('unvalid_value')
        self.input.send_keys(Keys.ENTER)
        self.alert_contains("m10c10z20")
        self.input.send_keys(Keys.ARROW_UP)
        self.input.send_keys('unvalid_value')
        self.input.send_keys(Keys.ENTER)
        self.alert_contains("unvalid_value")
        self.input.send_keys(Keys.ARROW_UP)
        retry(lambda: self.cell_value() == 'm0b0c0f0o1')

        # popup open
        self.double_click()
        retry(lambda: self.table_inner_verify("Eval\nJohn Doe\n\nk01", "Informatique\n\nDescription creer aplication\n\n"
                                            + "savoir coder\n\nCoder en java"))
        self.driver.find_element(By.CSS_SELECTOR, '#comp_searchbar').click()

        test_fine("LINE")
    def test_detailed_line(self) :
        self.driver.find_element(By.CSS_SELECTOR, '.depth4 .line_desc').click()
        grades = []

        # Write grades descriptions
        for i in range(1, 4):
            css_line = f'DIV.toggle_observations div:nth-child({i})'
            line = self.driver.find_element(By.CSS_SELECTOR, css_line)
            action = selenium.webdriver.ActionChains(self.driver)
            action.move_to_element_with_offset(line, 5, 5)
            action.perform()
            time.sleep(WAIT_KEY)

            self.driver.find_element(By.CSS_SELECTOR, css_line + ' .desc_edit_btn').click()
            paragraph = self.driver.find_element(By.CSS_SELECTOR, css_line + ' .editable')
            grades.append(paragraph.get_attribute('innerText'))
            grades[i - 1] ='bla' + str(i) + grades[i - 1]

            paragraph.send_keys('bla' + str(i))
            self.input.send_keys(Keys.ENTER)
            time.sleep(WAIT_KEY)

        action = selenium.webdriver.ActionChains(self.driver)
        action.move_to_element_with_offset(self.driver.find_element(By.CSS_SELECTOR, '#comp_searchbar'), 5, 5)
        action.perform()
        time.sleep(WAIT_KEY)

        # popup close and open
        self.driver.find_element(By.CSS_SELECTOR, 'DIV.import_export BUTTON.close').click()
        time.sleep(WAIT_KEY)
        self.double_click()
        time.sleep(WAIT_KEY)
        self.driver.find_element(By.CSS_SELECTOR, '.depth4 .line_desc').click()
        time.sleep(WAIT_KEY)

        descriptions = self.driver.find_element(By.CSS_SELECTOR, '.toggle_observations').get_attribute('innerText')
        if descriptions != grades[0] + "\n\n" + grades[1] + "\n\n" + grades[2]:
            raise ValueError('Grade descriptions are not saved !')

        test_fine("DETAILS")
    def test_subcomp_line(self) :
        self.driver.find_element(By.CSS_SELECTOR, '.subcompetence-add').click()
        time.sleep(WAIT_KEY)
        retry(lambda: self.count_lines(5))
        self.driver.find_element(By.CSS_SELECTOR, '.depth5 .line_desc').click()
        time.sleep(WAIT_KEY)

        text = self.driver.find_element(By.CSS_SELECTOR, '.depth5').get_attribute('innerText')
        expected = "Les compétences dans la formation\nLes compétences de l'UE\nLes compétences de cette colonne\n×\n\nNotes\n\n✎\n\nNouvelle Sous-compétence\n\nbla1Non noté\n\nbla2Mauvais\n\nbla3Bien"
        if text != expected:
            raise ValueError('New subcompetences are broken !\nexpected : ' + expected + '\nfound : ' + text)
        time.sleep(WAIT_KEY)

        self.driver.find_element(By.CSS_SELECTOR, '.title_edit_btn').click()
        subdesc = self.driver.find_element(By.CSS_SELECTOR, '.depth5 .editable.line_desc')
        subdesc.send_keys(Keys.CONTROL + "a")
        subdesc.send_keys('Maitriser le json')
        subdesc.send_keys(Keys.ENTER)
        time.sleep(WAIT_KEY)

        # popup close and open
        self.driver.find_element(By.CSS_SELECTOR, 'DIV.import_export BUTTON.close').click()
        time.sleep(WAIT_KEY)
        self.double_click()
        time.sleep(WAIT_KEY)
        self.driver.find_element(By.CSS_SELECTOR, '.depth5 .line_desc').click()
        time.sleep(WAIT_KEY)

        text = self.driver.find_element(By.CSS_SELECTOR, '.depth5 .line_desc').get_attribute('innerText')
        if text != 'Maitriser le json':
            raise ValueError('Edited subcompetences are broken')
        self.driver.find_element(By.CSS_SELECTOR, '.depth5 .toggle_observations div:nth-child(2) input').click()
        time.sleep(WAIT_KEY)

        for i in range(1, 4):
            css_line = f'.depth5 DIV.toggle_observations div:nth-child({i})'
            line = self.driver.find_element(By.CSS_SELECTOR, css_line)
            action = selenium.webdriver.ActionChains(self.driver)
            action.move_to_element_with_offset(line, 5, 5)
            action.perform()
            time.sleep(WAIT_KEY)

            self.driver.find_element(By.CSS_SELECTOR, css_line + ' .desc_edit_btn').click()
            paragraph = self.driver.find_element(By.CSS_SELECTOR, css_line + ' .editable')

            for i in range(4):
                paragraph.send_keys(Keys.DELETE)
            self.input.send_keys(Keys.ENTER)
            time.sleep(WAIT_KEY)

        self.driver.find_element(By.CSS_SELECTOR, '.subcompetence-delete').click()
        time.sleep(WAIT_KEY)
        retry(lambda: self.count_lines(4))

        test_fine("SUBCMP")
    def test_comp_jump(self) :

        # Start : Current cell : line 1
        # End :   Current cell : line 2

        self.goto(4, 2)

        # Table infos verify
        retry(lambda: self.table_inner_verify("Eval\nZac Zac\n\nk02", "Informatique\n\nDescription creer aplication\n\n"
                                            + "savoir coder\n\nCoder en java"))
        time.sleep(WAIT_KEY)

        # Boxes verify
        self.box_verify(False, self.driver.find_elements(By.CSS_SELECTOR, 'DIV.toggle_observations input'))
        self.box_verify(True, self.driver.find_elements(By.CSS_SELECTOR, 'DIV.toggle_FUC input'))

        # Arrows jump test
        self.input.send_keys(Keys.ARROW_UP)
        retry(lambda: self.table_inner_verify("Eval\nJohn Doe\n\nk01", "Informatique\n\nDescription creer aplication\n\n"
                                            + "savoir coder\n\nCoder en java"))
        time.sleep(WAIT_KEY)
        self.input.send_keys(Keys.PAGE_DOWN)

        test_fine("JUMP")
    def test_top_FUC(self) :

        # Start : Popup open
        # End : column and ue table changed

        searchbar = self.driver.find_element(By.CSS_SELECTOR, '#comp_searchbar')
        while searchbar.get_attribute('value'):
            searchbar.send_keys(Keys.BACKSPACE)
        self.hide_tip_popup()

        filter_buttons = self.driver.find_elements(By.CSS_SELECTOR, 'BUTTON.top_FUC')

        # Verify default filter
        self.filter_verify(filter_buttons[2])

        # Line filter refresh test
        self.click('DIV.toggle_FUC input', 2)
        retry(lambda: self.count_lines(0))

        # UE
        filter_buttons[1].click()
        retry(lambda: self.count_lines(4))
        self.filter_verify(filter_buttons[1])

        # Formation
        filter_buttons[0].click()
        retry(lambda: self.count_lines(4))

        locked_boxes = self.driver.find_elements(By.CSS_SELECTOR, '.toggle_FUC INPUT[disabled]') # -> locked parameters
        if len(locked_boxes) != 1 :
            raise ValueError('Formation problem, expected number of locked params : 1; find : '+str(len(locked_boxes)))
        locked_boxes = self.driver.find_elements(By.CSS_SELECTOR, '.toggle_observations INPUT[disabled]') # -> locked observations
        # It is the competence key that lock the observation, no more the UE
        # if len(locked_boxes) != 2 :
        #     raise ValueError('Formation problem, expected number of locked observations : 2; find : '+str(len(locked_boxes)))

        self.filter_verify(filter_buttons[1], False)
        self.filter_verify(filter_buttons[2], False)

        test_fine("FILTER")

    def count_lines(self, expected) :
        if expected != len(self.driver.find_elements(By.CSS_SELECTOR, 'DIV.comp_line')) :
            return False
        return True
    def table_inner_verify(self, expec_head, expec_lines) :
        for _ in range(10):
            popup = self.popup_visible()
            if popup:
                break
        else:
            raise ValueError('No popup')
        default_display = "\n\nF\nU\nC\n\nAcquisition\n\nCliquer sur l'intitulé pour voir plus. Taper ↑↓ pour changer d'étudiant\n\n"
        expec_table = expec_head + default_display + expec_lines + "\n\n×"
        ok = popup.get_attribute('innerText') == expec_table
        if not ok:
            ValueError("Found : " +repr(popup.get_attribute('innerText'))
            +   "\n\n Expected : " + expec_table)
        return ok
    def box_verify(self, check, boxes) :
        # boxes : selenium elements list
        # check : False for verify if all is uncheck, else True
        for i, box in enumerate(boxes) :
            if check and not box.get_attribute('checked') :
                raise ValueError('Checkbox problem. Expected have all boxes checked but number '+str(i+1)+' is not !')
            if not check and box.get_attribute('checked') :
                raise ValueError('Checkbox problem. Expected have all boxes unchecked but number '+str(i+1)+' is not !')
    def filter_verify(self, button, active=True) :
        if active and button.get_attribute('value') != "active" :
            return False
        if not active and button.get_attribute('value') == "active" :
            return False
        return True

    # Observation end

    def test_normal_table(self, base, ticket, year_semester):
        self.driver.get(f"{base}/{ticket}/{year_semester}/foobar")
        self.elements()
        self.remove_seq_column()
        self.deactivate_autosave()

        self.input.send_keys(Keys.ARROW_UP)
        self.input.send_keys('f')
        self.input.send_keys(Keys.ENTER)
        self.input.send_keys('g')
        self.input.send_keys(Keys.ENTER)
        self.input.send_keys('h')
        self.input.send_keys(Keys.ENTER)
        self.input.send_keys('i')
        self.input.send_keys(Keys.ENTER)
        self.input.send_keys('j')
        self.body.send_keys(Keys.ARROW_UP)
        self.body.send_keys(Keys.ARROW_UP)
        self.body.send_keys(Keys.ARROW_UP)
        self.body.send_keys(Keys.ARROW_UP)
        self.input.send_keys(Keys.ARROW_RIGHT)
        self.input.send_keys(Keys.ARROW_RIGHT)
        self.input.send_keys(Keys.ARROW_RIGHT)
        self.input.send_keys(Keys.ARROW_RIGHT)
        test_fine('INIT')

        def retry3(fct):
            """Retry three times the test"""
            for _ in "123":
                try:
                    fct()
                    return
                except: # pylint: disable=bare-except
                    try:
                        Alert(self.driver).accept() # Close garbage alert
                    except: # pylint: disable=bare-except
                        pass
                    print('fail ', end='', flush=True)
            fct()

        retry3(self.test_comment)
        retry3(self.test_enum_scroll)
        retry3(self.test_multiline)
        retry3(self.test_form)
        retry3(self.test_mcq)
        retry3(self.test_text)
        retry3(self.test_simple)
        retry3(self.test_prst)
        retry3(self.test_enum)
        retry3(self.test_columns)

    def test_comment(self):
        comment = self.driver.find_element(By.CSS_SELECTOR, '#comment')
        comment.click()
        comment.send_keys('A_COMMENT')
        retry(lambda: comment.get_attribute('value') == "A_COMMENT")
        comment.send_keys(Keys.ENTER)
        retry(lambda: len(self.driver.find_elements(By.CSS_SELECTOR, '.rw.comment')) == 1)
        self.input.send_keys(Keys.ARROW_DOWN)
        retry(lambda: len(self.driver.find_elements(By.CSS_SELECTOR, '.rw.comment')) == 1)
        self.input.send_keys(Keys.ARROW_UP)
        self.input.send_keys(Keys.ARROW_UP)

    def test_multiline(self):
        self.driver.find_element(By.CSS_SELECTOR, '#title_✎').click()
        multiline = self.driver.find_element(By.CSS_SELECTOR, '#t_editor')
        multiline.click()
        time.sleep(WAIT_KEY)
        multiline.send_keys("12.343")
        self.input.click()
        time.sleep(5*WAIT_KEY) # Why so much time after adding 'TAB_details' ?
        retry(lambda: self.cell_value() == "12.34")
        self.input.send_keys(Keys.BACKSPACE)
        multiline.click()
        retry(lambda: multiline.get_attribute('value') == "12.30")
        multiline.send_keys(Keys.BACKSPACE)
        multiline.send_keys(Keys.BACKSPACE)
        retry(lambda: multiline.get_attribute('value') == "12.")
        self.input.click()
        retry(lambda: self.cell_value() == "12.00")
        self.erase(self.input)
        test_fine('EDT✎')
        self.driver.find_element(By.CSS_SELECTOR, '#title_Cellule').click()

    def deactivate_autosave(self):
        self.driver.find_element(By.CSS_SELECTOR, '#title_Table').click()
        time.sleep(WAIT_KEY)
        choices = self.driver.find_element(By.CSS_SELECTOR, '#t_table_attr_nr_columns').get_attribute('innerHTML')
        if 'value="11"' not in choices:
            for choice in choices.split('value="')[1:]:
                if choice[0] >= '6':
                    self.driver.find_element(By.CSS_SELECTOR, '#t_table_attr_nr_columns OPTION[value="'
                        + choice[0] + '"]').click()
                    break
        retry(lambda: self.driver.find_element(By.CSS_SELECTOR, '#t_table_attr_nr_columns OPTION[value="11"]')).click()
        time.sleep(WAIT_KEY)
        self.driver.execute_script("table_autosave_toggle('t_table_attr_autosave')")

    def test_presence(self, base, year, ticket):
        self.driver.get(f"{base}/{ticket}/{year}/P/foobar")
        self.elements()
        self.deactivate_autosave()
        self.body.send_keys(Keys.ARROW_UP)
        self.input.send_keys('f')
        self.goto(3, 8)
        retry(lambda: self.cell_value() == '')
        self.input.send_keys('P')
        retry(lambda: self.cell_value() == 'PRST')
        self.input.send_keys(Keys.ENTER)
        while True:
            inputs = self.driver.find_elements(By.CSS_SELECTOR, '#P_inputs INPUT')
            if len(inputs) == 6:
                break
        feedback = self.driver.find_element(By.CSS_SELECTOR, '#P_feedback')
        feedback_last = feedback.get_attribute('innerText')
        def feedback_change():
            feedback_new = feedback_last
            while feedback_new == feedback_last:
                feedback_new = feedback.get_attribute('innerText')
            time.sleep(WAIT_KEY)
            return feedback_new
        inputs[0].click()
        inputs[0].send_keys('l')
        feedback_last = feedback_change()
        retry(lambda: inputs[0].get_attribute('value').startswith('Lundi'))
        inputs[0].send_keys(Keys.BACKSPACE)
        feedback_last = feedback_change()
        inputs[0].send_keys(Keys.BACKSPACE)
        feedback_last = feedback_change()
        inputs[0].send_keys('m')
        feedback_last = feedback_change()
        retry(lambda: self.select_values().count(' ') == 3) # Mardi Mercredi
        inputs[0].send_keys(Keys.ARROW_DOWN)
        inputs[0].send_keys(Keys.ENTER)
        retry(lambda: inputs[0].get_attribute('value').startswith('Mercredi'))
        feedback_last = feedback_change()

        inputs[0].send_keys(Keys.ARROW_RIGHT)
        time.sleep(WAIT_KEY)
        inputs[0].send_keys(Keys.ARROW_RIGHT)
        time.sleep(WAIT_KEY)
        inputs[0].send_keys(Keys.BACKSPACE)
        retry(lambda: len(inputs[0].get_attribute('value')) == 18)
        retry(lambda: self.select_values().count(' ') == 3) # Mardi Mercredi
        retry(lambda: self.driver.switch_to.active_element == inputs[0])
        inputs[0].send_keys(Keys.TAB)
        test_fine('P/I1')

        retry(lambda: self.driver.switch_to.active_element == inputs[1])
        inputs[1].send_keys("12")
        feedback_last = feedback_change()
        inputs[1].send_keys(Keys.TAB)
        retry(lambda: self.driver.switch_to.active_element == inputs[2])
        inputs[2].send_keys("13")
        feedback_last = feedback_change()
        inputs[2].send_keys(Keys.TAB)
        retry(lambda: self.driver.switch_to.active_element == inputs[3])
        inputs[3].send_keys("14")
        feedback_last = feedback_change()
        inputs[3].send_keys(Keys.TAB)
        retry(lambda: self.driver.switch_to.active_element == inputs[4])
        inputs[4].send_keys(Keys.ARROW_DOWN)
        time.sleep(WAIT_KEY)
        retry(lambda: self.driver.switch_to.active_element == inputs[4])
        inputs[4].send_keys("Course Title")
        feedback_last = feedback_change()
        inputs[4].send_keys(Keys.TAB)

        button = self.driver.find_element(By.CSS_SELECTOR, '#P_button')
        retry(lambda: button.get_attribute('disabled'))
        inputs[5].send_keys("Teacher")
        retry(lambda: not button.get_attribute('disabled'))
        time.sleep(WAIT_KEY)
        inputs[5].send_keys(Keys.TAB)
        retry(lambda: self.driver.switch_to.active_element == button)
        feedback_last = feedback_change()

        test_fine('P/I6')

        button.click()
        self.body.send_keys(Keys.ARROW_UP)
        retry(lambda: self.cell_value() == 'PRST')
        retry(lambda: self.driver.find_element(By.CSS_SELECTOR, '#comment').get_attribute('value') == 'Title Teacher ? ? 14#-2')
        test_fine('P/OK')

    def remove_seq_column(self):
        self.goto(3, 4)
        try:
            while True:
                time.sleep(WAIT_KEY)
                title = self.column_title.get_attribute('value')
                if title == 'Seq':
                    break
                if title == 'Grp':
                    return
            self.driver.find_element(By.CSS_SELECTOR, '#title_Affiche').click()
            while True:
                try:
                    time.sleep(WAIT_KEY)
                    self.driver.find_element(By.CSS_SELECTOR, '#t_column_hidden').click()
                    self.driver.find_element(By.CSS_SELECTOR, '#title_Colonne').click()
                    retry(lambda: self.column_title.get_attribute('value') == 'Grp')
                    return
                except selenium.common.exceptions.ElementNotInteractableException:
                    pass
        finally:
            self.goto(3, 1)
            retry(lambda: self.column_title.get_attribute('value') == 'ID')
            test_fine('HIDE')

    def test_mcq(self):
        self.set_column_type('Text')
        time.sleep(WAIT_KEY)
        self.set_column_type('MCQ')
        retry(lambda: 'Type → QCM' in self.messages.get_attribute('innerText'))
        self.input.send_keys(Keys.ARROW_UP)
        time.sleep(WAIT_KEY)
        while True:
            self.double_click()
            for _ in range(100):
                groups = self.driver.find_elements(By.CSS_SELECTOR, '#mcq_content .group')
                if len(groups) == 2:
                    break
            else:
                continue
            break
        groups[0].click()
        groups[0].send_keys('AAB')
        groups[1].click()
        groups[1].send_keys('a')
        retry(lambda: self.select_values() == 'a AAB')
        self.erase(groups[1])
        groups[1].send_keys('BBA')
        retry(lambda: self.select_values() == ' AAB')

        while True:
            groups = self.driver.find_elements(By.CSS_SELECTOR, '#mcq_groups SELECT OPTION[selected]')
            try:
                if (len(groups) == 2
                        and groups[0].get_attribute('innerText') == '1/1'
                        and groups[1].get_attribute('innerText') == '1/1'
                   ):
                    break
            except selenium.common.exceptions.StaleElementReferenceException:
                pass

        self.input.send_keys(Keys.ESCAPE)
        self.alert_contains('Enregistrer')

        self.input.send_keys(Keys.ESCAPE)
        test_fine('MCQ')

    def test_text(self):
        self.set_column_type('Text')
        time.sleep(WAIT_KEY)
        autocomplete = self.driver.find_element(By.CSS_SELECTOR, '#t_column_completion')
        if autocomplete.get_attribute('className') == 'linkstroked':
            autocomplete.click()
        self.erase_input()
        self.input.send_keys(Keys.ARROW_DOWN)
        self.input.send_keys(Keys.ARROW_DOWN)
        time.sleep(WAIT_KEY)
        self.input.send_keys('PRST')
        self.input.send_keys(Keys.ENTER)
        self.input.send_keys('ABJUS')
        self.input.send_keys(Keys.ENTER)
        self.input.send_keys('ABINJ')
        self.input.send_keys(Keys.ARROW_UP)
        self.input.send_keys(Keys.ARROW_UP)
        self.input.send_keys(Keys.ARROW_UP)
        self.input.send_keys(Keys.ARROW_UP)
        try:
            self.test_cell(text=True)
        finally:
            autocomplete.click()
        test_fine('TEXT')

    def test_simple(self):
        self.input.send_keys(Keys.ARROW_LEFT)
        time.sleep(WAIT_KEY)
        self.goto(4, 5)
        self.set_column_type('Note')
        self.erase_input()
        self.input.send_keys(Keys.ARROW_UP)
        self.erase_input()
        self.input.send_keys('1')
        self.input.send_keys(Keys.ARROW_DOWN)
        retry(lambda: self.cell_value() == '')
        self.input.send_keys(Keys.ARROW_UP)
        retry(lambda: self.cell_value() == '1.00')

        self.input.send_keys('1.994')
        self.input.send_keys(Keys.ENTER)
        retry(lambda: self.cell_value() == '')
        self.input.send_keys(Keys.ARROW_UP)
        self.input.send_keys(Keys.ARROW_UP)
        retry(lambda: self.cell_value() == '1.99')

        self.input.send_keys('1.996')
        retry(lambda: self.cell_value() == '1.996')

        self.input.send_keys(Keys.ARROW_RIGHT)
        retry(lambda: self.cell_value() == '')
        self.input.send_keys(Keys.ARROW_LEFT)
        retry(lambda: self.cell_value() == '2.00')

        time.sleep(2*WAIT_KEY)
        self.input.send_keys('3')
        self.goto(3, 1)
        retry(lambda: self.cell_value() == 'f')
        self.goto(3, 5)
        retry(lambda: self.cell_value() == '3.00')

        self.input.send_keys('4')
        self.input.send_keys(Keys.PAGE_DOWN)
        retry(lambda: self.cell_value() == '4.00')

        # In Xnest the focus is not lost
        # retry(lambda: not self.input_active())
        self.input.click()
        self.erase(self.input)
        self.input.send_keys('5')
        vertical_scrollbar = self.driver.find_element(By.CSS_SELECTOR, '#vertical_scrollbar')
        action = selenium.webdriver.ActionChains(self.driver)
        action.move_to_element_with_offset(vertical_scrollbar, 0, 300)
        action.click()
        action.perform()
        retry(lambda: self.cell_value() == '5.00')
        horizontal_scrollbar = self.driver.find_element(By.CSS_SELECTOR, '#horizontal_scrollbar')
        action = selenium.webdriver.ActionChains(self.driver)
        action.move_to_element_with_offset(horizontal_scrollbar, 20, 10)
        action.click()
        action.perform()
        retry(lambda: self.cell_value() == 'f')
        self.goto(3, 5)
        retry(lambda: self.cell_value() == '5.00')


        self.erase(self.input)
        self.input.send_keys(Keys.ENTER)
        self.input.send_keys(Keys.ARROW_UP)
        test_fine('BASE')

    def test_columns(self):
        self.goto(3, 5)
        self.driver.find_element(By.CSS_SELECTOR, '#title_Colonne').click()
        time.sleep(WAIT_KEY)
        for title in ('AAA', 'AAB', 'aAA', 'BAA', 'Average'):
            self.set_column_title(title)
            retry(lambda: self.column_title.get_attribute('value') == title) # pylint: disable=cell-var-from-loop
            self.input.send_keys(Keys.ARROW_RIGHT)
            time.sleep(WAIT_KEY)
            retry(lambda: self.column_title.get_attribute('value') != title)
        self.input.send_keys(Keys.ARROW_LEFT)

        self.set_column_type('Moy')
        retry(lambda: 'Type → Moyenne pondérée' in self.messages.get_attribute('innerText'))
        self.driver.find_element(By.CSS_SELECTOR, '#title_Formule').click()
        columns = self.driver.find_element(By.CSS_SELECTOR, '#t_column_columns')
        columns.click()
        columns.send_keys('b')
        retry(lambda: columns.get_attribute('value') == 'BAA')
        self.erase(columns)
        columns.send_keys('a')
        retry(lambda: self.select_values() == ' AAA AAB Average aAA')
        assert columns.get_attribute('value') == 'a'
        columns.send_keys(Keys.ENTER)
        retry(lambda: columns.get_attribute('value') == 'aAA')
        self.erase(columns)

        # Click on a cell without validation
        columns.send_keys('A')
        retry(lambda: self.select_values() == ' AAA AAB Average aAA')
        time.sleep(WAIT_KEY)
        self.goto(3, 2, False)
        self.alert_contains('«A»')
        self.goto(3, 9)
        retry(lambda: self.column_title.get_attribute('value') == 'Average')
        retry(lambda: columns.get_attribute('value') == '')

        # Click on a header without validation
        columns.click()
        time.sleep(WAIT_KEY)
        columns.send_keys('A')
        retry(lambda: columns.get_attribute('value') == 'A')
        time.sleep(WAIT_KEY)
        self.line_filter.click()
        self.alert_contains('«A»')
        retry(lambda: self.line_filter.get_attribute('class') == 'empty')
        retry(lambda: columns.get_attribute('value') == 'A')
        # Change column to restore value (empty)
        self.goto(3, 2)
        retry(lambda: self.column_title.get_attribute('value') == 'Nom')
        self.goto(3, 9)
        retry(lambda: self.column_title.get_attribute('value') == 'Average')
        retry(lambda: columns.get_attribute('value') == '')

        test_fine('COLS')

        time.sleep(WAIT_KEY)
        self.input.send_keys(Keys.ARROW_LEFT)
        time.sleep(WAIT_KEY)
        grps = self.driver.find_element(By.CSS_SELECTOR, '#t_column_groupcolumn')
        grps.click()
        time.sleep(2*WAIT_KEY)
        grps.send_keys('b')
        retry(lambda: grps.get_attribute('value') == 'BAA')
        self.erase(grps)
        time.sleep(WAIT_KEY)
        grps.send_keys('a')
        retry(lambda: self.select_values() == ' AAA AAB Average aAA')
        grps.send_keys(Keys.ESCAPE)
        grps.send_keys(Keys.ESCAPE)
        retry(lambda: grps.get_attribute('value') == '')

        test_fine('GRPS')

        self.goto(3, 5)
        self.driver.find_element(By.CSS_SELECTOR, '#title_Colonne').click()

    def test_enum(self):
        self.set_column_type('Enumeration')
        retry(lambda: 'Type → Énumération' in self.messages.get_attribute('innerText'))
        possibles = self.driver.find_element(By.CSS_SELECTOR, '#t_column_enumeration')
        self.erase(possibles)
        while True:
            try:
                possibles.send_keys('ABINJ ABJUS ABS PRST')
                break
            except selenium.common.exceptions.ElementNotInteractableException:
                time.sleep(0.1)
        self.test_cell()
        test_fine("ENUM")

    def test_enum_scroll(self):
        self.set_column_type('Enumeration')
        retry(lambda: 'Type → Énumération' in self.messages.get_attribute('innerText'))
        possibles = self.driver.find_element(By.CSS_SELECTOR, '#t_column_enumeration')
        while True:
            try:
                possibles.send_keys(' '.join(f'X{i}' for i in range(100)))
                break
            except selenium.common.exceptions.ElementNotInteractableException:
                time.sleep(0.1)
        self.body.send_keys(Keys.ARROW_UP)
        self.input.send_keys('x')
        select = retry(lambda: self.driver.find_element(By.CSS_SELECTOR, 'SELECT.login_list'))
        width = int(select.get_attribute('offsetWidth'))
        height = int(select.get_attribute('offsetHeight'))
        time.sleep(WAIT_KEY)
        action = selenium.webdriver.ActionChains(self.driver)
        # XXX Fragile
        action.move_to_element_with_offset(select, width - 5, height - 20)
        action.click()
        action.perform()
        time.sleep(WAIT_KEY)
        retry(lambda: self.cell_value() == 'x')
        self.input.send_keys(Keys.ENTER)
        retry(lambda: self.cell_value() == 'X0')
        self.erase(self.input)
        self.erase(possibles)
        self.set_column_type('Note')
        test_fine("SLCT")

    def test_prst(self):
        self.goto(3, 5)
        self.set_column_type('Prst')
        self.test_cell()
        test_fine("PRST")

    def test_form(self):
        self.set_column_type('Prst')
        self.driver.find_element(By.CSS_SELECTOR, '#table #title_Action').click()
        while True:
            time.sleep(WAIT_KEY)
            try:
                self.driver.find_element(By.CSS_SELECTOR, '#t_table_attr_forms').click()
                break
            except (IndexError, selenium.common.exceptions.ElementNotInteractableException):
                pass
        while True:
            time.sleep(WAIT_KEY)
            inputs = self.driver.find_elements(By.CSS_SELECTOR, '.formtable INPUT')
            if len(inputs) == 1:
                prst = inputs[0]
                grp = self.driver.find_element(By.CSS_SELECTOR, '.formtable TEXTAREA')
                break

        self.erase(prst)
        prst.send_keys('a')
        retry(lambda: self.select_values() == 'ABINJ ABJUS ABS')
        prst.send_keys(Keys.ENTER)
        retry(lambda: prst.get_attribute('value') == 'ABINJ')
        self.goto(4, 1)
        retry(lambda: prst.get_attribute('value') == '')
        self.goto(3, 1)
        prst.click()
        retry(lambda: prst.get_attribute('value') == 'ABINJ')
        prst.send_keys(Keys.BACKSPACE)
        prst.send_keys(Keys.BACKSPACE)
        prst.send_keys(Keys.BACKSPACE)
        retry(lambda: self.select_values() == 'ABINJ ABJUS ABS')
        prst.send_keys(Keys.ARROW_DOWN)
        prst.send_keys(Keys.ENTER)
        retry(lambda: prst.get_attribute('value') == 'ABJUS')
        grp.click()
        prst.click()
        retry(lambda: prst.get_attribute('value') == 'ABJUS')

        action = selenium.webdriver.ActionChains(self.driver)
        action.key_down(Keys.CONTROL)
        action.send_keys('a')
        action.key_up(Keys.CONTROL)
        action.perform()
        time.sleep(2*WAIT_KEY)
        retry(lambda: prst.get_attribute('value') == 'ABJUS')
        prst.send_keys(Keys.BACKSPACE) # Erase all
        time.sleep(WAIT_KEY)
        prst.send_keys('p')
        retry(lambda: prst.get_attribute('value') == 'PRST')

        self.goto(4, 1)
        retry(lambda: prst.get_attribute('value') == '')
        self.goto(3, 1)
        retry(lambda: prst.get_attribute('value') == 'PRST')

        self.body.send_keys(Keys.ESCAPE)
        self.driver.find_element(By.CSS_SELECTOR, '#title_Table').click()
        test_fine("FORM")
        self.goto(3, 5)
        retry(lambda: self.input.get_attribute('value') == 'PRST')
        self.erase_input()

    def test_cell(self, text=False): # pylint: disable=too-many-statements
        self.goto(3, 5)
        self.body.send_keys(Keys.ARROW_DOWN)
        self.body.send_keys(Keys.ARROW_UP)
        self.body.send_keys(Keys.ARROW_UP)
        time.sleep(2*WAIT_KEY)
        self.input.send_keys('p')
        retry(lambda: self.cell_value() == 'PRST')

        if text:
            abi = 'ABINJ ABJUS a'
            abi2 = 'ABINJ ABJUS a a'
            expected = 'a'
            expected2 = 'x'
        else:
            abi = abi2 = 'ABINJ ABJUS ABS'
            expected = expected2 = 'ABINJ'

        # Select by keyboard
        time.sleep(WAIT_KEY)
        self.erase_input()
        time.sleep(4*WAIT_KEY)
        self.input.send_keys('a')
        retry(lambda: self.select_values() == abi)
        assert self.selected_index() == '0'
        self.input.send_keys(Keys.ARROW_DOWN)
        assert self.selected_index() == '1'
        self.input.send_keys(Keys.ENTER)
        retry(lambda: self.cell_value() == 'ABJUS')

        # Select by mouse ABJUS
        time.sleep(2*WAIT_KEY)
        self.erase_input()
        time.sleep(2*WAIT_KEY)
        self.input.send_keys('a')
        retry(lambda: self.select_values() == abi)
        time.sleep(WAIT_KEY)
        self.driver.find_element(By.CSS_SELECTOR, '#tip SELECT OPTION[value="ABJUS"]').click()
        time.sleep(WAIT_KEY)
        retry(lambda: self.cell_value() == 'ABJUS')

        # Select by mouse ABINJ
        time.sleep(4*WAIT_KEY)
        self.erase_input()
        time.sleep(4*WAIT_KEY)
        self.input.send_keys('a')
        retry(lambda: self.select_values() == abi)
        self.driver.find_element(By.CSS_SELECTOR, '#tip SELECT OPTION[value="ABINJ"]').click()
        retry(lambda: self.cell_value() == 'ABINJ')

        # Cancel by Escape
        self.input.send_keys(Keys.ARROW_DOWN)
        self.input.send_keys(Keys.ARROW_UP)
        time.sleep(4*WAIT_KEY)
        self.input.send_keys('a')
        retry(lambda: self.cell_value() == 'a')
        retry(lambda: 'visibility: visible' in self.tip_style())
        self.input.send_keys(Keys.ESCAPE)
        retry(lambda: 'visibility: hidden' in self.tip_style())
        retry(lambda: self.cell_value() == 'a')
        self.input.send_keys(Keys.ESCAPE)
        retry(lambda: self.cell_value() == 'ABINJ')
        retry(lambda: 'visibility: hidden' in self.tip_style())

        # Cancel by clicking on another cell. Value is kept. No alert
        time.sleep(2*WAIT_KEY)
        self.erase_input()
        time.sleep(8*WAIT_KEY)
        self.input.send_keys('a')
        retry(lambda: self.select_values() == abi)
        retry(lambda: self.cell_value() == 'a')
        self.goto(3, 1, wait_ok=text)
        if not text:
            self.alert_contains('invalide')
        retry(lambda: self.cell_value() == 'f')
        self.goto(3, 5)
        retry(lambda: self.cell_value() == expected)

        # Cancel by clicking on an input header
        time.sleep(2*WAIT_KEY)
        self.erase_input()
        time.sleep(4*WAIT_KEY)
        self.input.send_keys('a')
        retry(lambda: self.select_values() == abi2)
        retry(lambda: self.cell_value() == 'a')
        self.driver.execute_script("console.log('BEFORE CLICK ' + element_focused)")
        self.line_filter.click()
        if not text:
            self.alert_contains('invalide')
        self.driver.execute_script("console.log('AFTER ALERT ' + element_focused)")
        self.driver.execute_script("console.log(['AFTER ALERT ', element_focused])")
        retry(lambda: not self.input_active())
        retry(lambda: self.cell_value() == expected)

        ## Enter an invalid value
        self.input.send_keys(Keys.ARROW_DOWN)
        self.input.send_keys(Keys.ARROW_UP)
        retry(lambda: self.cell_value() == expected)
        self.erase_input()
        self.input.send_keys('x')
        retry(lambda: self.cell_value() == 'x')
        self.goto(3, 1, wait_ok=text)
        if not text:
            self.alert_contains('invalide')
        self.goto(3, 5)
        retry(lambda: self.cell_value() == expected2)

        # Clear cell
        self.erase_input()
        self.input.send_keys(Keys.ENTER)
        self.input.send_keys(Keys.ARROW_UP)

    def alert_contains(self, text):
        text = ''
        for _ in range(10):
            try:
                alert = Alert(self.driver)
                text = alert.text
                break
            except ValueError:
                time.sleep(0.1)
        if text in alert.text:
            alert.accept()
        else:
            print(alert.text)
            raise ValueError("Alert")

    def close_alert(self):
        try:
            Alert(self.driver).accept()
        except selenium.common.exceptions.NoAlertPresentException:
            pass

    def input_active(self):
        return 'blue' in self.driver.find_element(By.CSS_SELECTOR, '#current_input_div').get_attribute('style')

    def cell_value(self):
        return self.input.get_attribute('value')

    def goto(self, line, column, wait_ok=True):
        if wait_ok:
            self.hide_tip_popup()
        self.close_alert()
        if (self.driver.execute_script('return the_current_cell.col + 1') == column
                and self.driver.execute_script('return the_current_cell.lin + 1') == line):
            # Move elsewhere
            self.driver.find_element(By.CSS_SELECTOR, 
                '#thetable TR:nth-child(10) TD:nth-child(2)').click()
            retry(lambda: self.driver.execute_script('return the_current_cell.col + 1') == 2)
            if wait_ok:
                self.close_alert()
                self.hide_tip_popup()

        self.driver.find_element(By.CSS_SELECTOR, 
            f'#thetable TR:nth-child({line}) TD:nth-child({column})').click()
        if wait_ok:
            self.close_alert()
            retry(lambda: self.driver.execute_script('return the_current_cell.col + 1') == column)

    def selected_index(self):
        return self.driver.find_element(By.CSS_SELECTOR, '#tip SELECT').get_attribute('selectedIndex')

    def select_values(self):
        return ' '.join(
            option.get_attribute('value')
            for option in self.driver.find_elements(By.CSS_SELECTOR, '#tip OPTION')
            )
    def tip_style(self):
        return self.driver.find_element(By.CSS_SELECTOR, '#tip').get_attribute('style')

    def erase(self, an_input): # pylint: disable=no-self-use
        an_input.click()
        time.sleep(WAIT_KEY)
        action = selenium.webdriver.ActionChains(self.driver)
        action.key_down(Keys.CONTROL)
        action.send_keys('a')
        action.key_up(Keys.CONTROL)
        action.send_keys(Keys.BACKSPACE)
        action.perform()
        retry(lambda: an_input.get_attribute('value') == '')

    def erase_input(self):
        self.erase(self.input)

    def set_column_type(self, typ):
        self.close_alert()
        test_fine(f'[{typ}', end='')
        column_type = self.driver.find_element(By.CSS_SELECTOR, '#t_column_type')
        time.sleep(WAIT_KEY)
        column_type.click()
        self.close_alert()
        while True:
            links = self.driver.find_elements(By.CSS_SELECTOR, '.type_chooser_div A')
            if links:
                for i in links:
                    if i.get_attribute('type') == typ:
                        i.click()
                        retry(lambda: not self.popup_visible())
                        retry(lambda: self.driver.execute_script('return the_current_cell.column.type') == typ)
                        test_fine(']')
                        return
                raise ValueError('Bug')
            time.sleep(0.01)

    def set_column_title(self, title):
        self.column_title.click()
        self.erase(self.column_title)
        self.column_title.send_keys(title)
        self.column_title.send_keys(Keys.ENTER)
        self.column_title.send_keys(Keys.ARROW_UP)

    def elements(self):
        while not self.driver.find_element(By.CSS_SELECTOR, '#linefilter'):
            time.sleep(0.1)
        self.body = self.driver.find_element(By.CSS_SELECTOR, 'BODY')
        self.input = self.driver.find_element(By.CSS_SELECTOR, '#current_input')
        self.line_filter = self.driver.find_element(By.CSS_SELECTOR, '#linefilter')
        self.column_title = self.driver.find_element(By.CSS_SELECTOR, '#t_column_title')
        self.messages = self.driver.find_element(By.CSS_SELECTOR, '#content_Messages')

        # click OK
        time.sleep(WAIT_KEY)
        self.input.send_keys(Keys.ENTER)
        retry(self.input_active)
        time.sleep(WAIT_KEY)
        self.input.send_keys(Keys.ARROW_UP)


with open('DBregtest/Y0/SDossiers/config_table.py', 'a', encoding='utf-8') as f:
    f.write("cell_change(1,'0_2','language','fr','20220321185042')\n")

LOGS = open('xxx.testgui.log', 'w', encoding='utf-8')

TOMUSS = subprocess.Popen( # pylint: disable=consider-using-with
    ['./tomuss.py', 'regtest'],
    stdout=LOGS, stderr=subprocess.STDOUT)

def get_xnest():
    """Get en invisible X11 server if possible"""
    if 'hidden' in sys.argv:
        x11 = ['Xvfb', ':?', '-noreset', '-screen', '0', '1024x768x24']
    else:
        x11 = ['Xnest', ':?', '-noreset', '-geometry', '1024x768']
    try:
        for port in '123456789':
            x11[1] = ':' + port
            print(x11, end=" ")
            xnest = subprocess.Popen( # pylint: disable=consider-using-with
                x11, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            time.sleep(0.1)
            if not xnest.poll():
                print("Started")
                os.environ['DISPLAY'] = x11[1]
                return xnest
            print("Can't start")
    except FileNotFoundError:
        print("X11 not found: run directly on your screen")

XNEST = get_xnest()
if XNEST:
    try:
        WINDOW_MANAGER = subprocess.Popen( # pylint: disable=consider-using-with
            ['i3', '-c', 'REGTEST_SERVER/i3config'],
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        print("«i3» window manager launched")
    except FileNotFoundError:
        print("«i3» window manager not found: test may fail")

try:
    EXIT_CODE = 1
    while True:
        if not IN_DOCKER and 'FF' not in sys.argv:
            # These options do not make Chrome driver working
            # options = selenium.webdriver.ChromeOptions()
            # options.add_argument("--password-store=basic")
            # options.add_argument("--disable-setuid-sandbox")
            # options.add_argument("--no-sandbox")
            # options.add_argument("--disable-webgl")
            # options.add_argument("--disable-gpu")
            # options.add_argument("--headless")
            # options.add_argument("--no-default-browser-check")
            # options.add_argument("--no-first-run")
            # options.add_argument("--disable-dev-shm-usage")
            # options.add_argument("--window-size=800,600")
            # options.add_argument("--user-data-dir=" + os.getcwd())
            # Tests(selenium.webdriver.Chrome(options=options))
            Tests(selenium.webdriver.Chrome())
        Tests(selenium.webdriver.Firefox())
        if '1' in sys.argv:
            EXIT_CODE = 0
            break
except: # pylint: disable=bare-except
    traceback.print_exc()
finally:
    if 'hidden' not in sys.argv:
        time.sleep(1000000)
    TOMUSS.kill()
    TOMUSS.wait()
    if XNEST:
        XNEST.kill()
        XNEST.wait()
    with open('DBregtest/Y0/SDossiers/config_table.py', 'a', encoding='utf-8') as f:
        f.write("cell_change(1,'0_2','language','en','20220321185042')\n")
    sys.exit(EXIT_CODE)
