#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2021 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

import os
import collections
import json
import re
import itertools
import datetime
import time
import html
import io
from typing import Dict, Tuple, Union
import xlsxwriter
from .. import configuration
from .. import utilities
from .. import document
from .. import plugin
from .. import teacher
from .. import cell
from .. import column

COLUMNS = {
    '0': {'type': 'Login',
          'title': utilities._("COL_TITLE_0_0"),
          'repetition': 1,
          'freezed': 'F'
         },
    '1': {'type': 'Surname',
          'title': utilities._("COL_TITLE_0_2"),
          'columns': utilities._("COL_TITLE_0_0"),
          'freezed': 'F'
         },
    '2': {'type': 'Firstname',
          'title': utilities._("COL_TITLE_0_1"),
          'columns': utilities._("COL_TITLE_0_0"),
          'freezed': 'F'
         },
    '4': {'type': 'Status',
          'columns': utilities._("COL_TITLE_0_0"),
          'title': utilities._("TITLE_student_type"),
          'freezed': 'F'
         },
    '6': {'type': 'Text',
          'columns': utilities._("COL_TITLE_0_0"),
          'title': utilities._("TITLE_student_contact"),
         },
    '8': {'type': 'Text',
          'columns': utilities._("COL_TITLE_0_0"),
          'title': utilities._("TITLE_student_contact_adm"),
         },
    '9': {'type': 'Nmbr',
          'columns': '*Prst',
          'test_filter': configuration.abi,
          'title': 'Σ'+configuration.abi,
          'rounding': 1
         },
    }

def create(table):
    """Create the P table"""
    page = table.get_nobody_page()
    table.update_columns(COLUMNS, ro_page=page)
    table.table_attr(page, 'default_sort_column', [1, 2, 0])
    _ = utilities._
    table.table_attr(page, 'table_title', _("MSG_P_title").format(table.ue))
    table.column_attr(page, '6', 'comment', json.dumps(
        {
            'subject': _('P_mail_subject'),
            'header': _('P_mail_header'),
            'item': _('P_mail_item'),
            'footer': _('P_mail_footer'),
            'duration': _('P_default_duration'),
            'abi_content': _('MSG_QRCode_ABI_mail'),
            'abi_cc': '',
            'to1': _('P_to1_default'),
            'to2': _('P_to2_default'),
            'to3': _('P_to3_default'),
            'mail_from': _('P_mail_from_default'),
            'student_rule': _('P_student_default_rule'),
            'to1_rule': _('P_to1_default_rule'),
            'to2_rule': _('P_to2_default_rule'),
            'to3_rule': _('P_to3_default_rule'),
            'pc_abi': _('P_pc_abi_default'),
            'send_to_tutor': 1,
            'send_to_admin': 1,
        }))

    # SVC plugin may create the table, so table.user does not exists
    # table.table_attr(page, 'masters', [table.user])

# pylint: disable=too-many-locals,too-many-branches,too-many-arguments,too-many-statements
def display_etape_table(file, students, nr_colors, the_date, students_of_etape):
    """Nice colored table with all the students"""
    columns = set()
    for user in students_of_etape:
        columns.update(students[user.student_id])
    columns = sorted(columns)

    teachers = collections.defaultdict(set)
    courses = {'': 'gray'}
    def get_color(comment):
        code, author, _ = (comment + ' ').split(' ', 2)
        author = author.split('@')[0]
        if code not in courses:
            courses[code] = len(courses) % nr_colors
        if author and code:
            teachers[code].add(author)
        return courses[code]

    txt = [f'''<table class="prst"><tr>
        <th rowspan="2">{utilities._('COL_TITLE_0_0')}
        <th rowspan="2">{utilities._('COL_TITLE_0_2')}
        <th rowspan="2">{utilities._('COL_TITLE_0_1')}
        <th rowspan="2">{utilities._('TITLE_student_type').replace('_', '<br>')}'''
        ]
    for day, names in itertools.groupby(columns, lambda x: x[:10]):
        if not day.startswith(the_date):
            continue # pragma: no cover
        txt.append(f'<th colspan="{len(tuple(names))}"><span class="day">{day}</span>')
    txt.append('</tr><tr>')
    for name in columns:
        if not name.startswith(the_date):
            continue # pragma: no cover
        txt.append(f'<th><span class="hour">{name[11:]}</span>')
    txt.append('</tr>')
    for user in students_of_etape:
        calendar = []
        for when, (value, what, late, _year, _table) in students[user.student_id].items():
            course, tname = what.split(' ', 2)[0:2]
            tname = tname.split('@')[0]
            calendar.append(f'<tr><td>{when}<td>{value}<td>{course:15}<td>{tname}<td>{late}</tr>\n')
        calendar.sort()
        details = f'''<table class="hidden"><tr>
            <td colspan=4>{user.student_id} {user.surname} {(user.firstname or "").title()}
            </tr>{"".join(calendar)}</table>'''

        txt.append(f'''<tr class="prst">
            <th>{details}{user.student_id}
            <th>{user.surname}
            <th>{user.firstname}
            <th>{user.status}''')
        last_column = ''
        for column in columns:
            if not column.startswith(the_date):
                continue # pragma: no cover
            value, comment, late, _year, _table = students[user.student_id].get(column, ('', '', 0, '', ''))
            if column[:10] == last_column[:10]:
                td_class = '' # pragma: no cover
            else:
                td_class = 'day'
            if value:
                td_class += f' c{get_color(comment)}'
            if value.startswith(configuration.pre):
                if late > 60:
                    value = '⚈' # pragma: no cover
                elif late > 45:
                    value = '◕' # pragma: no cover
                elif late > 30:
                    value = "◑" # pragma: no cover
                elif late > 15:
                    value = "◔" # pragma: no cover
                else:
                    value = " "
            elif value.startswith(configuration.abi):
                value = '⚈'
                td_class += ' abi'
            elif value.startswith(configuration.abj): # pragma: no cover
                value = "⚈"
                td_class += ' abj'
            if td_class:
                td_class = ' class="' + td_class + '"'
            txt.append(f'<td{td_class} TITLE="{comment} ({late}min)">{value}')
            last_column = column
        txt.append('</tr>')
    txt.append('</table><p>')
    for course, color in courses.items():
        teacher_names = '<br>'.join(sorted(teachers[course]))
        txt.append(
            f' <div class="inline"><span class="c{color}">{course}</span><br>{teacher_names}</div>')
    file.write(''.join(txt))

def display_etape_text(file, etape, students, the_date, students_of_etape,
                       header, footer, display_duration, hide_teacher):
    """Textual reporting with all the students"""
    _ = utilities._
    all_ues = teacher.all_ues()
    etp = all_ues.get('etape-' + str(etape), None)
    if isinstance(the_date, str):
        dates = the_date + '....' # pragma: no cover
    else:
        dates = the_date[0] + ' → ' + the_date[-1] # pragma: no cover
    if not etp:
        etp = all_ues.get(etape, None)
    if etp:
        etp = '« ' + html.escape(etp.intitule()) + ' »' # pragma: no cover
    else:
        etp = ''
    teachers = set()
    for user in students_of_etape:
        for when, (value, what, _late, _year, _table) in students[user.student_id].items():
            if when.startswith(the_date):
                teachers.add(what.split(' ', 2)[1].split('@')[0])
    teachers = configuration.Users(teachers)
    page_break = ''
    duration = ''
    if display_duration:
        label = utilities._('F_h').lower()
    else:
        label = utilities._('P_slot').lower()
    for user in students_of_etape:
        counters = collections.defaultdict(int)
        txt = []
        last_day = ''
        for when, (value, what, _late, _year, _table) in students[user.student_id].items():
            if when.startswith(the_date):
                ue_code, teacher_login = what.split(' ', 2)[0:2]
                teacher_login = teacher_login.split('@')[0]
                if '-' in ue_code:
                    ue_name = all_ues.get(ue_code.split('-', 1)[1], None)
                else:
                    ue_name = all_ues.get(ue_code, None) # pragma: no cover
                if ue_name:
                    ue_name = html.escape(ue_name.intitule()) # pragma: no cover
                else:
                    ue_name = '?'
                if display_duration:
                    what = re.split('[ #]', what) # ue_code, login, grp, semestre, duration, id
                    if len(what) == 6:
                        duration = '<td>' + what[4]
                    else:
                        duration = '<td>'
                if last_day != when[:10]:
                    new_day = ' class="new_day"'
                    last_day = when[:10]
                    date = when.replace('_', ' ')
                else: # pragma: no cover
                    new_day = ''
                    date = when[11:]
                value = value.strip('@')
                if display_duration:
                    if len(what) == 6:
                        counters[value] += int(what[4] or '0') / 60
                else:
                    counters[value] += 1
                txt.append(f'<!--{when}--><tr{new_day}><td>{date}{duration}<td>{value}'
                           f'<td class="ue">{ue_code.replace("UE-", "")}<br>{ue_name}')
                if hide_teacher:
                    txt[-1] += '</tr>\n'
                else:
                    user_teacher = teachers[teacher_login] # pylint: disable=unsubscriptable-object
                    if user_teacher.surname == 'Inconnu':
                        name = teacher_login.replace('.', ' ') # pragma: no cover
                    elif user_teacher.surname is None:
                        name = '?' # pragma: no cover
                    else:
                        name = user_teacher.firstname.title() + '<br>' + user_teacher.surname
                    txt[-1] += f'<td class="teacher">{name}</tr>\n'
        txt.sort()
        if user.firstname is None:
            continue # pragma: no cover
        txt.insert(
            0,
            f"""<h1{page_break}>{user.student_id} {user.surname} {user.firstname.title()}</h1>
            <p>{header}
            <p>{etape} {etp}
            <!-- <p>{dates} -->
            <p>
            <table class="text">
            <tr><th>{_("B_Date")}{'<th>⏱' if display_duration else ''}<th><th>UE
            """)
        if hide_teacher:
            txt[0] += '</tr>'
        else:
            txt[0] += f"<th>{_('COL_TITLE_teacher')}</tr>"
        txt.append("</table><p>")
        for abx in ('pre', 'abi', 'abj'):
            txt.append(utilities._('MSG_T_' + abx).title())
            txt.append(f" : {counters[getattr(configuration, abx)]:.1f} {label}. ")
        txt.append("<p>")
        txt.append(footer)
        file.write(''.join(txt))
        page_break = ' class="text"'

def display_etape(file, etape, students, students_of_etape, nr_colors,
                  date_begin, date_end, the_status, text_mode,
                  header, footer, display_duration, hide_teacher, all_status):
    """Display the students of an etape"""
    if not text_mode:
        file.write(f'<h1>{etape} {date_begin}→{date_end} {the_status}</h1>')
    if the_status:
        students_of_etape = [
            user
            for user in students_of_etape
            if re.match(the_status, all_status[user.student_id])
        ]
    students_of_etape.sort(key=lambda x: (x.surname or '?',
                                          x.firstname or '?',
                                          x.student_id or '?'))
    the_date = get_dates(date_begin, date_end)
    if text_mode:
        display_etape_text(file, etape, students, the_date, students_of_etape,
                           header, footer, display_duration, hide_teacher)
    else:
        display_etape_table(file, students, nr_colors, the_date, students_of_etape)

RE_DATE = re.compile('^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]_[0-9][0-9]:[0-9][0-9]$')

def p_get_students_of_a_table(year, table_name):
    """
    Returns:
       (student id, column_title) → (cell value, cell comment, late, year, table)
       errors
    """
    table = document.table(year, 'P', table_name)
    students = collections.defaultdict(dict)
    qrcodes = {}
    status = {}
    all_errors = []
    for column in table.columns:
        if not RE_DATE.match(column.title):
            continue
        if not hasattr(column, 'qrcode_prst'):
            continue # pragma: no cover
        datacol = column.data_col
        # Search duplicate QRCodes (in case of bug)
        for qrcode in column.qrcode_prst:
            if qrcode in qrcodes: # pragma: no cover
                all_errors.append(f'<tr><td>{table}<td>{qrcode}'
                                  f'<td>{qrcodes[qrcode].title}'
                                  f'<td>{qrcodes[qrcode].qrcode_prst}'
                                  f'<td>{column.title}'
                                  f'<td>{column.qrcode_prst}'
                                  f'</tr>')
            qrcodes[qrcode] = column
        # *Estimate* the first usage of QRCode in the column
        qrcodes_start = collections.defaultdict(list)
        for line in table.lines.values():
            status[line[0].value] = line[3].value
            cell = line[datacol]
            if cell.comment and cell.value:
                record_time = 60 * int(cell.date[-6:-4]) + int(cell.date[-4:-2])
                qrcode = cell.comment.split('#')[-1]
                qrcodes_start[qrcode].append(record_time)
        for qrcode, time_list in qrcodes_start.items():
            time_list.sort()
            # We assume 10% of the student scan before courses start
            # We never take the first time because it may be the teacher.
            qrcodes_start[qrcode] = time_list[min(len(time_list) // 10 + 1,
                                                    len(time_list) - 1)]
        when = column.title
        start = 60 * int(when[-5:-3]) + int(when[-2:])
        for line in table.lines.values():
            student_id = line[0].value
            if not student_id:
                continue # pragma: no cover
            cell = line[datacol]
            if not cell.value:
                continue # pragma: no cover
            record_time = 60 * int(cell.date[-6:-4]) + int(cell.date[-4:-2])
            if '#' in cell.comment and cell.comment.count(' ') >= 3:
                students[student_id][when] = (
                    cell.value, cell.comment,
                    record_time - qrcodes_start[cell.comment.split('#')[-1]],
                    year, table_name)
            else:
                students[student_id][when] = ( # pragma: no cover
                    cell.value, ' ' + cell.author, record_time - start,
                    year, table_name)
    return students, all_errors, status

class Merged:
    """
    Contains all P informations for a year.
    Caching is done per table.
    """
    def __init__(self, year):
        self.year = year
        self.students = {} # student → [table]
        self.tables = {} # table_name → {student → {date → (value, what, late, year, table)} }
        self.errors = {} # table_name → [error]
        self.parse_time = {} # table_name → timestamp
        self.status = {} # student → status # Assume the same for all tables
        self.last_parse_time = self.creation_time = time.time()
    def add_student(self, student, table_name):
        """Add a student to the table"""
        if student in self.students:
            if table_name not in self.students[student]:
                self.students[student].append(table_name)
        else:
            self.students[student] = [table_name]
    def parse_table(self, name):
        """Update self.students if necessary"""
        self.tables[name], self.errors[name], status = p_get_students_of_a_table(self.year, name)
        self.status.update(status)
        for student in self.tables[name]:
            self.add_student(student, name)
        self.last_parse_time = self.parse_time[name] = time.time()
    def scan_tables_for_change(self): # pragma: no cover
        """For 'suivi' server.
        If a table was modified on disk, parse it."""
        for name, timestamp in self.parse_time.items():
            filename = document.table_filename(self.year, 'P', name)
            try:
                if os.path.getmtime(filename) > timestamp:
                    self.clear_table(name)
            except FileNotFoundError:
                pass
    def clear_table(self, name):
        """Table will be reparsed when needed"""
        self.tables.pop(name, None)
    def __iter__(self):
        """Iterate on all students"""
        return iter(self.students)
    def items(self):
        """Simulate a dictionnary"""
        for student in self.students:
            yield student, self[student]
    def __getitem__(self, student):
        """Update if necessary and merge informations"""
        if student not in self.students:
            return {}
        merged = {}
        for table in self.students[student]:
            if table not in self.tables:
                self.parse_table(table)
            merged.update(self.tables[table].get(student, {}))
        return merged
    def get_errors(self):
        """Merge all table errors"""
        merged_errors = []
        for errors in self.errors.values():
            if errors: # pragma: no cover
                merged_errors.append('<p><table class="bug">')
                merged_errors.extend(errors)
                merged_errors.append('<p></table>')
        return ''.join(merged_errors)
    def __str__(self): # pragma: no cover
        students = '\n'.join(f"{student} {tables}" for student, tables in self.students.items())
        tables = '\n'.join(f"{table} {students}" for table, students in self.tables.items())
        return f'''MERGED {self.year}
Students:
{students}
Tables:
{tables}
'''


MERGEDS = {}
def p_get_students(year):
    """
    Returns: Merged
    """
    if year in MERGEDS:
        if configuration.read_only:
            merged = MERGEDS[year]
            now = time.time()
            if now - merged.creation_time < 86400: # No full scan # pragma: no cover
                if now - merged.last_parse_time >= 300:
                    merged.scan_tables_for_change()
                    merged.last_parse_time = now # In case of no table parsed
                return merged
            # Full rescan of the year to find new files
        else:
            return MERGEDS[year]
    # The first time full directory reading is needed.
    # Next time only modified or created tables will be parsed.
    MERGEDS[year] = merged = Merged(year)
    dirname = os.path.join(configuration.db, f'Y{year}', 'SP')
    if os.path.exists(dirname):
        for name in os.listdir(dirname):
            if not name.endswith('.py'):
                continue # pragma: no cover
            merged.parse_table(name[:-3])
    return merged

def get_student(year, student):
    """
    Returns a list of:
        ('%d/%m/%Y-%H:%M', (PRST_ABI_ABJ, comment, late_in_minutes, year, table))
    """
    return sorted(p_get_students(year)[student].items())

def cell_change(table, _page, _col, lin, _value, _date):
    """Clear cache"""
    merged = p_get_students(table.year)
    merged.clear_table(table.ue)
    # Really important to update the new table
    if table.lines[lin][0].value:
        merged.add_student(table.lines[lin][0].value, table.ue)

def fix_date(txt):
    """Translate 1/1/20 as 2020-01-01"""
    try:
        day, month, year = (int(i) for i in txt.split('/'))
        if year < 2000: # pragma: no cover
            year += 2000
        return f'{year:04d}-{month:02d}-{day:02d}' # pragma: no cover
    except ValueError:
        return txt

def get_dates(date_begin: str, date_end: str) -> Union[str, Tuple[str]]:
    """If a day or date_end given, then a tuple is returned with all the week dates"""

    date_begin = fix_date(date_begin)
    date_end = fix_date(date_end)
    try:
        year, month, date = date_begin.split('-')
        day = datetime.date(int(year), int(month), int(date))
        dates = []
        # day -= datetime.timedelta(days=day.weekday()) # goto Monday
        for i in range(1000):
            dates.append(day.strftime('%Y-%m-%d'))
            day += datetime.timedelta(days=1)
            if date_end == dates[-1]:
                break
            if not date_end and i == 6:
                break # pragma: no cover
        return tuple(dates)
    except ValueError:
        return date_begin

def P_more_reporting(_server): # pylint: disable=invalid-name
    """To be redefined"""
    return ''
configuration.P_more_reporting = P_more_reporting

def p_reporting(server):
    """Read ALL presences and consolidates the data"""
    if len(server.the_path) < 5:
        if server.the_path:
            etape = server.the_path[0]
        else:
            etape = ''
        server.the_file.write(
            f"""<!DOCTYPE html>
<title>Reporting</title>
<style>
BODY {{ font-family: sans-serif ; }}
</style>
<script>
function reporting(action, more) {{
    window.open('{configuration.server_url}/={server.ticket.ticket}/'
        + action + '/{server.the_year}/'
        + document.getElementById('etape').value + '/'
        + document.getElementById('datebegin').value.replace(/\\//g,'$2F') + '/'
        + document.getElementById('dateend').value.replace(/\\//g,'$2F') + '/'
        + document.getElementById('status').value + '/'
        + more);
}}
</script>
<script src="{configuration.url_files}/utilities.js"></script>
<h1>Reporting</h1>
<p>{utilities._('P_etape_filter')}<br>
   <input id="etape" value="{etape}">
<p>{utilities._('P_date_filter_begin')}<br>
   <input id="datebegin" value="{time.strftime('%d/%m/%Y')}">
<p>{utilities._('P_date_filter_end')}<br>
   <input id="dateend" value="{time.strftime('%d/%m/%Y')}">
<p>{utilities._('P_status_filter')}<br>
   <input id="status">
<p>{utilities._('P_header')}<br>
   <textarea id="header" cols="80" rows="4"></textarea>
<p>{utilities._('P_footer')}<br>
   <textarea id="footer" cols="80" rows="4"></textarea>
<p>{utilities._("P_counter")}<br>
<select id="count">
<option value="0">{utilities._("P_slot")}</option>
<option value="1" default>{utilities._("F_h").title()}</option>
</select>
<p>{utilities._("P_hide_teacher")}<br>
   <input type="radio" id="hide_teacher">
<hr>
<input type="submit" value="{utilities._('B_stat_colored').title()}"
       onclick="javascript:reporting('P_reporting','false/'
                + encode_uri(document.getElementById('header').value) + '/'
                + encode_uri(document.getElementById('footer').value) + '/'
                + document.getElementById('count').value + '/'
                + document.getElementById('hide_teacher').checked
                )">
<input type="submit" value="{utilities._('TITLE_table_attr_print')}"
       onclick="javascript:reporting('P_reporting','true/'
                + encode_uri(document.getElementById('header').value) + '/'
                + encode_uri(document.getElementById('footer').value) + '/'
                + document.getElementById('count').value + '/'
                + document.getElementById('hide_teacher').checked
                )">
<input type="submit" value="XLS"
onclick="javascript:reporting('P_reporting_xls',
                              document.getElementById('count').value + '/report.xlsx')">
{configuration.P_more_reporting(server)}
""")
        return
    merged = p_get_students(server.the_year)
    (the_etape, date_begin, date_end, the_status,
     text_mode, header, footer, display_duration, hide_teacher) = (
        server.the_path + ['', '', '', '', '', '', '', '', ''])[:9]
    text_mode = text_mode == 'true'
    display_duration = int(display_duration)
    hide_teacher = hide_teacher == 'true'
    if not the_etape:
        server.the_file.write(merged.get_errors()) # pragma: no cover
    server.the_file.write('''<!DOCTYPE html>
    <title>Reporting</title>
    <script>
    var displayed ;
    function show(event) {
        var t = event.target;
        while (t) {
            if(t.tagName == 'TABLE' && t.className.indexOf('hidden') != -1)
                return;
            t = t.parentNode;
        }
        if(displayed) {
            displayed.className = 'hidden';
            displayed = undefined;
        }
        t = event.target;
        while(t) {
            if (t.tagName == 'TR' && t.cells[0].firstChild.className == 'hidden' ) {
                displayed = t.cells[0].firstChild;
                displayed.className += ' show';
                return;
            }
            t = t.parentNode;
        }
    }
    window.onclick = show;
    </script>
    <style>
    TABLE.bug { border-spacing: 0px ; }
    TABLE.bug, TABLE.bug TD { border: 1px solid #000; }
    TABLE.prst, TABLE.text { border-spacing: 0px ; }
    TABLE.prst { border-right: 1px solid #000; border-bottom: 1px solid #000;  }
    TABLE.prst TH { border-left: 1px solid #000; border-top: 1px solid #000; font-weight: normal }
    TABLE.prst TD { border-left: 1px solid #DDD; border-top: 1px solid #DDD; text-align: center }
    TABLE.prst TR:nth-child(3) TD { border-top: 1px solid #000; }
    SPAN.hour, SPAN.day {
        display: block ;
        font-weight: normal;
        width: 1em;
        white-space: nowrap;
    }
    SPAN.hour {
        transform: translate(1.2em, 1em) rotate(-90deg);
        height: 3.5em;
        }
    SPAN.day {
        transform: translate(2.6em, 2em) rotate(-90deg);
        height: 6.5em;
        }
    TABLE.prst TD.day { border-left: 1px solid #000; }
    TABLE.prst TH:nth-child(1) { text-align: left }
    TABLE.prst TH:nth-child(2) { text-align: left }
    TABLE.prst TH:nth-child(3) { text-align: left }
    TABLE.prst TH:nth-child(4) { text-align: left }
    .abi { color: #F00 }
    .abj { color: #00F }
    .cgray { background: #AAA }
    TABLE.prst TR.prst:hover TD { filter: contrast(130%) ; }
    TABLE.prst TR.prst:hover TH { background: #EEE }
    TABLE.prst.labels TD { text-align: left }
    .inline { display: inline-block; vertical-align: top; background: #EEE }
    .hidden { display: none }
    .hidden.show { display: block ; position:fixed; top:0px ; right: 0px;
        background: #FFD; z-index: 1; bottom: 0px; overflow:auto; padding: 1em }
    TABLE.text { border-collapse: collapse; }
    TABLE.text TD { border: 1px solid #000; padding: 0.2em }
    TABLE.text TR.new_day TD { border-top: 2px solid #000; }
    TABLE.text TD:first-child { text-align: right; white-space: nowrap}
    TABLE.text TD.ue, TABLE.text TD.teacher { font-size: 70%}
    H1.text { page-break-before: always }
    </style>
    ''')
    nr_colors = 8
    if text_mode:
        pass
    else:
        server.the_file.write('<style>')
        def hue(i):
            """Compute course color"""
            return int(i * (nr_colors // 2 + 1) * 360 / nr_colors)
        server.the_file.write(
            ''.join(f'.c{i} {{ background: hsl({hue(i)}, 50%, 70%) }}'
                    for i in range(nr_colors)))
        server.the_file.write(f'''</style>
    <table class="prst labels">
    <tr><td class=""> <td>{server._('P_nothing')}</tr>
    <tr><td class="c2 abi">⚈<td>{configuration.abi}</tr>
    <tr><td class="c2 abj">⚈<td>{configuration.abj}</tr>
    <tr><td class="c2"> <td>{server._('P_present')}</tr>
    <tr><td class="c2">◔<td>{server._('P_late').format(15)}</tr>
    <tr><td class="c2">◑<td>{server._('P_late').format(30)}</tr>
    <tr><td class="c2">◕<td>{server._('P_late').format(45)}</tr>
    <tr><td class="c2">⚈<td>{server._('P_late').format(60)}</tr>
    <tr><td class="cgray"> <td>{server._('P_teacher')}</tr>
    </table>
    ''')
    users = configuration.Users(merged)
    date = fix_date(date_begin)
    year = int(date[:4])
    if date[4:8] < '-07-': # Before july it's the previous year
        year -= 1 # pragma: no cover
    if the_etape == '':
        by_etapes = collections.defaultdict(list)
        for user in users:
            by_etapes[tuple(user.get_etapes_list(year) or ())].append(user)
        for etape, students_of_etape in by_etapes.items():
            display_etape(server.the_file, etape, merged, students_of_etape,
                          nr_colors, date_begin, date_end, the_status, text_mode,
                          header, footer, display_duration, hide_teacher, merged.status)
    else:
        students_of_etape = []
        for user in users:
            etapes = user.get_etapes_list(year)
            if etapes and the_etape in etapes:
                students_of_etape.append(user.student_id) # pragma: no cover
        if not students_of_etape: # pragma: no cover
            # No an etape but something else.
            try:
                # If 'groups' attribute is defined for users, try this.
                for user in users:
                    for _nature, code, uyear, _date in user.groupes:
                        if code == the_etape and year == uyear:
                            students_of_etape.append(user.student_id)
            except AttributeError:
                pass
        # To not recompute attributes of unused students
        students_of_etape = configuration.Users(students_of_etape)
        display_etape(server.the_file, the_etape, merged, list(students_of_etape),
                      nr_colors, date_begin, date_end, the_status, text_mode,
                      header, footer, display_duration, hide_teacher, merged.status)

plugin.Plugin('P_reporting', '/P_reporting/{Y}/{*}',
              group="staff",
              function=p_reporting,
              launch_thread=True,
             )

@utilities.add_a_cache
def get_week(date):
    """Return the monday"""
    year, month, date = date.split('-')
    day = datetime.date(int(year), int(month), int(date))
    day -= datetime.timedelta(days=day.weekday()) # goto Monday
    return day.strftime('%Y-%m-%d')

def display_counter(count, nbr, minutes):
    """Compute the value to display in the XLS cell"""
    if count == 0:
        return nbr or ''
    if count == 1:
        return minutes / 60 or ''
    return -1 # pragma: no cover

class Counter:
    """Counters for one week"""
    __slots__ = ('pre_nr', 'pre_min', 'abj_nr', 'abj_min', 'abi_nr', 'abi_min')
    def __init__(self):
        self.pre_nr = self.pre_min = self.abj_nr = self.abj_min = self.abi_nr = self.abi_min = 0
    def pre(self, count):
        """Presence"""
        return display_counter(count, self.pre_nr, self.pre_min)
    def abj(self, count):
        """Abjus"""
        return display_counter(count, self.abj_nr, self.abj_min)
    def abi(self, count):
        """absence"""
        return display_counter(count, self.abi_nr, self.abi_min)

class Stat: # pylint: disable=too-few-public-methods
    """Compute student stats"""
    def __init__(self):
        self.counters = collections.defaultdict(Counter)
        self.abi_days = set()
    def add(self, when, value, what):
        """Add a QRCode state to the stats"""
        what = re.split('[ #]', what) # ue, login, grp, semestre, duration, id
        if len(what) == 6:
            duration = int(what[4])
        else:
            duration = 0
        counter = self.counters[get_week(when)]
        if value == configuration.pre:
            counter.pre_nr += 1
            counter.pre_min += duration
        elif value.startswith(configuration.abj):
            counter.abj_nr += 1
            counter.abj_min += duration
        elif value.startswith(configuration.abi):
            counter.abi_nr += 1
            counter.abi_min += duration
            self.abi_days.add(when)
        else:
            raise ValueError('bug') # pragma: no cover

def p_stats(student, dates) -> Stat:
    """Returns student stats"""
    stat = Stat()
    for when, (value, what, _late, _year, _table) in student.items():
        when = when.split('_')[0]
        if isinstance(dates, str):
            if not when.startswith(dates):
                continue
        else:
            if when not in dates:
                continue # pragma: no cover
        stat.add(when, value, what)
    return stat


def p_reporting_xls(server):
    """Send the XLS file with statistics"""
    etapes, date_begin, date_end, the_status, count, _filename = server.the_path
    merged = p_get_students(server.the_year)
    users = configuration.Users(merged)
    by_etape = collections.defaultdict(list)
    count = int(count)
    all_etapes = etapes == ''
    for user in users:
        if the_status and not re.match(the_status, merged.status.get(user.student_id, '')):
            continue # pragma: no cover
        for etape in user.get_etapes_list(server.the_year) or ():
            if etape in etapes or all_etapes:
                by_etape[etape].append(user)
    dates = get_dates(date_begin, date_end)
    file_content = io.BytesIO()
    workbook = xlsxwriter.Workbook(file_content)
    f_pre = workbook.add_format({'left': 1})
    f_abj = workbook.add_format({'color': '#0000FF'})
    f_abi = workbook.add_format({'color': '#FF0000', 'right': 1})
    first_col = 4
    for etape, students_of_etape in by_etape.items():
        worksheet = workbook.add_worksheet(etape)
        all_stats:Dict[str, Stat] = {}
        mondays = set()
        for user in students_of_etape:
            all_stats[user.student_id] = stats = p_stats(merged[user.student_id], dates)
            mondays.update(stats.counters)

        mondays = sorted(mondays)
        col_first = xlsxwriter.utility.xl_col_to_name(first_col)
        col_last = xlsxwriter.utility.xl_col_to_name(first_col + 3*len(mondays) - 1)
        col_sum_pre = xlsxwriter.utility.xl_col_to_name(first_col + 3*len(mondays))
        col_sum_abi = xlsxwriter.utility.xl_col_to_name(first_col + 3*len(mondays) + 2)
        worksheet.merge_range(0, 0, 0, 3, date_begin + ' → ' + date_end)
        if mondays:
            for col, monday in enumerate(mondays + ['Σ']):
                worksheet.merge_range(0, first_col + 3*col, 0, first_col + 3*col+2, monday)
                worksheet.write(1, first_col + 3*col, configuration.pre_short) # pylint: disable=no-member
                worksheet.write(1, first_col + 3*col+1, configuration.abj_short) # pylint: disable=no-member
                worksheet.write(1, first_col + 3*col+2, configuration.abi_short) # pylint: disable=no-member
                worksheet.set_column(first_col + 3*col, first_col + 3*col, 4, f_pre)
                worksheet.set_column(first_col + 3*col+1, first_col + 3*col+1, 4, f_abj)
                worksheet.set_column(first_col + 3*col+2, first_col + 3*col+2, 4, f_abi)
                worksheet.write(2, first_col + 3*col, 1)
                worksheet.write(3, first_col + 3*col, 0)
                worksheet.write(4, first_col + 3*col, 0)
                worksheet.write(2, first_col + 3*col+1, 0)
                worksheet.write(3, first_col + 3*col+1, 1)
                worksheet.write(4, first_col + 3*col+1, 0)
                worksheet.write(2, first_col + 3*col+2, 0)
                worksheet.write(3, first_col + 3*col+2, 0)
                worksheet.write(4, first_col + 3*col+2, 1)
            col = len(mondays)
            worksheet.set_column(first_col + 3*col+3, first_col + 3*col+3, 4)
            worksheet.write(1, first_col + 3*col+3, 'Σ')
            worksheet.set_column(first_col + 3*col+4, first_col + 3*col+4, 100)
            worksheet.write(1, first_col + 3*col+4, configuration.abi + ' ' + server._('F_d'))
        worksheet.set_row(2, None, None, {'hidden': True})
        worksheet.set_row(3, None, None, {'hidden': True})
        worksheet.set_row(4, None, None, {'hidden': True})
        for line, user in enumerate(sorted(students_of_etape,
                                           key=lambda x: (x.surname, x.firstname))):
            line += 5
            worksheet.write(line, 0, user.student_id)
            worksheet.write(line, 1, user.surname)
            worksheet.write(line, 2, user.firstname.title())
            worksheet.write(line, 3, user.status)
            if not mondays:
                continue
            stats = all_stats[user.student_id]
            for col, monday in enumerate(mondays):
                if monday not in stats.counters:
                    continue # pragma: no cover
                counter = stats.counters[monday]
                worksheet.write(line, first_col + 3*col, counter.pre(count))
                worksheet.write(line, first_col + 3*col+1, counter.abj(count))
                worksheet.write(line, first_col + 3*col+2, counter.abi(count))
            col = len(mondays)
            worksheet.write(
                line, first_col + 3*col,
                f"=sumproduct({col_first}{line+1}:{col_last}{line+1}, {col_first}3:{col_last}3)",
                None, '')
            worksheet.write(
                line, first_col + 3*col+1,
                f"=sumproduct({col_first}{line+1}:{col_last}{line+1}, {col_first}4:{col_last}4)",
                None, '')
            worksheet.write(
                line, first_col + 3*col+2,
                f"=sumproduct({col_first}{line+1}:{col_last}{line+1}, {col_first}5:{col_last}5)",
                None, '')
            worksheet.write(
                line, first_col + 3*col+3,
                f"=sum({col_sum_pre}{line+1}:{col_sum_abi}{line+1})", None, '')
            worksheet.write(line, first_col + 3*col+4, ' '.join(sorted(stats.abi_days)))
    workbook.close()
    server.the_file.write(file_content.getvalue())

plugin.Plugin('P_reporting_xls', '/P_reporting_xls/{Y}/{*}',
              group="staff",
              function=p_reporting_xls,
              mimetype="application/vnd.ms-excel",
              launch_thread=True,
             )

def attendance(server):
    """All the attendance for an UE"""
    merged = p_get_students(utilities.university_year(server.the_year, server.the_semester))
    code = server.the_ue
    by_student = collections.defaultdict(lambda: collections.defaultdict(list))
    by_date = collections.defaultdict(set)
    for student_id, all_dates in merged.items():
        for when, infos in all_dates.items():
            if infos[1].startswith(code):
                by_student[student_id][when].append(infos)
                by_date[when].add(student_id)
    dates = []
    last_students = set(by_student)
    # Merge students  hour1 and hour2 if the two student lists do not overlap.
    for when, students in sorted(by_date.items()):
        if students & last_students:
            last_students = students
            dates.append([when])
            continue # Not mergeable: student in common
        dates[-1].append(when) # pragma: no cover
        last_students |= students # pragma: no cover

    lines = []
    students = [user
                for user in configuration.Users(by_student)
                if user.firstname
               ]
    empty = cell.CellEmpty()
    nr_abi_max = 0
    for student in sorted(students, key=lambda x: (x.surname, x.firstname)):
        when = by_student[student.student_id]
        line = [
            cell.CellValue(student.student_id),
            cell.CellValue(student.surname),
            cell.CellValue(student.firstname.title()),
            empty,
        ]
        nr_abi = 0
        for some_dates in dates:
            the_cell = empty
            for date in some_dates:
                for info in when.get(date, ()):
                    the_cell = cell.Cell(info[0], comment=info[1].split(' ')[1])
                    if info[0].startswith(configuration.abi):
                        nr_abi += 1
                    break
                if the_cell is not empty:
                    break
            line.append(the_cell)
        line[3] = cell.Cell(str(nr_abi))
        nr_abi_max = max(nr_abi_max, nr_abi)
        lines.append(cell.Line(line))
    columns = [
        column.Column('0', '', freezed='F', width=2,
                      title=server._('COL_TITLE_ID')),
        column.Column('1', '', width=4,
                      title=server._('COL_TITLE_surname'),
                      ),
        column.Column('2', '', width=4,
                      title=server._('COL_TITLE_firstname'),
                      ),
        column.Column('3', '', width=2, type="Note",
                      rounding=1, minmax=f'[0;{nr_abi_max}]',
                      title=configuration.abi,
                      ),
        ]
    red = f'{configuration.abi}|{configuration.abj}'
    green = configuration.abj
    for some_dates in dates:
        columns.append(column.Column(str(len(columns)), '', width=1, type="Prst",
                                     red=red, green=green,
                                     comment=' '.join(some_dates),
                                     title=str(len(columns) - 3)))
    document.virtual_table(server, columns, lines,
                           table_attrs={
                                        'default_nr_columns': len(columns),
                                        'modifiable': 0,
                                        'title': f'{server.the_year}/{server.the_semester}/{code}',
                                        })


plugin.Plugin('attendance', '/{Y}/{S}/{U}/attendance',
              group="staff",
              function=attendance,
              launch_thread=True,
             )
