#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2009-2018 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr


"""This TEMPLATE can be modified and it will update table in usage
to match the modification. For example, the columns order.
"""

import html
from .. import utilities
from .. import data
from .. import configuration
from . import _ucbl_

prototype = "_ucbl_"

def referent_resume(table, login):
    s = []
    first = True
    for line in table.get_lines(login):
        for i, col in enumerate(table.columns):
            if i >= 3 and line[i].value:
                if first:
                    s.append('<div class="blocnote">'
                             + utilities._("MSG_Referents_suivi")
                             % (table.year, table.year+1) + '<br>')
                    first = False
                s.append('%s&nbsp;:&nbsp;<b>%s</b>,'
                         % ( col.title,
                             html.escape(str(line[i].value)) ))
    if first == False:
        s.append('</div>')
    return '\n'.join(s)

# The ID starting by 0_ are here for compatibility with _ucbl template
referent_columns = {
    '0_0' : {'position': 0,'type':'Text', "width":4, "freezed":'F',
             'repetition': 1},
    '0_1' : {'position': 1,'type':'Text', "width":6, "freezed":'F'},
    '0_2' : {'position': 2,'type':'Text', "width":6, "freezed":'F'},
    '0_3' : {'hidden':1},
    '0_4' : {'hidden':1},
    '0_5' : {'position': 3,'type':'Text', "width":1, "freezed":'C',
             "hidden":1, 'title':'Inscrit'},
    'FiRe': {'position': 4,'type':'Bool', "width":1, "freezed":'F'},
    'CON1': {'position': 5,'type':'Bool', "width":2},
    'REM1': {'position': 6,'type':'Text', "width":6},
    'RDV1': {'position': 7,'type':'Prst', "width":2},
    'RDV2': {'position': 8,'type':'Prst', "width":2},
    'JUR1': {'position': 9,'type':'Text', "width":6},
    'CON2': {'position':10,'type':'Bool', "width":2},
    'REM2': {'position':11,'type':'Text', "width":6},
    'RDV3': {'position':12,'type':'Prst', "width":2},
    'RDV4': {'position':13,'type':'Prst', "width":2},
    'JUR2': {'position':14,'type':'Text', "width":6},
    }

def init(table):
    _ucbl_.init(table)
    table.default_sort_column = [2, 1, 0]
    table.private = 1
    table.modifiable = int(utilities.university_year() == table.year)
    table.referent_resume = referent_resume
    table.referent_columns = referent_columns

def check_columns(table):
    if len(table.columns) != 0 and table.columns.from_id('FiRe') is None: # pragma: no cover
        table.modifiable = 0
        return # Old table : no more columns update
    cols = dict(referent_columns.items())
    cols.update(configuration.local_columns(table))
    for k, v in cols.items():
        x = utilities._('COL_TITLE_' + k)
        if x != 'COL_TITLE_' + k:
            v['title'] = x
        x = utilities._('COL_COMMENT_' + k)
        if x != 'COL_COMMENT_' + k:
            v['comment'] = x
    ro = {}
    rw = {}
    for k, v in cols.items():
        if k.startswith('0_'):
            ro[k] = v
        else:
            column = table.columns.from_id(k)
            if column and column.author == data.ro_user:
                # This is an old referent table
                continue # pragma: no cover
            rw[k] = v
    table.update_columns(ro)
    table.update_columns(rw, table.get_a_master_page())

def content(dummy_table):
    return _ucbl_.update_student_information

def onload(table):
    """This function is only here to correct a bug on old referent files"""
    user = utilities.module_to_login(table.ue)
    if user not in table.masters: # pragma: no cover
        utilities.warn('Missing master for ' + table.ue, what='warning')
        table.masters.append(user)

def create(table):
    p = table.get_ro_page()
    table.table_attr(p, 'masters', [utilities.module_to_login(table.ue)])
    table.table_attr(p, 'default_nr_columns', 9)
    table.table_attr(p, 'default_sort_column', [2, 1, 0])
    check_columns(table)

def update_inscrits_referents(the_ids, table, page):
    from .. import referent
    table.lock()
    try:
        check_columns(table)
    finally:
        table.unlock()
    if configuration.year_semester[1] == configuration.university_semesters[0]: # pragma: no cover
        contrat = 'CON1'
    else:
        contrat = 'CON2'

    fire = table.columns.from_id('FiRe').data_col
    done = {}
    for the_id in referent.students_of_a_teacher(utilities.module_to_login(table.ue)):
        login = utilities.the_login(the_id)
        done[login] = True
        user = configuration.User(login)
        the_ids[the_id] = user.mail
        # COpy/Paste with Favoris.py
        if user.need_a_charte: # pragma: no cover
            if utilities.charte_signed(login, year=table.year,
                                       semester=table.semester):
                s = configuration.yes
            else:
                s = configuration.no
        else:
            if user.manage_key('rsskey'):
                s = configuration.yes # pragma: no cover
            else:
                s = configuration.no
        pe = user.in_first_year
        table.lock()
        try:
            line = table.get_line_id_or_create(login)
            lin = tuple(table.get_items(login))
            if not lin or len(lin[0]) <= fire or lin[0][fire].value == '':
                if pe is True:
                    fr = configuration.yes # pragma: no cover
                elif pe is False:
                    fr = configuration.no
                else:
                    fr = '' # pragma: no cover
                table.cell_change(page, "FiRe" , line, fr)
            table.cell_change(page, "0_0"  , line, the_id)
            table.cell_change(page, "0_1"  , line, user.firstname)
            table.cell_change(page, "0_2"  , line, user.surname)
            table.cell_change(page, contrat, line, s)

        finally:
            table.unlock()

    students_to_remove = [the_id
                          for the_id in table.logins()
                          if utilities.the_login(the_id) not in done]
    _ucbl_.remove_students_from_table(table, students_to_remove)

def check(table, update_inscrits=update_inscrits_referents):
    _ucbl_.check(table, update_inscrits)
