#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2023 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

import time
import ast
import sys
import re
import json
from TOMUSS import configuration
from TOMUSS import document
from TOMUSS import utilities
from TOMUSS import column
from TOMUSS import sender

def create(table):
    """Create the table"""
    page = table.get_a_root_page()
    table.get_nobody_page() # Needed for update
    table.table_attr(page, 'masters', [table.user])
    table.table_attr(page, 'default_sort_column', [0])
    table.table_attr(page, 'comment', "")
    _ = utilities._
    attrs = [
        {'title': "ID",
          'comment': _("semester") + '/' + _('TH_table'),
          'repetition': 1, 'width': 16},
        {'width': 1, 'title': 'official_ue'       , 'rounding': 1},
        {'width': 1, 'title': 'private'           , 'rounding': 1},
        {'width': 1, 'title': 'modifiable'        , 'rounding': 1},
        {'width': 1, 'title': 'rounding'          , 'rounding': 1},
        {'width': 2, 'title': 'default_nr_columns', 'rounding': 1},
        {'width': 8, 'title': 'masters'},
        {'width': 8, 'title': 'teachers'},
        {'width': 8, 'title': 'table_title'},
        {'width': 8, 'title': 'dates'},
        {'width': 8, 'title': 'theme'},
        {'width': 8, 'title': 'comment'},
        {'width': 8, 'title': 'columns'},
    ]
    columns = {}
    for i, attr in enumerate(attrs):
        attr['position'] = i
        attr['type'] = 'Text'
        if 'comment' not in attr:
            key = 'TIP_table_attr_' + attr['title']
            comment = utilities._(key)
            if key != comment:
                attr['comment'] = re.sub('<[^>]*>', '', comment).replace('\n', ' ')
        columns[attr['title']] = attr
    table.update_columns(columns, table.get_a_master_page())

def init(table):
    table.contains_users = 0

def check(table):
    """Update all the values"""
    page = table.get_nobody_page()
    for line_id in table.lines:
        update_line(table, line_id, page)

def get_table_and_line(table, line_id):
    """Load the table"""
    line = table.lines[line_id]
    try:
        semester, ue_code = str(line[0].value).split('/')
    except ValueError:
        return None, None
    delta_year = configuration.semesters_year[configuration.semesters.index(semester)]
    return document.table(table.year - delta_year, semester, ue_code, create=False), line

def update_line(table, line_id, page):
    """Update the information for one table"""
    source, line = get_table_and_line(table, line_id)
    if not source:
        return
    with table.the_lock:
        for col in table.columns[1:]:
            if col.title == 'columns':
                columns = source.columns.js(False, python=True)
                while len(columns and columns[0]['author']) == 1:
                    columns.pop(0)
                columns = [ccol   for ccol in columns   if 'title' in ccol]
                columns.sort(key = lambda x: x['position'])
                for ccol in columns:
                    ccol.pop('position')
                    ccol.pop('the_id')
                    ccol.pop('author')
                value = (
                    '[\n'
                    + ',\n'.join(json.dumps(ccol, ensure_ascii=False, separators=(',', ':'))
                                 for ccol in columns)
                    + '\n]')
            elif col.title not in column.TableAttr.attrs:
                continue
            else:
                value = str(getattr(source, col.title))
            if value == str(line[col.data_col].value):
                continue
            table.cell_change(page, col.the_id, line_id, str(value), force_update=True)

TO_UPDATE = []

def check_new_value(old_value, value):
    """Returns evalued value and error message"""
    if isinstance(old_value, str):
        return value, None
    try:
        if isinstance(old_value, (float, int)):
            return type(old_value)(value), ''
        value = ast.literal_eval(value)
        if isinstance(value, type(old_value)):
            return value, None
        return None, 'BUG'
    except (ValueError, SyntaxError):
        return None, str(sys.exc_info()[0])

def update_columns(table, value, user_name):
    """Update or add columns to the table"""
    try:
        columns = json.loads(value)
    except ValueError:
        return 'Invalid JSON'
    page = table.get_a_page_for((user_name,))
    error = 'ok.png'
    for col in columns:
        if 'title' not in col:
            error = 'Column title missing'
            continue
        dest_col = table.columns.from_title(col['title'])
        if not dest_col:
            # Search non existent ID to create column
            for i in range(9999):
                col_id = f'T{i}'
                if not table.columns.from_id(col_id):
                    break
            dest_col = table.add_empty_column(page, col_id)
            if 'position' not in col:
                col['position'] = 1000 + 10 * i
        for attr, val in col.items():
            attribute = column.ColumnAttr.attrs.get(attr, None)
            if not attribute:
                error = "Unknown attribute: " + attr
                continue
            if getattr(dest_col, attr, None) == attribute.encode(val):
                continue
            is_ok = table.column_attr(page, dest_col.the_id, attr, val)
            if is_ok != 'ok.png' and attr != 'visibility_date':
                error = is_ok
    return error

def do_update():
    """Update the line content if first column change.
    Else update other table attributes."""
    last_test = time.time()
    while TO_UPDATE:
        table, page, col_id, line_id, value = TO_UPDATE.pop(0)
        if col_id == 'ID':
            update_line(table, line_id, table.get_nobody_page())
            continue
        destination, _line = get_table_and_line(table, line_id)
        if col_id == 'columns':
            with destination.the_lock:
                error = update_columns(destination, value, page.user_name)
        else:
            old_value = getattr(destination, col_id, None)
            if old_value is None:
                continue # Unknown attribute: user defined column
            new_value, error = check_new_value(old_value, value)
            if not error:
                with destination.the_lock:
                    dest_page = destination.get_a_page_for((page.user_name,))
                    is_ok = destination.table_attr(dest_page, col_id, new_value)
                    if is_ok != 'ok.png':
                        error = is_ok
        if error:
            if error == 'bad.png':
                error = "ERROR_value_not_modifiable"
            elif error == 'bug.png': # pragma: no cover
                error = "TIP_violet_square"
            sender.append(page.browser_file,
                f'<script>Alert({utilities.js(error)});</script>')
        last_test = time.time()
    return last_test

def cell_change(table, page, col_id, line_id, value, _date):
    """A cell changed"""
    if len(page.user_name) == 1:
        return
    TO_UPDATE.append((table, page, col_id, line_id, value))
    utilities.start_job(do_update, 1)

def content(_table):
    return r"""
function update_student_information(line)
{
   if ( ! t_student_picture.parentNode )
      return ;
   t_student_picture.parentNode.innerHTML = '<a href="' + url + '/doc_table.html#tables">Documentation</a>' ;
}
"""
