#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2021 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

"""
Management of external teachers.

python3-xlsxwriter
python3-qrcode

"""

import collections
import re
import io
import time
import html
import urllib.request
import xlsxwriter
import qrcode
from TOMUSS.ATTRIBUTES import tableinvitation
from .. import utilities
from .. import configuration
from .. import plugin
from .. import document
from .. import column

HEADER_RIGHT = '&R&P/&N'

def send_mail(mail_from, mail_to, subject, body, filecontents):
    """Send a plain text mail with an attachment"""
    # pylint: disable=import-outside-toplevel
    from email import encoders
    from email.mime.base import MIMEBase
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText

    message = MIMEMultipart(_subtype="related")
    message["From"] = mail_from
    message["To"] = mail_to
    message["Subject"] = subject

    message.attach(MIMEText(body, "html"))
    for filename, filecontent in filecontents.items():
        png = filename.endswith('.png')
        if png:
            part = MIMEBase("image", "png", name=filename, size=str(len(filecontent)))
        else:
            part = MIMEBase('application',
                            "vnd.openxmlformats-officedocument.spreadsheetml.sheet")
        part.set_payload(filecontent)
        encoders.encode_base64(part)
        if png:
            part.add_header("Content-Disposition", "attachment; filename= " + filename)
            part.add_header("Content-ID", '<' + filename + '>')
        else:
            part.add_header("Content-Disposition", "attachment; filename= "
                            + filename + '.xlsx')
        message.attach(part)

    return utilities.send_mail_smtp(mail_from, mail_to, message.as_string())


def student_lists(ue_table, codes_td): # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    """Get the list of students per TD group"""
    if codes_td:
        students_per_td = {}
        codes_td = re.split('[ \t\n,;/]+', codes_td)
        digits = collections.defaultdict(list)
        full = True
        for td_group in codes_td:
            if '+' in td_group:
                full = False
                students_per_td[tuple(td_group.split('+'))] = []
                continue
            students_per_td[td_group,] = [] # Singles: A1 A2 B2 B3
            number = re.findall('^(.*=)?([A-Z]+)([0-9]+).*$', td_group)
            if number:
                # Creates: digits {'1':['A1'], '2':['A2','B2'], '3':['B3']}
                digits[number[0][1]].append(td_group)
                digits[number[0][2]].append(td_group)
        if full:
            for number, groups in digits.items():
                if len(groups) > 1:
                    students_per_td[tuple(groups)] = [] # TP: A2+B2 or TD : A1+A2
            students_per_td[tuple(codes_td)] = [] # All: A1+A2+B2+B3
    else:
        students_per_td = {('',): []}

    grp_columns = collections.defaultdict(lambda: [False, False])
    need_sequence = False
    not_need_sequence = False
    for keys in students_per_td:
        for key in keys:
            if re.search('^[0-9]+-', key): # The sequence is a number
                need_sequence = True
            else:
                not_need_sequence = True
            if '=' in key:
                column_name, group = key.split('=')
                if '-' in group:
                    grp_columns[column_name][0] = True # Need sequence
                else:
                    grp_columns[column_name][1] = True # Not need sequence
    getters = []
    if need_sequence:
        getters.append(lambda line: line[4].value + '-' + line[3].value)
    if not_need_sequence:
        getters.append(lambda line: line[3].value)
    for column_name, (need_sequence, not_need_sequence) in grp_columns.items():
        col = ue_table.columns.from_title(column_name)
        if not col:
            raise ValueError(utilities._("ALERT_url_import_column")
                             + ' : «' + column_name + '»')
        data_col = col.data_col
        if need_sequence:
            def get_group_ns(line, data_col=data_col, column_name=column_name):
                return column_name + '=' + line[4].value + '-' + line[data_col].value
            getters.append(get_group_ns)
        if not_need_sequence:
            def get_group_nns(line, data_col=data_col, column_name=column_name):
                return column_name + '=' + line[data_col].value
            getters.append(get_group_nns)

    for line in ue_table.lines.values():
        if not line[0].value:
            continue # pragma: no cover
        for get_group in getters:
            td_group = get_group(line)
            if not codes_td:
                students_per_td[('',)].append((line[0].value, td_group))
                break # Only short group
            for key in students_per_td:
                if td_group in key:
                    students_per_td[key].append((line[0].value, td_group))

    if not students_per_td: # pragma: no cover
        return {'': []}
    return {'+'.join(groups): students
            for groups, students in students_per_td.items()}

def get_qrcode_id(presences, nr_qrcodes):
    """Allocated QRCode IDs"""
    with presences.the_lock:
        first_qrcode = int(presences.comment.split(' ')[0] or 1)
        column.TableAttr.attrs['comment'].set(
            presences, presences.pages[0],
            str(first_qrcode + nr_qrcodes) + ' 🚫', time.strftime('%Y%m%d%H%M%S'))
    return first_qrcode

def get_qrcode( # pylint: disable=too-many-arguments
        presences, qrcode_id, mailresp, mailteach, ue_table=None, mail=None,
        td_group=None, box_size=15, duration=''):
    """Returns the QRCode image and URL"""
    link = '%s/%s/%s/%s/%s' % (
        presences.year, presences.semester, presences.ue,
        'c' if mailresp else 'C',
        qrcode_id)
    if ue_table:
        if not mailteach:
            mail = mail.split('@')[0]
        link = '%s/%s/%s %s %s %s %s' % (
            link, configuration.pre, ue_table.ue, mail, td_group,
            ue_table.semester, duration)
    else:
        link += '/_/_'
    url = '%s/SVC/%s/%s' % (
        configuration.server_url, urllib.request.quote(link), tableinvitation.checksum_short(link))
    image = io.BytesIO()
    qrcode.make(url, box_size=box_size, border=1).save(image, format='PNG')
    icon = io.BytesIO()
    qrcode.make(url, box_size=6, border=1).save(icon, format='PNG')
    return image, icon, url

FIRST_HEIGHT = 560
FIRST_WIDTH = 50

def create_qrcode( # pylint: disable=too-many-arguments,too-many-locals
        workbook, i, title, message, qrcode_format, label_format,
        qrcode_image, qrcode_icon, qrcode_link, presence_url, nr_students, duration):
    """Create a sheet with a QRCode"""
    worksheet = workbook.add_worksheet('%s%s' % ('⌚' if message == 'Start' else '⛔', i+1))
    worksheet.set_landscape()
    worksheet.hide_gridlines(1)
    worksheet.hide_row_col_headers()
    worksheet.print_area(0, 0, 0, 1)
    worksheet.fit_to_pages(1, 1) # On one sheet
    # startstop, ue_code, ue_title = title.split('\n\n')
    # header = (utilities._(message) + startstop + '\n' + ue_code + ' ' + ue_title
    #          ).replace('&', '&&') + HEADER_RIGHT
    # worksheet.set_header(header)
    # worksheet.set_footer(header)
    worksheet.set_row(0, FIRST_HEIGHT)
    worksheet.set_column(0, 0, FIRST_WIDTH)
    worksheet.set_column(1, 1, 100)
    if duration:
        txt_duration = ('\n' + utilities._("COL_TITLE_Duration").replace('_', ' ')
                        + ' ' + str(duration)
                        + ' (' + utilities._("COL_COMMENT_Duration") + ')')
    else:
        txt_duration = ''
    worksheet.write('A1',
                    utilities._(message) + title
                    + '\n\n'
                    + utilities._("TITLE_referent_nr_students") % nr_students
                    + txt_duration,
                    qrcode_format)
    worksheet.insert_image(
        'B1', message + '.png', {'image_data': qrcode_image})
    worksheet.insert_image(
        'A1', message + '.png', {'image_data': qrcode_icon,
                                 'x_offset': 10,
                                 'y_offset': (7 * FIRST_HEIGHT) // 10})
    worksheet.write('A2', utilities._('QRCode_link'), label_format)
    worksheet.write_url('B2', qrcode_link, string=qrcode_link)
    worksheet.write(
        'A3',
        utilities._("Teachers_help").split('{1}>')[1].split('</a>')[0],
        label_format)
    worksheet.write_url('B3', presence_url, string=presence_url)

def create_xlsx( # pylint: disable=too-many-locals,too-many-arguments,too-many-statements
        ue_table, presences, td_group, students_td, mail, mailresp, mailteach,
        first_qrcode, nr_qrcodes, duration):
    """Generate an XLS file with one sheet per TD group"""
    file_content = io.BytesIO()
    workbook = xlsxwriter.Workbook(file_content)
    id_format = workbook.add_format({'valign': 'vcenter', 'align': 'center',
                                     'border': 1})
    mail_format = workbook.add_format({'valign': 'top', 'font_size': 7, 'text_wrap': True,
                                       'border': 1})
    name_format = workbook.add_format({'valign': 'vcenter', 'text_wrap': True,
                                       'border': 1})
    group_format = workbook.add_format({'valign': 'vcenter', 'align': 'center',
                                        'border': 1})
    url_top_format = workbook.add_format({'valign': 'top', 'bold': True})
    qrcode_format_dict = {'valign': 'top', 'bold': True, 'text_wrap': True,
                          'font_size': 24}
    qrcode_format = workbook.add_format(qrcode_format_dict)
    qrcode_format_dict['color'] = '#FF0000'
    qrcode_format_red = workbook.add_format(qrcode_format_dict)
    border_format = workbook.add_format({'border': 1})
    suivi = configuration.suivi.url(*configuration.year_semester)

    worksheet = workbook.add_worksheet(td_group or ue_table.ue)
    users = configuration.Users(i[0] for i in students_td)
    tds = {users[student].ID_: td
           for student, td in students_td
           if users[student].ID_
           }

    students = sorted(users, key=lambda i: (i.surname or '', i.firstname, i.student_id))
    worksheet.set_column(0, 0, 10)
    worksheet.set_column(1, 1, 3)
    worksheet.set_column(2, 2, 10)
    worksheet.set_column(3, 3, 15)
    worksheet.set_column(4, 4, 5)
    header = ue_table.ue + ' ' + ue_table.table_title
    worksheet.write_url(
        'A1',
        '{}/{}/{}/{}/=filters=0_3:{}='.format(
            configuration.server_url, ue_table.year, ue_table.semester, ue_table.ue, td_group),
        string=header
        )
    worksheet.merge_range('A1:Z1', header, url_top_format)
    first = 3
    nr_students = 0
    for student in students:
        if not student.student_id:
            continue # pragma: no cover
        line = str(first + nr_students)
        worksheet.write_url('A'+line,
                            suivi + '/' + student.student_id,
                            id_format,
                            string=student.student_id,
                            )
        worksheet.write('B'+line, '', border_format)
        worksheet.insert_image('B'+line, '{}.png'.format(line),
                               {'image_data': io.BytesIO(student.picture_icon),
                                'x_scale': 0.4, 'y_scale': 0.4,
                                'border': 1
                               })
        worksheet.write('C'+line, 'mailto:'+str(student.mail), mail_format)
        worksheet.write('D'+line, student.surname + '\n' + student.firstname.title(), name_format)
        worksheet.write('E'+line, tds[student.ID_], group_format)
        for letter in 'FGHI':
            worksheet.write(letter + line, '', border_format)
        worksheet.set_row(first + nr_students - 1, 26)
        nr_students += 1
    worksheet.print_area(first, 0, first + len(students) - 1, 8)
    header = header.replace('&', '&&')
    worksheet.set_header(header + HEADER_RIGHT)
    worksheet.set_footer(header + HEADER_RIGHT)

    presence_url = '%s/%s/P/%s' % (
        configuration.server_url, presences.year, presences.ue)
    label_format = workbook.add_format({'align': 'right'})
    for i in range(nr_qrcodes):
        title = ' %s %d.\n\n%s\n\n%s' % (
            utilities._('B_Prst').lower(), i+1, ue_table.ue, ue_table.table_title)
        qrcode_image, qrcode_icon, qrcode_link = get_qrcode(
            presences, first_qrcode + i, mailresp, mailteach, ue_table, mail, td_group, duration=duration)
        create_qrcode(
            workbook, i, title, 'Start', qrcode_format, label_format,
            qrcode_image, qrcode_icon, qrcode_link, presence_url, nr_students, duration)
        qrcode_image, qrcode_icon, qrcode_link = get_qrcode(presences, first_qrcode + i, mailresp, mailteach)
        create_qrcode(
            workbook, i, title, 'Stop', qrcode_format_red, label_format,
            qrcode_image, qrcode_icon, qrcode_link, presence_url, nr_students, duration)

    workbook.close()
    xlsx = file_content.getvalue()
    del file_content
    return xlsx

# pylint: disable=too-many-arguments,too-many-locals
def message_and_qrcodes(message, students_per_td, nr_qrcodes, first_qrcode,
                        presences, mailresp, mailteach, ue_table, mail, duration):
    """Returns the message in HTML and the QRCode images"""
    images = {}
    # box_size=1 does not work with Microsoft webmail because
    # the image style is stripped.
    style = 'width: calc(min(90vh, 90vw));image-rendering: crisp-edges;'
    message = [message, '<table><tr>%s<th>' %  utilities._('COL_COMMENT_0_3')]
    start = utilities._('TH_begin')
    end = utilities._('TH_end')
    for j in range(nr_qrcodes):
        message.append('<th>%s' % (j+1))
    message.append('</tr>')
    i = first_qrcode
    for code_td in students_per_td:
        message.append('<tr><th>%s' % code_td)
        for j in range(nr_qrcodes):
            message.append('<td><a href="#%s.png">%s</a>/<a href="#%s.png">%s</a> '
                           % (i, start, i+1, end))
            i += 2
        message.append('</tr>')
    message.append('</table>')

    def add_qrcode(what):
        filename = '{}.png'.format(first_qrcode)
        images[filename] = qrcode_image.getvalue()
        if duration:
            txt_duration = '%s : %s (%s)<br>' % (
                utilities._("COL_TITLE_Duration").replace('_', ' '),
                duration,
                utilities._("COL_COMMENT_Duration"))
        else:
            txt_duration = '' # pragma: no cover
        message.append('''
        <h1 style="margin-bottom: 0px"><a name="%s">%s %s %s%s</a></h1>
        «%s»<br>
        %s<br>%s
        <a href="%s"><img contenttype="image/png" src="cid:%s" style="%s"></a>
        ''' % (filename, utilities._(what), j + 1,
               ue_table.ue,
               '. {} «{}»'.format(utilities._("COL_COMMENT_0_3"), code_td) if code_td else '',
               ue_table.table_title,
               utilities._("TITLE_referent_nr_students") % len(students), # pylint: disable=undefined-loop-variable
               txt_duration,
               qrcode_link, filename, style))

    for code_td, students in students_per_td.items():
        for j in range(nr_qrcodes):
            qrcode_image, _icon, qrcode_link = get_qrcode(
                presences, first_qrcode, mailresp, mailteach, ue_table, mail, code_td,
                box_size=15, duration=duration)
            add_qrcode('Start')

            qrcode_image, _icon, qrcode_link = get_qrcode(
                presences, first_qrcode, mailresp, mailteach, box_size=15)
            add_qrcode('Stop')
            first_qrcode += 1
    return ''.join(message), images


# pylint: disable=too-many-locals,too-many-branches
def teacher_sendmail(server, table, presences, line_id):
    """
    Send a mail to the teacher containing an excel file per TD group.
    And QRCodes to start/stop attendance registration.
    """
    if line_id not in table.lines:
        server.the_file.write('line_id={}'.format(line_id))
        return
    line = table.lines[line_id]
    (mail, code_ue, semester, codes_td, _status, message,
     _comment, mailresp, nr_qrcodes, send_xls, duration, mailteach) = (cell.value for cell in line[:12])
    if not nr_qrcodes:
        nr_qrcodes = table.columns[8].empty_is
    nr_qrcodes = int(nr_qrcodes)
    if not send_xls:
        send_xls = table.columns[9].empty_is
    if not duration:
        duration = table.columns[10].empty_is or 0
    if not mailteach:
        mailteach = table.columns[11].empty_is or configuration.no
    if not semester:
        semester = table.columns[2].empty_is or ''
    duration = int(duration)
    mailresp = mailresp == configuration.yes
    mailteach = mailteach == configuration.yes
    send_xls = send_xls == configuration.yes
    if not message:
        message = (table.columns[5].empty_is + '###SEPARATOR###'
                  ).split('###SEPARATOR###')[int(not send_xls)]

    year = table.year
    try:
        i = configuration.semesters.index(semester)
        year -= configuration.semesters_year[i]
    except ValueError:
        pass

    ue_table = document.table(year, semester, code_ue, create=False)
    if not ue_table:
        server.the_file.write('{}/{}/{} {}\n'.format(
            year, semester, code_ue, server._("MSG_page_unload_no_table")))
        return
    if not ue_table.modifiable: # pragma: no cover
        server.the_file.write('{}/{}/{} {}\n'.format(
            year, semester, code_ue, server._("LABEL_table_ro")))
        return

    try:
        students_per_td = student_lists(ue_table, codes_td)
    except ValueError as err:
        server.the_file.write(str(err))
        return
    first_qrcode = get_qrcode_id(presences, nr_qrcodes * len(students_per_td))
    if send_xls:
        td_xlsx = {}
        for code_td, students in students_per_td.items():
            td_xlsx[code_ue + '-' + code_td] = create_xlsx(
                ue_table, presences, code_td, students, mail, mailresp, mailteach,
                first_qrcode, nr_qrcodes, duration)
            first_qrcode += nr_qrcodes
        errors = send_mail(server.ticket.user.mail, mail,
                           code_ue + ' ' + ue_table.table_title,
                           message, td_xlsx)
    else:
        message, images = message_and_qrcodes(
            message, students_per_td, nr_qrcodes,
            first_qrcode, presences, mailresp, mailteach, ue_table, mail, duration)
        errors = send_mail(server.ticket.user.mail, mail,
                           code_ue + ' ' + ue_table.table_title,
                           message, images)

    table.lock()
    try:
        table.cell_change(
            table.get_nobody_page(), '50Status', line_id, time.strftime("%Y-%m-%d %H:%M:%S"))
    finally:
        table.unlock()

    server.the_file.write(
        '{} {} {} <div style="display:inline-block; width: 10em">{}</div>\n'.format(
            mail, ue_table,
            ' '.join('{}:{}'.format(name, len(students))
                     for name, students in sorted(students_per_td.items())),
            html.escape(errors or '')
            ))

def teachers(server):
    """Send informative mail to the teachers"""
    data = server.uploaded
    if not data: # pragma: no cover
        server.the_file.write(server._("MSG_bad_ticket"))
        return
    year = data.getfirst('year')
    code_ue = data.getfirst('ue')
    server.the_file.write('<h1>{}/Teachers/{}</h1>\n'.format(year, code_ue))
    table = document.table(year, 'Teachers', code_ue, create=False)
    presences = document.table(year, 'P', code_ue)
    if not table or not presences:
        server.the_file.write(utilities._("ERROR_table_read_only"))
        return

    for line_id in data.getfirst('lin_ids').split(' '):
        server.the_file.write('<div>')
        teacher_sendmail(server, table, presences, line_id)
        server.the_file.write('</div>')
    server.the_file.write('<p>THE END')

def create(table):
    """Create a teacher table"""
    page = table.get_ro_page()
    table.table_attr(page, 'default_nr_columns', 11)
    table.table_attr(page, 'default_sort_column', [2, 1, 0])
    table.table_attr(page, 'private', 1)
    table.table_attr(page, 'masters', [table.user])
    table.table_attr(page, 'table_title',
                     utilities._("MSG_Teachers_title").format(table.ue))
    check(table, from_create=True)

configuration.get_qrcode_id = get_qrcode_id

def check(table, from_create=False):
    """Update table columns in case of template change"""
    _ = utilities._
    # XXX Do not change order
    columns = {
        '10Mail': { # 0
            'title': _("TH_home_mail"),
            'type': 'Text',
            'width': 8,
            'red': '!~@ | !~.'
            },
        '20UE': { # 1
            'type': 'Text',
            'title': _("COL_TITLE_code"),
            'comment': _("COL_COMMENT_code"),
            },
        '30Semester': { # 2
            'title': _("semester"),
            'type': 'Enumeration',
            'enumeration': ' '.join(configuration.semesters),
            },
        '40TD': { # 3
            'type': 'Text',
            'title': _("COL_TITLE_TD"),
            'comment': _("COL_COMMENT_TD"),
            },
        '50Status': { # 4
            'type': 'Date',
            'title': _("COL_TITLE_status"),
            'comment': _("COL_COMMENT_status"),
            },
        '60Message' : { # 5
            'type': 'Text',
            'title': _("COL_TITLE_message"),
            'comment': _("COL_COMMENT_message"),
            'width': 16,
            'position': 100, # Hide because not useful
            },
        '70Comment' : { # 6
            'type': 'Text',
            'title': _("TH_comment"),
            'width': 16,
            'position': 4.5,
            },
        '80MailResp': { # 7
            'title': _("COL_TITLE_ABI_mailresp"),
            'comment': _("COL_COMMENT_ABI_mailresp"),
            'type': 'Bool',
            'position': 3.5,
            'width': 1,
            'empty_is': configuration.no
            },
        '85NrQRCodes': { # 8
            'title': _("COL_TITLE_NrQRCodes"),
            'comment': _("COL_COMMENT_NrQRCodes"),
            'type': 'Note',
            'position': 3.6,
            'width': 1,
            'empty_is': 20,
            'rounding': 1,
            'minmax': '[0;100]',
            },
        '90SendXLS': { # 9
            'title': _("COL_TITLE_SendXLS"),
            'comment': _("COL_COMMENT_SendXLS"),
            'type': 'Bool',
            'position': 3.7,
            'width': 1,
            'empty_is': configuration.yes
            },
        '95Duration': { # 10
            'title': _("COL_TITLE_Duration"),
            'comment': _("COL_COMMENT_Duration"),
            'type': 'Note',
            'position': 3.1,
            'width': 1,
            'minmax': '[0;1000]',
            'rounding': 1,
            },
        '97MailTeacher': { # 11
            'title': _("COL_TITLE_mail_teacher"),
            'comment': _("COL_COMMENT_mail_teacher"),
            'type': 'Bool',
            'position': 3.05,
            'empty_is': configuration.no,
            'width': 1,
            },
    }
    if from_create:
        columns['60Message']['empty_is'] = utilities._("Teachers_message")

    table.update_columns(columns, table.get_nobody_page())

plugin.Plugin('teachers_mail', '/teachers_mail',
              group="staff",
              function=teachers,
              launch_thread=True,
              upload_max_size=100000,
             )
