#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2019 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

"""
A task scheduler, they can be unique or periodic.
Task are not running concurently.

So never planify a task taking more than one second.
"""

import time
import datetime
import os
import sys
import collections
import tomuss_init
from TOMUSS import configuration
from TOMUSS import utilities
from TOMUSS import plugin
from TOMUSS import ticket
from TOMUSS import document
from TOMUSS import files
from TOMUSS import cell
from TOMUSS import column

i_am_unique = (0, 'Dossiers')

time_format = '%Y-%m-%d %H:%M'

default_threads = (
    # Examples
    ('A',('2000-01-01 00:00', '24'   ,  '','print:A new day starts')),
    ('B',('2000-01-03 00:00','168'   ,  '',"print:It's monday")),
    ('C',('2000-01-01 00:00', '24'   ,'FM',"print:A new month starts")),
    ('D',('2000-01-01 00:00', '24'   ,  '',"shell:uptime;free")),
    ('E',('2000-01-01 00:00',  '0.01',  '',
          "nothing:plugin:thierry.excoffier@0/Dossiers/config_cron/cell/c/E/FM",
          "Remove «nothing:» and wait a minute to see the change")),
    # Needed to work
    #('0',('2001-01-01 00:00',  '0.01',  '',"python:utilities.do_print_lock_state()")),
    #('1',('2001-01-01 00:00',  '1'   ,  '',"python:utilities.do_clean_cache()")),
    #('2',('2001-01-01 00:00',  '0.02',  '',"python:document.check_down_connections()")),
    #('3',('2001-01-01 00:00',  '0.02',  '',"python:files.update_static_files()")),
    #('4',('2001-01-01 00:00',  '0.02',  '',"python:document.check_students_in_tables()")),
    ('6',('2001-01-01 00:00',  '1'   ,  '',
        "python:__import__('shutil').disk_usage('LOGS/').free < 1000000000 and utilities.send_mail((configuration.maintainer,), 'TOMUSS LOGS FULL', '')",
        "A mail is sent when LOGS disk is nearly full")),
    ('7',('2001-01-01 00:00',  '1'   ,  '',
        "python:__import__('shutil').disk_usage('DB/').free < 1000000000 and utilities.send_mail((configuration.maintainer,), 'TOMUSS DB FULL', '')",
        "A mail is sent when DB disk is nearly full")),
    ('8',('2001-01-01 00:00',  '1'   ,  '',
        "python:__import__('shutil').disk_usage('UPLOAD/').free < 50000000000 and utilities.send_mail((configuration.maintainer,), 'TOMUSS UPLOAD FULL', '')",
        "A mail is sent when UPLOAD disk is nearly full")),
    ('9',('2001-01-01 04:00',  '24'   ,  '',
        "plugin:%s@auto_update" % configuration.root[0],
        "Update all the student lists in current tables")),
    )

if configuration.real_regtest:
    loop_sleep_seconds = 1
    seconds_before_task_list_update = 1
else:
    loop_sleep_seconds = 60          # Check tasks every minute
    seconds_before_task_list_update = 60 # Wait a minute before updating after user edit

utilities.mkpath(os.path.join('LOGS', 'CRON'))

class Server:
    """Fake server"""
    def __init__(self, path):
        user, the_path = path.split('@', 1)
        self.the_path = the_path.split('/')
        self.ticket = ticket.Ticket('', user, '')
        self.ticket.password_ok = True
        self.the_file = open(os.path.join('LOGS', 'CRON', path.replace('/', ':')), 'a')
    def _(self, txt):
        return txt # pragma: no cover

class Protocol:
    """The action to run"""
    active = True
    def __init__(self, value=None):
        """Store the action in self.value"""
        self.value = value
        self.check()
    def check(self):
        """Verify if the action is fine"""
        pass
    def run(self):
        """Run the action"""
        pass
class Nothing(Protocol):
    """Do nothing : the url was commented with #"""
    active = False
class Print(Protocol):
    """Issue a warning"""
    def run(self):
        utilities.warn(self.value)
class Shell(Protocol):
    """Run a shell script"""
    def run(self):
        os.system(self.value)
class Python(Protocol):
    """Run the Python code"""
    def check(self):
        self.code = compile(self.value, 'cron', 'exec')
    def run(self):
        eval(self.code)
class Plugin(Protocol):
    """Execute a TOMUSS plugin :
           plugin:super.user@0/Dossiers/config_cron/cell/c/2_1/FM
    """
    def check(self):
        self.server = Server(self.value)
        self.plugin = plugin.search_plugin(self.server, True)
        if self.plugin == False:
            raise ValueError("config_cron: unknown plugin «{}»".format(self.value))
    def run(self):
        self.server.the_file.write('\n' + '='*77 + '\n')
        self.server.the_file.write(time.ctime() + '\n')
        if self.plugin.launch_thread:
            utilities.start_new_thread(self.plugin.function, (self.server,)) # pragma: no cover
        else:
            self.plugin.function(self.server)
        self.server.the_file.flush()

protocols = {c.__name__.lower():c for c in [Nothing, Print, Shell, Plugin, Python]}

class Task:
    """A line of the cron table"""
    wakeup = None
    def __init__(self, line_id, when, dt, constraint, action):
        """The definition parsed"""
        self.line_id    = line_id
        self.when       = when
        self.dt         = dt * 3600
        self.constraint = constraint
        self.action     = action
    def active(self, t):
        """Returns if the line will be activated in the future"""
        if not self.action.active:
            return False
        if self.dt:
            return True
        return self.when >= t
    def next_time(self, t):
        """Compute the next slot after t"""
        delta = t - self.when
        seconds = self.dt - delta.total_seconds() % self.dt
        return self.when + datetime.timedelta(delta.days, delta.seconds + seconds)
    def update_wakeup(self, t):
        """Compute the next wakeup date"""
        if self.dt == 0 or t < self.when:
            self.wakeup = self.when
        else:
            self.wakeup = self.next_time(t)
            if self.constraint == 'FM':
                t = datetime.datetime(t.year, t.month, 1, 0, 0)
                if self.next_time(t) != self.wakeup:
                    t = datetime.datetime(t.year, 1 + (t.month % 12), 1, 0 , 0)
                    self.wakeup = self.next_time(t)

def parse_time(value):
    """Returns the user defined time"""
    if value.count(":") == 2:
        return datetime.datetime.strptime(value, time_format + ':%S')
    return datetime.datetime.strptime(value, time_format)
def parse_dt(value):
    """The number of hours must be an integer if greater than 1"""
    if value == '':
        return 0
    value = float(value)
    if value >= 1:
        value = int(value)
    return value
def parse_url(value):
    """Check the URL"""
    if value == '':
        return Nothing()
    value = value.split(':', 1)
    return protocols[value[0]](value[1])

class Stats:
    def __init__(self):
        self.min = 1e9
        self.max = 0
        self.sum = 0
        self.nr = 0
        self.last = 0
    def add(self, start, now):
        t = now - start
        if t < self.min:
            self.min = t
        if t > self.max:
            self.max = t
        self.sum += t
        self.nr += 1
        self.last = now

class Cron:
    """Manage the tasks"""
    def __init__(self, table):
        """Keep the table pointer"""
        self.table = table
        self.todo = ()
        self.stats = collections.defaultdict(Stats)
    def full_update(self):
        """Generate the full task list"""
        todo = []
        t = datetime.datetime.now()
        for line_id, line in tuple(self.table.lines.items()):
            if line[0].value == '':
                continue
            task = Task(line_id,
                        parse_time(line[0].value),
                        parse_dt(line[1].value),
                        line[2].value ,
                        parse_url(line[3].value))
            task.update_wakeup(t)
            todo.append(task)
        self.todo = todo # XXX should be protected
        self.cleanup(t)
    def cleanup(self, t):
        """Remove tasks done and sort by date"""
        todo = [task
                for task in self.todo
                if task.active(t)]
        todo.sort(key = lambda task: task.wakeup )
        self.todo = todo
    def loop(self):
        """Task scheduler"""
        while True:
            need_cleanup = False
            t = datetime.datetime.now()
            for task in self.todo:
                if task.wakeup > t:
                    break
                start = time.time()
                try:
                    task.action.run()
                except: # pragma: no cover # pylint: disable=bare-except
                    utilities.send_backtrace('', 'bad cron action')
                self.stats[task.line_id].add(start, time.time())
                task.update_wakeup(t)
                need_cleanup = True
            if need_cleanup:
                self.cleanup(t)
            time.sleep(loop_sleep_seconds)

def create(table):
    """Create the table if it does not exists"""
    p = table.get_a_root_page()
    _ = utilities._
    table.update_columns({
        'a' : {'title': _("COL_TITLE_cron_time"),
               'comment': _("COL_COMMENT_cron_time"),
               'type':'Text', 'width': 8},
        'b' : {'title': _("COL_TITLE_cron_dt"),
               'comment': _("COL_COMMENT_cron_dt"),
               'type':'Text', 'width': 2},
        'c' : {'title': _("COL_TITLE_cron_constraint"),
               'comment': _("COL_COMMENT_cron_constraint"),
               'type':'Enumeration', 'width': 1,
               'enumeration': 'FM'},
        'd' : {'title': _("COL_TITLE_cron_url"),
               'comment': _("COL_COMMENT_cron_url"),
               'type':'Text', 'width': 16},
        'e' : {'title': _("COL_TITLE_cron_comment"),
               'comment': _("COL_COMMENT_cron_comment"),
               'type':'Text', 'width': 16},
        })
    table.table_attr(p, 'masters', list(configuration.root))
    table.table_attr(p, 'default_sort_column', [0,1])
    table.table_attr(p, 'default_nr_columns', 5)
    table.table_attr(p, 'private', 1)

def init(table):
    """Keep the table in memory"""
    table.do_not_unload_add('*config_cron')
    table.contains_users = 0

def full_update():
    """start_job does not accept methods"""
    configuration.cron.full_update()

def onload(table):
    """Start the scheduler"""
    if configuration.read_only:
        return # pragma: no cover
    if configuration.regtest and not configuration.real_regtest:
        table.modifiable = 0 # pragma: no cover
    table.cron = Cron(table)
    configuration.cron = table.cron
    table.lock()
    try:
        p = table.get_a_root_page()
        for lin_id, default in default_threads:
            if lin_id in table.lines:
                continue
            for col_id, value in zip("abcde", default):
                table.cell_change(p, col_id, lin_id, value)

    finally:
        table.unlock()
    utilities.start_job(full_update, seconds_before_task_list_update)
    utilities.start_new_thread_immortal(table.cron.loop, ())

def cell_change(_table, _page, col, _lin, value, _date):
    """Update task list on user change"""
    if configuration.regtest and not configuration.real_regtest:
        raise ValueError('Not allowed') # pragma: no cover
    if col == 'a':
        if value:
            parse_time(value)
    elif col == 'b':
        parse_dt(value)
    elif col == 'd':
        parse_url(value)
    utilities.start_job(full_update, seconds_before_task_list_update)

def cron(server):
    """Statistics about tasks scheduled by 'config_cron'"""
    cron = configuration.cron
    lines = []
    for task in cron.todo:
        stats = cron.stats.get(task.line_id, None)
        if not stats:
            continue
        lines.append(cell.Line((
            cell.CellValue(task.action.__class__.__name__.lower() + ':' + task.action.value),
            cell.CellValue(time.strftime(time_format, time.localtime(stats.last))),
            cell.CellValue(task.wakeup.strftime(time_format)),
            cell.CellValue(str(stats.nr)),
            cell.CellValue("{:.5f}".format(stats.min)),
            cell.CellValue("{:.5f}".format(stats.max)),
            cell.CellValue("{:.5f}".format(stats.sum/stats.nr)),
            ))) # pragma: no cover
    columns = [
        column.Column('0', '', width=4, title=server._("COL_TITLE_cron_url")),
        column.Column('1', '', width=4, title=server._("B_Date") + '_1'),
        column.Column('2', '', width=4, title=server._("B_Date") + '_2'),
        column.Column('3', '', width=1, title=server._("LABEL_number_of_access")),
        column.Column('4', '', width=2, title=server._("Minimum")),
        column.Column('5', '', width=2, title=server._("Maximum")),
        column.Column('6', '', width=2, title=server._("Average")),
        ]
    document.virtual_table(server, columns, lines,
                           table_attrs={
                            'default_nr_columns': 7,
                                       })

plugin.Plugin('cron', '/cron', function=cron, group='roots')

def content(dummy_table):
    return r"""
function update_student_information(line)
{
   if ( ! t_student_picture.parentNode )
      return ;
   t_student_picture.parentNode.href = add_ticket('cron') ;
   t_student_picture.parentNode.innerHTML =
      '<div style="width:15em">'
      + _("MSG_tablelinear_table_stats")
      + '</div>' ;

   document.getElementById('horizontal_scrollbar').parentNode.style.display = 'none' ;
}
"""

def regtest():
    """Test wakeup compute time"""
    task = Task('', parse_time("2019-03-30 23:00"), 1, '', Nothing())
    for t in (("2018-01-10 23:11", '2019-03-30 23:00'),
              ("2019-03-30 23:11", '2019-03-31 00:00'),
              ("2019-03-31 00:11", '2019-03-31 01:00'),
              ("2019-03-31 01:11", '2019-03-31 02:00'),
              ("2019-03-31 02:11", '2019-03-31 03:00'),
              ("2019-03-31 03:11", '2019-03-31 04:00'),
              ("2019-03-31 04:11", '2019-03-31 05:00'),
              ("2019-10-27 02:11", '2019-10-27 03:00'),
              ("2019-10-27 03:11", '2019-10-27 04:00'),
              ("2019-10-27 04:11", '2019-10-27 05:00'),
             ):
        task.update_wakeup(parse_time(t[0]))
        if task.wakeup != parse_time(t[1]):
            raise ValueError("{} {}".format(task.wakeup, t[1])) # pragma: no cover
    task = Task('', parse_time("2000-01-01 23:10"), 0.5, '', Nothing())
    for t in (("2019-01-01 23:11", '2019-01-01 23:40'),
              ("2019-01-01 23:41", '2019-01-02 00:10'),
             ):
        task.update_wakeup(parse_time(t[0]))
        if task.wakeup != parse_time(t[1]):
            raise ValueError("{} {}".format(task.wakeup, t[1])) # pragma: no cover
    task = Task('', parse_time("2000-01-01 08:00"), 7*24, 'FM', Nothing())
    for t in (("2000-01-01 07:00", '2000-01-01 08:00'),
              ("2000-02-01 09:00", '2000-02-05 08:00'),
              ("2000-01-01 09:00", '2000-02-05 08:00'),
              ("2019-08-01 09:00", '2019-08-03 08:00'),
             ):
        task.update_wakeup(parse_time(t[0]))
        if task.wakeup != parse_time(t[1]):
            raise ValueError("{} {}".format(task.wakeup, t[1])) # pragma: no cover

    messages = []
    def warn(*args, **kargs):
        messages.append(args[0])
    utilities.warn = warn
    t = parse_time("2000-01-01 08:00")
    Task('', t, 1, '', Nothing()).action.run() # for coverage
    Task('', t, 1, '', Print('Hello')).action.run()
    assert 'Hello' in messages
    Task('', t, 1, '', Shell('echo World >xxx.crontest')).action.run()
    assert utilities.read_file('xxx.crontest').strip() == 'World'

    print('OK '*100)

if __name__ == '__main__':
    regtest()
