#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2008-2012 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

# WARNING : Assume the table is locked, all this is not thread safe

import time
import math
from . import utilities
from . import configuration
from . import data

js = utilities.js

class CellVirtual(object):
    """The <b>set_</b> methods in cells returns a new object if needed.
    It is because an CellEmpty can become a Cell.
    """
    def history_string(self):
        return repr(self.history)
    def js_history(self):
        return '[' + ','.join('[{0},"{1}",{2},"{3}"]'.format(
                                  js(h[0]), h[1], js(h[2]), h[3])
                           for h in self.history) + ']'

class CellEmpty(CellVirtual):
    """Define an empty cell. It is only to use less memory than a Cell"""
    
    value = ''
    author = ''
    date = ''
    comment = ''
    history = []
    
    __slots__ = ()
    
    def set_comment(self, comment, author='', date=None):
        """Create a non empty cell with the comment and returns it."""
        return Cell().set_comment(comment, author, date)
    
    def set_value(self, value='', author='', date=None):
        """Create a non empty cell with the value and returns it."""
        if date == None:
            date = time.strftime('%Y%m%d%H%M%S')
        return CellValue(value=value, author=author, date=date)
    
    def js(self):
        """Create the javascript code for an empty cell."""
        return 'C()'

    def json(self):
        return ()

    def empty(self):
        """Returns True"""
        return True

    def previous_value(self):
        return ''

    def previous_comment(self): # pragma: no cover
        return ''

    def previous_author(self): # pragma: no cover
        return ''

    def previous_date(self): # pragma: no cover
        return ''

    def get_value(self, column):
        return column.empty_is

cellempty = CellEmpty()

class CellValue(CellVirtual):
    """Define a cell without history nor comment
    It is the most used cell content
    """
    
    __slots__ = ('value', 'author', 'date')
    comment = ''
    history = []

    def __init__(self, value='', author='', date=''):
        self.value = value
        self.author = author
        self.date = date

    def empty(self):
        """Returns True if the cell is empty of user value"""
        return (self.value == ''
                or self.author == data.ro_user
                or self.author == data.rw_user
                or self.author == data.no_user)
    
    def set_comment(self, comment, author='', date=None):
        """change the comment on the cell."""
        return Cell(self.value, self.author, self.date).set_comment(comment, author, date)

    def set_value(self, value='', author='', date=None):
        c = Cell(self.value, self.author, self.date)
        c.set_value(value, author, date)
        return c

    def previous_value(self):
        return ''

    def previous_comment(self): # pragma: no cover
        return ''

    def previous_author(self): # pragma: no cover
        return ''

    def previous_date(self): # pragma: no cover
        return ''

    def js(self):
        """Generate the Cell JavaScript object with the minimal code,
        in order to minimize file size."""
        if self.date:
            return 'C(%s,"%s","%s")' % (js(self.value),self.author,self.date)
        elif self.author:
            return 'C(%s,"%s")' % (js(self.value), self.author)
        elif self.value:
            return 'C(%s)' % js(self.value)
        else:
            return 'C()'

    def json(self):
        """Generate the Cell JavaScript object with the minimal code,
        in order to minimize file size."""
        if self.date:
            return (self.value, self.author, self.date)
        elif self.author:
            return (self.value, self.author)
        elif self.value: # pragma: no cover
            return (self.value,)
        else:
            return () # pragma: no cover

    def get_value(self, column):
        if self.value == '':
            return column.empty_is
        return self.value


class Cell(CellValue):
    """Define a cell with the given attributes.
    It also manage the cell history and a cache to minimize
    CPU time to generate the minimal JavaScript code for the cell.
    Attributes of the cell should not be modified directly.
    """

    __slots__ = ('value', 'author', 'date', 'comment', 'history')
    
    def __init__(self, value='', author='', date='', comment='', history=None):
        """Create a new cell with some attributes in the list:
        'value', 'author', 'date' and 'comment'.
        The date is formatted as YYYYMMDDHHMMSS.
        """
        self.value = value
        self.author = author
        self.date = date
        self.comment = comment
        if history:
            self.history = history # pragma: no cover # See compute_nmbr()
        else:
            self.history = []
        
    def set_comment(self, comment, author='', date=None):
        """change the comment on the cell."""
        if date == None:
            date = time.strftime('%Y%m%d%H%M%S')
        self.comment = comment
        self.history.append((comment, date, author, 'C'))
        return self

    def set_value(self, value='', author='', date=None):
        """change the value of the cell."""
        if date == None:
            date = time.strftime('%Y%m%d%H%M%S')

        self.history.append((self.value, self.date, self.author, 'V'))
        self.value = value
        self.author = author
        self.date = date
        return self

    def previous_value(self): # pragma: no cover
        for history in reversed(self.history):
            if history[-1] == 'V':
                return history[0]
        return ''

    def previous_comment(self): # pragma: no cover
        for history in reversed(self.history):
            if history[-1] == 'C':
                return history[0]
        return ''

    def previous_author(self): # pragma: no cover
        if self.history:
            return self.history[-1][2]
        return ''

    def previous_date(self): # pragma: no cover
        if self.history:
            return self.history[-1][1]
        return ''

    def js(self):
        """Generate the Cell JavaScript object with the minimal code,
        in order to minimize file size."""
        if self.history:
            return 'C(%s,"%s","%s",%s,%s)' % (
                js(self.value),
                self.author,
                self.date,
                js(self.comment),
                self.js_history()
                )
        elif self.comment:
            return 'C(%s,"%s","%s",%s)' % (
                js(self.value), self.author, self.date, js(self.comment))
        elif self.date: # pragma: no cover
            return 'C(%s,"%s","%s")' % (
                js(self.value), self.author, self.date)
        elif self.author: # pragma: no cover
            return 'C(%s,"%s")' % (
                js(self.value), self.author)
        elif self.value: # pragma: no cover
            return 'C(%s)' % (
                js(self.value),
                )
        else: # pragma: no cover
            return 'C()'

    def json(self):
        """Generate the Cell JavaScript object with the minimal code,
        in order to minimize file size."""
        if self.history:
            return (self.value, self.author, self.date, self.comment,
                    self.history)
        if self.comment: # pragma: no cover
            return (self.value, self.author, self.date, self.comment)
        if self.date: # pragma: no cover
            return (self.value, self.author, self.date)
        if self.author: # pragma: no cover
            return (self.value, self.author)
        if self.value: # pragma: no cover
            return (self.value, )
        return () # pragma: no cover

C = Cell # To be compatible with PythonJS translated javascript


class Line(object):
    """The Line object is usable as a Python list."""
    def __init__(self, cells):
        """Create the Line from a Cell list."""
        self.cells = cells

    def __getitem__(self, k):
        """Take a Cell in the line."""
        return self.cells[k]

    def __setitem__(self, k, v):
        """Replace a Cell in the line."""
        self.cells[k] = v

    def pop(self, k):
        """Remove the selected cell from the line."""
        self.cells.pop(k)

    def append(self, v):
        """Add a new cell in the line."""
        self.cells.append(v)

    def __len__(self):
        """The number of cells in the line."""
        return len(self.cells)

    def js(self, to_hide=None, columns=None):
        """Translate the line in JavaScript."""
        if to_hide:
            return '[' + ','.join("C('?','*')" if column in to_hide else cell.js()
             for cell, column in zip(self.cells, columns)) + ']'
        else:
            return '[' + ','.join([cell.js() for cell in self.cells]) + ']'

    def json(self, for_student=False, columns=None, see_private=False, ticket=None):
        """Translate the line in JavaScript"""
        if see_private:
            # for_student must be True
            return [cell.json()[:4] if col.visibility != 5 else () # Hide history
                    for cell, col in zip(self.cells, columns)
                    if col.visible(hide=1, line = self.cells, login = ticket and ticket.user_name)
                   ]
        return [['⛔'] if col.private or col.visibility == 6 else (cell.json()[:4] if col.visibility != 5 else ())
                for cell, col in zip(self.cells, columns)
                if col.visible(hide = 1 if for_student else True, line=self.cells,
                               login = ticket and ticket.user_name)
               ]

class Lines(object):
    """The Lines object is usable as a Python list of Line.
    But as it knows the columns it performs some high level functions.
    It is a dictionnary of lines, the key is the line_id.
    """
    
    def __init__(self, columns):
        """Create an empty set of lines for the given columns"""
        self.lines = {}
        self.columns = columns

    def __iter__(self):
        """Iterate over all the lines."""
        return self.lines.__iter__()

    def items(self):
        """Iterate on the line to get the pair: (line_id, line)."""
        return self.lines.items()

    def keys(self):
        """Iterate on the line to get the line keys: line_id."""
        return self.lines.keys()

    def values(self):
        """Iterate on the line to get the line values."""
        return self.lines.values()

    def __getitem__(self, line):
        """Access to a non existent Line create a new line
        with the good number of empty cells."""
        if line not in self.lines:
            self.lines[line] = Line( [cellempty] * len(self.columns) )
        return self.lines[line]

    def get(self, line, default):
        return self.lines.get(line, default)

    def __contains__(self, k):
        """Returns True if the line_id is known."""
        return k in self.lines

    def __len__(self):
        """Returns the number of lines."""
        return len(self.lines)


    def get_grp(self, line):
        """Get the 'Grp' value for the student"""
        grp = self.columns.get_grp()
        if grp is None:
            return None # pragma: no cover
        else:
            return line[grp].value

    def get_seq(self, line):
        """Get the 'Seq' value for the student"""
        seq = self.columns.get_seq()
        if seq is None:
            return None # pragma: no cover
        else:
            return line[seq].value

    def line_indicator(self, line, what):
        """Returns a list of (date, cell_indicator)
        for the column type given. It is used to compute icons."""
        status = []
        abi = configuration.all_abitnr
        pre = configuration.all_pre
        for value, column in tuple(zip(line, self.columns.columns))[6:]:
            v01 = value.value
            if v01 in abi:
                v01 = 0
            elif v01 in pre:
                v01 = 1 # pragma: no cover
            elif what == 'Prst':
                continue
            elif column.type.name not in configuration.grading_types:
                continue
            elif column.weight.startswith('+') or column.weight.startswith('-'):
                continue # pragma: no cover
            else:
                # A normal grade
                try:
                    v01 = float(v01)
                except ValueError:
                    continue
                if math.isnan(v01):
                    continue # pragma: no cover
                v_min, v_max = column.min_max()
                v01 = (v01 - v_min) / (v_max - v_min)
            status.append((value.date, v01))
        return status

    def line_stat(self, table, line, link, ticket):
        """Returns an HTML string displaying all the information in the line.
        This function is used to display the 'suivi'.
        The content is not the same for teacher and students because
        teacher have access to table links and boolean toggle to directly
        modify the tables.
        This function is expensive because the student rank is computed
        for all 'Note' columns.
        """
        grp = self.get_grp(line)
        seq = self.get_seq(line)
        lines = list(table.lines_of_grp(grp, seq))
        login = ticket and ticket.user_name
        # Put cells in column order
        val_col = zip(line.cells, self.columns.columns)
        next(val_col) # ID
        next(val_col) # Firstname
        next(val_col) # Surname
        d = {}
        for cell, column in val_col:
            if not column.visible(hide = link and True or 1, line=line, login=login):
                continue
            s = column.stat(cell, lines, line, (grp, seq), link, ticket)
            if s:
                d[column.the_id] = s

        # XXX
        # if empty_line and table.ue_code != table.ue:
        #     # If the table name ends with -1 -2 -3...
        #     # we don't display empty lines
        #     return {}

        return d

    def js(self, user_name):
        """Create JavaScript generating all the lines data."""
        to_hide = tuple(column
                        for column in self.columns
                        if not column.readable_by(user_name)
                        )
        s = []
        for line_id, line in self.lines.items():
            s.append('P(%s,' % utilities.js(line_id)
                     + line.js(to_hide, self.columns) + ');')
        return s

