#!/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet
#    Copyright (C) 2008-2018 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

import math
import os
import time
import threading
import socket
import re
import collections
import html
import glob
import weakref
import traceback
from . import utilities
from . import configuration
from . import column
from . import data
from . import files
from . import inscrits
from .cell import CellValue, Lines, cellempty
from . import sender
from . import plugins
from . import entity
from TOMUSS.ATTRIBUTES import columnalert

def date_time(date):
    if date is None:
        return '19700101010101' # pragma: no cover
    else:
        return time.strptime(date, '%Y%m%d%H%M%S')

warn = utilities.warn

tables = {}

the_head = files.add('FILES','head.html')

js = utilities.js

canceled_loads = []

class Page(object):
    request = 0
    last_request = 0
    mtime = 0
    nr_cell_change = 0
    browser_file = None

    def __init__(self, the_ticket, user_name, page_id, ttable,
                 user_ip, user_browser, date=None):
        self.ticket = the_ticket
        self.user_name = user_name
        self.page_id = page_id
        self._table = weakref.ref(ttable)
        self.user_ip = user_ip
        self.user_browser = user_browser
        self.date = date

    @property
    def table(self):
        return self._table()

    def check_identity(self, the_ticket, user_name, user_ip, user_browser):
        # Allow IP change with the same identity
        if (the_ticket == self.ticket
            and user_name == self.user_name
            and user_browser == self.user_browser):
            return
        warn('received ticket=%s user_name=%s user_browser=%s ip=%s' % (
            the_ticket, user_name, user_browser, user_ip), what="Info")
        warn('current  ticket=%s user_name=%s user_browser=%s ip=%s' % (
            self.ticket, self.user_name, self.user_browser, self.user_ip),
             what="Info")
        if configuration.regtest_sync or not configuration.regtest:
            raise ValueError('Cheater')

    def add_request(self, timestamp, actions, output_file):
        """Add at the end the last request only if newer"""
        if timestamp < self.last_request: # pragma: no cover
            output_file.close()
            return
        self.last_request = timestamp
        request_list.append((self, timestamp, actions, output_file))
        if configuration.regtest:
            check_requests()
            # We don't want asynchronous update when doing regtest
            if configuration.regtest_sync and self.table.ue != 'demo_animaux-1':
                update_students.append(self.table)
                check_new_students_real()
        else:
            utilities.start_job(check_requests, 0, important='request') # pragma: no cover

    def __str__(self):
        return '%s[%s]:%s(%s)' % (self.table, self.page_id, self.user_name,
                                  self.date)

    def date_time(self):
        return date_time(self.date)
    def day(self):
        return self.date[6:8] + '/' + self.date[4:6] + '/' + self.date[:4]

    def backtrace_html(self):
        return (html.escape(str(self)) + '<br>'
               + html.escape(self.user_browser) + '<br>'
                + self.user_ip)

    def is_a_teacher(self):
        if not hasattr(self, '_is_a_teacher'):
            self._is_a_teacher = configuration.is_member_of(self.user_name,
                                                            'staff')
        return self._is_a_teacher

def table_filename(year, semester, ue):
    return os.path.join(configuration.db, 'Y'+str(year), 'S'+semester, ue + '.py')

def link_to(filename, year, semester):
    ln = os.readlink(filename)[:-3].split(os.path.sep)
    ue = ln[-1]
    if len(ln) == 5:
        assert(ln[0] == '..')
        assert(ln[1] == '..')
        assert(ln[2][0] == 'Y')
        assert(ln[3][0] == 'S')
        year = int(ln[2][1:])
        semester = ln[3][1:]
    elif len(ln) == 3: # pragma: no cover
        assert(ln[0] == '..')
        assert(ln[1][0] == 'S')
        semester = ln[1][1:]
    else: # pragma: no cover
        assert(len(ln) == 1)
    return year, semester, ue

def create_link(year, semester, ue):
    "Returns True if a link is created"
    first_semester = configuration.university_semesters[0]
    if first_semester == semester:
        return
    filename = table_filename(year, semester, ue)
    if os.path.exists(filename):
        return
    first_table = table_filename(
        utilities.university_year(year, semester), first_semester, ue)
    if not os.path.exists(first_table):
        return
    # XXX Not working if UE ends with -1 -2...
    if not re.search(configuration.ue_not_per_semester, ue):
        return # UE per semester # pragma: no cover

    # Create a symbolic link from current semester to first semester
    os.symlink(
        os.path.join('..', '..',
                     os.path.sep.join(first_table.split(os.path.sep)[-3:])
                     ),
        filename)
    utilities.start_new_thread(update_indexes, (year, semester, ue))
    return True


def filter_language(language):
    # Remove not translated languages and duplicates
    t = []
    for x in language.strip(",").split(','):
        if x in plugins.languages:
            if x not in t:
                t.append(x)
    return ','.join(t)

def get_preferences(user_name, create_pref=True, the_ticket=None):
    from .PLUGINS import suivi_student
    p = utilities.display_preferences_get(user_name)
    for k, v in {'nr_lines'       : 0,
                 'nr_cols'        : 0,
                 'zebra_step'     : 5,
                 'page_step'      : 1,
                 'invert_name'    : 1,
                 'invert_grpseq'  : 1,
                 'scrollbar_right': 1,
                 'v_scrollbar'    : 1,
                 'v_scrollbar_nr' : 1,
                 'interface'      : "N",
                 'current_suivi'  : 0,
                 'debug_table'    : 0,
                 'home_3scrollbar': 1,
                 'one_line_more'  : 1,
                 'filter_freezed' : 0,
                 'disable_dbl_click' : 0,
                 'theme'          : "",
                 'forget_input'   : 0,
                 'compressed'     : 0, # For home page items
                 'compact'        : 0, # Padding around items
    }.items():
        if k not in p:
            p[k] = v

    warn('Language in preferences: (%s)' % p.get('language', ''), what="lang")
    if p.get('language', '') == '':
        if the_ticket is None: # pragma: no cover
            warn('Search old ticket to find language')
            for the_ticket in ticket.tickets.values():
                if the_ticket.user_name == user_name:
                    break
        warn('Language in ticket: (%s)' % the_ticket.language, what="lang")
        p['language'] = the_ticket.language

    p['language'] = filter_language(p['language'])
    warn('Language after filtering: ' + p['language'],what="lang")
    if p['language'] == '':
        warn('Language in server: (%s)' % configuration.language,
             what="lang")
        p['language'] = configuration.language

    return p


#REDEFINE
# This function returns javascript code to be included
# in the header of the 'ue' table
def table_head_more(dummy_ue):
    return ''

def translations_init(language):
    languages = []
    language = filter_language(language + ',' + configuration.language)
    for lang in language.split(','):
        languages.append(
            '<script onload="this.onloadDone=true;" src="%s/%s.js" charset="UTF-8"></script>'
            % (configuration.url_files, lang))
    return ('<script>var translations = {},'
            + 'preferences={"language":%s} ; </script>\n'
            % js(language)
            + '\n'.join(languages) + '\n')

@utilities.add_a_cache0
def the_years():
    return sorted(int(y[-4:])
                  for y in glob.glob(os.path.join(configuration.db, "Y2???"))
    )

@utilities.add_a_cache
def all_the_semesters(code_semester):
    code, table_semester = code_semester
    s = []
    to_remove = set()
    if table_semester not in configuration.semesters:
        for year in the_years():
            if os.path.exists(table_filename(year, table_semester, code)):
                s.append('<option>%s/%s</option>' % (year, table_semester))
    else:
        for dummy_url, dummy_port, year, semester, dummy_host in \
            configuration.suivi.urls_sorted():
            filename = table_filename(year, semester, code)
            if os.path.exists(filename):
                if os.path.islink(filename): # pragma: no cover
                    # Assumes that the link is always: ../../Yxxx/Sxxxx/Name.py
                    link = os.readlink(filename).split(os.path.sep)
                    s.append(
                       '<option value="%s/%s">%s/%s&amp;%s/%s</option>' % (
                        year, semester, year, semester,link[2][1:],link[3][1:]))
                    to_remove.add('<option>%s/%s</option>'
                                  % (link[2][1:], link[3][1:]))
                else:
                    s.append('<option>%s/%s</option>' % (year, semester))
        for x in to_remove:
            if x in s: # pragma: no cover
                s.remove(x)
        for option in configuration.special_semesters.split('<option>')[:0:-1]:
            option = option.split("<")[0]
            s.append('<option value="{}">{}</option>'.format(
                        option, utilities._(option)))
        s.reverse()
    return js(''.join(s))

def table_head(year=None, semester=None, the_ticket=None,
               user_name='', page_id=-1, ue='',
               create_pref=True,
               attrs_from=0, hide_more=False, table=None, hide_head=False):
    s = configuration.suivi.url(year, semester)
    t = ticket.tickets.get(the_ticket, None)
    prefs_table = get_preferences(user_name, create_pref, the_ticket=t)
    my_identity2 = utilities.login_to_module(user_name)

    bookmarked = configuration.User(user_name).manage_key('bookmarked')
    if bookmarked:
        bookmarked = (year, semester, ue) in eval(bookmarked)

    return (('' if hide_head else the_head.get_content()) +
            translations_init(prefs_table['language']) +
            (table.template.headers if table else '')
            + '<script>\n' +
            'page_id = "%d" ;\n' % page_id +
            'table_creation_date = "%s" ;\n' % (table.pages[0].date
                                                if table else "None") +
            'table_load_time = "%s" ;\n' % (table.load_time
                                                if table else "None") +
            'my_identity = %s ;\n' % repr(user_name) +
            'my_identity2 = %s ;\n' % repr(my_identity2) +
            'url = %s ;\n' % js(configuration.server_url) +
            'year = "%s" ;\n' % year +
            'semester = "%s" ;\n' % semester +
            'ticket = "%s" ;\n' % the_ticket +
            'ue = "%s" ;\n' % ue +
            'suivi = %s ;\n' % js(s) +
            'version = "%s" ;\n' % configuration.version +
            'root = %s ;\n' % js(list(configuration.root)) +
            'cas_url = %s ;\n' % repr(configuration.cas) +
            'preferences = %s ;\n' % prefs_table +
            'bookmarked = %d ;\n' % int(bookmarked)  +
            'lines = {};\n' +
            'columns = [];\n' +
            'lines_to_load = 0 ;\n' +
            'minors = %s ;\n' % js(configuration.major_of(user_name)) +
            'ticket_time_to_live = %d ;\n' % configuration.ticket_time_to_live +
            'upload_max = %d ;\n' % configuration.upload_max +
            'max_visibility_date = %d ;\n' % configuration.max_visibility_date +
            "rounding_avg = %f ;\n" % configuration.rounding_avg +
            "rounding_default = %f ;\n" % configuration.rounding_default +
            "rounding_min = %f ;\n" % configuration.rounding_min +
            'gui_record = %d ;\n' % int(configuration.gui_record) +
            'rgpd_link = "%s" ;\n' % configuration.rgpd_link +
            'all_the_semesters = %s ;\n' % all_the_semesters(
                (table.ue, table.semester) if table else ('', '')) +
            "table_headers = %s;\n" % js(table.template.headers
                                         if table else '') +
            'table_attr = {\n' +
                ',\n'.join(attr.name+':'+js(
                getattr(attrs_from, attr.name,
                        attr.get_default_value(table)))
                for attr in column.TableAttr.attrs.values()
                ) + '} ;\n' +
            (hide_more and '' or table_head_more(ue)) +
            '</script>\n')

class Template(object):
    loaded = {}
    module = None
    update_time = 0

    def __init__(self, name):
        self.name = name
        self.splited = name.split(os.path.sep)

    def update_template(self):
        if time.time() - self.update_time < 2:
            if not configuration.regtest:
                return # pragma: no cover
        self.update_time = time.time()
        if self.name == "NONE":
            return
        self.module = utilities.import_reload(self.name)[0]
        if hasattr(self.module, 'prototype'):
            prototype = import_template(self.module.prototype)
        else:
            prototype = None
        for item in ('init', 'create', 'onload', 'cell_change',
                     'comment_change', 'content', 'css', 'check',
                     'i_am_unique'):
            if hasattr(self.module, item):
                self.__dict__[item] = getattr(self.module, item)
            elif prototype:
                self.__dict__[item] = getattr(prototype, item)

        self.headers = ''
        if prototype:
            self.headers += prototype.headers
        for extension in ('.css', '.js'):
            if os.path.exists(self.name[:-3] + extension):
                name = self.splited[-1].replace(".py", extension)
                if name not in files.files:
                    path = list(self.splited)
                    path[-1] = name
                    files.add(*path).update_content()
                self.headers += {
                    ".css":
                    '<link rel="stylesheet" href="{}/{}" type="text/css">',
                    ".js":
                    '<script src="{}/{}" onload="this.onloadDone=true"></script>'
                    }[extension].format(configuration.url_files, name)

    def init(*args, **keys):
        pass
    def create(self, ttable):
        ttable.table_attr(ttable.pages[0], 'masters', [ttable.user])
    def onload(*args, **keys):
        pass
    def cell_change(*args, **keys):
        pass
    def comment_change(*args, **keys):
        pass
    css = ''
    headers = ''
    i_am_unique = False
    def check(*args, **keys):
        pass
    def content(*args, **keys):
        return ''

def search_template(name):
    if name in Template.loaded:
        return Template.loaded[name]
    for path in (
            ('LOCAL', 'LOCAL_TEMPLATES', name.lstrip('/')),
            ('TEMPLATES', name.lstrip('/')),
        ):
        if ((configuration.regtest and not configuration.regtest_load_local)
              or name.startswith('/')
            ) and path[0] == 'LOCAL':
            continue
        filename = os.path.join(*path) + '.py'
        if os.path.exists(filename):
            t = Template(filename)
            Template.loaded[name] = t
            return t

def import_template(ue, semester=''):
    t = search_template(ue) or search_template(semester) or Template('NONE')
    t.update_template()
    return t

class Table(object):
    new_abjs = None
    update_content = ()
    
    def __init__(self, year, semester, ue, ro=True, user=None):

        warn('%s/%s/%s ro=%s user=%s' % (year, semester, ue, ro, user),
             what="table")
        self.year = year
        self.semester = semester
        self.ue = ue
        self.loading = False # The table is not being loaded
        self.sent_to_browsers = [] # Nothing has been sent
        self.filter_cache = {}

        if self.ue == 'abj' or self.ue == 'data':
            # Break the import in the file
            raise ValueError("Invalid table name: '%s'" % self.ue) # pragma: no cover

        # Remove -1 -2 -3 ... -9999... at the end of the UE name
        ue_code = ue.split('-')
        if len(ue_code) >= 2 and utilities.is_an_int(ue_code[-1]):
            self.ue_code = '-'.join(ue_code[:-1])
        else:
            self.ue_code = ue

        # Get the filter
        if '@' in ue:
            self.ue_code, self.seqgrp_filter = self.ue_code.split('@', 1)
        else:
            self.seqgrp_filter = None

        for attr in column.TableAttr.attrs.values():
            d = attr.get_default_value(self)
            if isinstance(d, list):
                d = list(d)
            elif isinstance(d, dict):
                d = dict(d)
            setattr(self, attr.name, d)

        self.pages = []
        self.active_pages = []
        self.columns = column.Columns(self)
        self.lines = Lines(self.columns)
        self.the_lock = threading.Lock()
        self.ro = ro
        self.mtime = 0
        self.the_key_dict = collections.defaultdict(list)
        self.unloaded = False
        self.do_not_unload = []
        self.last_cell_change = None
        self.update_time = self.load_time = int(time.time())
        dirname = os.path.join(configuration.db,
                               'Y'+str(self.year), 'S'+self.semester)
        self.filename = os.path.join(dirname, ue + '.py')

        self.template = import_template(self.ue_code, self.semester)
        self.template.init(self)
        self.is_unique() # Raise an error before creating the table

        if self.template.module:
            for v in self.template.module.__dict__.values():
                if isinstance(v, utilities.Variables):
                    self.group = v._group # pragma: no cover

        created = False
        self.on_disc = True
        if self.modifiable:
            utilities.mkpath(dirname)
            if not os.path.exists(self.filename):
                s = ['# -*- coding: utf8 -*- ']
                s.append('from TOMUSS.data import *') # Do not remove
                s = '\n'.join(s) + '\n'
                utilities.mkpath(dirname)
                utilities.append_file(self.filename, s)
                created = True
        else:
            if not os.path.exists(self.filename):
                self.on_disc = False
                # Read only table and file does not exists
                return

        stat = os.stat(self.filename)
        self.mtime = stat.st_mtime
        self.inode = stat.st_ino

        if not created:
            self.loading = True
            try:
                with open(self.filename, "r") as f:
                    while True:
                        line = f.readline()
                        if line.startswith('from') or line == '':
                            break
                    self.compile(f)
            finally:
                self.loading = False
        else:
            if ro:
                 # Do not create the table
                return # pragma: no cover
            warn('Create start', what='table')
            self.user = user # Usable by template.create()
            # We lock because table change methods expects to be locked.
            # But this lock does not protect anything
            self.lock()
            self.get_ro_page() # The first page is always ro one
            try:
                self.template.create(self)
            except: # pragma: no cover
                utilities.send_backtrace("", "Template.create %s/%s/%s" % (
                                year, semester, ue))
                utilities.unlink_safe(self.filename, do_backup=False)
                raise
            finally:
                self.unlock()
                warn('Unlock', what='table')
            warn('Create end', what='table')

        self.extensions = []
        if self.semester == configuration.university_semesters[0]:
            for dy, sem in zip(configuration.semesters_year,
                               configuration.semesters):
                if sem == self.semester:
                    continue
                if os.path.islink(table_filename(year-dy, sem, self.ue)):
                    self.extensions.append((year-dy, sem))

        self.terminate_init()

        if not ro:
            # Check mails or new students
            # Only mail if allow_modification is False
            warn('Update student list', what='table')
            update_students.append(self)

    def compile(self, table_file):
        """The file must not contain: from data import *"""
        _cell_change = self.cell_change
        _comment_change = self.comment_change
        pages = self.pages
        new_page = self.new_page
        cell_change = lambda page_id, col, lin, value, date: _cell_change(
            pages[page_id], col, lin, value, date)
        column_delete = lambda page_id, col: self.column_delete(pages[page_id], col)
        def comment_change(page_id, col, lin, value, date=None):
            _comment_change(pages[page_id], col, lin, value, date)
        def table_attr(attr, page_id, value, date=''):
            """Set a table attribute"""
            column.TableAttr.attrs[attr].set(self, pages[page_id], value, date)
        def column_attr(attr, page_id, col_id, value, date=''):
            """Set a column attribute"""
            attr = column.ColumnAttr.attrs[attr]
            col = self.columns.from_id(col_id)
            page = pages[page_id]
            if col is None:
                col = self.add_empty_column(page, col_id)
            attr.set(self, page, col, value, date)
        # DEPRECATED !!!!!!!!!
        # The following function are only used to load table created
        # with old TOMUSS version
        def column_empty_is(page_id, col, empty_is): # pragma: no cover
            """DEPRECATED: Change the column empty_is value."""
            column_attr('empty_is', page_id, col, empty_is)
        def column_visibility_date(page_id, col, date): # pragma: no cover
            """DEPRECATED: Specify the date of visibility"""
            column_attr('visibility_date', page_id, col, date)
        def column_comment(page_id, col, comment): # pragma: no cover
            """DEPRECATED: Change the column comment."""
            column_attr('comment', page_id, col, comment)
        def column_position(page_id, col, position): # pragma: no cover
            """DEPRECATED: Define the column position (the columns are sorted by position."""
            column_attr('position', page_id, col, position)
        def column_change(page_id, col, title, ttype, test, weight, freezed,
                          hidden, width): # pragma: no cover
            """DEPRECATED: Change some attributes of the column."""
            self.column_change(pages[page_id],
                               col, title, ttype, test, weight, freezed,
                               hidden, width)
        def add_master(name, page_id=None): # pragma: no cover
            """DEPRECATED: Add or remove a master on the page."""
            self.add_master(name, page_id)
        def date_change(page_id, dates): # pragma: no cover
            """DEPRECATED: Change the table dates."""
            self.date_change(pages[page_id], dates)
        def private_toggle(page_id): # pragma: no cover
            """DEPRECATED: Change the privacy state of the table."""
            self.private_toggle(pages[page_id])
        def table_comment(page_id, comment): # pragma: no cover
            """DEPRECATED: Change the table comment."""
            self.table_comment(pages[page_id], comment)
        default_nr_columns = self.default_nr_columns_change # DEPRECATED

        while True:
            lines = table_file.read(100000) + table_file.readline()
            if not lines:
                return
            compiled = compile(lines, self.filename, 'exec')
            eval(compiled, {}, locals())
            self.nb_bytes_compiled = table_file.tell()

    def terminate_init(self):
        self.template.onload(self)
        if configuration.read_only:
            self.compute_columns()

    @utilities.add_a_lock
    def compute_columns(self, lines=None, hook=None):
        """ """
        if lines is None:
            lines = tuple(self.lines.values())
        lines = [line for line in lines if line[0].value != '']
        if configuration.read_only:
            for column in self.columns:
                column.type.update_for_suivi(column)

        from .PYTHON_JS import tomuss_python
        from . import cell
        tomuss_python.columns = self.columns.columns
        tomuss_python.C = cell.Cell
        for i in ('abi', 'abj', 'ppn', 'tnr', 'pre'):
            setattr(tomuss_python, i, getattr(configuration, i, ''))
            setattr(tomuss_python, i+'_short',
                        getattr(configuration, i+'_short', ''))
        tomuss_python.allowed_grades = configuration.allowed_grades
        for column in self.columns.columns_ordered():
            column.real_weight = tomuss_python.to_float(column.weight)
            column.real_weight_add = not column.weight.startswith(('+','-'))
            column.min, column.max = column.min_max()
            column.best_of = -int(tomuss_python.to_float(column.best))
            column.mean_of = -int(tomuss_python.to_float(column.worst))
            column.floats = None # Erase note.py/stat cache
            column.replacements = None # Erase cache
            if column.rounding == '':
                column.round_by = 0
            else:
                column.round_by = float(column.rounding)

            column.test_filter_filter = self.get_filter(column.test_filter,
                                                        column.type.name)
            column.test_if_filter = self.get_filter(column.test_if,
                                                    column.type.name)
            if column.is_computed():
                try:
                    column.average_columns = [
                        self.columns.from_title(title).data_col
                        for title in column.get_columns()
                    ]
                except: # pragma: no cover
                    column.average_columns = []
                cell_compute = eval(column.type.cell_compute,
                                    tomuss_python.__dict__
                                )
                for line in lines:
                    tomuss_python.compute_cell_safe(column.data_col, line,
                                                    cell_compute, 'nobody')
        if hook:
            hook()

    def update(self):
        """Update the table if the file on disc changed.
        It is used only by 'suivi' servers
        """
        # warn('Start ro=%s dt=%f p=%s' % (self.ro, time.time() - self.mtime, os.path.exists(self.filename)), what="table")
        if not self.ro:
            return self

        try:
            stat = os.stat(self.filename)
        except FileNotFoundError: # pragma: no cover
            # Do not work (to not recreate it)
            # tables_manage("del", self.year, self.semester, self.ue)
            # Should return None, but as it is very intrusive
            # Make a non dangerous kludge
            self.official_ue = False
            return self

        if stat.st_mtime <= self.mtime:
            return self

        if time.time() - self.mtime < configuration.maximum_out_of_date:
            return self # pragma: no cover

        if stat.st_size < self.nb_bytes_compiled or stat.st_ino != self.inode:
            # The table has been edited using a text editor : reload from start
            return tables_manage("replace", self.year, self.semester, self.ue,
                             new_table = Table(self.year, self.semester,
                                               self.ue, ro=True))

        self.mtime = stat.st_mtime
        if stat.st_size == self.nb_bytes_compiled:
            # wtime changed but not the file
            return self # pragma: no cover

        computed_columns = set(column.the_id for column in self.columns if column.is_computed())
        self.loading = True
        try:
            with open(self.filename, 'r') as f:
                f.seek(self.nb_bytes_compiled)
                self.compile(f)
        finally:
            self.loading = False
        if computed_columns - set(column.the_id for column in self.columns if column.is_computed()):
            # A column is no more computed: old values must be restored
            # The only way to do this is to reload the table
            return tables_manage("replace", self.year, self.semester, self.ue,
                                 new_table = Table(self.year, self.semester,
                                                   self.ue, ro=True))
        self.terminate_init()

        return self


    def change_mails(self, mails):
        self.table_attr_computed('mails', mails)
    def change_portails(self, portails):
        self.table_attr_computed('portails', portails)

    def table_attr_computed(self, attr, value):
        if getattr(self, attr) == value:
            return
        setattr(self, attr, value)
        t = '<script>Xtable_attr(%s,%s);</script>\n' % (repr(attr), js(value))
        self.send_update(None, t)

    def update_mail(self, login, mail):
        warn('update mail of ' + self.location(), what="table")
        self.mails[login] = mail
        self.send_update(None,
                         ('<script>update_mail(' + repr(login)
                         + ',' + js(mail) + ');</script>\n')
                         )

    def update_portail(self, login, portail):
        if not portail:
            return
        self.portails[login] = portail # pragma: no cover
        self.send_update(None,
                         ('<script>update_portail(' + repr(login)
                         + ',' + js(portail) + ');</script>\n')
                        ) # pragma: no cover
        
    def lock(self):
        warn('start', what='debug')
        self.the_lock.acquire()
        warn('acquired', what='debug')

    def unlock(self):
        warn('start', what='debug')
        self.the_lock.release()

    def panic(self, txt): # pragma: no cover
        utilities.send_backtrace('UE: %s ' % self.ue + txt,
                                 exception=False)
        self.send_update(None, "<script>window.close();</script>\n")
        raise ValueError('PANIC')

    def state_is_fine(self):
        if self.unloaded:
            self.panic('Modification on an unloaded table') # pragma: no cover
        if not self.the_lock.locked():
            self.panic('Modification on an unlocked table') # pragma: no cover
        if configuration.read_only:
            self.panic('Modification in a readonly process') # pragma: no cover
                
    def log(self, text):
        self.state_is_fine()
        utilities.append_file(self.filename, text + '\n')
        warn(self.filename + ' ' + text)
        self.mtime = time.time()

    def new_page(self, ticket, user_name, user_ip, user_browser, date=None,
                 recycle=False):
        if recycle:
            for p in self.pages:
                if (p.user_name == user_name
                    and p.user_ip == user_ip
                    and p.user_browser == user_browser
                    and p.ticket == ticket
                    ):
                    return p
        if not self.loading and self.modifiable:
            if date is None:
                date = time.strftime('%Y%m%d%H%M%S')
            self.log('new_page(%s ,%s, %s, %s, %s) # %d' % (
                repr(ticket),
                repr(user_name),
                repr(user_ip),
                repr(user_browser),
                repr(date),
                len(self.pages),
                ))
        p = Page(ticket, user_name, len(self.pages), self,
                 user_ip, user_browser, date)
        p.logged = self.loading or self.modifiable
        p.index = len(self.sent_to_browsers)
        self.pages.append(p)
        return p

    def active_page(self, page, ffile):
        warn('page=%s file=%s(%s)' % (page, ffile, ffile.closed), what="table")
        if page.browser_file and not page.browser_file.closed:
            page.browser_file.close() # pragma: no cover
        page.browser_file = ffile
        if page not in self.active_pages:
            self.active_pages.append(page)

    def remove_active_page(self, page):
        self.active_pages.remove(page)
        page.browser_file = None # stop a memory leak
            
    def send_update(self, page, value, store=True):
        warn('%s actives: %s' % (page, str(self.active_pages)), what='table')
        if store:
            self.sent_to_browsers.append(value)
        for p in tuple(self.active_pages):
            if p == page:
                continue
            if p.browser_file.closed:
                self.remove_active_page(p)
                if hasattr(p, 'start_load') and not hasattr(p, 'end_of_load'):
                    global canceled_loads
                    # Update list of canceled page load
                    now = time.time()
                    canceled_loads.append((now, p.user_name))
                    old = now - 3600
                    canceled_loads = [i for i in canceled_loads
                                     if i[0] > old]
                    user = p.user_name
                    nr = list([i for i in canceled_loads if i[1] == user])
                    if len(nr) == 4: # pragma: no cover
                        # 4 cancels in less than one hour
                        user_mail = configuration.User(user).mail
                        if user_mail:
                            utilities.send_mail_in_background(
                                user_mail,
                                # XXX need translation
                                utilities._("MSG_document_problems"),
                                utilities.read_file(
                                        os.path.join('FILES', 'mail_cancel'))
                                )

                    # Send a mail to the maintainer
                    utilities.send_backtrace(
                        repr(nr),
                        utilities._("MSG_document_canceled") + user,
                        exception=False)
                        
                continue
            sender.append(p.browser_file, value,
                          index=len(self.sent_to_browsers), page=p)

    def get_filter_(self, the_filter, column_type):
        if the_filter not in self.filter_cache:
            from .PYTHON_JS import tomuss_python
            self.filter_cache[the_filter] = tomuss_python.Filter(
                the_filter, column_type, self.columns.columns)
        return self.filter_cache[the_filter]

    def get_filter(self, the_filter, column_type):
        return self.get_filter_(the_filter, column_type).evaluate

    def authorized(self, user_name, value, column=None, line=None, is_a_teacher=False):
        # Authorized because the test have yet be done in the past
        if self.loading:
            return True # pragma: no cover
        # RO and RW users have all the rights
        if user_name == data.ro_user or user_name == data.rw_user:
            return True
        # Values setted by user '*' are not modifiable
        if value.author == data.ro_user and value.value != '':
            return False
        # The masters of the UE may change any value setted by another user
        if user_name in self.masters:
            return True
        if column and line and column and column.cell_writable:
            return self.get_filter(
                column.cell_writable, column.type.name)(line, value, user_name, is_a_teacher)
        # Values setted by user '' are modifiable
        if value.author == data.rw_user:
            return True
        # Empty values are modifiable by anyone
        if value.value == '':
            return True
        # The user can change its values
        if value.author == user_name:
            return True
        return False

    def authorized_column(self, user_name, a_column, is_a_teacher=False):
        return self.authorized(
            user_name, CellValue(a_column.title, a_column.author),
            is_a_teacher=is_a_teacher)

    def cell_change(self, page, col, lin, value=None,
                    date=None, force_update=False,
                    change_author=True):
        if not self.loading and not self.modifiable:
            return self.bad_ro(page)
        a_column = self.columns.from_id(col)
        if a_column == None:
            raise ValueError(utilities._("MSG_document_bug_column") % col)

        on_a_new_line = lin not in self.lines
        line = self.lines[lin]
        cell = line[a_column.data_col]

        if value is None:
            value = cell.value

        if not self.loading and not force_update:
            if a_column.is_computed():
                self.compute_columns([line])
                cell = line[a_column.data_col] # Class change is possible
                # Now the cell contains the right value
                if (isinstance(cell.value, float)
                    and not math.isnan(cell.value)
                    and not a_column.cell_is_modifiable
                    ):
                    return 'bad.png' # pragma: no cover
            if not self.authorized(page.user_name, cell, a_column, line,
                                   is_a_teacher=page.is_a_teacher()):
                utilities.warn('cell value = (%s)' % cell.value)
                return self.bad_auth(page, "cell_change %s/%s/%s" % (
                        col, lin, value))
            if a_column.locked and page.user_name != data.ro_user:
                self.error(page, utilities._("MSG_document_column_locked"))
                return 'bad.png'
        old_value = str(cell.value)
        new_value = str(value)
        if new_value == 'nan':
            value = "nan" # pragma: no cover
        # Do not use 'a_column.equals': look at 'git blame'
        equal = old_value == new_value
        if equal:
            if cell.author == page.user_name or not change_author:
                if len(page.user_name) > 1 and a_column.type.cell_compute != 'undefined':
                    line[a_column.data_col].date = (date
                                               or time.strftime('%Y%m%d%H%M%S'))
                return 'ok.png'

        if a_column.type.should_be_a_float:
            try:
                if value != "nan":
                    value = float(value)
            except ValueError:
                pass

        if (not self.loading
            and a_column.repetition
            and value != ''
            and not equal
            # 'repetition' is checked only for the first member
            and not getattr(a_column, 'groupcolumn_running', False)
        ):
            data_col = a_column.data_col
            group = None
            if a_column.groupcolumn:
                group = self.columns.from_title(a_column.groupcolumn)
                if group:
                    group = group.data_col
            if a_column.repetition > 0:
                verify_lines = self.lines.values()
            else:
                grp = line[self.columns.get_grp()].value
                seq = line[self.columns.get_seq()].value
                verify_lines = self.columns.table.lines_of_grp(grp, seq)
            value_striped = str(value).rstrip('_') # For enumeration
            if group is None:
                counters = collections.defaultdict(int)
                for a_line in verify_lines:
                    if str(a_line[data_col].value).rstrip('_') == value_striped:
                        counters[a_line[data_col].value] += 1
            else:
                groups = collections.defaultdict(set)
                for a_line in verify_lines:
                    if str(a_line[data_col].value).rstrip('_') == value_striped:
                        groups[a_line[data_col].value].add(str(line[group].value))
                counters = {key: len(val) for key, val in groups.items()}
            if a_column.type.name == 'Enumeration':
                for item in a_column.enumeration.split(' '):
                    if item.startswith(value) and item not in counters:
                        counters[item] = 0
            if value in counters:
                for a_value, nr_use in sorted(counters.items()):
                    if nr_use < abs(a_column.repetition):
                        value = a_value # First slot usable
                        break
                else:
                    sender.append(
                        page.browser_file,
                        '<script>alert(_("ALERT_repetition")+%s+"/"+%s);</script>\n'
                        % (js(value), js(a_column.title)))
                    return 'bad.png'

        if not self.loading and not equal:
            self.template.cell_change(self, page, col, lin, value, date)

        if a_column.data_col == 0:
            login = utilities.the_login(old_value)
            if login in self.the_key_dict:
                try:
                    self.the_key_dict[login].remove(lin)
                    if len(self.the_key_dict[login]) == 0:
                        del self.the_key_dict[login]
                except:
                    if login:
                        utilities.warn(str(page) + ' old_login=' + login
                                       + ' new_login=' + new_value
                                       + ' lin=' + lin
                                       + ' ' + repr(self.the_key_dict[login])) # pragma: no cover
            new_login = utilities.the_login(new_value)
            self.the_key_dict[new_login].append(lin)
            if not self.loading and self.official_ue:
                indexes_to_update.append((self, login, new_login))
                for y, s in self.extensions:
                    # XXX The regression tests must cover this line
                    indexes_to_update_append(y, s, self.ue, login, new_login) # pragma: no cover

        elif on_a_new_line:
            self.the_key_dict[''].append(lin)

        # The class may change on value change
        cell = line[a_column.data_col] = cell.set_value(value=value,
                                                        author=page.user_name,
                                                        date=date)
        self.last_cell_change = date
        page.nr_cell_change += 1
        if not self.loading:
            self.log('cell_change(%s,%s,%s,%s,"%s")' % (
                page.page_id,
                repr(col),
                repr(lin),
                repr(value),
                cell.date))
            t = '<script>Xcell_change(%s,%s,%s,%s,%s,%s);</script>\n' % (
                js(col),
                js(lin),
                js(value),
                js(cell.date),
                js(cell.author),
                cell.js_history(),
                )
            if force_update:
                self.send_update(None, t)
            else:
                self.send_update(page, t)
            cell_changed_list_fast.append((self, lin, a_column))

            if (a_column.groupcolumn
                and not getattr(a_column, 'groupcolumn_running', False)
                ):
                a_column.groupcolumn_running = True
                is_a_teacher = page.is_a_teacher()
                try:
                    for line_key, i_line in a_column.lines_of_the_group(line):
                        if self.authorized(page.user_name,
                                           i_line[a_column.data_col],
                                           a_column, i_line,
                                           is_a_teacher=is_a_teacher):
                            self.cell_change(page, a_column.the_id,
                                             line_key, value,
                                             force_update=True)
                finally:
                    a_column.groupcolumn_running = False

            if a_column.alert:
                columnalert.send_mail(page, a_column, line)

        return 'ok.png'

    def cell_set_change(self, page, col, lin, add=(), remove=(), **kargs):
        """The cell contains a string of sorted space separated values.
        The cell change indicates what to add and remove to the current string"""
        current_value = self.lines[lin][self.columns.from_id(col).data_col].value
        if current_value == '':
            new_set = add
        else:
            new_set = set(current_value.split(' '))
            for i in remove:
                new_set.discard(i)
            new_set.update(add)
        new_value = ' '.join(sorted(new_set))
        if new_value != current_value:
            self.cell_change(page, col, lin, new_value, **kargs)

    def default_nr_columns_change(self, n): # pragma: no cover
        """Deprecated"""
        self.default_nr_columns = n

    def add_master(self, name, page_id=0): # pragma: no cover
        """Deprecated"""
        name = name.lower()
        if name in self.masters:
            self.masters.remove(name)
        else:
            self.masters.append(name)
        if not self.loading:
            # For old TEMPLATES files
            self.log('table_attr("masters",%d,%s)' % (page_id, repr(name))) 

    def get_a_page_for(self, user_list):
        """Returns a page for one of the users in the list.
        If not exists, create one with the first user.
        If the user list is empty, create a page for the rw_user
        """
        rw = None
        for p in self.pages:
            if p.user_name in user_list:
                return p
            if p.user_name == data.rw_user:
                rw = p # pragma: no cover
        if user_list:
            return self.new_page('', user_list[0], '', '')
        if rw:
            return rw # pragma: no cover
        return self.new_page('', data.rw_user, '', '')

    def get_a_master_page(self):
        """Returns a page modifiable by the masters.
        Create one if there is none"""
        return self.get_a_page_for(self.masters)

    def get_ro_page(self):
        """Returns a page modifiable by nobody."""
        return self.get_a_page_for((data.ro_user,))

    def get_a_root_page(self):
        """Returns a page for root."""
        return self.get_a_page_for((configuration.root[0],))

    def get_rw_page(self):
        """Returns a page modifiable by anybody."""
        return self.get_a_page_for(())

    def get_nobody_page(self):
        """Returns a page for the nobody user."""
        return self.get_a_page_for((data.no_user,))

    def private_toggle(self, page): # pragma: no cover
        """Deprecated"""
        self.private = 1 - self.private
        if not self.loading: # compatibility with old TEMPLATES
            self.log('table_attr("private",%d,%d)' % (page.page_id,
                                                      self.private))
        return 'ok.png'

    def error(self, page, message, more_in_mail="", backtrace=True,
                error_comment='ERROR_server_bug'):
        if backtrace:
            utilities.send_backtrace(
                'UE: %s, Page: %s' % (self.ue, page)
                + '\n' + more_in_mail, subject='###' + message,
                exception=False)
        if '_(' not in message:
            # The message is not javascript program
            message = js(message)
        message = message + '+"\\n\\n"+_("{}")'.format(error_comment)
        sender.append(page.browser_file,
                      '<script>alert(%s);</script>\n' % message)
        return "bad.png"

    def bad_column(self, page):
        return self.error(page, '_("ALERT_column_not_exist")')

    def bad_auth(self, page, more_in_mail=""):
        return self.error(page, '_("ALERT_not_authorized")', more_in_mail)

    def bad_ro(self, page):
        return self.error(page, '_("ALERT_value_ro")')

    def comment_change(self, page, col, lin, value, date=None):
        if not self.loading and not self.modifiable:
            return self.bad_ro(page) # pragma: no cover

        a_column = self.columns.from_id(col)
        if a_column == None:
            return self.bad_column(page)

        line = self.lines[lin]
        if value == line[a_column.data_col].comment:
            return 'ok.png'

        if self.loading:
            if date is None: # Old files
                date = self.last_cell_change
        else:
            if not self.authorized(page.user_name,
                                   line[a_column.data_col], a_column, line,
                                   is_a_teacher = page.is_a_teacher()):
                return self.bad_auth(page, "comment_change %s/%s/%s" % (
                        col, lin, value))
            self.template.comment_change(self, page, col, lin, value)

        cell = line[a_column.data_col] = line[a_column.data_col].set_comment(
                        value, author=page.user_name, date=date)

        if not self.loading:
            self.log('comment_change(%s,%s,%s,%s,%s)' % (
                page.page_id,
                repr(col),
                repr(lin),
                repr(value),
                repr(cell.history[-1][1])
                ))
            t = '<script>Xcomment_change(%s,%s,%s,%s,%s);</script>\n' % (
                js(page.user_name),
                js(col),
                js(lin),
                js(value),
                js(list(cell.history[-1]))
                )
            self.send_update(page, t)

        return 'ok.png'

    def empty(self, empty_even_if_used_page=False,
              empty_even_if_created_today=False,
              empty_even_if_column_created=False,
              ):
        if not empty_even_if_used_page and self.active_pages:
            return False, utilities._("MSG_document_table_active") # pragma: no cover
        if not empty_even_if_created_today and time.time()-self.mtime <24*3600:
            return False, utilities._("MSG_document_table_today")
        if self.comment:
            return False, utilities._("MSG_document_table_comment"
                                      ) % html.escape(self.comment)
        if not empty_even_if_column_created:
            for c in self.columns:
                if c.author != data.ro_user and c.author != data.no_user:
                    return False, utilities._("MSG_document_table_title") # pragma: no cover
        for line in self.lines.values():
            for j in line:
                if not j.empty():
                    return False, utilities._("MSG_document_table_cell") # pragma: no cover
                if j.comment:
                    return False, utilities._("MSG_document_table_cell_comment") # pragma: no cover

        return True, utilities._("MSG_document_table_lines") % len(self.lines)

    def problem_in_column_name(self):
        names = set()
        for c in self.columns:
            if c.title in names:
                return utilities._("MSG_document_table_duplicate") + c.title # pragma: no cover
            names.add(c.title)
        for c in self.columns:
            for name in c.depends_on():
                if name not in names:
                    return utilities._("MSG_document_table_no_column") % (
                        c.title, name) # pragma: no cover
        return ""
        

    def table_comment(self, page, comment): # pragma: no cover
        """Deprecated"""
        return self.table_attr(page, 'comment', comment)

    def add_empty_column(self, page, the_id=None):
        """If the id is not given, create a new one"""
        if self.columns.columns:
            position = max([c.position for c in self.columns]) + 1
        else:
            position = 0
        if the_id is None:
            i = len(self.columns)
            ids = set(c.the_id for c in self.columns)
            while True:
                the_id = str(i)
                if the_id not in ids:
                    break
                i += 1 # pragma: no cover
        a_column = column.Column(the_id, page.user_name, position=position)

        self.columns.append(a_column)
        for line in self.lines.values():
            line.append(cellempty)
        return a_column

    def column_attr(self, page, col_id, attr, value):
        col = self.columns.from_id(col_id)
        if col is None:
            col = self.add_empty_column(page, col_id)
        date = time.strftime('%Y%m%d%H%M%S')
        return column.ColumnAttr.attrs[attr].set(self, page, col, value, date)

    def column_changed(self, a_column, attribute):
        "An column attribute change, column content may be updated"
        column_changed_list.append((self, a_column, attribute))

    def table_attr(self, page, attr, value):
        date = time.strftime('%Y%m%d%H%M%S')
        return column.TableAttr.attrs[attr].set(self, page, value, date)

    def column_comment(self, page, col_id, comment):
        """DEPRECATED : use column_attr"""
        self.column_attr(page, col_id, 'comment', comment)

    def column_change(self, page, col_id, title, ttype, test, weight,
                      freezed, hidden=0, width=1): # pragma: no cover
        """DEPRECATED : use column_attr"""

        col = self.columns.from_id(col_id)
        if col is None:
            col = self.add_empty_column(page, col_id)
        attrs = column.ColumnAttr.attrs
        attrs['type'].set(self, page, col, ttype, '')
        attrs['title'].set(self, page, col, title, '')
        attrs['freezed'].set(self, page, col, freezed, '')
        attrs['hidden'].set(self, page, col, hidden, '')
        attrs['width'].set(self, page, col, width, '')

        t = test.split(';')
        if len(t) == 4:
            test = '[' + t[0].strip('[') + ';' + t[3].strip(']') + ']'
            attrs['red'].set(self, page, col, t[1], '')
            attrs['green'].set(self, page, col, t[2], '')
        if ttype == 'Nmbr':
            attrs['test_filter'].set(self, page, col, test, '')
        else:
            if test:
                attrs['minmax'].set(self, page, col, test, '')

        t = re.split(' +', weight)
        if len(t) > 1:
            weight = t[0]
            attrs['columns'].set(self, page, col, ' '.join(t[1:]), '')

        if weight:
            attrs['weight'].set(self, page, col, weight, '')


    def column_delete(self, page, col):
        if not self.loading and not self.modifiable:
            return self.bad_ro(page)
        a_column = self.columns.from_id(col)
        if a_column == None:
            if self.loading:
                return self.bad_column(page)
            else:
                # User delete a not created column
                return 'bad.png'
        if (a_column.type.cell_is_modifiable
            and not a_column.empty_of_user_values()):
            return self.error(page, '_("ALERT_delete_not_empty")')

        if not self.loading:
            self.log('column_delete(%s,%s)' % (page.page_id, repr(col)))
            t = '<script>Xcolumn_delete(%s,%s);</script>\n' % (
                js(page.user_name), js(col))
            self.send_update(page, t)
            self.filter_cache = {} # filters are computed using data_col
            self.columns.column_ordered_cache = None

        for line in self.lines.values():
            line.pop(a_column.data_col)
        self.columns.pop(a_column.data_col)
            
        return 'ok.png'

    def column_inscrit(self):
        for i, c in enumerate(self.columns):
            if c.freezed == 'C':
                return i
        return None

    def content_head(self, page, hide_head=False):
        return table_head(
            self.year, self.semester, page.ticket, page.user_name,
            page.page_id, self.ue,
            attrs_from=self, table=self, hide_head=hide_head
            )

    def date_change(self, page, date):
        """Deprecated"""
        return column.TableAttr.attrs['dates'].set(self, page, date, '')

    def logins(self):
        for v in self.lines.values():
            if v[0].value != '':
                yield v[0].value

    def logins_valid(self):
        c = self.column_inscrit()
        if c is None:
            for v in self.lines.values(): # pragma: no cover
                if v[0].value != '':
                    yield v[0].value
        else:
            for v in self.lines.values():
                if v[0].value != '' and v[c].value != 'non':
                    yield v[0].value

    def the_keys(self):
        """Returns a dictionary for fast access by student login"""
        return self.the_key_dict

    def get_lines(self, login):
        login = utilities.the_login(login)
        if login in self.the_key_dict:
            for lin in self.the_key_dict[login]:
                yield self.lines[lin]
        return

    def create_line_id(self):
        i = len(self.lines)
        while True:
            line_id = "0_" + str(i)
            if line_id not in self.lines:
                break
            i += 1 # pragma: no cover
        return line_id

    def get_line_id_or_create(self, login):
        login = utilities.the_login(login)
        line_id = self.the_key_dict.get(login, None)
        if line_id is not None:
            return line_id[0]
        return self.create_line_id()

    def lines_of_grp(self, grp, seq):
        if grp == '':
            return # Do not auto compute ABINJ??? if there is no real groups
        grp_col = self.columns.get_grp()
        seq_col = self.columns.get_seq()
        if grp_col and seq_col:
            for line in self.lines.values():
                if line[grp_col].value == grp and line[seq_col].value == seq:
                    yield line

    def lines_of_grp_seq(self, grp, seq):
        """Set 'grp' 'seq' as empty string to not test it.
        If both are empty, returns all the lines.
        Returns only registered students lines.
        """
        inscrit_col = self.column_inscrit()
        grp_col = self.columns.get_grp()
        seq_col = self.columns.get_seq()
        for line in self.lines.values():
            if not line[0].value:
                continue
            if inscrit_col and line[inscrit_col].value == 'non':
                continue
            if grp and line[grp_col].value != grp:
                continue
            if seq and line[seq_col].value != seq:
                continue
            yield line

    def get_items(self, login):
        login = utilities.the_login(login)
        if login in self.the_key_dict:
            for lin in self.the_key_dict[login]:
                yield lin, self.lines[lin]
        return
    
    def authors(self):
        """List of all the logins that have modified a cell value"""
        a = set()
        for v in self.lines.values():
            for cell in v:
                if cell.author not in ('', data.ro_user, data.no_user):
                    a.add(cell.author)
        a.update(self.masters)
        a.update(self.teachers)
        return a

    def update_columns(self, columns, ro_page=None):
        """Update the default columns of the table.
        This can be called by the TEMPLATE 'check' method
        """
        if not self.modifiable:
            return # pragma: no cover
        if ro_page is None:
            ro_page = self.pages[0]
        locked = self.the_lock.locked()
        try:
            if not locked:
                self.lock()
            for col in sorted(columns):
                a_column = self.columns.from_id(col)
                if 'title' in columns[col]:
                    from_title = self.columns.from_title(columns[col]['title'])
                    if from_title and from_title != a_column:
                        # Duplicate column
                        continue # pragma: no cover
                for attr, value in columns[col].items():
                    value = str(value)
                    if a_column is None or value != str(getattr(a_column, attr)):
                        if a_column is None or a_column.author == ro_page.user_name:
                            self.column_attr(ro_page, col, attr, value)
                        else:
                            # The column is modified in the name of its author.
                            # The column definition must be updated but not the author
                            self.column_attr(self.get_a_page_for((a_column.author,)),
                                             col, attr, value)
        finally:
            if not locked:
                self.unlock()

    def content(self, page, server, hide_head=False):
        warn('%s content for page %d' % (
            self.filename, page.page_id))
        s = []

        s.append(self.content_head(page, hide_head=hide_head))

        try:
            s.append('  <script><!--')
            s.append(utilities.wait_scripts())
            s.append('var page_index = %d ;' % page.index)
            s.append('--></script> ')
            lines = self.lines.js(server.ticket.user_name)
            i = 0
            nb = 100
            to_load = len(lines)
            while lines:
                s.append('<script><!--')
                s.append('\n'.join(lines[:nb]))
                del lines[:nb]
                i += nb
                s.append('--></script><div class="loading_bar"><div style="width:%d%%">&nbsp;</div></div>'
                         % int(i*100./to_load) )
            s.append('  <style>DIV.loading_bar { display: none;}</style>')
            s.append('<style id="template_style">'
                     + self.template.css + '</style>')
            s.append('''  <script><!--
            function initialize()
            {
            if ( ! wait_scripts("initialize()") )
               return ;
            var t = document.getElementsByTagName("DIV");
            for(var i=0; i<t.length; i++)
                if ( t[i].className == "loading_bar" )
                    t[i].parentNode.removeChild(t[i]) ;
            document.write(head_html()) ;
            insert_middle();
            ''')
            if page.user_name in self.masters:
                hide = None
            else:
                hide = False
            s.append(self.columns.js(hide=hide, ticket=server.ticket))

            s.append('document.write(tail_html());')
            s.append('runlog(columns, lines) ;')
            s.append('set_updating(%s) ;' % int(self in update_students))
            s.append('}')
            if self.new_abjs:
                s.append(self.new_abjs)
            s.append(self.template.content(self))
            s.append('initialize();')
            s.append('--></script>  \n')
        finally:
            pass

        try:
            return '\n'.join(s)
        except UnicodeDecodeError: # pragma: no cover
            utilities.send_backtrace('\n'.join(repr(i)
                                               for i in s),
                                     "UNICODE PROBLEM")
            raise

    def location(self):
        return '%d %s %s' % (self.year, self.semester, self.ue)

    def rewrite(self, only_columns=False, user_name=''):
        s = ['# -*- coding: utf8 -*-']
        s.append('from data import *') # Do not remove
        s.append("new_page('' ,'" + data.ro_user + "', '', '')")
        pages = collections.defaultdict(lambda: 1)
        pages[data.ro_user] = 0
        authors = set()
        for c in self.columns:
            authors.add(c.author)
        if user_name:
            authors.add(user_name)
        if not only_columns:
            for line in self.lines.values():
                for cell in line:
                    authors.add(cell.author)
        authors.discard(data.ro_user)
        for author in authors:
            pages[author] = len(pages)
            s.append("new_page('' ,'%s', '', '')" % author)
            
        for c in self.columns:
            for attr in column.ColumnAttr.attrs.values():
                if attr.name == 'author':
                    continue
                attr_value = attr.decode(getattr(c, attr.name))
                if attr_value != attr.default_value:
                    s.append('column_attr(%s,%d,%s,%s)' % (
                        repr(attr.name), pages[c.author],
                        repr(c.the_id), repr(attr_value)))
                         
        for attr in column.TableAttr.attrs.values():
            if attr.computed:
                continue
            if attr.name in ('modifiable', 'dates'):
                continue
            attr_value = getattr(self, attr.name)
            if attr.name == 'masters':
                if user_name not in attr_value:
                    attr_value.append(user_name)
                if '' in attr_value:
                    attr_value.remove('')
            attr_value = attr.decode(attr_value)
            if attr_value != attr.default_value:
                s.append('table_attr(%s,0,%s)' % (
                    repr(attr.name),
                    repr(attr_value)))

        if only_columns:
            return '\n'.join(s) + '\n'
                         
        for line_key, line in self.lines.items():
            for col, cell in zip(self.columns, line):
                if cell.value != '':
                    if cell.date != '': # Not a computed value
                        s.append('cell_change(%d,%s,%s,%s,%s)' % (
                            pages[cell.author],
                            repr(col.the_id),
                            repr(line_key),
                            repr(cell.value),
                            repr(cell.date),
                            ))
                    if cell.comment:
                        s.append('comment_change(%d,%s,%s,%s)' % (
                                pages[cell.author],
                                repr(col.the_id),
                                repr(line_key),
                                repr(cell.comment),
                                ))
                                                    
        return '\n'.join(s) + '\n'

    def close_active_pages(self): # pragma: no cover
        for page in self.active_pages:
            try:
                page.browser_file.write('<script>window.parent.close();</script>')
                page.browser_file.close()
                page.browser_file = None
            except:
                pass
        self.active_pages = []

    def do_not_unload_add(self, value):
        if value:
            self.do_not_unload.append(value)
    def do_not_unload_remove(self, value):
        self.do_not_unload.remove(value)

    def unload(self, force=False):
        if force:
            self.close_active_pages()
            self.do_not_unload = []
        if '*' in ''.join(self.do_not_unload):
            return
        warn(str(self.ue), what="table")

        # XXX
        # In rare cases the table is unloaded while the student list
        # is being updated. So we finish with a table half updated
        # But it is not important.
        deleted = tables_manage("del", self.year, self.semester, self.ue)
        try:
            if deleted:
                update_students.remove(self) # No update student list.
        except ValueError:
            pass

        return deleted

    def delete(self):
        warn(str(self.ue), what="table")
        utilities.manage_key('CLOSED', self.ue, separation=5,
                             content='%d/deleted_table_%s' % (self.year,
                                                              self.semester)
                             )
        self.unload(force=True)
        if not self.unloaded:
            return # pragma: no cover
        # XXX Not locked, so the table may be reloaded before deletion....
        utilities.unlink_safe(self.filename)

        for name in self.masters:
            self.master_of_update('-', name)
        for login in self.the_keys():
            indexes_to_update.append((self, login, ''))
            for y, s in self.extensions:
                indexes_to_update_append(y, s, self.ue, login, '')

        ue_to_delete.append((self.year, self.semester, self.ue))
        utilities.start_job(remove_bookmarks, 0 if configuration.real_regtest else 60 * 5,
                            important='update_bookmarks')

    def send_alert(self, text):
        self.send_update(None, '<script>alert(_("ALERT_message_for")+%s);</script>\n' %
                         js(self.ue + '\n\n' + text))

    def __str__(self):
        return '%d/%s/%s' % (self.year, self.semester, self.ue)

    def backtrace_html(self):
        return "Table: " + str(self)

    def readable_by(self, the_ticket):
        if the_ticket.user_name in self.masters:
            return True
        if the_ticket.user_name in self.teachers:
            return True
        if the_ticket.user_name in self.managers:
            return True # pragma: no cover
        if the_ticket.user_name in configuration.root:
            return True
        if self.private:
            warn('Unauthorized access (private)', what='table')
            return False
        if self.teachers:
            warn('Unauthorized access (not teacher)', what='table')
            return False
        return True

    def master_of_update(self, what, name):
        if self.semester in configuration.master_of_exceptions:
            return
        user = configuration.User(name)
        d = user.manage_key('master_of')
        if d is False:
            d = []
        else:
            d = eval(d)

        t = (str(self.year), self.semester, self.ue)
        if (t in d) and what == '-':
            d.remove(t)
        elif (t not in d) and what == '+':
            d.append(t)
        else:
            return # pragma: no cover

        user.manage_key('master_of', content=utilities.stable_repr(d))

    def the_abjs(self):
        from . import abj

        grp_col = self.columns.get_grp()
        seq_col = self.columns.get_seq()
        inscrit = self.column_inscrit()

        t = []
        users = configuration.Users(self.logins())
        for login, user in users.items():
            tt = abj.tierstemps(user)
            student = abj.Abj(self.year, self.semester, login)
            items = tuple(self.get_items(login))
            first_sec = self.dates[0]
            last_sec = self.dates[1] + 86400
            if items:
                line_id, line = items[0]
                the_abjs= abj.do_prune(
                    student.abjs,
                    first_sec, last_sec,
                    grp_col and line[grp_col].value or '',
                    seq_col and line[seq_col].value or '',
                    self.ue)
            else:
                the_abjs = () # pragma: no cover
            da = student.da
            if not da:
                da = [
                    infos
                    for infos in user.da or ()
                    if first_sec < infos[1] < last_sec
                ]

            if tt or the_abjs or da:
                t.append("%s:[[%s],[%s],%s]" % (
                    js(login),
                    ','.join(['[%s,%s,%s]' % (js(a),js(b),
                             js(d if configuration.display_abj_comment else ''))
                              for a,b,dummy_c,d in the_abjs]),
                    ','.join(['[%s,%s,%s]' % (js(a),js(b),js(d))
                              for a,b,dummy_c,d in da]),
                    js(tt)))
                da_here = False
                if da:
                    for infos in da:
                        if infos[0] in self.ue:
                            da_here = True
                if inscrit and line[inscrit].value != 'non' and self.modifiable:
                    to_add = []
                    to_remove = []
                    (to_add if tt else to_remove).append('tierstemps')
                    (to_add if da else to_remove).append('dispense_assiduite')
                    (to_add if da_here else to_remove).append('DA')
                    self.lock()
                    try:
                        self.cell_set_change(self.pages[0],
                                             self.columns[inscrit].the_id, line_id,
                                             add=to_add, remove=to_remove)
                    finally:
                        self.unlock()

        self.new_abjs = 'change_abjs({%s});\n' % ',\n'.join(t)
        return self.new_abjs

    def update_the_abjs(self):
        if not self.contains_users:
            return
        old_abjs = self.new_abjs
        if self.the_abjs() != old_abjs:
            self.send_update(None,'<script>' + self.new_abjs + '</script>')

    def retrieve_student_list(self):
        """This function allows to call old code in order
        to upgrade TOMUSS without having to update customized code.
        """
        return retrieve_student_list(self.ue_code,
                                     self.year, self.semester, self)

    def allow_modification_of_system_columns(self, line_id, data_col):
        p = None
        line = self.lines[line_id]
        for i, column in enumerate(self.columns):
            if (line[i].value != ''
                and line[i].author == data.ro_user
                and i != data_col):
                if p == None:
                    p = self.get_nobody_page()
                self.cell_change(p, column.the_id, line_id, force_update=True)

    def is_unique(self):
        if self.template.i_am_unique is False:
            return False
        assert '-' not in self.ue
        assert self.template.i_am_unique == (self.year, self.semester)
        return True

    def unload_on_suivi_(self, action, clear):
        entity.TO_POP.write('{}\n'.format(
            (action, (self.year, self.semester, self.ue))))
        if clear:
            entity.TO_POP.write('{}\n'.format((' CLEAR ', clear)))
            if clear == 'ACLS':
                # The cache clear was done before the cell value was changed
                # so if the acls was checked too early, the cache
                # contains a bad value.
                # To be sure, clear it again.
                configuration.config_acls_clear_cache()
        entity.TO_POP.flush()

    def unload_on_suivi(self, action=' RELOAD ', clear=''):
        """Wait a second before asking the suivi servers to reload the table.
        Because when 'cell_change' calls this method, the new value is not
        yet in the file."""
        utilities.start_job(lambda: self.unload_on_suivi_(action, clear), 1)

def get_cell_from_table_ro(server, allowed_types=None):
    """server.the_path must starts by 'col_id/lin_id'
    Return an error string or the tuple (table, column, lin_id)
    """
    t = table(server.the_year, server.the_semester,
              server.the_ue, create=False)
    if not t:
        return "Can't find table"
    col_id = server.the_path[0]
    lin = server.the_path[1]
    column = t.columns.from_id(col_id)
    if not column:
        return "Can't find column"
    if lin not in t.lines:
        if lin == server.ticket.user_name and server.ticket.is_a_teacher:
            # For teacher annotate to place column names on PDF
            return t, column, None
        if lin != '_':
            return "Can't find line"
        lines = t.the_key_dict.get(server.ticket.user.ID_, ())
        if len(lines) != 1:
            return "Can't find line (none or duplicate)"
        lin = lines[0]
    if allowed_types  and  column.type.name not in allowed_types:
        return "Not an %s column type" % str(allowed_types)
    t.compute_columns((t.lines[lin],)) # visible and cell_writable may depend on averages
    if column.visibility != 3: # Not a public column
        if server.ticket.ticket == "none":
            return "No anonymous access allowed"
        if (column.visibility != 4
            and not server.ticket.is_a_teacher
            and lin not in t.the_keys()[server.ticket.user_name]
            and not server.table_access_allowed(str(t))
            ):
            return 'You are not allowed to read/modify this value'
        if not column.visible(True if server.ticket.is_a_teacher else 1,
                              t.lines[lin], server.ticket.user_name):
            return 'You are not allowed to read this value'
        if server.ticket.is_a_teacher and not column.readable_by(server.ticket.user_name):
            return 'This is private data'
    return t, column, lin

def get_cell_from_table(server, allowed_types=None):
    """server.the_path must starts by 'col_id/lin_id'
    Return an error string or the tuple (table, page, column, lin_id)
    Once the cell value is modified, call:
         table.do_not_unload_remove('cell_change')
    """
    err = get_cell_from_table_ro(server, allowed_types)
    if isinstance(err, str):
        return err
    t, column, lin = err
    if lin and not column.is_modifiable(server.ticket, t.lines[lin]):
        return server._("ERROR_value_not_modifiable")

    t, page = table(server.the_year, server.the_semester,
                    server.the_ue, None, server.ticket,
                    do_not_unload='cell_change', recycle=True)

    if t is None:
        return server._("MSG_bad_ticket") # pragma: no cover

    return t, page, column, lin

def retrieve_student_list(ue, year=None, semester=None, the_table=None):
    options = {}
    if year is None:
        year, semester = configuration.year_semester
    try:
        student_list = []
        done = set()
        if '__' in ue:
            # Merge students lists the '__' usage if deprecated and will be removed
            codes = ue.split('__') # pragma: no cover
        else:
            codes = [ue]
            if not the_table:
                the_table = table(year, semester, ue, create=False)
            if the_table:
                codes.extend(the_table.merge_students.split(' '))
        for code in codes:
            if not code:
                continue
            for line in inscrits.L_batch.students(code, year=year, semester=semester,
                                                  table=the_table): # NO_ENTITY
                if line[0] in done:
                    continue
                done.add(line[0])
                if (the_table and the_table.seqgrp_filter
                    and line[5] + '_' + line[4] != the_table.seqgrp_filter):
                    continue
                student_list.append(line)
        return student_list
    except: # pragma: no cover
        utilities.send_backtrace("", "Can't get student list for %s %s %s"
                                 % (year, semester, ue))
        return ()

def send_alert(text):
    for atable in tables_values():
        atable.send_alert(text)

@utilities.add_a_lock
def tables_manage(action, year, semester, ue, do_not_unload=0, new_table=None):
    if action == 'get':
        try:
            t = tables[year, semester, ue]
            if t and do_not_unload:
                t.do_not_unload_add(do_not_unload)
            return t
        except KeyError:
            tables[year, semester, ue] = None
            return False
    elif action == 'del':
        try:
            t = tables[year, semester, ue]
            if not t:
                # Yet destroyed
                return # pragma: no cover
            if t.do_not_unload:
                warn('Unload of do_not_unload: '
                     + str(t.ue) + repr(t.do_not_unload), what="warning")
                return False
            if t.active_pages: # pragma: no cover
                warn('Unload with active pages.', what="warning")
                return False
            # write access to the table will make an error.
            t.unloaded = True

            # A loop because extended table have multiple keys
            for k, v in list(tables.items()):
                if v is t:
                    del tables[k]
            return True
        except KeyError: # pragma: no cover
            # Yet destroyed
            return
    elif action == 'replace':
        old = tables.get((year, semester, ue), None)
        tables[year, semester, ue] = new_table
        return new_table
    else:
        raise ValueError("Unknown action '%s'" % action) # pragma: no cover

def tables_values():
    # Use list because the tables may change while creating the list.
    # Do not use a tuple because it must be sortable
    # XXX Need a lock ?
    return [t
            for k, t in tuple(tables.items())
            if t not in (None, False, True)
            and k == (t.year, t.semester, t.ue)
            ]

def table(year, semester, ue, page=None, ticket=None, ro=False, create=True,
          do_not_unload=0, recycle=False, update=True):
    """With a ticket : return table and page
    Without ticket : return only the table
    """
    # utilities.warn('%s/%s/%s page=%s ticket=%s ro=%s create=%s' % (
    #    year, semester, ue, page, ticket, ro, create))
    if configuration.read_only:
        ro = True
    year = int(year)
    t = tables_manage('get', year, semester, ue, do_not_unload)
    if t is None:
        # Somebody else is creating the table
        while t is None: # pragma: no cover
            time.sleep(0.1)
            t = tables_manage('get', year, semester, ue, do_not_unload)

    if t is False:
        # I must create the table
        # Only one thread can be here at the same time.
        try:
            filename = table_filename(year, semester, ue)
            if os.path.islink(filename):
                y, s, u = link_to(filename, year, semester)
                t = table(y, s, u, page=page, ticket=ticket, ro=ro,
                          create=create, do_not_unload=do_not_unload)
                if (y, s, u) in tables:
                    tables[year, semester, ue] = tables[y, s, u]
                else:
                    del tables[year, semester, ue] # pragma: no cover
                return t
            if ticket == None:
                if (not create
                    and not os.path.exists(table_filename(year, semester, ue))):
                    del tables[year, semester, ue]
                    return None
                t = Table(year, semester, ue, ro)
            else:
                t = Table(year, semester, ue, ro, user=ticket.user_name)
        except: # pragma: no cover
            # The table import failed, allow the new one to retry
            # tables[year, semester, ue] = False
            # raise
            utilities.send_backtrace("", "Can't load %s/%s/%s" % (
                    year, semester, ue))
            # tables[year, semester, ue] contains None
            # So no need to call table_manage
            del tables[year, semester, ue]
            if ticket:
                return None, None
            else:
                return None
        t.do_not_unload_add(do_not_unload) # Only on this case
        tables[year, semester, ue] = t
    else:
        # Table yet loaded
        if update:
            t = t.update()

    t.atime = time.time()

    if page is not None and not t.modifiable and page >= len(t.pages):
        # Bug raised by TOMUSS restart (page list is lost)
        # XXX : if 2 users are on the same unmodifiable table,
        # then there is 50% chance that the second one is
        # with the bad page and so the bad ticket.
        # It will be accused of hacking.
        page = None

    if page == None:
        if ticket == None:
            return t
        # It is allowed to change a cell on a private table
        if do_not_unload != 'cell_change' and not t.readable_by(ticket):
            try:
                t.do_not_unload_remove('new_page')
            except ValueError:
                pass
            return None, None
        # The new page may append work to check_students_in_tables
        # And the work is done without call to 'table'.
        # So possible simultaneous execution is possible.
        t.lock()
        try:
            page = t.new_page(ticket.ticket,
                              ticket.user_name,
                              ticket.user_ip,
                              ticket.user_browser,
                              time.strftime('%Y%m%d%H%M%S'),
                              recycle=recycle)
        finally:
            t.unlock()
        return t, page

    page = t.pages[page]
    page.check_identity(ticket.ticket, ticket.user_name,
                        ticket.user_ip, ticket.user_browser)
    return t, page


def remove_unused_tables():
    """Do not call from a thread"""
    now = time.time()
    tables = tables_values()
    if configuration.read_only:
        for atable in tables:
            if now - getattr(atable,"rtime",0) > configuration.unload_interval:
                atable.unload()
                yield
    else:
        for atable in tables:
            if atable.active_pages:
                continue # pragma: no cover
            if now - atable.atime < configuration.unload_interval:
                continue
            atable.unload() # pragma: no cover
            yield # pragma: no cover

# continuous update of students lists for all tables
def check_students_in_tables():
    t = time.time()
    for table in tables_values():
        yield
        if t - table.update_time > configuration.students_check_interval:
            update_students.append(table) # pragma: no cover

# continuous update of students lists
update_students = []

@utilities.add_a_lock # The lock is here for regression tests
def check_new_students_real():
    try:
        while update_students:
            for t in update_students:
                # Interactively used table are prioritary
                if t.active_pages:
                    update_students.remove(t)
                    break
            else:
                t = update_students.pop()
            t.update_time = time.time()
            t.do_not_unload_add('check_new_students_real')
            # XXX Not a good test, but a lock is overkill
            if t.unloaded:
                continue # pragma: no cover
            try:
                utilities.important_job_add('check_new_students_real')
                t.send_update(None, "<script>set_updating(1);</script>")
                utilities.bufferize_this_file(t.filename)
                warn('start update students of %s' % t.ue, what="table")

                try:
                    t.template.check(t)
                except UserWarning: # pragma: no cover
                    pass
                except: # pragma: no cover
                    utilities.send_backtrace('', 'Student list %s' % t)
                if not t.update_content:
                    if getattr(t, 'update_inscrits', True):
                        t.update_content = ('column', 'mail', 'abj')
                    else:
                        t.update_content = ('mail', 'abj')

                warn('done %s' % t.ue, what="table")
                if 'mail' in t.update_content:
                    logins_to_get = t.authors()
                    if t.contains_users:
                        logins_to_get |= set(t.logins())
                    users = configuration.Users(logins_to_get)
                    mails = {user.student_id: user.mail
                             for user in users
                             if user.mail
                            }
                    t.change_mails(mails)
                if t.modifiable:
                    if 'column' in t.update_content:
                        for a_column in t.columns:
                            if not a_column.locked:
                                a_column.type.update_all(t, a_column)
                if 'abj' in t.update_content:
                    t.update_the_abjs()
            finally:
                t.update_content = ()
                utilities.important_job_remove('check_new_students_real')
                t.do_not_unload_remove('check_new_students_real')
                utilities.bufferize_this_file(None)
                t.send_update(None, "<script>set_updating(0);</script>")
            
    except IndexError:
        if not configuration.regtest:
            # Real problem
            raise # pragma: no cover
        # XXX The regtest raise a bug, why ?

indexes_to_update = []

def update_index(login, action=None):
    """
    'action' may be:
        * None: return the index content
        * list: set the index content to the list
        * function: transform the current index content with the function and save it
    """
    if not login:
        return
    if utilities.safe(login) != login:
        return # pragma: no cover
    user = configuration.User(login)
    c = user.manage_key('index')
    if c:
        c = eval(c)
    else:
        c = []
    if action is None:
        return c
    c = action(c) # Update index content
    user.manage_key('index', content = utilities.stable_repr(c))

def indexes_to_update_append(the_year, the_semester, the_ue, old, new):
    tmp = TMP()
    tmp.year = the_year
    tmp.semester = the_semester
    tmp.ue = the_ue
    indexes_to_update.append((tmp, old, new))

def check_indexes_to_update():
    t = time.time()
    while indexes_to_update:
        tabl, old, new = indexes_to_update.pop(0)
        def remove(x):
            try:
                x.remove((tabl.year, tabl.semester, tabl.ue))
            except ValueError: # pragma: no cover
                pass
            return x
        def append(x):
            x.append((tabl.year, tabl.semester, tabl.ue))
            return x
        update_index(old, remove)
        update_index(new, append)
        t = time.time()
    return t

def update_indexes(the_year, the_semester, the_ue):
    tmp = TMP()
    tmp.year = the_year
    tmp.semester = the_semester
    tmp.ue = the_ue
    the_table = table(the_year, the_semester, the_ue)
    extension = (the_year, the_semester)
    if ((the_table.year, the_table.semester) != extension
        and extension not in the_table.extensions):
            the_table.extensions.append(extension)

    for login in the_table.logins_valid():
        indexes_to_update.append((tmp, '', utilities.the_login(login)))

def check_new_students():
    while True:
        if configuration.regtest_sync:
            time.sleep(0.01)
        else:
            time.sleep(1)

        check_indexes_to_update()
        check_new_students_real()

ue_to_delete = []

def remove_bookmarks():
    to_delete = set()
    while ue_to_delete:
        to_delete.add(ue_to_delete.pop())
        t = time.time()
    for filename in glob.glob(os.path.join(configuration.db, 'LOGINS',
                                           '*', '*', 'bookmarked')):
        login = filename.split(os.path.sep)[-2]
        user = configuration.User(login)
        bookmarked = set(eval(user.manage_key('bookmarked')))
        cleaned = bookmarked - to_delete
        if bookmarked != cleaned:
            user.manage_key('bookmarked', content=utilities.stable_repr(list(cleaned)))
    warn("Remove bookmarks done")
    return t

def login_list(page, name):
    # XXX Not very clean the : configuration.teachers[-1]
    # XXX copy paste in PLUGINS/login_list.py
    t = list(inscrits.L_slow.firstname_or_surname_to_logins(
        name.replace('_',' '),
        base=configuration.teachers[-1],
        attributes = [configuration.attr_login,
                      configuration.attr_surname,
                      configuration.attr_firstname,
                      'cn']
        ))
    t.sort(key = lambda x: x[1] + x[2])
    s = []
    if t:
        for lo, surname, firstname, cn in t:
            s.append('[' + utilities.js(lo)
                     + ',' + utilities.js(surname.upper())
                     + ',' + utilities.js(firstname.title())
                     + ',' + utilities.js(cn)
                     + ']')
    s = '<script>login_list(' + utilities.js(name) + \
        ',[' + ',\n'.join(s) + ']);</script>\n'
    sender.append(page.browser_file, s)
            
def process_request(page, action):
    warn(str(page) + ' ' + str(action))
    page.answer = 'bug.png'
    action_name = action[0]
    action = action[1:]
    if action_name.startswith('column_attr_'):
        page.answer = page.table.column_attr(page, action[0],
                                        action_name[12:], action[1])
        page.mtime = time.time()
    elif action_name.startswith('table_attr_'):
        page.answer = page.table.table_attr(page,
                                        action_name[11:], action[0])
        page.mtime = time.time()
    elif action_name == 'cell_change':
        col, lin, value = action
        page.answer = page.table.cell_change(page, col, lin, value)
        page.mtime = time.time()
    elif action_name == 'comment_change':
        col, lin, value = action
        page.answer = page.table.comment_change(page, col, lin, value)
        page.mtime = time.time()
    elif action_name == 'column_delete':
        col = action[0]
        page.answer = page.table.column_delete(page, col)
        page.mtime = time.time()
    elif action_name == 'login_list':
        login_list(page, utilities.safe(action[0]))
        page.answer = 'ok.png'
    elif action_name == 'update_content':
        if page.user_name in page.table.masters:
            for what in action:
                if what.startswith('merge_students='):
                    page.table.table_attr(page, 'merge_students', what.split('=')[1])
        page.table.update_content = action
        if page.table not in update_students:
            update_students.append(page.table)
        page.answer = 'ok.png'
    else:
        warn('BUG: %s %s' % (action_name, action), what="error")

request_list = []
def check_requests():
    def process():
        if timestamp != page.last_request:
            return b'["%d!=%d"]' % (timestamp, page.last_request) # pragma: no cover
        bugs = []
        bad = False
        with page.table.the_lock:
            for action in actions:
                try:
                    process_request(page, action)
                    if page.answer in ('bad.png', 'bug.png'):
                        bad = page.answer
                except:
                    bugs.append(repr(action) + '\n' + traceback.format_exc() + '\n')
        page.table.do_not_unload_remove('page_action')

        if bugs:
            warn('bug raised', what="error")
            utilities.send_backtrace('check_requests: %s\n' % page +'\n'.join(bugs))
            page.answer = "bug.png"
            return b'"bug.png"'
        if bad:
            return b'"%s"' % bad.encode('ascii')
        return b'["%d",%d]' % (timestamp, len(page.table.sent_to_browsers))

    while request_list:
        page, timestamp, actions, output_file = request_list.pop()
        answer = process()
        # XXX il faut logguer print(actions, answer)
        if answer:
            try:
                warn('Send %s(%s) %s %s' % (output_file, output_file.closed,
                                            page.answer, page.browser_file),
                    what="table")
                if not output_file.closed:
                    output_file.write(answer)
                    output_file.close()
            except socket.error: # pragma: no cover
                pass
        test = time.time()
    return test

# continuous send of packets to check connections
from . import ticket
def check_down_connections():
    while True:
        yield
        for ttable in tables_values(): # pragma: no cover
            if not ttable.active_pages:
                continue
            yield
            try:
                ttable.send_update(None, '<!-- connected -->\n',
                                   store=False)
            except: # pylint: disable=bare-except
                # Any problem will freeze TOMUSS if this loop is broken
                pass

# Update computed values

cell_changed_list_fast = []

def update_computed_values_fast():
    while True:
        time.sleep(0.1)
        while cell_changed_list_fast:
            the_table, lin, a_column = cell_changed_list_fast.pop()
            for col in the_table.columns.use(a_column):
                col.type.update_one(the_table, lin, col)
            # Update mail if new teacher
            login = the_table.lines[lin][a_column.data_col].author
            if login not in the_table.mails and len(login) > 1:
                m = configuration.User(login).mail
                if m:
                    the_table.update_mail(login, m)
            # Update mail if login changed
            if a_column.data_col == 0 and the_table.contains_users:
                login = the_table.lines[lin][0].value
                if isinstance(login, str) and login:
                    m = configuration.User(login).mail
                    if m:
                        the_table.update_mail(login, m)

column_changed_list = []

def update_computed_values_slow():
    while True:
        time.sleep(0.1)
        while column_changed_list:
            the_table, a_column, attr = column_changed_list.pop()
            a_column.type.update_all(the_table, a_column, attr)
            for col in the_table.columns.use(a_column):
                col.type.update_all(the_table, col)

class TMP:
    pass

def virtual_table(server, the_columns, the_lines, table_attrs={}, js="",css=""):
    """Send the table to the browser without storage.
    Do not use in a not threaded plugin.
    """
    if not hasattr(server, "year"):
        server.year = server.__dict__.get("the_year", 0)
    if not hasattr(server, "semester"):
        if hasattr(server, 'the_semester'):
            server.semester = server.the_semester # pragma: no cover
        else:
            server.semester = "?"
    for i, a_column in enumerate(the_columns):
        if not a_column.position:
            a_column.position = i
    tmp = TMP()
    tmp.__dict__ = table_attrs
    
    server.the_file.write(table_head(server.year,
                                     server.semester,
                                     server.ticket.ticket,
                                     create_pref = False,
                                     attrs_from=tmp,
                                     user_name=server.ticket.user_name
                                     )
                          )
    lines = []
    for i, line in enumerate(the_lines):
        lines.append('"%d": %s' % (i, line.js()))
    lines = '{' + ',\n'.join(lines) + '}'
    columns = '[' + ',\n'.join([the_column.js(hide=False)
                                for the_column in the_columns]) + ']'
    server.the_file.write("""
    <style id="template_style">%s</style>
    <script>
    %s
    %s
    var page_index = 0 ;
    function initialize()
    {
    if ( ! wait_scripts("initialize()") )
               return ;
    document.write(head_html()) ;
    insert_middle();
    columns = %s ;
    lines = %s ;
    document.write(tail_html()) ;
    table_attr.table_title = %s ;
    runlog(columns, lines) ;
    }
    initialize() ;
    </script>
    """ % (css, utilities.wait_scripts(),
           js, columns, lines, utilities.js(repr(server.the_path)) ))

    logins = (line[0].value
              for line in the_lines
              if line[0].value
             )
    mails = configuration.Users(logins).get_not_none('mail')
    server.the_file.write("<script>Xtable_attr('mails',%s);</script>\n"
                          % utilities.js(mails))
       
def start_threads():
    utilities.start_new_thread_immortal(check_new_students, ())
    utilities.start_new_thread_immortal(update_computed_values_fast, ())
    utilities.start_new_thread_immortal(update_computed_values_slow, ())

