#!/usr/bin/python3
"""
A framework to transparently cache and vectorize attribute computations.

The framework allows multiple unique identifiers for the same entity.

Internaly, there is only one canonical identifier.
For 'User' it is the value given by 'utilities.the_login'


The programmer creates each vector using the identifier of its choice,
but it is programmer's responsability to ensure coherence.
The following vector of users contains twice the same user:
    users = Users(("john.doe", "John.Doe", "super.user"))
Nevertheless, the framework will use internaly the canonical identifier,
so the 2 vector items will point to the same user.
'users["JOHN.DOE"] ' will raise an exception.

Usage:
    users = Users(("a", "b", "c"))
    for user in users:
        print(user.surname, user.firsname)
    print(users["b"].surname)
    print(User("b").surname)
    users["b"].pop('surname') # Force attribute clear on all the servers
    users["b"]['surname'] # Get the attribute but update its value on exceeded TTL

Cache management :
    * If attribute 'foo' is updated, then 'foo_update_ ← time'
    * If an instance is used, 'atime_' is updated.
    * Unused instances are removed from memory and stored in : TMP/ENTITY/ID_
    * On instance creation, disc cache is loaded without being updated
    * A thread walks over all instances to recomputed out of date attributes.

If a cache must be cleared immediatly, call «item.pop(attr)» on the server,
the suivi will also pop just before executing any request

key : a not canonized key
ID_ : a canonized key
attributes : a dictionnary of attributes
VectorItem : an attribute+Vector

The attributes are private, only Vector and VectorItem must be used.

BEWARE a weakref is used to break the cycle. So do not do this :
    users = Users(("a", "b", "c"))
    users = tuple(users) # Users instance is deleted because no more used

+----------------------+
| Vector               |
|======================|
|                      |
| class attribute      |
| Dict[ID_,dict]       |                        +------------------------+
| «instances»>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>| instance Dict[str,Any] |
|                      |                        |========================|
|----------------------|   +----------------+   | ID_                    |
|                      |   | VectorItem     |   | atime_                 |
| Dict|ID_,VectorItem] |   |================|   |                        |
| «proxies»>>>>>>>>>>>>>>>>|                |   | attrname               |
|                      |   | «attributes_»>>>>>>| attrname_update_       |
| Dict|key,VectorItem] |   |                |   +------------------------+
| «content»>>>>>>>>>>>>>>>>| «vector_»      |     ^
|                      |   |     V          |     ^
|                      |   +-----V----------+     ^
|                      |         V                ^
|                      |<<<<<<<<<V                ^
+----------------------+ WEAKREF                  ^
                                                  ^
                           +----------------+     ^
                           | VectorSingle   |     ^
VectorSingle use the same  |================|     ^
API than VectorItem.       |                |     ^
But there is not a Vector  | class attribute|     ^
instance associated.       | Vector class   |     ^
It contains a single item. | «vector_»      |     ^
                           |                |     ^
                           |----------------|     ^
                           |                |     ^
                           | «attributes_»>>>>>>>>^
                           |                |
                           +----------------+

"""

 # pylint: disable=no-self-use,wrong-import-order,global-statement

import tomuss_init # pylint: disable=unused-import
import time
import os
import sys
import io
import collections
import html
import re
import weakref
import PIL.Image
import TOMUSS
from . import configuration
from . import utilities
from . import document

###############################################################################
# FRAMEWORK
###############################################################################

def backtrace_html(attributes):
    """Affichage HTML d'un dictionnaire d'instance"""
    lines = []
    now = time.time()
    for key in sorted(attributes):
        if key.endswith('_update_'):
            continue
        update = key + '_update_'
        value = attributes[key]
        lines.append(
            '<tr><td{}>{}<td>{}<td>{}</tr>\n'.format(
                ' style="font-weight:bold"'
                if isinstance(value, (str, bytes, int, list, tuple, dict))
                else '',
                key,
                int(now - attributes[update]) if update in attributes else '',
                '{} bytes'.format(len(value)) if isinstance(value, bytes)
                else html.escape(str(value)))
        )
    return '<table border>{}</table>'.format(''.join(lines))


class VectorItem:
    """A proxy to translate a single attribute access
    to a vectorized attribute update.
    'vector_' is a weakref.
    """
    __slots__ = ("attributes_", "vector_")
    def __init__(self, attributes, vector):
        "The current attributes and the vector of Entity"
        super().__setattr__('attributes_', attributes)
        super().__setattr__('vector_', vector)

    def __getattr__(self, attr):
        if attr not in self.attributes_:
            self.vector_().get_attribute(attr) # Update the vector
        return self.attributes_.get(attr, None)

    def __setattr__(self, attr, value):
        self.attributes_[attr] = value
        self.attributes_[attr + '_update_'] = int(time.time())
        return value

    def pop(self, *attrs):
        """Remove some (or all) attributes (to force update)"""
        if not configuration.read_only and TO_POP:
            if isinstance(self, VectorSingle):
                self.vector_.save(self.attributes_)
            else:
                self.vector_().save(self.attributes_)
            TO_POP.write("{}\n".format((self.ID_, attrs)))
            TO_POP.flush()
        if not attrs:
            attrs = tuple(i
                          for i in self.attributes_
                          if not i.endswith('_'))
        for attr in attrs:
            self.attributes_.pop(attr, None)

    def __getitem__(self, attr):
        if self.vector_().attr_is_alive(self.attributes_, attr):
            return self.attributes_.get(attr, None)
        self.pop(attr, None)
        return getattr(self, attr)

    def backtrace_html(self):
        """For description of the user"""
        return backtrace_html(self.attributes_)

class Vector:
    """A set of items updated at the same time.

    Real data is shared between all the vectors.
    """
    # Data common to all vector instances.
    # Must be redefined in the subclass
    instances = {} # type: Dict[str, Dict[str, Any]]

    # Creator of vector items
    proxy = lambda x, y: None # type: Callable[[Dict[str, Any], Vector], VectorItem]

    # In seconds
    default_ttl = 30000

    def __init__(self, keys):
        """The identifier to be managed by the vector.
        They are not canonized: John.Doe and john.doe can be defined.
        They will reference the same user.
        """
        self.content = collections.OrderedDict()
        self.proxies = {}
        now = int(time.time())
        if False: # pylint: disable=using-constant-test
            keys = tuple(keys)
            if len(keys) == 1:
                print(keys)
                import traceback  # pylint: disable=import-outside-toplevel
                import threading  # pylint: disable=import-outside-toplevel
                frames = sys._current_frames()
                print(''.join(traceback.format_stack(frames[threading.current_thread().ident])))
        self_weak = weakref.ref(self)
        for key in keys:
            self.add(key, now, self_weak)

    def __iter__(self):
        yield from self.content.values()

    def items(self):
        "On the content"
        yield from self.content.items()

    def values(self):
        "On the content"
        yield from self.content.values()

    @classmethod
    def is_an_identifier(cls, identifier):
        """Attributes are not computed if it is not an identifier"""
        return 2 < len(identifier) < 100 and bool(re.fullmatch('[-_. @a-z0-9]+', identifier))

    @classmethod
    def canonize(cls, identifier):
        """Clean the identifier to get an unique one.
        A bad identifier must not be translated in a valid identifier.
        For example «john's doe» must not be translated has «john_s_doe».
        """
        return str(identifier).lower() # pragma: no cover

    def __len__(self):
        "On the content"
        return len(self.content)

    def __getitem__(self, key):
        return self.content[key]

    def __contains__(self, key):
        return key in self.content

    @classmethod
    def get_attributes(cls, identifier, now):
        """Get the attribute dict for the identifier, update access time"""
        attrs = cls.instances.get(identifier, None)
        if attrs is None:
            try:
                old = utilities.read_file(os.path.join('TMP', 'ENTITY', identifier[:3], identifier))
                attrs = utilities.literal_eval(old)
            except (FileNotFoundError, IsADirectoryError, OSError):
                attrs = {"ID_": identifier, "valid_ID_": cls.is_an_identifier(identifier)}
            cls.instances[identifier] = attrs
        attrs["atime_"] = now
        return attrs

    def add(self, key, now, self_weak):
        """Add a new item in the vector.
        Create a new instance if it does not exists.
        """
        identifier = self.canonize(key)
        if key in self.content:
            # raise ValueError("Duplicate key: " + key)
            return
        self.proxies[identifier] = self.content[key] = self.proxy(
            self.get_attributes(identifier, now), self_weak)

    def get_attribute(self, attr, items=None):
        """Recompute only the needed attributes."""
        if items is None:
            items = (i
                     for i in self.proxies.values() # 'content' contains duplicates
                     if attr not in i.attributes_
                    )
        to_get = []
        for i in items:
            if i.attributes_['valid_ID_']:
                to_get.append(i)
                # Check if the User uses an out of date 'attributes_'
                if i.attributes_ is not self.instances[i.attributes_['ID_']]: # pragma: no cover
                    message = ['='*77, '\n', repr(i.attributes_), '\n\n',
                               repr(self.instances[i.attributes_['ID_']])]
                    utilities.send_backtrace(''.join(message),
                                             'Out of date attribute «{}»'.format(attr))
            else:
                setattr(i, attr, None)
        if to_get:
            if ' ' in attr:
                args = attr.split(' ')
                getattr(self.__class__, args[0])(to_get, args)
            else:
                getattr(self.__class__, attr)(to_get)

    def get(self, *attributes, default=None) -> "Dict[str, Any]":
        """If a single attribute is asked, it returns a simple dict.
        If more than one: it returns a list of attribute value
        """
        if len(attributes) == 1:
            attr = attributes[0]
            mapping = {k: getattr(i, attr, default)
                       for k, i in self.content.items()
                      }
        else:
            mapping = {k: [getattr(i, attr, default)  for attr in attributes]
                       for k, i in self.content.items()
                      }
        return mapping

    def get_not_none(self, attr, default=None) -> "Dict[str, Any]":
        """Do not return items with a None attribute"""
        return {k:v for k, v in self.get(attr, default=default).items() if v}

    @classmethod
    def save(cls, attributes):
        """Save the attributes"""
        key = attributes.get('ID_', None)
        if attributes.get('valid_ID_', False):
            attrs = {
                attr: value
                for attr, value in attributes.items()
                if isinstance(value, (str, bytes, int, float, list, tuple, dict))
                }
            dirname = os.path.join('TMP', 'ENTITY', key[:3])
            filename = os.path.join(dirname, key)
            if not os.path.exists(dirname):
                os.mkdir(dirname)
            elif os.path.exists(filename):
                old = utilities.literal_eval(utilities.read_file(filename))
                for attr, value in old.items():
                    attr_update = attr + '_update_'
                    if old.get(attr_update, 0) > attributes.get(attr_update, 0):
                        # The attribute on disk is newer
                        attributes[attr] = attrs[attr] = old[attr]
                        attributes[attr_update] = attrs[attr_update] = old[attr_update]
            utilities.write_file(filename, repr(attrs))
        return key

    @classmethod
    def cleanup(cls, unused=3600):
        """Instance used more than 'unused' seconds before are unloaded
        """
        unused = int(time.time()) - unused
        for attributes in utilities.failsafe_tuple(cls.instances.values): # pylint: disable=too-many-nested-blocks
            if attributes["atime_"] < unused:
                cls.instances.pop(cls.save(attributes), None)
                yield

    @classmethod
    def cleanup_pop(cls):
        """Remove attributes poped on the tomuss process"""
        global TO_POP
        done = set()
        for line in TO_POP:
            if line in done:
                continue # Do not do twice the same thing
            done.add(line)
            id_, attrs = utilities.literal_eval(line.strip())
            if id_ == ' REOPEN ':
                TO_POP = open('xxx.pop', 'r')
                cls.cleanup_pop()
                return
            if id_ == ' VARIABLE ':
                variables = document.table(*attrs, ro=True)
                variables.unload(force=True)
                try:
                    for variable in variables.variables:
                        variable._clean_() # pylint: disable=protected-access
                except AttributeError: # pragma: no cover
                    pass
                continue # pragma: no cover # pylint bug
            if id_ == ' RELOAD ':
                document.table(*attrs, ro=True, update=False).unload(force=True)
                document.table(*attrs, ro=True)
                continue
            if id_ == ' CLEAR ':
                if attrs == 'ACLS':
                    configuration.config_acls_clear_cache() # pylint: disable=no-member
                continue

            item = cls.instances.get(id_, None)
            if item:
                for attr in attrs:
                    item.pop(attr, None)

    @classmethod
    def attr_is_alive(cls, attributes, attr, now=None):
        """Return True if the TTL is not exceeded"""
        last_update = attributes.get(attr + '_update_', 0)
        method = getattr(cls, attr.split(' ', 1)[0])
        if now is None:
            now = time.time()
        return last_update > now - getattr(method, 'ttl', cls.default_ttl)

    @classmethod
    def update(cls):
        """Update any attribute too old to be valid.
        The old value is not erased, so if the new one can't be computed its ok.
        """
        # Done by attributes because updating one may update others.
        # Bug: may update a just unloaded item
        items = utilities.failsafe_tuple(cls.instances.items)
        for attr in cls.get_attributes_name():
            if configuration.read_only and attr in configuration.attr_to_not_update_on_suivi:
                continue # pragma: no cover
            now = time.time()
            to_update = []
            for key, attributes in items:
                if attr in attributes and not cls.attr_is_alive(attributes, attr, now):
                    to_update.append(key)
            vector = cls(to_update)
            vector.get_attribute(attr, vector.proxies.values())

    @classmethod
    def add_attribute(cls, fct):
        """Add or replace an item attribute"""
        setattr(cls, fct.__name__, fct)

    @classmethod
    def get_attributes_name(cls):
        """Returns the computed attributes of the class"""
        attrs = []
        for attr in sorted(cls.__dict__, key=lambda x: cls.__dict__[x].__doc__ or ''):
            fct = cls.__dict__[attr]
            code = getattr(fct, '__code__', None)
            if not code:
                if not hasattr(fct, '__func__'):
                    continue
                code = fct.__func__.__code__
            if 'items' not in code.co_varnames:
                continue
            if 'args' in code.co_varnames and getattr(fct, '__defaults__', None) != (None,):
                continue
            attrs.append(attr)

        # This takes a lot of time
        attrs2 = set(attrs)
        for attributes in utilities.failsafe_tuple(cls.instances.values):
            attrs2.update(attr
                          for attr in attributes
                          if ' ' in attr and not attr.endswith('_')
                         )

        return list(attrs2)

class VectorSingle(VectorItem):
    """A vector of only one item"""
    __slots__ = ("attributes_",)
    vector_ = Vector
    def __init__(self, key): # pylint: disable=super-init-not-called
        identifier = self.vector_.canonize(key)
        object.__setattr__(self, 'attributes_',
                           self.vector_.get_attributes(identifier, int(time.time())))

    def __getattr__(self, attr):
        if attr not in self.attributes_ and self.valid_ID_:
            if ' ' in attr:
                args = attr.split(' ')
                getattr(self.vector_, args[0])((self,), args)
            else:
                getattr(self.vector_, attr)((self,))
        return self.attributes_.get(attr, None)

    def __getitem__(self, attr):
        if self.vector_.attr_is_alive(self.attributes_, attr):
            return self.attributes_.get(attr, None)
        self.pop(attr, None)
        return getattr(self, attr)

    def forget(self):
        """Delete the item because it will never be reused"""
        self.vector_.instances.pop(self.ID_)

###############################################################################
# USERS
###############################################################################

class UsersItem(VectorItem):
    """Item for user"""
    def is_in_one_of_the_groups(self, groups):
        """Returns true if the login is one of the groups or sub group"""
        member_of_list = self.member_of_list
        for group in groups:
            for a_group in member_of_list:
                if a_group.endswith(group):
                    return True
        return False

    def get_etapes_list(self, year=None):
        """Get the etape list for the indicated year"""
        if year is None:
            return self.etapes_list
        return self['etapes_list {}'.format(year)] # Check date

    def referent_old(self, year, semester):
        """Get the referent for the indicated semester"""
        return getattr(self, 'referent_old {} {}'.format(year, semester))

    def is_member_of(self, groups):
        """ACLS: Return True if member of one of the groups"""
        return configuration.is_member_of(self.ID_, groups) # pylint: disable=no-member

    def manage_key(self, name, *args, **kargs):
        """Abstract manakey access"""
        return utilities.manage_key('LOGINS',
                                    os.path.join(self.ID_, name),
                                    *args, **kargs)

    def a_mailto(self):
        """Return an HTML link to the mail"""
        return '<a href="mailto:%s">%s %s</a><!-- %s -->' % (
            self.mail, (self.firstname or '').title(), self.surname, self.ID_)

    def __str__(self):
        return self.backtrace_html()

def query_logins(logins, *args, **kargs):
    """LDAP: Generic query function"""
    if len(logins) < 2:
        ldap = TOMUSS.inscrits.L_fast
    else:
        ldap = TOMUSS.inscrits.L_slow # pragma: no cover
    return ldap.query_logins(logins, *args, **kargs)

def get_one_attribute(the_items, ldap_attr, user_attr, default=None):
    """LDAP: Get a single attribute"""
    todo = {i.ID_: i
            for i in the_items}
    if default != ():
        for login, value in query_logins(
                tuple(todo), (configuration.attr_login, ldap_attr),
        ):
            setattr(todo.pop(login.lower()), user_attr, value)
    else:
        for data in query_logins(
                tuple(todo), (configuration.attr_login, ldap_attr),
                only_first_value=False
        ):
            login = data[configuration.attr_login][0].lower()
            setattr(todo.pop(login), user_attr, data.get(ldap_attr, default))
    for item in todo.values():
        if user_attr not in item.attributes_:
            setattr(item, user_attr, default)

class Users(Vector): # pylint: disable=too-many-public-methods
    """This subclass computes attributes of UserItems"""
    instances = {}
    proxy = UsersItem

    @classmethod
    def canonize(cls, identifier):
        """Canonized version of the user identifier"""
        return utilities.the_login(identifier)

    @staticmethod
    def student_id(items):
        """TOMUSS.inscrits.login_to_student_id():
        If the student login in LDAP is not the same as the student ID.
        This function translate student login to student ID.
        The returned value must be usable safely."""
        for i in items:
            i.student_id = TOMUSS.inscrits.login_to_student_id(i.ID_)

    @staticmethod
    def firstname(items):
        """LDAP: Get surname and firstname"""
        todo = {i.ID_: i for i in items}

        for login, firstname, surname, mail in query_logins(
                tuple(todo),
                (configuration.attr_login,
                 configuration.attr_firstname,
                 configuration.attr_surname,
                 configuration.attr_mail,
                )):
            login = login.lower()
            user = todo[login]
            user.firstname = firstname
            user.surname = surname
            user.mail = mail
            todo.pop(login)
        for item in todo.values():
            item.mail = None
            item.firstname = None
            item.surname = None

    surname = firstname # Exchangeable attributes
    mail = firstname  # Exchangeable attributes

    @staticmethod
    def member_of_list(items):
        """LDAP: Get member_of list"""
        get_one_attribute(items, 'memberOf', 'member_of_list', ())

    @staticmethod
    def default_password(items):
        """LDAP: Get the default password"""
        get_one_attribute(items, configuration.attr_default_password,
                          'default_password')

    @staticmethod
    def password_ok(items):
        """TOMUSS.utilities.stupid_password():
        True if the password is not trivial. Cached only for True"""
        for i in items:
            passwords = [i.ID_] + i.ID_.split('.')
            default_password = i.default_password
            if default_password:
                passwords.append(default_password)
            # Do not store if not ok: can be changed by the user
            if not utilities.stupid_password(i.ID_, passwords):
                i.password_ok = True

    @staticmethod
    def phone(items):
        """LDAP: Get the phone"""
        get_one_attribute(items, configuration.attr_phone, 'phone')

    @staticmethod
    def etapes_list(items, args=None):
        """LDAP: Get etapes, assuming 'OU' with ' etape-XXXX' inside"""
        if args:
            attr = 'etapes_list {}'.format(args[1])
        else:
            attr = 'etapes_list'
        for i in items:
            member_of = i.member_of_list
            if member_of:
                setattr(i, attr, sorted(aa.split(' etape-')[1].split(',')[0]
                                        for aa in member_of
                                        if ' etape-' in aa))
            else:
                setattr(i, attr, ())
    @staticmethod
    def ues_long(items): # pragma: no cover
        """LDAP: List of UEs where the student is registered."""
        for i in items:
            i.ues_long = [aa
                          for aa in i.member_of_list
                          if (configuration.ou_ue_contains in aa)
                          and (aa.startswith(configuration.ou_ue_starts)
                               or aa.startswith(configuration.ou_ue_starts2))
                         ]
    @staticmethod
    def ues_short(items): # pragma: no cover
        """LDAP: Extract from the OU the UE short name (code). Uses 'ues_long'"""
        for i in items:
            i.ues_short = sorted(x.replace(' ', ',').split(',')[0].split('=')[1]
                                 for x in i.ues_long
                                )
    @staticmethod
    def portails(items):
        "LDAP: Portails of students"
        for i in items:
            i.portails = [configuration.is_a_portail(aa)
                          for aa in i.member_of_list or ()
                          if configuration.is_a_portail(aa)
                         ]
    @staticmethod
    def referent_teacher(items):
        """TOMUSS.referent.referents_students():
        Teacher object with all the data about a referent teacher"""
        table = TOMUSS.referent.referents_students()
        for i in items:
            i.referent_teacher = None
            for line_key, line in table.get_items(i.ID_):
                if i.referent_teacher:
                    # Should never happen (duplicate teacher)
                    for student in line[2:]:
                        i.referent_teacher.append(User(student.value)) # XXX Must not be User
                else:
                    i.referent_teacher = TOMUSS.referent.Teacher(i, line_key, table)

    @staticmethod
    def ues_with_groups(items): # pragma: no cover
        """LDAP: List of UEs of the student as triplets (ue, group, sequence)."""
        for i in items:
            ues = set()
            result = []
            # Get UE with groups
            for group in i.member_of_list:
                if 'GRP-GRP' in group:
                    group = group.split(',')[0].split('-')
                    ue_code = group[-2]
                    grp_seq = group[-1].split('_')
                    try:
                        result.append((ue_code, grp_seq[0], grp_seq[1][1:]))
                    except IndexError:
                        result.append((ue_code, grp_seq[0], ''))
                    ues.add(ue_code)
            # Add UE without groups
            for group in i.member_of_list:
                if (group.startswith(configuration.ou_ue_starts)
                        or group.startswith(configuration.ou_ue_starts2)
                   ):
                    ue_code = group.split(' ')[0][6:]
                    if ue_code not in ues: # Do not erase existing group
                        result.append((ue_code, '', ''))
            i.ues_with_groups = result

    @staticmethod
    def in_first_year(items):
        """TOMUSS.configuration.student_in_first_year():
        True if the student is registered for the first time"""
        for i in items:
            i.in_first_year = configuration.student_in_first_year(
                i.student_id) or False

    @staticmethod
    def current_tt(items):
        """TOMUSS.document.table(tt): The current examination exceptions"""
        table = TOMUSS.document.table(utilities.university_year(), 'Dossiers', 'tt')
        current_tt = table.the_current_tt(table)
        for i in items:
            i.current_tt = i.student_id in current_tt and current_tt[i.student_id].current()

    @staticmethod
    def home_class(items):
        """TOMUSS.configuration.student_class():
        The display classes of the students on the homepage for referent"""
        for i in items:
            i.home_class = configuration.student_class(i)

    @staticmethod
    def inscrit_value(items):
        """TOMUSS.configuration.student_inscrit_value():
        The content to add to the Inscrit column in tables"""
        for i in items:
            i.inscrit_value = configuration.student_inscrit_value(i)

    @staticmethod
    def referent(items):
        "TOMUSS.referent.referents_students(): The student's teacher referent login"
        id_to_referent = TOMUSS.referent.referents_students(
            *configuration.year_semester).id_to_referent
        for i in items:
            i.referent = id_to_referent.get(i.student_id, None)

    @staticmethod
    def referent_old(items, args):
        "TOMUSS.referent.referents_students(): The student's previous teacher referent login"
        _, year, semester = args
        id_to_referent = TOMUSS.referent.referents_students(
                        year, semester).id_to_referent
        attr = "referent_old {} {}".format(year, semester)
        for i in items:
            setattr(i, attr, id_to_referent.get(i.student_id, None))

    @staticmethod
    def need_a_charte(items):
        """TOMUSS.referent.need_a_charte():
        'True' if the student must sign an interactive contract document."""
        for i in items:
            i.need_a_charte = TOMUSS.referent.need_a_charte(i.student_id)

    @staticmethod
    def need_a_referent(items):
        "TOMUSS.referent.need_a_referent(): 'True' if the student needs a referent teacher"
        for i in items:
            i.need_a_referent = TOMUSS.referent.need_a_referent(i.ID_)

    @staticmethod
    def preferences(items):
        """TOMUSS.utilities.manage_key('preferences'): The user preferences"""
        for i in items:
            prefs = i.manage_key('preferences')
            if prefs:
                i.preferences = utilities.literal_eval(prefs)
            else:
                i.preferences = {}

    # Not clean, but does not break old scripts
    # This function computes for the student some attributes needed to
    # assign it a referent teacher.
    @staticmethod
    def licence(items):
        """TOMUSS.referent.analyse_groups(): Misc informations for referents affectation"""
        for i in items:
            i.key = i.ID_
            if i.portails:
                i.portail = next(iter(i.portails)) # pragma: no cover
            else:
                i.portail = None
            i.ues = []
            i.discipline = ()
            i.licence = False
            i.licence_first_year = False
            TOMUSS.referent.analyse_groups(i, i.member_of_list)
    licence_first_year = licence
    discipline = licence
    key = licence
    ues = licence
    portail = licence

    def get_etapes_list(self, year=None):
        """The registration etapes for a given year"""
        return {k: i.get_etapes_list(year)
                for k, i in self.content.items()
               }

    @staticmethod
    def is_a_student(items):
        """configuration.is_a_student() : True if is a student"""
        for i in items:
            i.is_a_student = configuration.is_a_student(i.ID_)

    # Read/compute the student picture from a database.
    # Store it in the filename indicated if it needs to be cached.
    # Returns the filename containing the picture to display
    @staticmethod
    def picture(items):
        """TOMUSS.configuration.get_the_picture(): Return the JPG data"""
        for i in items: # pragma: no cover
            data = configuration.get_the_picture_data(i) # pylint: disable=no-member
            if len(data) > 14000:
                # Downsize the image if it is possible
                try:
                    img = PIL.Image.open(io.BytesIO(data))
                    width = img.size[0]
                    if width:
                        height = img.size[1]
                        img = img.resize((
                            4 * configuration.icon_picture_width,
                            height * (4 * configuration.icon_picture_width) // width
                            ))
                        icon_content = io.BytesIO()
                        img.save(icon_content, format='JPEG')
                        data = icon_content.getbuffer().tobytes()
                    else:
                        data = utilities.read_file(os.path.join('FILES', 'tip.png'), 'bytes')
                except: # pylint: disable=bare-except
                    data = utilities.read_file(os.path.join('FILES', 'tip.png'), 'bytes')
            i.picture = data

    @staticmethod
    def picture_icon(items):
        """TOMUSS.configuration.get_the_picture(): Return the JPG data for the student icon"""
        for i in items:
            if len(i.picture) < 200:
                i.picture_icon = i.picture
                continue
            try:
                img = PIL.Image.open(io.BytesIO(i.picture))
                img = img.resize(TOMUSS.PLUGINS.picture.find_size(
                    img, configuration.icon_picture_width))
                icon_content = io.BytesIO()
                img.save(icon_content, format='JPEG')
                i.picture_icon = icon_content.getbuffer().tobytes()
            except: # pylint: disable=bare-except # pragma: no cover
                i.picture_icon = utilities.read_file(os.path.join('FILES', 'bug.png'), 'bytes')

    @staticmethod
    def status(items, args=None):
        """The status of the student.
        This function always returns '?', you must redefine it (but not here...)"""
        if args:
            year = args[1]
            attr = 'status {}'.format(year)
        else: # pragma: no cover
            year = utilities.university_year()
            attr = 'status'
        for item in items:
            setattr(item, attr, '?')

    @staticmethod
    def civilite(items):
        """Set M. to MME (redefine DisplayCivilite if needed)"""
        for item in items:
            item.civilite = '?'

    @staticmethod
    def da(items):
        """Return a list of assiduity exemptations from an external source,
           the item of the list are tuples:
            (UECode, start(seconds1970), author, comment)
        """
        for item in items:
            item.da = ()

class User(UsersItem, VectorSingle):
    """A vector of a single User"""
    vector_ = Users

###############################################################################
# TESTS
###############################################################################

def test(): # pylint: disable=too-many-statements
    """Regression tests"""
    def new_query_logins(logins, *args, **kargs):
        """Specials"""
        attr_logins = []
        for login in logins:
            if login == 'forgetme':
                continue
            the_attr = {}
            for attr in args[0]:
                if attr.lower() == 'samaccountname':
                    value = [login]
                elif attr.lower() == 'mail':
                    if login == 'tata':
                        value = [None]
                    else:
                        value = [login + "@x.com"]
                elif attr.lower() == 'memberof':
                    if login == 'titi':
                        value = ()
                    else:
                        value = [login + "1", login + "2", 'CN=42 etape-' + login]
                else:
                    value = [login + attr]
                the_attr[attr] = value
            attr_logins.append(the_attr)
        if kargs.get('only_first_value', True):
            return [[v[0] for v in x.values()] for x in attr_logins]
        return attr_logins
    global query_logins # pylint: disable=global-variable-undefined,invalid-name
    query_logins = new_query_logins
    del Users.current_tt
    del Users.referent_teacher
    del Users.referent
    from TOMUSS.PLUGINS import picture # pylint: disable=unused-import,import-outside-toplevel
    global TO_POP
    TO_POP = sys.stderr

    for user in ('toto', 'titi'):
        cache = os.path.join('TMP', 'ENTITY', user[:3], user)
        if os.path.exists(cache):
            os.unlink(cache) # pragma: no cover

    users1 = Users(('toto',))
    assert users1['toto'].firstname == 'totogivenName'
    assert tuple(users1) == (users1['toto'],)
    assert users1.get('surname') == {'toto': 'totosn'}
    assert users1.get('surname', 'mail') == {'toto': ['totosn', 'toto@x.com']}

    # Change underlying data structure
    users1['toto'].mail = 'toto@toto'
    assert users1['toto'].mail == 'toto@toto'
    assert User('toto').mail == 'toto@toto'

    # Do not recompute computed things
    users2 = Users(('toto', 'titi'))
    assert users1['toto'].mail == 'toto@toto'
    assert users2['toto'].mail == 'toto@toto'
    assert users2.get('mail', 'firstname') == {"toto": ['toto@toto', 'totogivenName'],
                                               "titi": ['titi@x.com', 'titigivenName']}
    # Cleanup
    utilities.mkpath('TMP/ENTITY/tot')
    utilities.write_file(
        'TMP/ENTITY/tot/toto',
        '{"firstname": "TOTO", "firstname_update_":9999999999, "mail":"", "mail_update_":1}'
        )
    for _ in Users.cleanup(unused=-1):
        pass
    # Values used when a cleanup is done are kept.
    assert users2.get('mail', 'firstname') == {"toto": ['toto@toto', 'TOTO'],
                                               "titi": ['titi@x.com', 'titigivenName']}
    users3 = Users(('toto', 'titi'))
    assert users3['toto'].portails == []

    Users.update()
    assert users3.get('mail', 'firstname') == {"toto": ['toto@toto', 'TOTO'],
                                               "titi": ['titi@x.com', 'titigivenName']}
    Users.mail.ttl = 0
    Users.update()
    assert users3.get('mail', 'firstname') == {"toto": ['toto@x.com', 'totogivenName'],
                                               "titi": ['titi@x.com', 'titigivenName']}

    users3['toto'].mail = 'badmail'
    assert users3['toto'].mail == 'badmail'
    assert users3['toto']['mail'] == 'toto@x.com' # TTL check dead
    assert users3['toto']['portails'] == [] # TTL check alive

    assert users3.get('portail', default="XYZ") == {'toto': 'XYZ', 'titi': 'XYZ'}
    assert User('toto').is_in_one_of_the_groups(['toto1'])
    assert not User('toto').is_in_one_of_the_groups(['h1', 'h2'])
    assert "mailto:toto@toto" in users1['toto'].a_mailto()

    users4 = Users(('forgetme', 'tata'))
    assert users4['forgetme'].mail is None
    assert users4['forgetme'].phone is None
    assert users4.get('mail') == {'forgetme': None, 'tata': None}
    assert users4.get('mail', default='XYZ') == {'forgetme': None, 'tata': None}
    assert users4.get('portail', default='XYZ') == {'forgetme': 'XYZ', 'tata': 'XYZ'}
    assert users4.get('picture', default='XYZ') == {'forgetme': b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x02\x00\x00\x00\x90wS\xde\x00\x00\x00\x0cIDATx\x9ccx\xc1\x7f\x11\x00\x03\xab\x01\xc9\x92\xc1\xccG\x00\x00\x00\x00IEND\xaeB`\x82', 'tata': b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x02\x00\x00\x00\x90wS\xde\x00\x00\x00\x0cIDATx\x9ccx\xc1\x7f\x11\x00\x03\xab\x01\xc9\x92\xc1\xccG\x00\x00\x00\x00IEND\xaeB`\x82'} # pylint: disable=line-too-long
    assert not users1.is_an_identifier('9@=')
    print(str(users1['toto']))

    Users.mail.ttl = 1
    utilities.write_file(
        'TMP/ENTITY/tot/toti',
        '{"ID_": "toti","mail":"bad","valid_ID_":True,"mail_update_":1}')
    assert User('toti').mail == 'bad'
    assert User('toti')['mail'] == 'toti@x.com'
    assert User('toti')['mail'] == 'toti@x.com'# for coverage

    assert getattr(User('toti'), 'etapes_list 2000 Automne') is None

    def number(items):
        """Assign a number from 42"""
        for i, item in enumerate(items):
            item.number = 42 + i
    Users.add_attribute(number)
    for i, user in enumerate(users3):
        assert user.number == 42 + i

    assert users3['toto'].etapes_list == ['toto']
    assert users3['toto'].get_etapes_list() == ['toto']
    assert users3['toto'].get_etapes_list(2021) == ['toto']
    assert len(users3['titi'].etapes_list) == 0

    Users.update()

    assert User('*invalid_ID*').mail is None
    users4 = Users(('toto', '*invalid_ID*'))
    assert users4['toto'].mail == 'toto@x.com'
    assert users4['*invalid_ID*'].mail is None

    print('ok')

TO_POP = None

def init(read):
    """Open the communication ways between tomuss and 'suivi' servers.
    The tomuss server modifies files and asks all the suivi servers to pop values.
    """
    if not ('tomuss.py' in sys.argv[0] or 'suivi.py' in sys.argv[0]):
        return
    if 'recompute_the_ue_list' in sys.argv:
        return

    global TO_POP
    if read:
        # Suivi server
        while not TO_POP:
            try:
                TO_POP = open("xxx.pop", "r")
            except FileNotFoundError: # pragma: no cover
                print('Wait TOMUSS start', file=sys.stderr)
                time.sleep(1)
    else:
        # TOMUSS server
        if os.path.exists("xxx.pop"):
            old_pop = open("xxx.pop", "a")
            os.unlink("xxx.pop")
        else:
            old_pop = None # Random coverage # pragma: no cover
        TO_POP = open("xxx.pop", "w")
        if old_pop:
            old_pop.write("(' REOPEN ', ())\n")
            old_pop.close()

    utilities.mkpath(os.path.join("TMP", "ENTITY"))

    def update():
        while True:
            time.sleep(600)
            Users.update() # pragma: no cover
    utilities.start_new_thread_immortal(update, ())

if __name__ == "__main__":
    test()
