#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet)
#    Copyright (C) 2008-2016 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@univ-lyon1.fr

import ldap
import ssl
import re
import time
import threading
import gc
import tomuss_init
from . import configuration
from . import utilities
from . import sender

time_limit = 5 # seconds

safe_re = re.compile('[^0-9a-zA-Z-. _@]')
def safe(txt):
    """Values safe in an LDAP request"""
    if not isinstance(txt, str):
        return '' # pragma: no cover
    return safe_re.sub('_', txt)

class LDAP_Logic(object):
    search = None

    def get_ldap_ues(self): # pragma: no cover
        """Full list of UE"""
        dues = {}
        ues = self.query('(|(%s*)(%s*))' % (configuration.ou_ue_starts,
                                            configuration.ou_ue_starts2),
                         base=configuration.ou_groups)
        for i in [i[0] for i in ues]:
            dues[i[3:].split(' ')[0]] = True

        dues = list(dues.keys())
        dues.sort()
        return dues

    def firstname_and_surname_to_login(self, firstname, surname): # pragma: no cover
        firstname = safe(utilities.flat(firstname))
        surname = safe(utilities.flat(surname))
        aa = self.query(
            search='(|(&('+configuration.attr_surname +'='
            + surname.replace('-','*')
            + ')(' +configuration.attr_firstname + '='
            + firstname.replace('-','*')
            + '))(' + configuration.attr_login + '=' +
            firstname.replace(' ','-') + '.' + surname.replace(' ','-') +
            '))',
            attributes = (configuration.attr_login, configuration.attr_mail,
                          'lastLogonTimestamp',),
            base=configuration.ou_top)
        a = []
        for x in aa: # For all the answers
            if x[0] == None:
                continue
            if configuration.attr_login not in x[1]:
                continue
            i = 0
            n = x[1][configuration.attr_login][0]
            if '.' in n:
                i += 10000
            if not n[-1].isdigit():
                i += 10001
            if ' ' not in n:
                i += 10010
            if 'lastLogonTimestamp' in x[1]:
                i += 10100
            if configuration.attr_mail in x[1]:
                i += 11000
            a.append( (i,n) )
        a.sort()
        if len(a):
            return a[-1][1].lower()

        aa = self.query_login(surname.replace(' ','-') + '.'
                             + firstname.replace(' ','-'))
        if aa:
            return aa[configuration.attr_login][0].lower()
        return None

    @utilities.add_a_method_cache
    def firstname_surname_to_login(self, firstname_surname): # pragma: no cover
        """Give a login from a firstname and a surname"""
        firstname_surname = firstname_surname.split(' ')
        for i in range(1, len(firstname_surname)):
            login = self.firstname_and_surname_to_login(' '.join(firstname_surname[:i]),
                                                   ' '.join(firstname_surname[i:]))
            if login:
                return login
        return None

    def firstname_or_surname_to_logins(self, name, base=None, attributes=None): # pragma: no cover
        """Retrieve possible logins from a surname, a firstname
        or an incomplete login (with a dot inside).
        Returns the attributes needed."""
        name = name.strip()
        if len(name) == 1:
            return
        if (name[0].isdigit() or name[1].isdigit()) and len(name) < 6:
            return # XXX To not overload LDAP server
        q = []
        if ' ' in name and len(name) >= 3:
            both = '(%s=%s*)' % (configuration.attr_surname,
                                 utilities.safe_space_quote(name))
        else:
            both = ''
        for start in name.split(' '):
            start = utilities.safe_quote(start.strip())
            if start == '':
                continue
            q.append('(|(%s=%s*)(%s=%s*)(%s=%s*)%s)' % (
                configuration.attr_surname, start,
                configuration.attr_firstname, start,
                configuration.attr_login, start,
                both))
        if len(q) == 0:
            return
        q = '(&' + ''.join(q) + ')'

        if '@' in name:
            q = '(%s=%s)' % (configuration.attr_mail, name)

        if base:
            if base.startswith("CN="):
                q = '(&(memberOf=' + base + ')' + q + ')'
                base = configuration.ou_top
        else:
            base = configuration.ou_top
                
        q = '(&(objectClass=person)' + q + ')' # To accelerate query

        if attributes == None:
            attributes = [configuration.attr_login,
                          configuration.attr_surname,
                          configuration.attr_firstname]
        aa = self.query(q, base=base, attributes=attributes)
        i = attributes.index(configuration.attr_login)
        for cn, x in aa:
            if cn is None:
                continue
            if x.get(configuration.attr_login) == None:
                continue
            r = [x.get(attr, ('',))[0]
                 for attr in attributes
                 ]
            r[i] = r[i].lower() # login must be in lower case
            yield r

    def query_logins(self, logins, attributes, only_first_value=True): # pragma: no cover
        if not logins:
            return ()
        chunk_size = 100
        if len(logins) > chunk_size:
            r = []
            chunk = logins[:]
            n = 1 + len(logins) // int(1 + len(logins)//chunk_size)
            while chunk:
                r += self.query_logins(chunk[:n], attributes, only_first_value)
                chunk = chunk[n:]
            return r
        
        logins = ''.join(['(%s=%s)' % (configuration.attr_login,
                                      utilities.the_login(login))
                                        for login in logins
                         ])
        a = self.query(search='(|' + logins + ')',
                       base=configuration.ou_top, attributes=attributes
                       )
        r = []
        for i in a:
            if i[0] != None:
                i = i[1]
                if only_first_value:
                    r.append([i.get(attr,('',))[0]
                              for attr in attributes])
                else:
                    r.append(i)
        return r


    @utilities.add_a_method_cache
    def students(self, ue, year=None, semester=None, table=None): # pragma: no cover
        """Iterator giving the student list for an UE"""
        alls = self.query(base=configuration.ou_students,
                          search='(memberOf=*%s*)' % ue,
                          attributes=(configuration.attr_login,
                                      configuration.attr_firstname,
                                      configuration.attr_surname,
                                      configuration.attr_mail,
                                      configuration.attr_login_alt))
        for i in alls:
            i = i[1]
            yield (login_to_student_id(i[configuration.attr_login][0]),
                   i[configuration.attr_firstname][0],
                   i[configuration.attr_surname][0],
                   i[configuration.attr_mail][0],
                   '', # Group
                   '', # Sequence
                   )

    # XXX All the following methods are deprecated: DO NOT USE THEM
    def password_ok(self, login): # pragma: no cover
        return configuration.User(login).password_ok
    def member_of_list(self, login): # pragma: no cover
        return configuration.User(login).member_of_list or ()
    def etapes_of_student(self, login, year=None): # pragma: no cover
        return configuration.User(login).get_etapes_list(year)
    def etapes_of_students(self, logins, year=None): # pragma: no cover
        return configuration.Users(logins).get_etapes_list(year)
    def is_in_one_of_the_groups(self, login, groups): # pragma: no cover
        return configuration.User(login).is_in_one_of_the_groups(groups)
    def ues_of_a_student(self, login): # pragma: no cover
        return configuration.User(login).ues_long
    def ues_of_a_student_short(self, login): # pragma: no cover
        return configuration.User(login).ues_short
    def ues_of_a_student_with_groups(self, login): # pragma: no cover
        return configuration.User(login).ues_with_groups
    def firstname_and_surname(self, login): # pragma: no cover
        user = configuration.User(login)
        return [user.firstname or 'Inconnu', user.surname or 'Inconnu']
    def mail(self, login): # pragma: no cover
        return configuration.User(login).mail or login
    def mails(self, logins): # pragma: no cover
        return configuration.Users(logins).get('mail', default='')
    def phone(self, login): # pragma: no cover
        return configuration.User(login).phone
    def phone_from_logins(self, logins): # pragma: no cover
        return configuration.Users(logins).get('phone')
    def portail(self, login): # pragma: no cover
        return configuration.User(login).portails
    def portails(self, logins): # pragma: no cover
        return configuration.Users(logins).get('portails')
    def firstname_and_surname_and_mail_from_logins(self, logins): # pragma: no cover
        return configuration.Users(logins).get('firstname', 'surname', 'mail')
    def firstname_and_surname_and_mail(self, login): # pragma: no cover
        user = configuration.User(login)
        return [user.firstname or 'Inconnu', user.surname or 'Inconnu',
                user.mail or user.ID_]
    get_student_info = firstname_and_surname_and_mail

class Empty(LDAP_Logic):
    """Fake LDAP handler"""
    def __init__(self, name='LDAP'):
        self.nr = 0
    def query(self, search='', attributes=None, base=''):
        self.nr += 1
        if attributes == None:
            attributes = (configuration.attr_login,) # pragma: no cover
        d = {}
        for i in attributes:
            d[i] = (i, configuration.teachers[1])
        d['base'] = base
        d['search'] = search
        return (('ldapdisabled', d),)
    def query_login(self, login, attributes=(configuration.attr_login,)):
        self.nr += 1
        if login in demo_animaux:
            d = demo_animaux[login]
            d = {configuration.attr_login: [login],
                 configuration.attr_mail: [d[3]],
                 configuration.attr_surname: [d[2]],
                 configuration.attr_firstname: [d[1]],
                 'memberOf': ['CN=UE-%d,' % i + configuration.ou_ue_contains
                              for i in range(10)],
                 }
        else:
            d = {configuration.attr_login: [login],
                 configuration.attr_mail: ['???@???.???'],
                 configuration.attr_surname: [login + 'surname'],
                 configuration.attr_firstname: [login + 'firstname'],
                 'memberOf': (),
                 }
        for k in attributes:
            if k not in d:
                d[k] = [k]
        return d

#REDEFINE
# This class can be replaced by a subclass of itself in order to replace
# the 'students' method returning the student list for an UE.
# The student list may be computed without using LDAP.
class LDAP(LDAP_Logic): # pragma: no cover
    """An LDAP connection."""
    connection = True
    connect_time = 0
    def __init__(self):
        self.connect()

    def connect(self):
        if time.time() - self.connect_time < 1:
            # The reconnection was done by another thread.
            # We must not reinitialize the connection once again.
            return
        if ( len(configuration.ldap_server) == 0
             or configuration.ldap_server[0].endswith('.domain.org')):
            # Fake LDAP server, do not use it
            return

        url = ('ldaps://' if configuration.ldap_server_port in (636, 6360) else 'ldap://')
        url += configuration.ldap_server[0] + ':' + str(configuration.ldap_server_port) + '/'
        connection = ldap.ldapobject.ReconnectLDAPObject(url, retry_max=10, retry_delay=5.,
            trace_stack_limit=0)
        connection.set_option(ldap.OPT_TIMELIMIT, time_limit)
        connection.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
        connection.set_option(ldap.OPT_REFERRALS, 0)
        connection.set_option(ldap.OPT_X_KEEPALIVE_IDLE, True)
        connection.simple_bind_s(
            configuration.ldap_server_login,
            configuration.ldap_server_password)
        old_connection = self.connection
        self.connection = connection
        if old_connection and old_connection is not True:
            try:
                old_connection.unbind()
            except:
                pass
        self.connect_time = time.time()

    @utilities.add_a_lock
    def query(self, search, attributes=(configuration.attr_login,),
              base=configuration.ou_groups):
        """Returns a list of answers.
        A answer is a pair : (CN, dictionnary)
        """
        if self.connection is True:
            # Fake connection
            t = {}
            for a in attributes:
                t[a] = [a + '?']
            return (('cn?', t),)
        self.search = search
        while True:
            try:
                infos = self.connection.search_s(base, ldap.SCOPE_SUBTREE, search, attributes)
                coding = configuration.ldap_encoding
                for key, attributes in infos:
                    if key:
                        for values in attributes.values():
                            for i, value in enumerate(values):
                                values[i] = value.decode(coding, errors='replace')
                self.search = None
                return infos
            except ldap.NO_SUCH_OBJECT:
                utilities.warn("No such object: {}".format(base))
                return ()
            except:
                utilities.send_backtrace(
                    repr(configuration.ldap_server) + '\n' +
                    'QUERY=' + search + '\n' +
                    'ATTRIBUTES=' + repr(attributes) + '\n' +
                    'BASE=' + base + '\n',
                    subject='LDAP Error',
                    exception=True
                )
                self.connect()


    def query_login(self, login, attributes=(configuration.attr_login,),
                    star_is_safe=False):
        if ('*' in login) and not star_is_safe:
            return {}
        if not star_is_safe:
            login = utilities.the_login(login)
        if len(login) <= 2:
            return {}
        a = self.query(search='(%s=%s)' % (configuration.attr_login, login),
                    base=configuration.ou_top, attributes=attributes
                    )
        if not a or a[0][0] == None:
            a = self.query(search='(%s=%s)'% (configuration.attr_login, login),
                        base=configuration.ou_top, attributes=attributes
                        )
            if a and a[0][0] != None:
                utilities.send_backtrace('THIS CODE IS USEFUL!',
                                         exception=False)
        if a and a[0][0] != None:
            if len(a[0]) >= 2:
                return a[0][1]
        return {}

L_fast  = None # Fast interactive access
L_slow  = None # Slow request interactive access
L_batch = None # Any request in a batch thread

def init():
    global L_fast, L_slow, L_batch
    utilities.warn('Create LDAP connector {}'.format(configuration.ldap_server))
    L_fast = LDAP()
    L_slow = LDAP()
    L_batch = LDAP()

#DEPRECATED: you should subclass entity.py:Users
def login_to_student_id(login): return utilities.safe(login)

#DEPRECATED: you should subclass entity.py:Users
def login_from_ldap(i): # pragma: no cover
    if configuration.attr_login_alt in i:
        return i[configuration.attr_login_alt][0]
    else:
        return i[configuration.attr_login][0]

#DEPRECATED
def a_mailto(login): return configuration.User(login).a_mailto() # pragma: no cover

demo_animaux = {
    'k01':('k01','Bernard' ,'BONOBO'      ,'bbonobo@africa.net'     ,'M',''),
    'k02':('k02','Georges' ,'ROUGE GORGE' ,'grouge-gorge@europe.net','O',''),
    'k03':('k03','Magalie' ,'MIGALE'      ,'mmigale@africa.net'     ,'A',''),
    'k04':('k04','Lucien'  ,'LEZARD'      ,'llezard@france.net'     ,'R',''),
    'k05':('k05','Théodore','TIGRE'       ,'ttigre@asia.net'        ,'M',''),
    'k06':('k06','Simon'   ,'SCORPION'    ,'sscorpion@africa.net'   ,'A',''),
    'k07':('k07','Cécilia' ,'CHEVAL'      ,'ccheval@europe.net'     ,'M',''),
    'k08':('k08','Tatiana' ,'TORTUE'      ,'ttortue@ocean.net'      ,'R',''),
    'k09':('k09','Ambroise','AIGLE'       ,'aaigle@america.net'     ,'O',''),
    'k10':('k10','Bill'    ,'BOA'         ,'bboa@africa.net'        ,'R',''),
    'k11':('k11','Merlin'  ,'MYRIAPODE'   ,'mmerlin@europe.net'     ,'A',''),
    'k12':('k12','Fanny'   ,'FLAMANT ROSE','fflamant-rose@europ.net','O',''),
    'k13':('k13','Olivier' ,'OURS'        ,'oours@us.net'           ,'M',''),
    }
