#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    TOMUSS: The Online Multi User Simple Spreadsheet)
#    Copyright (C) 2008-2018 Thierry EXCOFFIER, Universite Claude Bernard
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact: Thierry.EXCOFFIER@bat710.univ-lyon1.fr

import os
from . import inscrits
from . import document
from . import utilities
from . import configuration

#DEPRECATED: you should subclass entity.py:Users
# These functions are called by default to not break existing code.
def need_a_charte(login):    return False
def need_a_referent(login):  return True
def Student(v, portail): pass
def analyse_groups(student, groups):
    # Update list of the UE of the student
    for ue in groups:
        if ue.startswith(configuration.ou_ue_starts): # pragma: no cover
            student.ues.append(ue[6:].split(' ')[0])

#DEPRECATED: you should use User instances
def get(login):
    # Return the student's referent for the current semester
    return configuration.User(login).referent # pragma: no cover


#REDEFINE
# List of LDAP OU of student to be affected to a 'referent'
def port():
    return configuration.the_portails['UFRFST'] 

#REDEFINE
# List of LDAP OU of student not to be affected to a 'referent'
def not_in():
    return ()

#REDEFINE
# Compute the login list of students login needing a referent
def student_list(f, pportails, not_in_list): # pragma: no cover
    f.write('<h1>' + utilities._("TITLE_referent_student_list") + '</h1>\n')
    students = set()
    for portail in pportails:
        f.write('<h2>' + portail + '</h2>')
        f.flush()
        portail_students = configuration.Users(
            i[1][configuration.attr_login][0]
            for i in inscrits.L_slow.query(
                base=configuration.ou_students,
                search='(memberOf=%s)' % portail,
                attributes=(configuration.attr_login,))
        )
        for student in portail_students:
            f.write(student.ID_ + ' ')
            for p in student.member_of_list:
                if p in not_in_list:
                    break
            else:
                # Not in any forbidden list
                students.add(student.ID_)
    return configuration.Users(students)

#REDEFINE
# Triggered when a student is removed from a teacher
def remove_student_from_referent_hook(referent, student_id):
    return

#REDEFINE
# Triggered when a student is added to a teacher
def add_student_to_referent_hook(referent, student_id):
    return

#REDEFINE
# This function returns the teacher to assign to a student.
# It is only needed for special case of the generic algorithm.
def search_best_teacher_local(student, sorted_teachers, f, all_teachers):
    if 'INFO_L3' in student.discipline and student.primo_entrant:
        return all_teachers['elodie.desseree'] # pragma: no cover

#REDEFINE
# Returns True if teacher want to keep the student even if
# the student must not have a referent teacher
def teacher_keep_student(teacher:'Teacher', student_id):
    return False

def les_blocsnotes(year):
    dirname = os.path.join(configuration.db, 'Y' + str(year), 'SReferents')
    if not os.path.exists(dirname):
        return # pragma: no cover

    for ue in utilities.python_files(dirname):
        if ue == 'abjs.py':
            continue # pragma: no cover
        ue = utilities.login_to_module(ue[:-3])
        yield document.table(str(year), 'Referents', ue, ro=True)

def html(student):
    return '%s %s %s %s ues=%s disciplines=%s' % (student.ID_,
            student.firstname, student.surname, student.mail,
            student.ues, student.discipline)

class Teacher(object):
    def __init__(self, user, line_key, table):
        self.table = table
        self.line = table.lines[line_key]
        discipline = self.line[1].value
        self.no_more_students = '*' in discipline
        self.discipline = set(discipline.replace('*', '').split(' '))
        self.line_key = line_key
        self.page = self.table.pages[0]
        assert user.ID_ == self.line[0].value
        self.user_ID = user.ID_
        self.user_firstname = user.firstname.title()
        self.user_surname = user.surname
        self.user_mail = user.mail
        self.students = []
        for cell in self.line[2:]:
            if cell.value:
                self.students.append(cell.value)
        self.reset()

    def reset(self):
        self.nr_weight = 0
        for student in self.students:
            self.nr_weight += self.student_weight(student)
        self.message = []
        self.updated_students = list(self.students)

    @classmethod
    def student_weight(cls, student_id):
        "Some students, need more work for teachers"
        if student_id.startswith('1%02d' % (utilities.university_year()%100)):
            return 4 # pragma: no cover
        else:
            return 1

    def find_cell(self, value=''):
        for cell, column in list(zip(self.line, self.table.columns))[2:]:
            if cell.value == value:
                return column.the_id
        if value: # pragma: no cover
            return None # The cell does not exists, not yet synchronized...
        add_column(self.table)
        return self.table.columns[-1].the_id

    def append(self, student, in_table=False, mail=False):
        if in_table:
            column_key = self.find_cell()
            self.table.lock()
            try:
                self.table.cell_change(self.table.pages[1],
                                       column_key, self.line_key,
                                       student.student_id)
            finally:
                self.table.unlock()

            add_student_to_referent_hook(self.user_ID, student.student_id)
            student.pop('referent')
            if mail:
                utilities.send_mail_in_background(
                    student.mail,
                    utilities._("MSG_referent_change_subject"),
                    utilities._("MSG_referent_change_message") % (
                        student.firstname.title(), student.surname,
                        self.user_firstname, self.user_surname,
                        self.user_mail,
                        ),
                    show_to=True,
                    reply_to=self.user_mail,
                    error_to=self.user_mail) # pragma: no cover
        if in_table:
            self.students.append(student.student_id)
        else:
            self.updated_students.append(student.student_id)
        self.nr_weight += self.student_weight(student.student_id)

    def remove(self, student, in_table=False):
        if in_table:
            remove_student_from_referent_hook(self.user_ID, student.student_id)
            column_id = self.find_cell(student.student_id)
            if column_id:
                # In some cases the table was not synchronized
                # So the student is not yet in the table
                with self.table.the_lock:
                    self.table.cell_change(self.page, column_id, self.line_key, '')
            student.pop('referent')
            student.manage_key('old_referent', content=self.user_ID)

        if in_table:
            self.students.remove(student.student_id)
        else:
            if student.student_id in self.updated_students:
                self.updated_students.remove(student.student_id)
        self.nr_weight -= self.student_weight(student.student_id)

    def common_disciplines(self, student):
        return self.discipline.intersection(student.discipline)

    def merge_line_in_the_table(self, line_key):
        # The lines are already merged in the Teacher instance
        table = self.table
        page = self.page
        table.lock()
        try:
            table.cell_change(page, 'a', line_key, '')
            table.cell_change(page, 'b', line_key, '')
        finally:
            table.unlock()
        for cell, column in list(zip(table.lines[line_key],
                                     table.columns))[2:]:
            if cell.value == '':
                continue # pragma: no cover
            column_key = self.find_cell()
            table.lock()
            try:
                table.cell_change(page, column_key, self.line_key, cell.value)
                table.cell_change(page, column.the_id, line_key, '')
            finally:
                table.unlock()

    def __str__(self):
        return '%s %s (%f) %s' % (self.user_ID,
                                  ' '.join(self.discipline),
                                  len(self.students), self.students)



def students_of_a_teacher(teacher):
    try:
        return configuration.User(teacher).referent_teacher.students
    except AttributeError:
        return () # Not a referent teacher

def referent(year, semester, login):
    """Returns the 'referent' of a student."""
    return referents_students(
        year, semester
        ).id_to_referent.get(configuration.User(login).student_id)

def referents_login(year, semester):
    """Returns the referent list"""
    return tuple(referents_students(year, semester).logins())

def referent_dict(year, semester):
    """For each student, its referent teacher"""
    d = {}
    for line in referents_students(year, semester).lines.values():
        for cell in line[2:]:
            if cell.value:
                d[cell.value] = line[0].value
    return d

def the_students(year, semester):
    """The list of student WITH a 'referent'"""
    students = []
    for line in referents_students(year, semester).lines.values():
        if line[0].value == '':
            continue
        for cell in line[2:]:
            if cell.value != '':
                students.append(cell.value)
    return students

def nr_students(year, semester):
    """Number of student with a 'referent'"""
    return len(the_students(year, semester))

def remove_student_from_referent(referent, student):
    configuration.User(referent).referent_teacher.remove(
        configuration.User(student), in_table=True)

def add_column(table):
    table.lock()
    try:
        i = len(table.columns) - 2
        table.column_change(table.pages[1],'%d'%i,str(i+1),'Text','','','',0,4)
    finally:
        table.unlock()

def add_student_to_referent(referent, student):
    configuration.User(referent).referent_teacher.append(
        configuration.User(student), in_table=True)

def search_best_teacher(student, sorted_teachers, f, all_teachers):        
    """Search the teacher of a student"""

    teacher = search_best_teacher_local(student, sorted_teachers,
                                        f, all_teachers)
    if teacher:
        return teacher # pragma: no cover

    # If an old referent is possible.
    k = student.manage_key('old_referent')
    if k:
        # \n at the end if hand edited
        k = k.strip() # pragma: no cover
    if (k in all_teachers
        and  all_teachers[k].referent_teacher.common_disciplines(student)
       ):
        return all_teachers[k] # pragma: no cover

    # The teacher is in the same discipline with the less students.
    
    for teacher in sorted_teachers:
        if not teacher.referent_teacher.common_disciplines(student):
            continue

        # No new students for this teacher
        if teacher.referent_teacher.no_more_students:
            continue # pragma: no cover
        # OK for this teacher.
        return teacher
    return None


def referents_students(year=None, semester=None):
    if year is None:
        year, semester = configuration.year_semester
    return document.table(year, semester, 'referents_students',
                          ro=configuration.read_only)

def students_as_users(f):
    students = student_list(f, port(), not_in())
    if not isinstance(students, configuration.Users):
        # old scripts may return dict
        students = configuration.Users(students) # pragma: no cover
    return students

def dump_state(table, students, all_teachers):
    print('*'*99)
    for line in table.lines.values():
        print('REFERENT:', ' '.join(cell.value for cell in line))
        login = line[0].value
        if login in all_teachers:
            teacher = all_teachers[line[0].value]
            print(teacher)
    for student in students:
        print(student)

def update_referents(ticket, f, really_do_it = False, add_students=True):
    # Get the referent table
    table = referents_students()
    page = table.pages[0]
    f.write('''
    <link rel="stylesheet" href="../style.css" type="text/css">
    <script src="../utilities.js"></script>
    ''')

    f.write('<h1>' + configuration.year_semester[1] + ' '
            + str(configuration.year_semester[0]) + '</h1>\n')
    f.write(table.filename + '\n')

    if not table.modifiable: # pragma: no cover
        f.write('<p>%s\n' % utilities._("MSG_referent_unmodifiable"))
        return

    students = students_as_users(f)

    f.write('<h1>' + utilities._("TITLE_referent_nr_students") % len(students)
            + '</h1>\n')

    f.write('<h1>%s</h1>\n' % utilities._("TITLE_referent_remove_duplicates"))
    all_cells = {} # The key is the login
    all_teachers = configuration.Users(set(table.logins()))
    for teacher in all_teachers:
        teacher.referent_teacher.reset()
    if configuration.real_regtest:
        dump_state(table, students, all_teachers)
    # teacher.discipline and not teacher.no_more_students:
    # utilities._("MSG_referent_orphan_student"),
    for line_key, line in table.lines.items():
        login = line[0].value
        if not login:
            continue
        teacher = all_teachers[login]
        if teacher.referent_teacher.line_key != line_key:
            f.write('%s%s<br>\n' % (
                    utilities._("MSG_referent_remove_teacher"),
                    login))
            if really_do_it:
                teacher.referent_teacher.merge_line_in_the_table(line_key)

        for cell in line[2:]:
            if cell.value == '': # XXX Assume no students on the line
                continue
            student = configuration.User(cell.value)
            the_student = student.student_id
            if the_student in all_cells: # pragma: no cover
                f.write(utilities._("MSG_referent_remove_duplicate_student")
                        + '%s<br>\n' % cell.value)
                teacher.referent_teacher.remove(student, in_table=really_do_it)
            else:
                if (the_student in students
                    or teacher_keep_student(teacher, cell.value)
                    ):
                    all_cells[the_student] = teacher
                else:
                    f.write(utilities._("MSG_referent_student_not_in_list")
                        % (cell.value, login, student.etapes_list))
                    f.write(utilities._("MSG_referent_student_removed"))
                    teacher.referent_teacher.remove(student,
                                                    in_table=really_do_it)
                    teacher.referent_teacher.message.append(
                        utilities._("MSG_referent_remove_student") +
                        '%s %s %s' % (the_student, student.firstname,
                                        student.surname))
                    f.write('<br>\n')

    f.write('<h1>%s</h1>\n' % utilities._("TITLE_referent_students_in_need"))
    missing = []
    if add_students:
        for student in students.values():
            if student.student_id not in all_cells and student.need_a_referent:
                missing.append(student)
                f.write(html(student) + '<br>\n') # XXX

    f.write('<h1>%s</h1>\n' % utilities._("TITLE_referent_teacher_list"))
    sorted_teachers = list(all_teachers.values())
    sorted_teachers.sort(key=lambda x: (x.referent_teacher.nr_weight, x.ID_))
    for teacher in sorted_teachers:
        f.write('<li> ' + str(teacher.referent_teacher) + '\n')

    f.write('<h1>%s</h1>\n' % utilities._("TITLE_referent_affectations"))
    for student in missing:
        sorted_teachers.sort(key=lambda x: x.referent_teacher.nr_weight)

        teacher = search_best_teacher(student, sorted_teachers, f, all_teachers)
        if teacher == None:
            f.write('<li> ' + utilities._("MSG_referent_no_teacher") %
                    html(student) + '\n')
            continue
        if really_do_it:
            s = teacher.referent_teacher.students
        else:
            s = teacher.referent_teacher.updated_students
        f.write('<li><b>' + teacher.ID_
                + '[' + str(len(s))
                + ']</b> (%s): ' % (teacher.referent_teacher.discipline,)
                + student.student_id + ' ' + repr(student.discipline) + '\n')
        teacher.referent_teacher.append(student, in_table=really_do_it)
        teacher.referent_teacher.message.append(
            utilities._("MSG_referent_add_student") % html(student))

    if really_do_it:
        for teacher in sorted_teachers:
            teacher.referent_teacher.students = teacher.referent_teacher.updated_students

    f.write('<h1>%s</h1>\n' % utilities._("TITLE_referent_resume"))
    for teacher in sorted_teachers:
        if not teacher.referent_teacher.message:
            continue
        f.write('<li> %s<br>\n' % teacher.ID_
                + '<br>\n'.join(teacher.referent_teacher.message) + '\n')

        if really_do_it:
            utilities.send_mail_in_background(
                teacher.mail,
                utilities._("MSG_referent_mail_subject"),
                utilities._("MSG_referent_mail_body") %
                ('<br>\n'.join(teacher.referent_teacher.message),
                 configuration.server_url),
                frome=ticket.user.mail)

